/******************************************************************************
 *                  QLOGIC LINUX SOFTWARE
 *
 * QLogic ISP2x00 device driver for Linux 2.2.x and 2.4.x
 * Copyright (C) 2001 Qlogic Corporation
 * (www.qlogic.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/


#define	QLA_PT_CMD_TOV			(30)
#define QLA_IOCTL_ACCESS_WAIT_TIME	30 /* seconds */

#ifdef INTAPI
#include "inioct.h"
/* from qla_inioct.c */
extern int qla2x00_read_nvram(adapter_state_t *, EXT_IOCTL *, int);
extern int qla2x00_update_nvram(adapter_state_t *, EXT_IOCTL *, int);
extern int qla2x00_write_nvram_word(adapter_state_t *, uint8_t, uint16_t);
extern int qla2x00_send_loopback(adapter_state_t *, EXT_IOCTL *, int);
#endif

/*
 * Local prototypes
 */
STATIC int qla2x00_find_curr_ha(int, scsi_qla_host_t **);

STATIC int qla2x00_aen_reg(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_aen_get(adapter_state_t *, EXT_IOCTL *, int);
STATIC void qla2x00_enqueue_aen(adapter_state_t *, uint16_t, void *);

STATIC int qla2x00_query(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_hba_node(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_hba_port(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_disc_port(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_disc_tgt(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_chip(adapter_state_t *, EXT_IOCTL *, int);

STATIC int qla2x00_get_data(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_get_statistics(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_get_fc_statistics(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_get_port_summary(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_driver(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_query_fw(adapter_state_t *, EXT_IOCTL *, int);

STATIC int qla2x00_send_fcct(adapter_state_t *, EXT_IOCTL *, int);
STATIC int qla2x00_start_ms_cmd(adapter_state_t *, EXT_IOCTL *, srb_t *);

STATIC int qla2x00_wwpn_to_scsiaddr(scsi_qla_host_t *, EXT_IOCTL *, int);

STATIC void qla2x00_waitq_sem_timeout(unsigned long);
STATIC uint8_t qla2x00_get_ioctl_access(adapter_state_t *, uint32_t);
STATIC uint8_t qla2x00_release_ioctl_access(adapter_state_t *);

STATIC void qla2x00_wait_q_memb_alloc(adapter_state_t *, wait_q_t **);
STATIC void qla2x00_wait_q_memb_free(adapter_state_t *, wait_q_t *);
STATIC uint8_t qla2x00_wait_q_add(adapter_state_t *, wait_q_t **);
STATIC void qla2x00_wait_q_remove(adapter_state_t *, wait_q_t **);


/*************************************************************************
 * qla2x00_ioctl_sleep_done
 *
 * Description:
 *   Wakeup ioctl completion semaphore for request that is waiting.
 *
 * Returns:
 *************************************************************************/
STATIC
void qla2x00_ioctl_sleep_done (struct semaphore * sem)
{
	DEBUG9(printk("ioctl_sleep: entered.\n");)

	if (sem != NULL) {
		DEBUG9(printk("ioctl_sleep: wake up sem.\n");)
		up(sem);
	}

	DEBUG9(printk("ioctl_sleep: exiting.\n");)
}

/*************************************************************************
 * qla2x00_ioctl_sem_init
 *
 * Description:
 *   Initialize timer and semaphore used while waiting for passthru
 *   completion.
 *
 * Returns:
 *   None.
 *************************************************************************/
void qla2x00_ioctl_sem_init (scsi_qla_host_t *ha)
{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	ha->ioctl->cmpl_sem = MUTEX_LOCKED;
#else
	init_MUTEX_LOCKED(&ha->ioctl->cmpl_sem);
#endif

	init_timer(&(ha->ioctl->cmpl_timer));
	ha->ioctl->cmpl_timer.data = (unsigned long)&ha->ioctl->cmpl_sem;
	ha->ioctl->cmpl_timer.function =
	    (void (*)(unsigned long))qla2x00_ioctl_sleep_done;

}

/*************************************************************************
 * qla2x00_ioctl_sleep
 *
 * Description:
 *   Release io_request_lock and schedule a timeout for the ioctl request.
 *
 * Returns:
 *   None.
 *************************************************************************/
void qla2x00_ioctl_sleep (scsi_qla_host_t *ha, int timeout, uint8_t got_lock)
{
	ha->ioctl->cmpl_timer.expires = jiffies + timeout;
	add_timer(&ha->ioctl->cmpl_timer);

	/* Now ready to receive interrupt */
	if (got_lock) {
		DEBUG(printk(
		    "qla2x00_ioctl_sleep: releasing io_request_lock.\n");)
		spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);
	}
	down(&ha->ioctl->cmpl_sem);

	del_timer(&ha->ioctl->cmpl_timer);
}

/*************************************************************************
 * qla2x00_scsi_pt_done
 *
 * Description:
 *   Sets completion flag.
 *
 * Returns:
 *************************************************************************/
void
qla2x00_scsi_pt_done(Scsi_Cmnd *pscsi_cmd)
{
	struct Scsi_Host *host;
	adapter_state_t  *ha;
	srb_t *sp;

	host = pscsi_cmd->host;
	ha = (scsi_qla_host_t *) host->hostdata;
	sp = (srb_t *) CMD_SP(pscsi_cmd);

	/* printk("[[qla2x00_scsi_pt_done post function called OK]]\n"); */
	DEBUG4(printk("qla2x00_scsi_pt_done post function called OK\n");)

	/* save detail status for IOCTL reporting */
	sp->flags = sp->flags | SRB_ISP_COMPLETED;
	ha->ioctl->SCSIPT_InProgress = 0;
	ha->ioctl->ioctl_tov = 0;

	up(&ha->ioctl->cmpl_sem);

	return;
}


/*************************************************************************
 * qla2x00_fcct_done
 *
 * Description:
 *   Sets completion flag.
 *
 * Returns:
 *************************************************************************/
void
qla2x00_fcct_done(Scsi_Cmnd *pscsi_cmd)
{
	struct Scsi_Host *host;
	adapter_state_t  *ha;
	srb_t *sp;

	host = pscsi_cmd->host;
	ha = (scsi_qla_host_t *) host->hostdata;
	sp = (srb_t *) CMD_SP(pscsi_cmd);

	/* printk("[[qla2x00_fcct_done post function called OK]]\n"); */
	DEBUG4(printk("qla2x00_fcct_done post function called OK\n");)

	sp->flags = sp->flags | SRB_ISP_COMPLETED;
	ha->ioctl->FCCT_InProgress = 0;
	ha->ioctl->ioctl_tov = 0;

	up(&ha->ioctl->cmpl_sem);

	return;
}

/*************************************************************************
 * qla2x00_ioctl
 *
 * Description:
 *   Performs ioctl requests not satified by the upper levels.
 *
 * Returns:
 *   ret  = 0    Success
 *   ret != 0    Failed; detailed status copied to EXT_IOCTL structure
 *               if applicable
 *************************************************************************/
int
qla2x00_ioctl(Scsi_Device *dev, int cmd, void *arg)
{
	int		mode = 0;
	int		rval = 0;
	int		ret = EINVAL;

	uint8_t		*usr_temp, *kernel_tmp;
	uint8_t		*temp;
	uint8_t		tempbuf[8];
	uint32_t	i, b, t, l, status;
	uint32_t	tgt;
	uint32_t	which_pass, scsi_direction;
#if BITS_PER_LONG <= 32
	uint32_t	handle;
#else
	uint64_t	handle;
#endif

	EXT_FC_SCSI_PASSTHRU	fc_scsi_pass;
	EXT_FC_SCSI_PASSTHRU	*pfc_scsi_pass = &fc_scsi_pass;
	EXT_IOCTL		ext;
	PEXT_IOCTL		pext = &ext;
	EXT_SCSI_PASSTHRU	scsi_pass;
	EXT_SCSI_PASSTHRU	*pscsi_pass = &scsi_pass;

	Scsi_Cmnd		scsi_cmd;
	Scsi_Cmnd		*pscsi_cmd = &scsi_cmd;
	Scsi_Device		scsi_device;
	struct Scsi_Host	*host;

	adapter_state_t		*ha;
	fc_lun_t		temp_fclun;
	fc_lun_t		*fclun = NULL;
	fc_port_t		*fcport;
	os_lun_t		*lq;
	os_tgt_t		*tq;
	srb_t			ioctl_sp;
	srb_t			*sp = &ioctl_sp;


	DEBUG9(printk("qla2x00_ioctl: entry to command (%x), arg (%p)\n",
	    cmd, arg);)

	BZERO(sp, sizeof(srb_t));

	host = dev->host;
	ha = (scsi_qla_host_t *) host->hostdata; /* midlayer chosen instance */

	ret = verify_area(VERIFY_READ, (void *)arg, sizeof(EXT_IOCTL));
	if (ret) {
		DEBUG9_10(printk("qla2x00_ioctl: ERROR VERIFY_READ EXT_IOCTL "
		    "sturct. ha=%8x.\n", (uint32_t)ha);)
		return(ret);
	}

	/* copy in application layer EXT_IOCTL */
	os_copyin(arg, pext, sizeof(EXT_IOCTL), 0);

	/* check signature of this ioctl */
	temp = (uint8_t *) &pext->Signature;

#if BITS_PER_LONG <= 32
	for (i = 0; i < 4; i++, temp++)
		tempbuf[i] = *temp;

	if ((tempbuf[0] == 'Q') && (tempbuf[1] == 'L') &&
	    (tempbuf[2] == 'O') && (tempbuf[3] == 'G'))
		status = 0;
	else
		status = 1;
#else
	if ((tempbuf[0] == 'Q') && (tempbuf[1] == 'L') &&
	    (tempbuf[2] == 'O') && (tempbuf[3] == 'G') &&
	    (tempbuf[4] == 'I') && (tempbuf[5] == 'C'))
		status = 0;
	else
		status = 1;
#endif

	if (status != 0) {
		DEBUG9_10(printk("qla2x00_ioctl: signature did not match. "
		    "ha=%8x\n", (uint32_t)ha);)
		ret = EXT_STATUS_INVALID_PARAM;
		return(ret);
	}

	/* check version of this ioctl */
	if (pext->Version > EXT_VERSION) {
		printk(KERN_WARNING
		    "qla2x00: interface version not supported = %d.\n",
		    pext->Version);
		pext->Status = EXT_STATUS_UNSUPPORTED_VERSION;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		ret = EXT_STATUS_ERR;
		return(ret);
	}

	/* check for API setting HBA Instance for subsequent operations */
	if (cmd == (int)EXT_CC_STARTIOCTL) {
		DEBUG9(printk("qla2x00_ioctl: got startioctl command.\n");)

		pext->Instance = num_hosts;
		pext->Status = EXT_STATUS_OK;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		return(EXT_STATUS_OK);

	} else if (cmd == (int)EXT_CC_SETINSTANCE) {
		/*
		 * Since API opens devices once and uses handle for
		 * subsequent calls, we keep a parameter to designate
		 * the "active HBA" for ioctls.
		 */
		if (pext->HbaSelect < num_hosts) {
			apiHBAInstance = pext->Instance;
			/*
			 * Return host number in pext->HbaSelect for
			 * reference by IOCTL caller.
			 */
			if (qla2x00_find_curr_ha(apiHBAInstance, &ha) != 0) {
				DEBUG9_10(printk("qla2x00_ioctl: ERROR finding "
				    "ha in EXT_SETINSTANCE. Instance=%d "
				    "num_hosts=%d ha=%8x.\n",
				    pext->Instance, num_hosts, (uint32_t)ha);)
				pext->Status = EXT_STATUS_DEV_NOT_FOUND;
				os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
				return (EXT_STATUS_ERR);
			}

			pext->HbaSelect = ha->host_no;
			pext->Status = EXT_STATUS_OK;
			os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

			DEBUG9(printk("qla2x00_ioctl: Setting instance to "
			    "%d.\n", apiHBAInstance);)

			ret = EXT_STATUS_OK;
		} else {
			DEBUG9_10(printk("qla2x00_ioctl: ERROR EXT_SETINSTANCE."
			    " Instance=%d num_hosts=%d ha=%8x.\n",
			    pext->Instance, num_hosts, (uint32_t)ha);)

			pext->Status = EXT_STATUS_DEV_NOT_FOUND;
			os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
			ret = EXT_STATUS_ERR;
		}

		return (ret);
	}

	/*
	 * Check for valid apiHBAInstance (set previously by EXT_SETINSTANCE
	 * or default 0)  and set ha context for this IOCTL.
	 */
	if (qla2x00_find_curr_ha(apiHBAInstance, &ha) != 0) {

		DEBUG9_10(printk("qla2x00_ioctl: ERROR matching apiHBAInstance "
		    "%d to an HBA Instance.\n", apiHBAInstance);)

		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		ret = EXT_STATUS_ERR;
		return(ret);
	}

	/*
	 * Get permission to process ioctl command. Only one will proceed
	 * at a time.
	 */
	if (qla2x00_get_ioctl_access(ha, QLA_IOCTL_ACCESS_WAIT_TIME) != 0) {
		/* error timed out */
		DEBUG9_10(printk("qla2x00_ioctl: ERROR timeout getting "
		    "ioctl access. Inst=%d.\n", apiHBAInstance);)

		pext->Status = EXT_STATUS_BUSY;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		ret = EXT_STATUS_ERR;
		return(ret);
	}

	/*
	 * Set EXT_IOCTL.HbaSelect to host number for reference by IOCTL
	 * caller.
	 */
	pext->HbaSelect = ha->host_no;

	DEBUG9(printk("qla2x00_ioctl: active apiHBAInstance=%d CC=%x SC=%x.\n",
	    apiHBAInstance, cmd, pext->SubCode);)

        
	while( ha->cfg_active || ha->dpc_active ) {
       		if( signal_pending(current) )
               		break;   /* get out */

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(1);
	};

#ifdef FC_MP_SUPPORT
	/* handle qlmultipath ioctls and return to user */
	if (cmd > SDM_QLMULTIPATH_IOCTLS) {
		printk(KERN_INFO
		    "qla2x00: calling handle_mp_ioctl ha=%8x\n\n",
		    (uint32_t)ha);
		ret = handle_mp_ioctl(ha, cmd, pext, arg);
		qla2x00_release_ioctl_access(ha);
		return(ret);
	}
#endif

	switch (cmd) { /* switch on EXT IOCTL COMMAND CODE */

	case EXT_CC_QUERY:

		DEBUG9(printk("qla2x00_ioctl: got query command.\n");)
		rval = qla2x00_query(ha, pext, 0);
		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case EXT_CC_GET_DATA:

		DEBUG9(printk("qla2x00_ioctl: got get_data command.\n");)
		rval = qla2x00_get_data(ha, pext, 0);
		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case EXT_CC_SEND_FCCT_PASSTHRU:
		DEBUG(printk("qla2x00_ioctl: got CT passthru cmd.\n"));
		rval = qla2x00_send_fcct(ha, pext, 0);
		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		copy_to_user((void *)arg, (void *)pext, sizeof(EXT_IOCTL));

		break;

    case EXT_CC_SEND_SCSI_PASSTHRU:
      /* clear ioctl_sp and scsi_cmd and scsi_device to be used */
      kernel_tmp = (uint8_t *)sp;
      for (i=0;i<sizeof(srb_t);i++)       *kernel_tmp = 0;
      kernel_tmp = (uint8_t *)pscsi_cmd;
      for (i=0;i<sizeof(Scsi_Cmnd);i++)   *kernel_tmp = 0;
      kernel_tmp = (uint8_t *)&scsi_device;
      for (i=0;i<sizeof(Scsi_Device);i++) *kernel_tmp = 0;
      kernel_tmp = (uint8_t *)ha->ioctl_mem;
      for (i=0;i<PAGE_SIZE;i++) *kernel_tmp = 0;
      pscsi_cmd->sense_buffer[0]  = 0;
      pscsi_cmd->sense_buffer[1]  = 0;
      pscsi_cmd->sense_buffer[2]  = 0;

      /* printk("[start EXT_CC_SEND_SCSI_PASSTHRU]\n"); */

      switch(pext->SubCode) { /* get target specification accordingly */
        case EXT_SC_SEND_SCSI_PASSTHRU:
		  BZERO(sp, sizeof(*sp));
          /* printk("[start EXT_SC_SEND_SCSI_PASSTHRU]\n"); */
          ret = verify_area(VERIFY_READ, (void *)pext->RequestAdr,
                sizeof(EXT_SCSI_PASSTHRU));
          if (ret) {
            pext->Status = EXT_STATUS_COPY_ERR;
            os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
            DEBUG9_10(printk("qla2x00_ioctl: ERROR verify READ "
              "SCSI_PASSTHRU.\n");)
	    qla2x00_release_ioctl_access(ha);
            return(ret);
          }

          for (i=0 ; i < sizeof(EXT_SCSI_PASSTHRU) ; i++) {
            usr_temp   = (uint8_t *)pext->RequestAdr + i;
            kernel_tmp = (uint8_t *)pscsi_pass + i;
            __get_user(*kernel_tmp, usr_temp);
          }
          /* set target coordinates */
          scsi_cmd.target               = pscsi_pass->TargetAddr.Target;
          scsi_cmd.lun                  = pscsi_pass->TargetAddr.Lun;
	  printk(KERN_INFO "[start EXT_SC_SEND_SCSI_PASSTHRU T=%x L=%x cmd=%x] ",pscsi_pass->TargetAddr.Target,pscsi_pass->TargetAddr.Lun,pscsi_pass->Cdb[0]);
      if (pscsi_pass->CdbLength == 6) {
        scsi_cmd.cmd_len = 6;
        scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pscsi_pass->Cdb[0];
        scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pscsi_pass->Cdb[1];
        scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pscsi_pass->Cdb[2];
        scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pscsi_pass->Cdb[3];
        scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pscsi_pass->Cdb[4];
        scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pscsi_pass->Cdb[5];
        scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = 0;
        scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = 0;
        scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = 0;
        scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = 0;
      } else if (pscsi_pass->CdbLength == 10) {
        scsi_cmd.cmd_len = 0x0A;
        scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pscsi_pass->Cdb[0];
        scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pscsi_pass->Cdb[1];
        scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pscsi_pass->Cdb[2];
        scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pscsi_pass->Cdb[3];
        scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pscsi_pass->Cdb[4];
        scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pscsi_pass->Cdb[5];
        scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = pscsi_pass->Cdb[6];
        scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = pscsi_pass->Cdb[7];
        scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = pscsi_pass->Cdb[8];
        scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = pscsi_pass->Cdb[9];
      } else if (pscsi_pass->CdbLength == 12) {
        scsi_cmd.cmd_len = 0x0C;
        scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pscsi_pass->Cdb[0];
        scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pscsi_pass->Cdb[1];
        scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pscsi_pass->Cdb[2];
        scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pscsi_pass->Cdb[3];
        scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pscsi_pass->Cdb[4];
        scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pscsi_pass->Cdb[5];
        scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = pscsi_pass->Cdb[6];
        scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = pscsi_pass->Cdb[7];
        scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = pscsi_pass->Cdb[8];
        scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = pscsi_pass->Cdb[9];
        scsi_cmd.data_cmnd[10] = scsi_cmd.cmnd[10] = pscsi_pass->Cdb[10];
        scsi_cmd.data_cmnd[11] = scsi_cmd.cmnd[11] = pscsi_pass->Cdb[11];
      } else {
          printk(KERN_WARNING "qla2x00_ioctl: SCSI_PASSTHRU Unknown Cdb Length=%x.\n",pscsi_pass->CdbLength);
      }
      which_pass = EXT_SC_SEND_SCSI_PASSTHRU;
          break;

        case EXT_SC_SEND_FC_SCSI_PASSTHRU:
          DEBUG4(printk("[start EXT_SC_SEND_FC_SCSI_PASSTHRU]\n");)
          ret = verify_area(VERIFY_READ, (void *)pext->RequestAdr,
                sizeof(EXT_FC_SCSI_PASSTHRU));
          if (ret) {
            pext->Status = EXT_STATUS_COPY_ERR;
            os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
            DEBUG9_10(printk("qla2x00_ioctl: ERROR verify READ "
              "SCSI_FC_PASSTHRU.\n");)
	    qla2x00_release_ioctl_access(ha);
            return(ret);
          }
          for (i=0 ; i < sizeof(EXT_FC_SCSI_PASSTHRU) ; i++) {
            usr_temp   = (uint8_t *)pext->RequestAdr + i;
            kernel_tmp = (uint8_t *)pfc_scsi_pass + i;
            __get_user(*kernel_tmp, usr_temp);
          }
          if (fc_scsi_pass.FCScsiAddr.DestType != EXT_DEF_DESTTYPE_WWPN) {
              pext->Status = EXT_STATUS_DEV_NOT_FOUND;
              os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
              DEBUG9_10(printk("qla2x00_ioctl: EXT_FC_SCSI_PASSTHRU ERROR - "
                "wrong Dest type. \n");)
              ret = EXT_STATUS_ERR;
	      qla2x00_release_ioctl_access(ha);
              return(ret);
          }

        fclun = NULL;
	tgt=MAX_FIBRE_DEVICES;
        for (fcport= ha->fcport; (fcport) ; fcport = fcport->next) {
           if (memcmp(fcport->port_name, 
		fc_scsi_pass.FCScsiAddr.DestAddr.WWPN, 8) == 0) {
		for (fclun = fcport->fclun; fclun; fclun = fclun->next) {
			if (fclun->lun == fc_scsi_pass.FCScsiAddr.Lun) {
				break;
			}
		}
		break;
	    }
  	}

         if (fcport == NULL) {
             pext->Status = EXT_STATUS_DEV_NOT_FOUND;
             os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
             DEBUG9_10(printk("qla2x00_ioctl: SCSI_PASSTHRU FC AddrFormat - "
               "DID NOT FIND Port for WWPN.\n");)
	     qla2x00_release_ioctl_access(ha);
             return(pext->Status);
         }

	 /* Ensure the port has a valid loop id before issuing the 
	    command to the firmware */
	 if (fcport->loop_id == FC_NO_LOOP_ID) {
             pext->Status = EXT_STATUS_DEVICE_OFFLINE;
             os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
             DEBUG9_10(printk("qla2x00_ioctl: SCSI_PASSTHRU FC AddrFormat - "
               "Port has no loop id.\n");)
	     qla2x00_release_ioctl_access(ha);
             return(pext->Status);
	 }

	/* v5.21b9 - use a temporary fclun */
         if (fclun == NULL) {
  		fclun = &temp_fclun;
		fclun->fcport = fcport;
		fclun->lun = fc_scsi_pass.FCScsiAddr.Lun;
		fclun->flags = 0;
		fclun->next = NULL;
	}
          	/* set target coordinates */
          	scsi_cmd.target               = tgt;
          	scsi_cmd.lun                  = fc_scsi_pass.FCScsiAddr.Lun;
		DEBUG4(printk("[SCSI_PASSTHRU FC AddrFormat  T=%x L=%x]\n",tgt,scsi_cmd.lun); )
			if (pfc_scsi_pass->CdbLength == 6) {
				scsi_cmd.cmd_len = 6;
				scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pfc_scsi_pass->Cdb[0];
				scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pfc_scsi_pass->Cdb[1];
				scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pfc_scsi_pass->Cdb[2];
				scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pfc_scsi_pass->Cdb[3];
				scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pfc_scsi_pass->Cdb[4];
				scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pfc_scsi_pass->Cdb[5];
				scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = 0;
				scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = 0;
				scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = 0;
				scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = 0;
			} else if (pfc_scsi_pass->CdbLength == 10) {
				scsi_cmd.cmd_len = 0x0A;
				scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pfc_scsi_pass->Cdb[0];
				scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pfc_scsi_pass->Cdb[1];
				scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pfc_scsi_pass->Cdb[2];
				scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pfc_scsi_pass->Cdb[3];
				scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pfc_scsi_pass->Cdb[4];
				scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pfc_scsi_pass->Cdb[5];
				scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = pfc_scsi_pass->Cdb[6];
				scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = pfc_scsi_pass->Cdb[7];
				scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = pfc_scsi_pass->Cdb[8];
				scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = pfc_scsi_pass->Cdb[9];
			} else if (pfc_scsi_pass->CdbLength == 12) {
				scsi_cmd.cmd_len = 0x0C;
				scsi_cmd.data_cmnd[0] = scsi_cmd.cmnd[0] = pfc_scsi_pass->Cdb[0];
				scsi_cmd.data_cmnd[1] = scsi_cmd.cmnd[1] = pfc_scsi_pass->Cdb[1];
				scsi_cmd.data_cmnd[2] = scsi_cmd.cmnd[2] = pfc_scsi_pass->Cdb[2];
				scsi_cmd.data_cmnd[3] = scsi_cmd.cmnd[3] = pfc_scsi_pass->Cdb[3];
				scsi_cmd.data_cmnd[4] = scsi_cmd.cmnd[4] = pfc_scsi_pass->Cdb[4];
				scsi_cmd.data_cmnd[5] = scsi_cmd.cmnd[5] = pfc_scsi_pass->Cdb[5];
				scsi_cmd.data_cmnd[6] = scsi_cmd.cmnd[6] = pfc_scsi_pass->Cdb[6];
				scsi_cmd.data_cmnd[7] = scsi_cmd.cmnd[7] = pfc_scsi_pass->Cdb[7];
				scsi_cmd.data_cmnd[8] = scsi_cmd.cmnd[8] = pfc_scsi_pass->Cdb[8];
				scsi_cmd.data_cmnd[9] = scsi_cmd.cmnd[9] = pfc_scsi_pass->Cdb[9];
				scsi_cmd.data_cmnd[10] = scsi_cmd.cmnd[10] = pfc_scsi_pass->Cdb[10];
				scsi_cmd.data_cmnd[11] = scsi_cmd.cmnd[11] = pfc_scsi_pass->Cdb[11];
			} else {
    			printk(KERN_WARNING "qla2x00_ioctl: FC_SCSI_PASSTHRU Unknown Cdb Length=%x.\n", pfc_scsi_pass->CdbLength);
			}
			which_pass = EXT_SC_SEND_FC_SCSI_PASSTHRU;
          	break;

        default:
          	/* printk("[SCSI_PASSTHRU UNKNOWN ADDRESS FORMAT]\n");*/
         	pext->Status = EXT_STATUS_INVALID_PARAM;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		DEBUG9_10(printk("qla2x00_ioctl: ERROR in SubCode decode of "
		    "EXT_CC_SEND_SCSI_PASSTHRU ha=%8x]\n",(uint32_t)ha);)
		ret = EXT_STATUS_ERR;
	        qla2x00_release_ioctl_access(ha);
		return(ret);

      } /* end of switch for EXT_SC_SEND_SCSI_PASSTHRU or EXT_SC_SEND_FC_SCSI_PASSTHRU */

	DEBUG(printk("Dump of cdb buffer:\n");)
        DEBUG(qla2100_dump_buffer((uint8_t *)&scsi_cmd.data_cmnd[0], 16);)
      /******** Common Portion of SCSI PassThru Operations ********/
      if (pext->ResponseLen > PAGE_SIZE) {
        pext->Status = EXT_STATUS_INVALID_PARAM;
        os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
        DEBUG9_10(printk("qla2x00_ioctl: ERROR size of requested "
          "EXT_SCSI_PASSTHRU.\n");)
	qla2x00_release_ioctl_access(ha);
        return(ret);
      }
      pscsi_cmd->host       = ha->host;
      sp->ha                = ha;
      sp->cmd               = pscsi_cmd;
      sp->flags             = 0; //SRB_WATCHDOG;
      /* set local scsi_cmd's sp pointer to sp */
      CMD_SP(pscsi_cmd)     = (void *) sp;

      /* mark this as a special delivery and collection command */
      scsi_cmd.flags     = 0;
      scsi_cmd.scsi_done = qla2x00_scsi_pt_done;

      scsi_cmd.device               = &scsi_device;
      scsi_cmd.device->tagged_queue = 0;
      scsi_cmd.use_sg               = 0; /* no ScatterGather */
      scsi_cmd.request_bufflen      = pext->ResponseLen;
      scsi_cmd.request_buffer       = ha->ioctl_mem;
      scsi_cmd.timeout_per_command  = QLA_PT_CMD_TOV * HZ;

      DEBUG4(printk("[start FCSCSI IOCTL look at direction t=%x l=%x]\n",
       scsi_cmd.target,scsi_cmd.lun); )
      if (pscsi_pass->Direction == EXT_DEF_SCSI_PASSTHRU_DATA_OUT ||
          pfc_scsi_pass->Direction == EXT_DEF_SCSI_PASSTHRU_DATA_OUT) {
          /* sending user data from pext->ResponseAdr to device */
          ret = verify_area(VERIFY_READ, (void *)pext->ResponseAdr,
                            pext->ResponseLen);
          if (ret) {
            pext->Status = EXT_STATUS_COPY_ERR;
            os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
            DEBUG9_10(printk("qla2x00_ioctl: ERROR verify READ "
              "SCSI_PASSTHRU.\n");)
	    qla2x00_release_ioctl_access(ha);
            return(pext->Status);
          }
/* 03/20 -- missing direction */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
          scsi_cmd.sc_data_direction = SCSI_DATA_WRITE;
#endif
            usr_temp   = (uint8_t *)pext->ResponseAdr;
            kernel_tmp = (uint8_t *)ha->ioctl_mem;
            copy_from_user(kernel_tmp, usr_temp, pext->ResponseLen);
      }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
      else
          scsi_cmd.sc_data_direction = SCSI_DATA_READ;
#endif
      /* Generate LU queue on bus, target, LUN */
      b = SCSI_BUS_32(pscsi_cmd);
      t = SCSI_TCN_32(pscsi_cmd);
      l = SCSI_LUN_32(pscsi_cmd);

      tq = ha->ioctl->ioctl_tq;
      lq = ha->ioctl->ioctl_lq;
	DEBUG4(printk("[FC_SCSI_PASSTHRU: ha instance=%d tq=%p lq=%p fclun=%p]\n",
		ha->instance,tq,lq,fclun); )
	if (which_pass == EXT_SC_SEND_FC_SCSI_PASSTHRU && 
		fclun && tq && lq ) {
		BZERO(tq, sizeof(os_tgt_t));
		BZERO(lq, sizeof(os_lun_t));

		tq->olun[fclun->lun] = lq;
		tq->ha = ha;
		lq->fclun = fclun;
		fcport = fclun->fcport;
    	} else if ((tq = (os_tgt_t *) TGT_Q(ha, t)) != NULL &&
	    	(lq = (os_lun_t *) LUN_Q(ha, t, l)) != NULL ) {

		fcport = lq->fclun->fcport;
	} else {
		lq = NULL;
		fcport = NULL;
    	}

	DEBUG4(printk("qla2x00_ioctl: FC_SCSI_PT CDB=%x %x %x %x; "
	    "b=%x t=%x l=%x\n",
	    scsi_cmd.cmnd[0],scsi_cmd.cmnd[1],scsi_cmd.cmnd[2],
	    scsi_cmd.cmnd[3],b,t,l);)

		/* set sp->target for 32bit/64bit delivery */
		sp->wdg_time                   = 0;

		/* Setup device queue pointers. */
		sp->tgt_queue          = tq;
		sp->lun_queue          = lq;
		sp->s_cmd.base_address = sp;

		/* check presense of requested target and other conditions */
		if( fcport == NULL || ( ha->loop_down_timer == 0 && 
		    ha->loop_state == LOOP_DOWN)) {

			DEBUG9_10(printk("qla2x00_ioctl: SCSI PT IOCTL "
			    "Target/Lun MISSING fcport=%p.\n",fcport);)

			if( fcport == NULL ) {
				printk(KERN_INFO "scsi(%2d:%2d:%2d:%2d): "
				    "qla2x00 PassThur port is unavailable\n",
				    (int)ha->host_no,b,t,l);
			} else {
				printk(KERN_INFO "scsi(%2d:%2d:%2d:%2d): "
				    "qla2x00 PassThur - loop is down\n",
				    (int)ha->host_no,b,t,l);
			}
			pext->Status = EXT_STATUS_DEV_NOT_FOUND;
			os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
	                qla2x00_release_ioctl_access(ha);
			return(pext->Status);
		}

	/* If lun is suspended wait for it */
	while( (lq->q_flag & LUN_QUEUE_SUSPENDED) ){
		qla2x00_ioctl_sem_init(ha);
		qla2x00_ioctl_sleep (ha, 10, FALSE);
	};

      /* get spin lock for this operation */
      spin_lock_irqsave(&io_request_lock, ha->cpu_flags);

        /* Set an invalid handle until we issue the command to ISP */
        /* then we will set the real handle value.                 */
        handle = INVALID_HANDLE;
        CMD_HANDLE(pscsi_cmd) = (unsigned char *)handle;

        if( sp->flags ) {
            sp->port_down_retry_count = ha->port_down_retry_count - 1;
            sp->retry_count = ha->retry_count;
            DEBUG9(printk("qla2x00_ioctl: PT Set retry counts =0x%x,0x%x\n\r",
                   sp->port_down_retry_count,sp->retry_count);)
        }
        ha->qthreads++;
        ha->total_ios++;
        qla2100_putq_b(lq,sp);

        DEBUG5(printk("qla2x00 ioctl scsi passthru: queue hndl=0x%x\n\r",
	    handle);)

        /* set flag to indicate IOCTL SCSI PassThru in progress */
        ha->ioctl->SCSIPT_InProgress = 1;
        ha->ioctl->ioctl_tov = (int)QLA_PT_CMD_TOV + 1;

	/* prepare for receiving completion. */
	qla2x00_ioctl_sem_init(ha);

        DEBUG5(qla2x00_print_scsi_cmd(pscsi_cmd);)

        //printk("[start FC_SCSI IOCTL restart queues]\n");
        /* send command to adapter */
	qla2x00_next(tq, lq);

	/* Look for successful completion */
	qla2x00_ioctl_sleep (ha, ha->ioctl->ioctl_tov * HZ, TRUE);

	if (ha->ioctl->SCSIPT_InProgress == 1) {

	  printk(KERN_WARNING
	      "qla2x00: scsi%ld ERROR passthru command timeout.\n",
	      ha->host_no);

	  pext->Status = EXT_STATUS_DEV_NOT_FOUND;
	  os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
	  ret = EXT_STATUS_ERR;

	} else {

	switch (sp->ccode) {
	    case CS_INCOMPLETE:
	    case CS_ABORTED:
	    case CS_PORT_UNAVAILABLE:
	    case CS_PORT_LOGGED_OUT:
	    case CS_PORT_CONFIG_CHG:
	    case CS_PORT_BUSY:
		DEBUG10(printk("qla2x00_ioctl: scsi passthru cs err = %x.\n",
		  sp->ccode);)
	        ret = EXT_STATUS_ERR;
	        pext->Status = EXT_STATUS_BUSY;
	        break;
	}
        if ( !(sp->flags & SRB_ISP_COMPLETED) ) {
          /* We waited and post function did not get called */
          pext->Status = EXT_STATUS_DEV_NOT_FOUND;
          DEBUG10(printk("qla2100_ioctl: scsi%ld ERROR command timeout\n",
                     ha->host_no);)
          os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
          ret = EXT_STATUS_ERR;
        } else if ((sp->ccode == CS_DATA_UNDERRUN) ||(sp->scode != 0))  {

	  /* have done the post function */
          pext->Status       = EXT_STATUS_SCSI_STATUS;
          pext->DetailStatus = sp->scode;
	  DEBUG10(printk("qla2x00_ioctl: data underrun or scsi err. host "
	      "status =0x%x, scsi status = 0x%x.\n",
	      sp->ccode, sp->scode);)
          if (which_pass == EXT_SC_SEND_FC_SCSI_PASSTHRU) {
	    /* printk("[[ioctl completion which_pass from FC_SCSI_PT]]\n"); */
               scsi_direction = pfc_scsi_pass->Direction;
	  } else {
	    /* printk("[[ioctl completion which_pass from SCSI_PT]]\n"); */
               scsi_direction = pscsi_pass->Direction;
	  }
          if (scsi_direction == EXT_DEF_SCSI_PASSTHRU_DATA_IN) {
	     DEBUG9(printk("qla2x00_ioctl: copying data.\n");)

             /* getting device data and putting in pext->ResponseAdr */
             ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
                               pext->ResponseLen);
             if (ret) {
               pext->Status = EXT_STATUS_COPY_ERR;
               os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
               DEBUG9_10(printk("qla2100_ioctl: ERROR verify WRITE for "
		 "IOCTL PT ha=%8x]\n",
                     (uint32_t)ha);)
	       qla2x00_release_ioctl_access(ha);
               return(ret);
             }
             /* now copy up the READ data to user */
             for (i=0 ; i < pext->ResponseLen ; i++) {
                usr_temp   = (uint8_t *)pext->ResponseAdr + i;
                kernel_tmp = (uint8_t *)ha->ioctl_mem + i;
		/*  printk("[%x]",*kernel_tmp); */
                __put_user(*kernel_tmp, usr_temp);
             }
	  }

	  /* printk("[[sense0=%x sense2=%x]]\n",
             pscsi_cmd->sense_buffer[0],
	     pscsi_cmd->sense_buffer[2]); */
          /* copy up structure to make sense data available to user */
          for (i=0;i<16;i++)
               pfc_scsi_pass->SenseData[i] = pscsi_cmd->sense_buffer[i];
          ret = verify_area(VERIFY_WRITE, (void *)pext->RequestAdr,
                            sizeof(EXT_SCSI_PASSTHRU));
          if (ret) {
            pext->Status = EXT_STATUS_COPY_ERR;
            os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
            DEBUG9_10(printk("qla2100_ioctl: ERROR verify WRITE "
              "FC_SCSI_PASSTHRU]\n");)
	    qla2x00_release_ioctl_access(ha);
            return(ret);
          }
          for (i=0 ; i < sizeof(EXT_FC_SCSI_PASSTHRU) ; i++) {
            usr_temp   = (uint8_t *)pext->RequestAdr + i;
            kernel_tmp = (uint8_t *)pfc_scsi_pass + i;
            __put_user(*kernel_tmp, usr_temp);
          }
          DEBUG4(printk("[finished QLA2100 IOCTL EXT_FC_SCSI_REQ - Status= %d]\n",pext->Status); )
          os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
          ret = EXT_STATUS_OK;
        } else if (sp->ccode != 0) {
	    DEBUG9_10(printk("qla2x00_ioctl: scsi passthru cs err = %x. "
		"copying ext stat %x\n", sp->ccode, pext->Status);)
            os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
	    qla2x00_release_ioctl_access(ha);
	    return (pext->Status);
	} else {
	    DEBUG9(printk("qla2x00_ioctl: complete. host status =0x%x, scsi "
		"status = 0x%x.\n", sp->ccode, sp->scode);)

            if (which_pass == EXT_SC_SEND_FC_SCSI_PASSTHRU) {
                DEBUG(printk("ioctl completion which_pass from FC_SCSI_PT\n");)
                scsi_direction = pfc_scsi_pass->Direction;
            } else {
                DEBUG(printk("ioctl completion which_pass from SCSI_PT\n");)
                scsi_direction = pscsi_pass->Direction;
            }
            /* copy up structure to make sense data available to user */
            for (i=0;i<16;i++)
                 pfc_scsi_pass->SenseData[i] = pscsi_cmd->sense_buffer[i];

            if (scsi_direction == EXT_DEF_SCSI_PASSTHRU_DATA_IN &&
                pscsi_cmd->sense_buffer[2] == 0) {
		DEBUG9(printk("qla2x00_ioctl: copying data.\n");)

                /* getting device data and putting in pext->ResponseAdr */
                ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
                    pext->ResponseLen);
                if (ret) {
                    pext->Status = EXT_STATUS_COPY_ERR;
                    os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
                    DEBUG9_10(printk("qla2x00_ioctl: ERROR verify write "
			"resp buf.\n");)
	            qla2x00_release_ioctl_access(ha);
                    return(ret);
                }
                /* now copy up the READ data to user */
                usr_temp = (uint8_t *)pext->ResponseAdr;
                kernel_tmp = (uint8_t *)ha->ioctl_mem;
                copy_to_user(usr_temp, kernel_tmp, pext->ResponseLen);
                ret = EXT_STATUS_OK;
	    }
	}
        }
      break;

	case EXT_CC_REG_AEN:
		rval = qla2x00_aen_reg(ha, pext, mode);

		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case EXT_CC_GET_AEN:
		rval = qla2x00_aen_get(ha, pext, mode);

		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case EXT_CC_WWPN_TO_SCSIADDR:
		rval = qla2x00_wwpn_to_scsiaddr(ha, pext, 0);
		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		copy_to_user(arg, pext, sizeof(EXT_IOCTL));
		break;

#ifdef INTAPI
	case INT_CC_READ_NVRAM:
		rval = qla2x00_read_nvram(ha, pext, mode);

		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case INT_CC_UPDATE_NVRAM:
		rval = qla2x00_update_nvram(ha, pext, mode);

		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;

	case INT_CC_LOOPBACK:
		rval = qla2x00_send_loopback(ha, pext, mode);

		ret = (rval == QL_STATUS_SUCCESS) ?
		    EXT_STATUS_OK : EXT_STATUS_ERR;
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);

		break;
#endif /* INTAPI */

/* all others go here */
/*
    case EXT_CC_ELS_RNID_SEND:
      break;
    case EXT_CC_ELS_RTIN_SEND:
      break;
    case EXT_CC_PLATFORM_REG:
      break;
*/

#if MPIO_SUPPORT
/* Failover IOCTLs */
	case FO_CC_GET_PARAMS:
	case FO_CC_SET_PARAMS:
	case FO_CC_GET_PATHS:
	case FO_CC_SET_CURRENT_PATH:
	case FO_CC_RESET_HBA_STAT:
	case FO_CC_GET_HBA_STAT:
	case FO_CC_GET_LUN_DATA:
	case FO_CC_SET_LUN_DATA:
	case FO_CC_GET_TARGET_DATA:
	case FO_CC_SET_TARGET_DATA:
		DEBUG4(printk("qla2x00_ioctl: failover arg (%p):\n", arg);)
		qla2x00_fo_ioctl(ha,cmd, arg, mode);
		os_copyout(pext, arg, sizeof(EXT_IOCTL), mode);
		break;
#endif

	default:
		ret = EXT_STATUS_ERR;
		break;
	} /* end of CC decode switch */

	DEBUG9(printk("qla2x00_ioctl: exiting. rval(%d) ret(%d)\n", rval, ret);)

	qla2x00_release_ioctl_access(ha);

	return(ret);
}

/*
 * qla2x00_alloc_ioctl_mem
 *	Allocates memory needed by IOCTL code.
 *
 * Input:
 *	ha = adapter state pointer.
 *
 * Returns:
 *	qla2x00 local function return status code.
 *
 * Context:
 *	Kernel context.
 */
int
qla2x00_alloc_ioctl_mem(adapter_state_t *ha)
{
	DEBUG9(printk("qla2x00_alloc_ioctl_mem entered.\n");)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	ha->ioctl_mem = (void *)KMEM_ZALLOC(PAGE_SIZE, 10);
	ha->ioctl_mem_phys =  VIRT_TO_BUS(ha->ioctl_mem);
#else
	/* Get consistent memory allocated for ioctl I/O operations. */
	ha->ioctl_mem = pci_alloc_consistent(ha->pdev, PAGE_SIZE,
	    &ha->ioctl_mem_phys);
#endif
	if (ha->ioctl_mem == NULL) {
		printk(KERN_WARNING
		    "qla2x00: ERROR in ioctl physical memory allocation\n");
		return (QL_STATUS_RESOURCE_ERROR);
	}

	/* Allocate context memory buffer */
	ha->ioctl = KMEM_ZALLOC(sizeof(hba_ioctl_context), 11);
	if (ha->ioctl == NULL) {
		/* error */
		printk(KERN_WARNING
		    "qla2x00: ERROR in ioctl context allocation.\n");
		return (QL_STATUS_RESOURCE_ERROR);
	}

	/* Allocate AEN tracking buffer */
	ha->ioctl->aen_tracking_queue =
	    KMEM_ZALLOC(EXT_DEF_MAX_AEN_QUEUE * sizeof(EXT_ASYNC_EVENT), 12);
	if (ha->ioctl->aen_tracking_queue == NULL) {
		printk(KERN_WARNING
		    "qla2x00: ERROR in ioctl aen_queue allocation.\n");
		return (QL_STATUS_RESOURCE_ERROR);
	}

	ha->ioctl->ioctl_tq = KMEM_ZALLOC(sizeof(os_tgt_t), 13);
	if (ha->ioctl->ioctl_tq == NULL) {
		printk(KERN_WARNING
		    "qla2x00: ERROR in ioctl tgt queue allocation.\n");
		return (QL_STATUS_RESOURCE_ERROR);
	}

  	ha->ioctl->ioctl_lq = KMEM_ZALLOC(sizeof(os_lun_t), 14);
	if (ha->ioctl->ioctl_lq == NULL) {
		printk(KERN_WARNING
		    "qla2x00: ERROR in ioctl lun queue allocation.\n");
		return (QL_STATUS_RESOURCE_ERROR);
	}

	/* Init wait_q fields */
	ha->ioctl->wait_q_lock = SPIN_LOCK_UNLOCKED;

	DEBUG9(printk("qla2x00_alloc_ioctl_mem exiting.\n");)

	return (QLA2X00_SUCCESS);
}

/*
 * qla2x00_free_ioctl_mem
 *	Frees memory used by IOCTL code.
 *
 * Input:
 *	ha = adapter state pointer.
 *
 * Context:
 *	Kernel context.
 */
void
qla2x00_free_ioctl_mem(adapter_state_t *ha)
{
	DEBUG9(printk("qla2x00_free_ioctl_mem entered.\n");)

	if (ha->ioctl != NULL) {

		if (ha->ioctl->ioctl_tq != NULL) {
			KMEM_FREE(ha->ioctl->ioctl_tq, sizeof(os_tgt_t));
			ha->ioctl->ioctl_tq = NULL;
		}

		if (ha->ioctl->ioctl_lq != NULL) {
  			KMEM_FREE(ha->ioctl->ioctl_lq, sizeof(os_lun_t));
			ha->ioctl->ioctl_lq = NULL;
		}

		if (ha->ioctl->aen_tracking_queue != NULL) {
			KMEM_FREE(ha->ioctl->aen_tracking_queue,
			    EXT_DEF_MAX_AEN_QUEUE * sizeof(EXT_ASYNC_EVENT));
			ha->ioctl->aen_tracking_queue = NULL;
		}

  		KMEM_FREE(ha->ioctl, sizeof(hba_ioctl_context));
		ha->ioctl = NULL;
	}

	/* free memory allocated for ioctl operations */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	KMFREE(ha->ioctl_mem, PAGE_SIZE);
#else
	pci_free_consistent(ha->pdev, PAGE_SIZE, ha->ioctl_mem,
	    ha->ioctl_mem_phys);
#endif

	DEBUG9(printk("qla2x00_free_ioctl_mem exiting.\n");)
}

STATIC int
qla2x00_find_curr_ha(int inst, scsi_qla_host_t **ret_ha)
{
	int	rval = QL_STATUS_SUCCESS;
	scsi_qla_host_t *search_ha = NULL;

	/*
	 * Check for valid apiHBAInstance (set previously by EXT_SETINSTANCE 
	 * or default 0)  and set ha context for this IOCTL
	 */    
	for (search_ha = qla2100_hostlist; 
	    (search_ha != NULL) && search_ha->instance != inst;
	     search_ha = search_ha->next)
		continue;

	if (search_ha == NULL) { 
		DEBUG10(printk("qla2x00_ioctl: ERROR matching apiHBAInstance "
		    "%d to an HBA Instance.\n", apiHBAInstance);)
		rval = QL_STATUS_ERROR;
	} else {
		*ret_ha = search_ha;
	}

	return(rval);
}

/*
 */
STATIC int
qla2x00_get_statistics(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	EXT_HBA_PORT_STAT tmp_stat;
	int ret;
	link_stat_t stat_buf;
	uint8_t i, rval;
	uint8_t  *usr_temp, *kernel_tmp;
	uint16_t mb_stat[1];
	uint32_t transfer_size;

	DEBUG9(printk("entered qla2x00_get_statistics function.\n");)

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
	    sizeof(EXT_HBA_PORT_STAT));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_get_statistics(%ld): ERROR "
		    "VERIFY_WRITE EXT_HBA_PORT_STAT.\n", ha->instance);)
		return(pext->Status);
	}

	/* Send mailbox cmd to get more. */
	spin_lock_irqsave(&io_request_lock, ha->cpu_flags);
	if ((rval = qla2x00_get_link_status(ha, ha->loop_id, &stat_buf,
	    mb_stat)) != QL_STATUS_SUCCESS) {
		spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

		if (rval == BIT_0) {
			pext->Status = EXT_STATUS_NO_MEMORY;
		} else if (rval == BIT_1) {
			pext->Status = EXT_STATUS_MAILBOX;
			pext->DetailStatus = EXT_DSTATUS_NOADNL_INFO;
		} else {
			pext->Status = EXT_STATUS_ERR;
		}

		DEBUG9_10(printk("qla2x00_get_statistics(%ld): ERROR "
		    "mailbox failed. mb[0]=%x.\n", ha->instance, mb_stat[0]);)
		printk(KERN_WARNING
		    "qla2x00_get_statistics(%ld): ERROR "
		    "mailbox failed. mb[0]=%x.\n", ha->instance, mb_stat[0]);

		return(pext->Status);
	}
	spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

	tmp_stat.ControllerErrorCount   =  ha->total_isp_aborts;
	tmp_stat.DeviceErrorCount       =  ha->total_dev_errs;
	tmp_stat.TotalIoCount           =  ha->total_ios;
	tmp_stat.TotalMBytes            =  ha->total_bytes / (1024*1024);
	tmp_stat.TotalLipResets         =  ha->total_lip_cnt;
/*
	tmp_stat.TotalInterrupts        =  ha->total_isr_cnt;
*/

	tmp_stat.TotalLinkFailures               = stat_buf.link_fail_cnt;
	tmp_stat.TotalLossOfSync                 = stat_buf.loss_sync_cnt;
	tmp_stat.TotalLossOfSignals              = stat_buf.loss_sig_cnt;
	tmp_stat.PrimitiveSeqProtocolErrorCount  = stat_buf.prim_seq_err_cnt;
	tmp_stat.InvalidTransmissionWordCount    = stat_buf.inval_xmit_word_cnt;
	tmp_stat.InvalidCRCCount                 = stat_buf.inval_crc_cnt;

	/* now copy up the STATISTICS to user */
	if (pext->ResponseLen < sizeof(EXT_HBA_PORT_STAT))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_HBA_PORT_STAT);


	for (i=0 ; i < transfer_size ; i++) {
		usr_temp   = (uint8_t *)pext->ResponseAdr + i;
		kernel_tmp = (uint8_t *)&tmp_stat + i;
		__put_user(*kernel_tmp, usr_temp);
	}
	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("finished qla2x00_get_statistics function.\n");)

	return (pext->Status);
}

/*
 */
STATIC int
qla2x00_get_fc_statistics(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	EXT_HBA_PORT_STAT tmp_stat;
	EXT_DEST_ADDR addr_struct;
	int ret;
	link_stat_t stat_buf;
	uint8_t i, rval, tgt;
	uint8_t  *usr_temp, *kernel_tmp;
	uint8_t  *req_name;
	uint16_t mb_stat[1];
	uint32_t transfer_size;

	DEBUG9(printk("entered qla2x00_get_fc_statistics function.\n");)

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
	    sizeof(EXT_HBA_PORT_STAT));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_get_fc_statistics(%ld): ERROR "
		    "VERIFY_WRITE.\n", ha->instance);)
		return(pext->Status);
	}

	rval = os_copyin((void*)(pext->RequestAdr), &addr_struct,
	    pext->RequestLen, 0);
	if (rval != 0) {
		pext->Status = EXT_STATUS_COPY_ERR;
		return (pext->Status);
	}

	/* find the device's loop_id */
	switch (addr_struct.DestType) {
	case EXT_DEF_DESTTYPE_WWPN:
		req_name = addr_struct.DestAddr.WWPN;
		for (tgt = 0; tgt < MAX_FIBRE_DEVICES; tgt++) {
			if (BCMP(ha->fc_db[tgt].wwn, req_name,
			    EXT_DEF_WWN_NAME_SIZE) == 0)
				break;
		}
		break;

	case EXT_DEF_DESTTYPE_WWNN:
	case EXT_DEF_DESTTYPE_PORTID:
	case EXT_DEF_DESTTYPE_FABRIC:
	case EXT_DEF_DESTTYPE_SCSI:
	default:
		pext->Status = EXT_STATUS_INVALID_PARAM;
		pext->DetailStatus = EXT_DSTATUS_NOADNL_INFO;
		DEBUG9_10(printk("qla2x00_get_statistics(%ld): ERROR "
		    "Unsupported subcode address type.\n", ha->instance);)
		return(pext->Status);
		break;
	}

	if (tgt == MAX_FIBRE_DEVICES) {
		/* not found */
		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		pext->DetailStatus = EXT_DSTATUS_TARGET;
		return (pext->Status);
	}

	/* check for suspended/lost device */
/*
	if (ha->fcport is suspended/lost) {
		pext->Status = EXT_STATUS_SUSPENDED;
		pext->DetailStatus = EXT_DSTATUS_TARGET;
		return (pext->Status);
	}
*/


	/* Send mailbox cmd to get more. */
	spin_lock_irqsave(&io_request_lock, ha->cpu_flags);
	if ((rval = qla2x00_get_link_status(ha, ha->fc_db[tgt].loop_id,
	    &stat_buf, mb_stat)) != QL_STATUS_SUCCESS) {
		spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

		if (rval == BIT_0) {
			pext->Status = EXT_STATUS_NO_MEMORY;
		} else if (rval == BIT_1) {
			pext->Status = EXT_STATUS_MAILBOX;
			pext->DetailStatus = EXT_DSTATUS_NOADNL_INFO;
		} else {
			pext->Status = EXT_STATUS_ERR;
		}

		DEBUG9_10(printk("qla2x00_get_fc_statistics(%ld): ERROR "
		    "mailbox failed. mb[0]=%x.\n", ha->instance, mb_stat[0]);)
		return(pext->Status);
	}
	spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

	tmp_stat.ControllerErrorCount   =  ha->total_isp_aborts;
	tmp_stat.DeviceErrorCount       =  ha->total_dev_errs;
	tmp_stat.TotalIoCount           =  ha->total_ios;
	tmp_stat.TotalMBytes            =  ha->total_bytes / (1024*1024);
	tmp_stat.TotalLipResets         =  ha->total_lip_cnt;
/*
	tmp_stat.TotalInterrupts        =  ha->total_isr_cnt;
*/

	tmp_stat.TotalLinkFailures               = stat_buf.link_fail_cnt;
	tmp_stat.TotalLossOfSync                 = stat_buf.loss_sync_cnt;
	tmp_stat.TotalLossOfSignals              = stat_buf.loss_sig_cnt;
	tmp_stat.PrimitiveSeqProtocolErrorCount  = stat_buf.prim_seq_err_cnt;
	tmp_stat.InvalidTransmissionWordCount    = stat_buf.inval_xmit_word_cnt;
	tmp_stat.InvalidCRCCount                 = stat_buf.inval_crc_cnt;

	/* now copy up the STATISTICS to user */
	if (pext->ResponseLen < sizeof(EXT_HBA_PORT_STAT))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_HBA_PORT_STAT);


	for (i=0 ; i < transfer_size ; i++) {
		usr_temp   = (uint8_t *)pext->ResponseAdr + i;
		kernel_tmp = (uint8_t *)&tmp_stat + i;
		__put_user(*kernel_tmp, usr_temp);
	}
	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("finished qla2x00_get_fc_statistics function.\n");)

	return (pext->Status);
}

/*
 * qla2x00_aen_reg
 *	IOCTL management server Asynchronous Event Tracking Enable/Disable.
 *
 * Input:
 *	ha = adapter state pointer.
 *	cmd = EXT_IOCTL cmd struct pointer.
 *	mode = flags.
 *
 * Returns:
 *	0 = success
 *
 * Context:
 *	Kernel context.
 */
/* ARGSUSED */
STATIC int
qla2x00_aen_reg(adapter_state_t *ha, EXT_IOCTL *cmd, int mode)
{
	int		rval = 0;
	EXT_REG_AEN	reg_struct;

	DEBUG9(printk("qla2x00_aen_reg: entered.\n");)

	rval = os_copyin((void*)(cmd->RequestAdr), &reg_struct,
	    cmd->RequestLen, mode);

	if (rval == 0) {
		cmd->Status = EXT_STATUS_OK;
		if (reg_struct.Enable) {
			ha->ioctl->flags |= IOCTL_AEN_TRACKING_ENABLE;
		} else {
			ha->ioctl->flags &= ~IOCTL_AEN_TRACKING_ENABLE;
		}
	} else {
		cmd->Status = EXT_STATUS_COPY_ERR;
		rval = EFAULT;
	}
	DEBUG9(printk("qla2x00_aen_reg: reg_struct. Enable(%d) "
	    "ha->ioctl_flag(%x) cmd->Status(%d) cmd->DetailStatus (%d).",
	    reg_struct.Enable, ha->ioctl->flags, cmd->Status,
	    cmd->DetailStatus);)

	DEBUG9(printk("qla2x00_aen_reg: exiting.\n");)

	return (rval);
}

/*
 * qla2x00_aen_get
 *	IOCTL management server Asynchronous Event Record Transfer.
 *
 * Input:
 *	ha = adapter state pointer.
 *	cmd = EXT_IOCTL cmd struct pointer.
 *	mode = flags.
 *
 * Returns:
 *	0 = success
 *
 * Context:
 *	Kernel context.
 */
STATIC int
qla2x00_aen_get(adapter_state_t *ha, EXT_IOCTL *cmd, int mode)
{
	int		rval = 0;
	EXT_ASYNC_EVENT	*tmp_q;
	EXT_ASYNC_EVENT	aen[EXT_DEF_MAX_AEN_QUEUE];
	uint8_t		i;
	uint8_t		queue_cnt;
	uint8_t		request_cnt;
	uint32_t	stat = EXT_STATUS_OK;
	uint32_t	dstat = EXT_STATUS_OK;
	uint32_t	ret_len = 0;

	DEBUG9(printk("qla2x00_aen_get: entered.\n");)

	request_cnt = (uint8_t)(cmd->ResponseLen / sizeof(EXT_ASYNC_EVENT));

	if (request_cnt < EXT_DEF_MAX_AEN_QUEUE) {
		/* We require caller to alloc for the maximum request count */
      		cmd->Status       = EXT_STATUS_BUFFER_TOO_SMALL;
		rval = 0;
		DEBUG9_10(printk(
		    "qla2x00_aen_get: Buffer too small. Exiting normally.");)
		return (rval);
	}

	/* 1st: Make a local copy of the entire queue content. */
	tmp_q = (EXT_ASYNC_EVENT *)ha->ioctl->aen_tracking_queue;
	queue_cnt = 0;

	DRIVER_LOCK
        spin_lock_irqsave(&io_request_lock, ha->cpu_flags);
	i = ha->ioctl->aen_q_head;

	for (; queue_cnt < EXT_DEF_MAX_AEN_QUEUE;) {
		if (tmp_q[i].AsyncEventCode != 0) {
			BCOPY((void *)&tmp_q[i], (void *)&aen[queue_cnt],
			    sizeof(EXT_ASYNC_EVENT));
			queue_cnt++;
			tmp_q[i].AsyncEventCode = 0; /* empty out the slot */
		}

		if (i == ha->ioctl->aen_q_tail) {
			/* done. */
			break;
		}

		i++;

		if (i == EXT_DEF_MAX_AEN_QUEUE) {
			i = 0;
		}
	}

	/* Empty the queue. */
	ha->ioctl->aen_q_head = 0;
	ha->ioctl->aen_q_tail = 0;

        spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);
	DRIVER_UNLOCK

	/* 2nd: Now transfer the queue content to user buffer */
	/* Copy the entire queue to user's buffer. */
	ret_len = (uint32_t)(queue_cnt * sizeof(EXT_ASYNC_EVENT));
	if (queue_cnt != 0) {
		rval = os_copyout((void *)&aen[0],
		    (void *)(cmd->ResponseAdr), ret_len, mode);

	}
	cmd->ResponseLen = ret_len;

	if (rval != 0) {
		stat = EXT_STATUS_COPY_ERR;
		rval = EFAULT;
		DEBUG10(printk("qla2x00_aen_get: FAILED. error = %d\n", stat);)
	} else {
		stat = EXT_STATUS_OK;
		rval = 0;
		DEBUG9(printk("qla2x00_aen_get: exiting normally.\n");)
	}

	cmd->Status       = stat;
	cmd->DetailStatus = dstat;

	DEBUG9(printk("qla2x00_aen_get: exiting. rval= %d\n", rval);)

	return (rval);
}

/*
 * qla2x00_enqueue_aen
 *
 * Input:
 *	ha = adapter state pointer.
 *	event_code = async event code of the event to add to queue.
 *	payload = event payload for the queue.
 *	INTR_LOCK must be already obtained.
 *
 * Context:
 *	Interrupt context.
 */
void
qla2x00_enqueue_aen(adapter_state_t *ha, uint16_t event_code, void *payload)
{
	uint8_t			new_entry; /* index to current entry */
	uint16_t		*mbx;
	EXT_ASYNC_EVENT		*aen_queue;

	DEBUG9(printk("qla2x00_enqueue_aen: entered.\n");)

	aen_queue = (EXT_ASYNC_EVENT *)ha->ioctl->aen_tracking_queue;
	if (aen_queue[ha->ioctl->aen_q_tail].AsyncEventCode != 0) {
		/* Need to change queue pointers to make room. */

		/* Increment tail for adding new entry. */
		ha->ioctl->aen_q_tail++;
		if (ha->ioctl->aen_q_tail == EXT_DEF_MAX_AEN_QUEUE) {
			ha->ioctl->aen_q_tail = 0;
		}
		if (ha->ioctl->aen_q_head == ha->ioctl->aen_q_tail) {
			/*
			 * We're overwriting the oldest entry, so need to
			 * update the head pointer.
			 */
			ha->ioctl->aen_q_head++;
			if (ha->ioctl->aen_q_head == EXT_DEF_MAX_AEN_QUEUE) {
				ha->ioctl->aen_q_head = 0;
			}
		}
	}

	DEBUG(printk("qla2x00_enqueue_aen: Adding code 0x%x to aen_q %p @ %d\n",
		event_code, aen_queue, ha->ioctl->aen_q_tail);)
	new_entry = ha->ioctl->aen_q_tail;
	aen_queue[new_entry].AsyncEventCode = event_code;

	DEBUG(printk("qla2x00_enqueue_aen: Adding code 0x%8x\n",
	    aen_queue[new_entry].AsyncEventCode);)

	/* Update payload */
	switch (event_code) {
	case MBA_LIP_OCCURRED:
	case MBA_LOOP_UP:
	case MBA_LOOP_DOWN:
	case MBA_LIP_RESET:
	case MBA_PORT_UPDATE:
		/* empty */
		break;

	case MBA_SCR_UPDATE:
		mbx = (uint16_t *)payload;
		/* al_pa */
		aen_queue[new_entry].Payload.RSCN.RSCNInfo[0] = LSB(mbx[2]);
		/* area */
		aen_queue[new_entry].Payload.RSCN.RSCNInfo[1] = MSB(mbx[2]);
		/* domain */
		aen_queue[new_entry].Payload.RSCN.RSCNInfo[2] = LSB(mbx[1]);
		/* save in big endian */
		BIG_ENDIAN_24(aen_queue[new_entry].Payload.RSCN.RSCNInfo[0]);

		aen_queue[new_entry].Payload.RSCN.AddrFormat = MSB(mbx[1]);

		break;

	default:
		/* Not supported */
		aen_queue[new_entry].AsyncEventCode = 0;
		break;
	}

	DEBUG9(printk("qla2x00_enqueue_aen: exiting.\n");)
}

STATIC int
qla2x00_query(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int rval;

	DEBUG9(printk("qla2x00_query: entered.\n");)

	/* All Query type ioctls are done here */
	switch(pext->SubCode) {

	case EXT_SC_QUERY_HBA_NODE:
		/* fill in HBA NODE Information */
		rval = qla2x00_query_hba_node(ha, pext, mode);
		break;

	case EXT_SC_QUERY_HBA_PORT:
		/* return HBA PORT related info */
		rval = qla2x00_query_hba_port(ha, pext, mode);
		break;

        case EXT_SC_QUERY_DISC_PORT:
		/* return discovered port information */
		rval = qla2x00_query_disc_port(ha, pext, mode);
		break;

        case EXT_SC_QUERY_DISC_TGT:
        	/* printk("[Start SC_QUERY_DISC_TGT active ha=%x]\n",ha); */
		rval = qla2x00_query_disc_tgt(ha, pext, mode);
		break;

        case EXT_SC_QUERY_CHIP:
		rval = qla2x00_query_chip(ha, pext, mode);
		break;

        case EXT_SC_QUERY_DISC_LUN:
		pext->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		rval = pext->Status;
		break;

        default:
		DEBUG10(printk("qla2x00_query: unknown SubCode %d.\n",
		    pext->SubCode);)
		pext->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		rval = pext->Status;
		break;
	}

	DEBUG9(printk("qla2x00_query: exiting.\n");)
	return (rval);
}

STATIC int
qla2x00_query_hba_node(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EXT_STATUS_OK;
	uint32_t	i, transfer_size;
	EXT_HBA_NODE	tmp_hba_node;
	qla_boards_t	*bdp;

	DEBUG9(printk("qla2x00_query_hba_node: entered.\n");)

	BZERO(&tmp_hba_node, sizeof(EXT_HBA_NODE));

	/* fill all available HBA NODE Information */
	bdp = &QLBoardTbl_fc[ha->devnum];
	for (i = 0; i < 8 ; i++)
		tmp_hba_node.WWNN[i] = ha->node_name[i];

	sprintf((char *)(tmp_hba_node.Manufacturer),"Qlogic Corp.");
	sprintf((char *)(tmp_hba_node.Model),(char *)&bdp->bdName[0]);

	tmp_hba_node.SerialNum[0] = ha->serial0;
	tmp_hba_node.SerialNum[1] = ha->serial1;
	tmp_hba_node.SerialNum[2] = ha->serial2;

	sprintf((char *)(tmp_hba_node.DriverVersion),QLA2100_VERSION);
	sprintf((char *)(tmp_hba_node.FWVersion),"%2d.%02d.%02d",
	    bdp->fwver[0], bdp->fwver[1], bdp->fwver[2]);

	sprintf((char *)(tmp_hba_node.OptRomVersion),"%d.%d",
	    ha->optrom_major, ha->optrom_minor);

	tmp_hba_node.InterfaceType = EXT_DEF_FC_INTF_TYPE;
	tmp_hba_node.PortCount = 1;

#if MPIO_SUPPORT
	tmp_hba_node.DriverAttr = (ha->flags.failover_enabled) ?
	    DRVR_FO_ENABLED : 0;
#endif

	ret = verify_area(VERIFY_WRITE, (void  *)pext->ResponseAdr,
	    sizeof(EXT_HBA_NODE));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_query_hba_node: ERROR verify write "
		    "rsp buffer.\n");)
		return(pext->Status);
	}

	/* now copy up the HBA_NODE to user */
	if (pext->ResponseLen < sizeof(EXT_HBA_NODE))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_HBA_NODE);

	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&tmp_hba_node,
	    transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("qla2x00_query_hba_node: exiting.\n");)
	return (ret);
}

STATIC int
qla2x00_query_hba_port(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int	ret = EXT_STATUS_OK;
	uint32_t	tgt_cnt, tgt, transfer_size;
	uint32_t	port_cnt;
	fc_port_t	*fcport;
	EXT_HBA_PORT	tmp_hba_port;

	DEBUG9(printk("qla2x00_query_hba_port: entered.\n");)

	BZERO(&tmp_hba_port, sizeof(EXT_HBA_PORT));

	/* reflect all HBA PORT related info */
	tmp_hba_port.WWPN[7] = ha->init_cb->port_name[7];
	tmp_hba_port.WWPN[6] = ha->init_cb->port_name[6];
	tmp_hba_port.WWPN[5] = ha->init_cb->port_name[5];
	tmp_hba_port.WWPN[4] = ha->init_cb->port_name[4];
	tmp_hba_port.WWPN[3] = ha->init_cb->port_name[3];
	tmp_hba_port.WWPN[2] = ha->init_cb->port_name[2];
	tmp_hba_port.WWPN[1] = ha->init_cb->port_name[1];
	tmp_hba_port.WWPN[0] = ha->init_cb->port_name[0];
	tmp_hba_port.Id[0] = 0;
	tmp_hba_port.Id[1] = ha->d_id.r.d_id[2];
	tmp_hba_port.Id[2] = ha->d_id.r.d_id[1];
	tmp_hba_port.Id[3] = ha->d_id.r.d_id[0];
	tmp_hba_port.Type =  EXT_DEF_INITIATOR_DEV;

	switch (ha->current_topology) {
	case ISP_CFG_NL:
	case ISP_CFG_FL:
		tmp_hba_port.Mode = EXT_DEF_LOOP_MODE;
		break;
	case ISP_CFG_N:
	case ISP_CFG_F:
		tmp_hba_port.Mode = EXT_DEF_P2P_MODE;
		break;
	default:
		tmp_hba_port.Mode = EXT_DEF_UNKNOWN_MODE;
		break;
	}

	port_cnt = 0;
	for (fcport = ha->fcport; (fcport); fcport = fcport->next) {
		/* if removed or missing */
		if (fcport->state != FC_ONLINE)
			continue;
	    	port_cnt++;
	}

	tgt_cnt  = 0;
	for (tgt = 0; tgt < MAX_TARGETS; tgt++) {
		if (ha->otgt[tgt] == NULL) {
			continue;
		}
		tgt_cnt++;
	}
	tmp_hba_port.DiscPortCount   = port_cnt;
	tmp_hba_port.DiscTargetCount = tgt_cnt;

	if (ha->loop_down_timer == 0 && ha->loop_state == LOOP_DOWN) {
		tmp_hba_port.State = EXT_DEF_HBA_LOOP_DOWN;
	} else {
		tmp_hba_port.State = EXT_DEF_HBA_OK;
	}

	tmp_hba_port.DiscPortNameType = EXT_DEF_USE_PORT_NAME;

	/* Return supported FC4 type depending on driver support. */
	tmp_hba_port.PortSupportedFC4Types = EXT_DEF_FC4_TYPE_SCSI;
#ifdef FC_IP_SUPPORT
	tmp_hba_port.PortSupportedFC4Types |= EXT_DEF_FC4_TYPE_IP;
#endif
#ifdef FC_SCTP_SUPPORT
	tmp_hba_port.PortSupportedFC4Types |= EXT_DEF_FC4_TYPE_SCTP;
#endif
#ifdef FC_VI_SUPPORT
	tmp_hba_port.PortSupportedFC4Types |= EXT_DEF_FC4_TYPE_VI;
#endif

	tmp_hba_port.PortActiveFC4Types = ha->active_fc4_types;

	/* Return supported speed depending on adapter type */
#ifdef ISP2100
	tmp_hba_port.PortSupportedSpeed = EXT_DEF_PORTSPEED_1GBIT;
#elif defined(ISP2200)
	tmp_hba_port.PortSupportedSpeed = EXT_DEF_PORTSPEED_1GBIT;
#elif defined(ISP2300)
	tmp_hba_port.PortSupportedSpeed = EXT_DEF_PORTSPEED_2GBIT;
#else
	/* invalid */
	tmp_hba_port.PortSupportedSpeed = 0;
#endif

	tmp_hba_port.PortSpeed = ha->current_speed;

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	      sizeof(EXT_HBA_PORT));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_query_hba_port: ERROR verify write "
		    "rsp buffer.\n");)
		return(ret);
	}

	/* now copy up the HBA_PORT to user */
	if (pext->ResponseLen < sizeof(EXT_HBA_PORT))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_HBA_PORT);

	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&tmp_hba_port,
	    transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("qla2x00_query_hba_port: exiting.\n");)
	return (ret);
}

STATIC int
qla2x00_query_disc_port(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EXT_STATUS_OK;
	uint32_t	tgt, transfer_size, inst = 0;
	fc_port_t	*fcport;
	os_tgt_t	*tq;
	EXT_DISC_PORT	tmp_disc_port;

	DEBUG9(printk("qla2x00_query_disc_port: entered.\n");)

	for (fcport = ha->fcport; fcport != NULL; fcport = fcport->next) {
		if (fcport->state != FC_ONLINE)
			continue;

		if (inst != pext->Instance) {
			inst++;
			continue;
		}
		break;
	}

	if (fcport == NULL) {
		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		return(ret);
	}

	BZERO(&tmp_disc_port, sizeof(EXT_DISC_PORT));

	BCOPY(&fcport->node_name[0], &tmp_disc_port.WWNN[0], 8);
	BCOPY(&fcport->port_name[0], &tmp_disc_port.WWPN[0], 8);

	tmp_disc_port.Id[0] = 0;
	tmp_disc_port.Id[1] = fcport->d_id.r.d_id[2];
	tmp_disc_port.Id[2] = fcport->d_id.r.d_id[1];
	tmp_disc_port.Id[3] = fcport->d_id.r.d_id[0];

	/* Currently all devices on fcport list are target capable devices */
	/* This default value may need to be changed after we add non target
	 * devices also to this list.
	 */
	tmp_disc_port.Type = EXT_DEF_TARGET_DEV;

	if (fcport->flags & FC_FABRIC_DEVICE) {
		tmp_disc_port.Type |= EXT_DEF_FABRIC_DEV;
	}
	if (fcport->flags & FC_TAPE_DEVICE) {
		tmp_disc_port.Type |= EXT_DEF_TAPE_DEV;
	}
	if (fcport->flags & FC_INITIATOR_DEVICE) {
		tmp_disc_port.Type |= EXT_DEF_INITIATOR_DEV;
	}

	tmp_disc_port.LoopID = fcport->loop_id;
	tmp_disc_port.Status = 0;
	tmp_disc_port.Bus    = 0;

	for (tgt = 0; tgt < MAX_TARGETS; tgt++) {
		if ((tq = ha->otgt[tgt]) == NULL) {
			continue;
		}

		if (tq->vis_port == NULL)  /* dg 08/14/01 */
			continue;

		if (BCMP(&fcport->port_name[0], &tq->vis_port->port_name[0],
		    EXT_DEF_WWN_NAME_SIZE) == 0) {
			tmp_disc_port.TargetId = tgt;
			break;
		}
	}

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	        sizeof(EXT_DISC_PORT));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_query_disc_port: ERROR verify write "
		    "rsp buffer.\n");)
		return(ret);
	}

	/* now copy up the DISC_PORT to user */
	if (pext->ResponseLen < sizeof(EXT_DISC_PORT))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_DISC_PORT);

	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&tmp_disc_port,
	    transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;


	DEBUG9(printk("qla2x00_query_disc_port: exiting.\n");)
	return (ret);
}

STATIC int
qla2x00_query_disc_tgt(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EXT_STATUS_OK;
	uint32_t	tgt, transfer_size, inst;
	uint32_t	cnt, i;
	fc_port_t	*tgt_fcport;
	os_tgt_t	*tq;
	EXT_DISC_TARGET	tmp_disc_target;

	DEBUG9(printk("qla2x00_query_disc_tgt: entered.\n");)

	tq = NULL;
	for (tgt = 0, inst = 0; tgt < MAX_TARGETS; tgt++) {
		if (ha->otgt[tgt] == NULL) {
			continue;
		}
		/* if wrong target id then skip to next entry */
		if (inst != pext->Instance) {
			inst++;
			continue;
		}
		tq = ha->otgt[tgt];
		break;
        }

	if (tq == NULL || tgt == MAX_TARGETS) {
		DEBUG9_10(printk("qla2x00_query_disc_tgt: target dev not "
		    "found. tq=%x, tgt=%x.\n",
		    (unsigned int)tq, tgt);)
		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		return(pext->Status);
	}

	if (tq->vis_port == NULL) { 	/* dg 08/14/01 */
		DEBUG9_10(printk("qla2x00_query_disc_tgt: target dev not "
		    "found. tq=%x, tgt=%x.\n",
		    (unsigned int)tq, tgt);)
		pext->Status = EXT_STATUS_BUSY;
		return(pext->Status);
	}

	BZERO(&tmp_disc_target, sizeof(EXT_DISC_TARGET));

	tgt_fcport = tq->vis_port;
	BCOPY (&tgt_fcport->node_name[0], &tmp_disc_target.WWNN[0], 8);
	BCOPY (&tgt_fcport->port_name[0], &tmp_disc_target.WWPN[0], 8);

	tmp_disc_target.Id[0] = 0;
	tmp_disc_target.Id[1] = tgt_fcport->d_id.r.d_id[2];
	tmp_disc_target.Id[2] = tgt_fcport->d_id.r.d_id[1];
	tmp_disc_target.Id[3] = tgt_fcport->d_id.r.d_id[0];

	/* All devices on ha->otgt list are target capable devices. */
	tmp_disc_target.Type = EXT_DEF_TARGET_DEV;

	if (tgt_fcport->flags & FC_FABRIC_DEVICE) {
		tmp_disc_target.Type |= EXT_DEF_FABRIC_DEV;
	}
	if (tgt_fcport->flags & FC_TAPE_DEVICE) {
		tmp_disc_target.Type |= EXT_DEF_TAPE_DEV;
	}
	if (tgt_fcport->flags & FC_INITIATOR_DEVICE) {
		tmp_disc_target.Type |= EXT_DEF_INITIATOR_DEV;
	}

	tmp_disc_target.LoopID   = tgt_fcport->loop_id;
	tmp_disc_target.Status   = 0;
	tmp_disc_target.Bus      = 0;
	tmp_disc_target.TargetId = tgt;

	cnt = 0;
	/* enumerate available LUNs under this TGT (if any) */
	if (ha->otgt[tgt] != NULL) {
		for (i = 0; i < MAX_LUNS ; i++) {
			if ((ha->otgt[tgt])->olun[i] !=0)
				cnt++;
		}
	}

	tmp_disc_target.LunCount = cnt;

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
	    sizeof(EXT_DISC_TARGET));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_query_disc_tgt: ERROR verify write "
		    "rsp buffer.\n");)
		return(pext->Status);
	}

	/* now copy up the DISC_PORT to user */
	if (pext->ResponseLen < sizeof(EXT_DISC_PORT))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_DISC_TARGET);

	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&tmp_disc_target,
	    transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("qla2x00_query_disc_tgt: exiting.\n");)
	return (ret);
}

STATIC int
qla2x00_query_chip(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EXT_STATUS_OK;
	uint32_t	transfer_size, i;
	EXT_CHIP		tmp_isp;
	struct Scsi_Host	*host;

	DEBUG9(printk("qla2x00_query_chip: entered.\n");)

	BZERO(&tmp_isp, sizeof(EXT_CHIP));

	host = ha->host;
	tmp_isp.VendorId       = QLA2X00_VENDOR_ID;
	tmp_isp.DeviceId       = ha->device_id;
	tmp_isp.SubVendorId    = ha->subsystem_vendor;
        tmp_isp.SubSystemId    = ha->subsystem_device;
	tmp_isp.PciBusNumber   = ha->pci_bus;
	tmp_isp.PciDevFunc     = ha->pci_device_fn;
	tmp_isp.PciSlotNumber  = (ha->pci_device_fn & 0xf8) >> 3;
	tmp_isp.IoAddr         = host->io_port;
	tmp_isp.IoAddrLen      = 512;
	tmp_isp.MemAddr        = 0; /* ? */
	tmp_isp.MemAddrLen     = 0; /* ? */
	tmp_isp.ChipType       = 0; /* ? */
	tmp_isp.InterruptLevel = host->irq;

	for (i = 0; i < 8; i++)
		tmp_isp.OutMbx[i] = 0;

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	    sizeof(EXT_CHIP));
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_query_chip: ERROR verify write rsp "
		    "buffer.\n");)
		return(pext->Status);
	}

	/* now copy up the ISP to user */
	if (pext->ResponseLen < sizeof(EXT_CHIP))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_CHIP);

	copy_to_user((uint8_t *)pext->ResponseAdr, (uint8_t *)&tmp_isp,
	    transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	DEBUG9(printk("qla2x00_query_chip: exiting.\n");)
	return (ret);
}

STATIC int
qla2x00_get_data(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int	tmp_rval = 0;

	switch(pext->SubCode) {
	case EXT_SC_GET_STATISTICS:
		tmp_rval = qla2x00_get_statistics(ha, pext, mode);
		break;

	case EXT_SC_GET_FC_STATISTICS:
		tmp_rval = qla2x00_get_fc_statistics(ha, pext, mode);
		break;

	case EXT_SC_GET_PORT_SUMMARY:
		tmp_rval = qla2x00_get_port_summary(ha, pext, mode);
		break;

	case EXT_SC_QUERY_DRIVER:
		tmp_rval = qla2x00_query_driver(ha, pext, mode);
		break;

	case EXT_SC_QUERY_FW:
		tmp_rval = qla2x00_query_fw(ha, pext, mode);
		break;

        default:
		DEBUG10(printk("qla2x00_get_data: unknown SubCode %d.\n",
		    pext->SubCode);)
		pext->Status = EXT_STATUS_UNSUPPORTED_SUBCODE;
		tmp_rval = pext->Status;
		break;
	}

	return (tmp_rval);
}

STATIC int
qla2x00_get_port_summary(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EXT_STATUS_OK;
	uint8_t		*usr_temp, *kernel_tmp;
	uint32_t	b;
	uint32_t	i, port_cnt, entry;
	uint32_t	number_of_entries = 0;
	uint32_t	tgt_cnt, transfer_size;
	fc_port_t	*fcport;
	void		*start_of_entry_list, *current_offset;

	EXT_DEVICEDATA		devicedata;
	EXT_DEVICEDATAENTRY	dd_entry;

	port_cnt = 0;
	tgt_cnt  = 0;
	BZERO(&devicedata, sizeof(EXT_DEVICEDATA));

	for (fcport = ha->fcport; fcport != NULL; fcport = fcport->next) {
		port_cnt++;
	}
	devicedata.TotalDevices = port_cnt;

	number_of_entries = pext->ResponseLen / sizeof(EXT_DEVICEDATAENTRY);

	/* we want the lesser of port_cnt and number_of_entries */
	if (number_of_entries > port_cnt)
		number_of_entries = port_cnt;
	devicedata.ReturnListEntryCount  = number_of_entries;

	DEBUG9(printk("qla2x00_ioctl: EXT_SC_GET_PORT_SUMMARY port_cnt=%x, "
	    "return entry cnt=%x.\n", port_cnt, number_of_entries);)

	transfer_size = sizeof(devicedata.ReturnListEntryCount) +
	    sizeof(devicedata.TotalDevices);

	/* copy top of devicedata here */
	ret = verify_area(VERIFY_WRITE, (void *)(pext->ResponseAdr),
	    transfer_size);
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG10(printk("qla2x00_ioctl: ERROR verify_area WRITE ha=%8x\n",
		    (uint32_t)ha);)
		return(pext->Status);
	}

	for (i=0 ; i < transfer_size ; i++) {
		usr_temp   = (uint8_t *)pext->ResponseAdr + i;
		kernel_tmp = (uint8_t *)&devicedata + i;
		__put_user(*kernel_tmp, usr_temp);
	}

	start_of_entry_list = (void *)(pext->ResponseAdr) + transfer_size;

	for (entry = 0, fcport = ha->fcport; (entry < number_of_entries) &&
	    (fcport); entry++, fcport = fcport->next) {

		/* copy from fc_db of this target (port) to dd_entry */

		BCOPY(&fcport->node_name[0], &dd_entry.NodeWWN[0], 8);
		BCOPY(&fcport->port_name[0], &dd_entry.PortWWN[0], 8);

		for (b = 0; b < 3 ; b++)
			dd_entry.PortID[b] = fcport->d_id.r.d_id[2-b];

		if (fcport->flags & FC_FABRIC_DEVICE) {
			dd_entry.ControlFlags = EXT_DEF_GET_FABRIC_DEVICE;
		} else {
			dd_entry.ControlFlags = 0;
		}

		dd_entry.TargetAddress.Bus    = 0;
		dd_entry.TargetAddress.Target = 0;
		/* Retrieve 'Target' number for port via fc_db */
		for (i = 0; i < MAX_TARGETS; i++) {
			if (ha->fc_db[i].loop_id == PORT_UNUSED)
				continue;

			if (BCMP(&fcport->port_name[0], &ha->fc_db[i].wwn[0],
			    EXT_DEF_WWN_NAME_SIZE) == 0) {
				dd_entry.TargetAddress.Target = i;
				break;
			}
		}
		dd_entry.TargetAddress.Lun    = 0;
		dd_entry.DeviceFlags          = 0;
		dd_entry.LoopID               = fcport->loop_id;
		dd_entry.BaseLunNumber        = 0;

		current_offset = (void *)(entry * sizeof(EXT_DEVICEDATAENTRY));

		transfer_size = sizeof(EXT_DEVICEDATAENTRY);
		ret = verify_area(VERIFY_WRITE,
		    (void *)((start_of_entry_list + (u_long)current_offset)),
		    transfer_size);

		if (ret) {
			pext->Status = EXT_STATUS_COPY_ERR;
			DEBUG10(printk("qla2x00_ioctl: ERROR verify_area WRITE "
			    "ha=%8x\n",(uint32_t)ha);)
		 	return(pext->Status);
		}

		/* now copy up this dd_entry to user */
		for (i=0 ; i < transfer_size ; i++) {
			usr_temp = (uint8_t *)((u_long) start_of_entry_list +
			    (u_long)current_offset + i);
			kernel_tmp = (uint8_t *)&dd_entry + i;
			__put_user(*kernel_tmp, usr_temp);
		}
	} /* for number_of_entries */

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;
	/* printk("[finished QLA2100 EXT_SC_GET_PORT_SUMMARY ]\n"); */

	return (pext->Status);
}

STATIC int
qla2x00_query_driver(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EINVAL;
	uint8_t		*usr_temp, *kernel_tmp;
	uint32_t	transfer_size;
	EXT_DRIVER	driver_prop;

	BZERO(&driver_prop, sizeof(EXT_DRIVER));

	/* printk("[started EXT_SC_QUERY_DRIVER]\n"); */
	sprintf(driver_prop.Version, QLA2100_VERSION);
	driver_prop.NumOfBus = MAX_BUSES;
	driver_prop.TargetsPerBus = MAX_FIBRE_DEVICES;
	driver_prop.LunsPerTarget = MAX_LUNS;
	driver_prop.MaxTransferLen  = 0xffffffff;
	driver_prop.MaxDataSegments = 0xffffffff;

	if (ha->flags.enable_64bit_addressing == 1)
 		driver_prop.DmaBitAddresses = 64;
	else
		driver_prop.DmaBitAddresses = 32;

	if (pext->ResponseLen < sizeof(EXT_DRIVER))
		transfer_size = pext->ResponseLen;
	else
		transfer_size = sizeof(EXT_DRIVER);

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	    transfer_size);
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG10(printk("qla2x00_ioctl: ERROR verify_area WRITE ha=%8x\n",
		    (uint32_t)ha);)
		return(pext->Status);
	}

	/* now copy up the ISP to user */
	usr_temp   = (uint8_t *)pext->ResponseAdr;
	kernel_tmp = (uint8_t *)&driver_prop;
	copy_to_user(usr_temp, kernel_tmp, transfer_size);

	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	return (pext->Status);
}

STATIC int
qla2x00_query_fw(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = EINVAL;
	qla_boards_t	*bdp;
	uint8_t		*usr_temp, *kernel_tmp;
	uint32_t	transfer_size;
	EXT_FW		fw_prop;

	BZERO(&fw_prop, sizeof(EXT_FW));

	/* printk("[started EXT_SC_QUERY_FW]\n"); */
	bdp = &QLBoardTbl_fc[ha->devnum];
	fw_prop.Version[0] = bdp->fwver[0];
	fw_prop.Version[1] = bdp->fwver[1];
	fw_prop.Version[2] = bdp->fwver[2];

	transfer_size = sizeof(EXT_FW);

	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	    transfer_size);
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG10(printk("qla2x00_ioctl: ERROR verify_area WRITE ha=%8x\n",
		    (uint32_t)ha);)
		return(pext->Status);
	}

	usr_temp   = (uint8_t *)pext->ResponseAdr;
	kernel_tmp = (uint8_t *)&fw_prop;
	copy_to_user(usr_temp, kernel_tmp, transfer_size);

	/*  printk("[finished EXT_SC_QUERY_FW]\n"); */
	pext->Status       = EXT_STATUS_OK;
	pext->DetailStatus = EXT_STATUS_OK;

	return(pext->Status);
}

STATIC int
qla2x00_send_fcct(adapter_state_t *ha, EXT_IOCTL *pext, int mode)
{
	int		ret = 0;
	uint16_t	mb[MAILBOX_REGISTER_COUNT];

	Scsi_Cmnd	scsi_cmd;
	Scsi_Cmnd	*pscsi_cmd = &scsi_cmd;
	srb_t		ioctl_sp;
	srb_t		*sp = &ioctl_sp;

	DEBUG9(printk("qla2x00_send_fcct(%ld): entered.\n", ha->host_no);)

	/* check on current topology */
	if ((ha->current_topology != ISP_CFG_F) &&
	    (ha->current_topology != ISP_CFG_FL)) {
		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		DEBUG9_10(printk(
		    "qla2x00_send_fcct(%ld): ERROR not in F or FL mode.\n",
		    ha->host_no);)
		return(ret);
	}

	if (pext->ResponseLen > PAGE_SIZE) {
		pext->ResponseLen = PAGE_SIZE;
		DEBUG10(printk("qla2x00_send_fcct(%ld): rsp buf length larger "
		    "than current limitation. Truncating response length to "
		    "%ldB.\n", ha->host_no, PAGE_SIZE);)
	}

	if (pext->RequestLen > PAGE_SIZE) {
		pext->Status = EXT_STATUS_INVALID_PARAM;
		DEBUG9_10(printk(
		    "qla2x00_send_fcct(%ld): ERROR RequestLen too big=%x.\n",
		    ha->host_no, pext->RequestLen);)
		return(pext->Status);
	}

	ret = verify_area(VERIFY_READ, (void *)pext->RequestAdr,
	    pext->RequestLen);
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk(
		    "qla2x00_send_fcct(%ld): ERROR verify read request buf.\n",
		    ha->host_no);)
		return(ret);
	}

	/* clear ioctl_sp and scsi_cmd to be used */
	BZERO(sp, sizeof(srb_t));
	BZERO(&scsi_cmd, sizeof(Scsi_Cmnd));
	BZERO(ha->ioctl_mem, PAGE_SIZE);

	/* copy request buffer */
	copy_from_user((uint8_t *)ha->ioctl_mem, (uint8_t *)pext->RequestAdr,
	    pext->RequestLen);

	/* setup sp for this FCCT pass thru */
	sp->cmd = pscsi_cmd;
	sp->flags = SRB_WATCHDOG;

	/* init scsi_cmd */
	scsi_cmd.host = ha->host;
	scsi_cmd.scsi_done = qla2x00_fcct_done;

	/* check on management server login status */
	if (ha->flags.management_server_logged_in == 0) {
		/* login to management server device */

		ret = qla2x00_login_fabric(ha, MANAGEMENT_SERVER, 0xff, 0xff,
		    0xfa, &mb[0], 0);

		if (ret != 0 || mb[0] != 0x4000) {
			pext->Status = EXT_STATUS_DEV_NOT_FOUND;

			DEBUG10(printk(
			    "qla2x00_send_fcct(%ld): ERROR login to MS.\n",
			    ha->host_no);)
			return(pext->Status);
		}

		ha->flags.management_server_logged_in = 1;
	}

	DEBUG9(printk("qla2x00_send_fcct(%ld): success login to MS.\n",
	    ha->host_no);)

	/* check on loop down */
	if (ha->loop_state != LOOP_READY)  {
		pext->Status = EXT_STATUS_BUSY;
		DEBUG10(printk("qla2x00_send_fcct(%ld): loop not ready.\n",
		    ha->host_no);)
		return(pext->Status);
	}

	DEBUG9(printk("qla2x00_send_fcct(%ld): going to issue command.\n",
	    ha->host_no);)

	ret = qla2x00_start_ms_cmd(ha, pext, sp);

	if (ret != 0) {
		/* We waited and post function did not get called */
		DEBUG10(printk("qla2x00_send_fcct(%ld): command timed out.\n",
		    ha->host_no);)
		pext->Status = EXT_STATUS_MS_NO_RESPONSE;
		return(pext->Status);
	}

	/* getting device data and putting in pext->ResponseAdr */
	ret = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr ,
	    pext->ResponseLen);
	if (ret) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_send_fcct(%ld): ERROR verify write "
		    "response buffer. ha=%x.\n", ha->host_no, (uint32_t)ha);)
		return(ret);
	}

	/* sending back data returned from Management Server */
	copy_to_user((uint8_t *)pext->ResponseAdr,
	    (uint8_t *)ha->ioctl_mem, pext->ResponseLen);

	DEBUG9(printk("qla2x00_send_fcct(%ld): exiting normally.\n",
	    ha->host_no);)

	return 0;
}

STATIC int
qla2x00_start_ms_cmd(adapter_state_t *ha, EXT_IOCTL *pext, srb_t *sp)
{
	ms_iocb_entry_t	*pkt;

	/* get spin lock for this operation */
	spin_lock_irqsave(&io_request_lock, ha->cpu_flags);

	/* Get MS request packet. */
	pkt = (ms_iocb_entry_t *)qla2100_ms_req_pkt(ha, sp);
	if (pkt == NULL) {
		/* release spin lock and return error. */
		spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

		pext->Status = EXT_STATUS_NO_MEMORY;
		DEBUG10(printk("qla2x00_ioctl: FCCT_PASSTHRU - could not get "
		    "Request Packet.\n");)
		return(pext->Status);
	}

	pkt->entry_type  = MS_IOCB_TYPE;
	pkt->entry_count = 1;
	pkt->loop_id     = MANAGEMENT_SERVER;
	pkt->timeout     = 10;
	pkt->DSDcount    = 1;
	pkt->RespDSDcount = 2;
	pkt->Response_bytecount = pext->ResponseLen;
	pkt->Request_bytecount  = pext->RequestLen;

	/* loading command payload address */
	pkt->dseg_req_address[0] = LS_64BITS(ha->ioctl_mem_phys);
	pkt->dseg_req_address[1] = MS_64BITS(ha->ioctl_mem_phys);
	pkt->dseg_req_length = pext->RequestLen;

	/* loading command response address */
	pkt->dseg_rsp_address[0] = LS_64BITS(ha->ioctl_mem_phys);
	pkt->dseg_rsp_address[1] = MS_64BITS(ha->ioctl_mem_phys);
	pkt->dseg_rsp_length = pext->ResponseLen;

	/* set flag to indicate IOCTL FCCT PassThru in progress */
	ha->ioctl->FCCT_InProgress = 1;
	ha->ioctl->ioctl_tov = pkt->timeout + 1; /* 1 second more */

	/* Issue command to ISP */
	qla2100_isp_cmd(ha);

	/* prepare for receiving completion. */
	qla2x00_ioctl_sem_init(ha);

	/* release io_request_lock and wait for completion */
	qla2x00_ioctl_sleep (ha, ha->ioctl->ioctl_tov * HZ, TRUE);

        if (ha->ioctl->FCCT_InProgress == 1) {
		DEBUG9_10(printk(
		    "qla2x00_start_ms_cmd: timed out. exiting.\n");)
		return 1;
	} else {
		DEBUG9(printk("qla2x00_start_ms_cmd: done. exiting.\n");)
		return 0;
	}
}

STATIC int
qla2x00_wwpn_to_scsiaddr(scsi_qla_host_t *ha, EXT_IOCTL *pext, int mode)
{
	int		rval;
	fc_port_t	*tgt_fcport;
	os_tgt_t	*tq;
	uint8_t		tmp_wwpn[EXT_DEF_WWN_NAME_SIZE];
	uint32_t	b, tgt, l;
	EXT_SCSI_ADDR	tmp_addr;


	DEBUG9(printk("qla2x00_wwpn_to_scsiaddr(%ld): entered.\n",
	    ha->host_no);)

	if (pext->RequestLen != EXT_DEF_WWN_NAME_SIZE ||
	    pext->ResponseLen < sizeof(EXT_SCSI_ADDR)) {
		/* error */
		DEBUG9_10(printk("qla2x00_wwpn_to_scsiaddr(%ld): invalid WWN "
		    "buffer size %d received.\n",
		    ha->host_no, pext->ResponseLen);)
		pext->Status = EXT_STATUS_INVALID_PARAM;
		return (pext->Status);
	}

	rval = verify_area(VERIFY_READ, (void *)pext->RequestAdr,
	    pext->RequestLen);
	if (rval) {
		DEBUG9_10(printk("qla2x00_wwpn_to_scsiaddr(%ld): ERROR "
		    "VERIFY_READ request buf.\n", ha->host_no);)
		pext->Status = EXT_STATUS_COPY_ERR;
		return(pext->Status);
	}

	copy_from_user(&tmp_wwpn[0], (uint8_t *)pext->RequestAdr,
	    pext->RequestLen);

	tq = NULL;
	for (tgt = 0; tgt < MAX_TARGETS; tgt++) {
		if (ha->otgt[tgt] == NULL) {
			continue;
		}

		tq = ha->otgt[tgt];
		if (tq->vis_port == NULL) {
			break;
		}

		tgt_fcport = tq->vis_port;
		if (BCMP(tmp_wwpn, tgt_fcport->port_name,
		    EXT_DEF_WWN_NAME_SIZE) == 0) {
			break;
		}
        }

	if (tq == NULL || tgt >= MAX_TARGETS) {
		DEBUG9_10(printk("qla2x00_query_disc_tgt: target dev not "
		    "found. tq=%x, tgt=%x.\n",
		    (unsigned int)tq, tgt);)
		pext->Status = EXT_STATUS_DEV_NOT_FOUND;
		return(pext->Status);
	}

	if (tq->vis_port == NULL) { 	/* dg 08/14/01 */
		DEBUG9_10(printk("qla2x00_query_disc_tgt: target port not "
		    "found. tq=%x, tgt=%x.\n",
		    (unsigned int)tq, tgt);)
		pext->Status = EXT_STATUS_BUSY;
		return(pext->Status);
	}

	/* Currently we only have bus 0 and no translation on LUN */
	b = 0;
	l = 0;

	/*
	 * Return SCSI address. Currently no translation is done for
	 * LUN.
	 */
	tmp_addr.Bus = b;
	tmp_addr.Target = tgt;
	tmp_addr.Lun = l;
	if (pext->ResponseLen > sizeof(EXT_SCSI_ADDR))
		pext->ResponseLen = sizeof(EXT_SCSI_ADDR);

	rval = verify_area(VERIFY_WRITE, (void *)pext->ResponseAdr,
	    pext->ResponseLen);
	if (rval) {
		pext->Status = EXT_STATUS_COPY_ERR;
		DEBUG9_10(printk("qla2x00_wwpn_to_scsiaddr(%ld): ERROR "
		    "VERIFY_WRITE response buf.\n", ha->host_no);)
		return(pext->Status);
	}

	copy_to_user((uint8_t *)pext->ResponseAdr, &tmp_addr,
	    pext->ResponseLen);

	DEBUG9(printk(KERN_INFO
	    "qla2x00_wwpn_to_scsiaddr: Found t%d l%d for %02x%02x%02x%02x"
	    "%02x%02x%02x%02x.\n",
	    tmp_addr.Target, tmp_addr.Lun,
	    tmp_wwpn[0], tmp_wwpn[1], tmp_wwpn[2], tmp_wwpn[3],
	    tmp_wwpn[4], tmp_wwpn[5], tmp_wwpn[6], tmp_wwpn[7]);)

	pext->Status = EXT_STATUS_OK;

	DEBUG9(printk("qla2x00_wwpn_to_scsiaddr(%ld): exiting.\n",
	    ha->host_no);)

	return (pext->Status);
}

STATIC void
qla2x00_waitq_sem_timeout(unsigned long data)
{
	wait_q_t *tmp_ptr = (wait_q_t *)data;

	DEBUG9(printk("qla2x00_sem_timeout: entered.\n");)

	if (tmp_ptr != NULL)
		up(&tmp_ptr->wait_q_sem);

	DEBUG9(printk("qla2x00_sem_timeout: exiting.\n");)
}

/*
 *  tov = timeout value in seconds
 */
STATIC uint8_t
qla2x00_get_ioctl_access(adapter_state_t *ha, uint32_t tov)
{
	int		prev_val = 1;
	uint8_t		ret;
	unsigned long	cpu_flags;
	struct timer_list	tmp_access_timer;
	wait_q_t	*ptmp_wq = NULL;


	DEBUG9(printk("qla2x00_get_ioctl_access(%ld): entered.\n",
	    ha->host_no);)

	while (1) {
		if (test_bit(IOCTL_WANT, (void *)&(ha->ioctl->access_bits)) ==
		    0) {

			DEBUG9(printk("qla2x00_get_ioctl_access(%ld): going "
			    " to test access_bits.\n", ha->host_no);)

			/* No one else is waiting. Go ahead and try to
			 * get access.
			 */
			if ((prev_val = test_and_set_bit(IOCTL_ACTIVE,
			    (void *)&ha->ioctl->access_bits)) == 0) {
				break;
			}
		}

		/* wait for previous command to finish */
		DEBUG9(printk("qla2x00_get_ioctl_access(%ld): access_bits=%x. "
		    "busy. Waiting for access. curr time=0x%lx.\n",
		    ha->host_no, ha->ioctl->access_bits, jiffies);)
 
		/*
		 * Init timer and get semaphore from wait_q. if we got valid
		 * semaphore pointer the IOCTL_WANT flag would also had
		 * been set.
		 */
		qla2x00_wait_q_add(ha, &ptmp_wq);

		if (ptmp_wq == NULL) {
			/* queue full? problem? can't proceed. */
			DEBUG9_10(printk("qla2x00_get_ioctl_access(%ld): ERROR "
			    "no more wait_q allowed. exiting.\n", ha->host_no);)

			break;
		}

		init_timer(&tmp_access_timer);

		tmp_access_timer.data = (unsigned long)ptmp_wq;
		tmp_access_timer.function =
		    (void (*)(unsigned long))qla2x00_waitq_sem_timeout;
		tmp_access_timer.expires = jiffies + tov * HZ;

		DEBUG9(printk("get_ioctl_access(%ld): adding timer. "
		    "curr time=0x%lx timeoutval=0x%lx.\n",
		    ha->host_no, jiffies, tmp_access_timer.expires);)

		/* wait. */
		add_timer(&tmp_access_timer);

		DEBUG9(printk("get_ioctl_access(%ld): going to sleep. current "
		    "time=0x%lx.\n", ha->host_no, jiffies);)

		down_interruptible(&ptmp_wq->wait_q_sem);

		DEBUG9(printk("get_ioctl_access(%ld): woke up. current "
		    "time=0x%lx.\n", ha->host_no, jiffies);)

		del_timer(&tmp_access_timer);

		/* try to get lock again. we'll test later to see
		 * if we actually got the lock.
		 */
		prev_val = test_and_set_bit(IOCTL_ACTIVE,
		    (void *)&(ha->ioctl->access_bits));

		/*
		 * After we tried to get access then we check to see
		 * if we need to clear the IOCTL_WANT flag. Don't clear
		 * this flag before trying to get access or another
		 * new thread might grab it before we did.
		 */
		spin_lock_irqsave(&ha->ioctl->wait_q_lock, cpu_flags);
		if (ha->ioctl->wait_q_head == NULL) {
			/* We're the last thread in queue. */
			clear_bit(IOCTL_WANT, (void *)&ha->ioctl->access_bits);
		}
		qla2x00_wait_q_memb_free(ha, ptmp_wq);
		spin_unlock_irqrestore(&ha->ioctl->wait_q_lock, cpu_flags);

		break;
	}

	if (prev_val == 0) {
		/* We got the lock */

		DEBUG9(printk("qla2x00_get_ioctl_access(%ld): got access.\n",
		    ha->host_no);)

		ret = QL_STATUS_SUCCESS;
	} else {
		/* Timeout or resource error. */
		DEBUG9_10(printk("qla2x00_get_ioctl_access(%ld): timed out "
		    "or wait_q error.\n", ha->host_no);)

		ret = QL_STATUS_TIMEOUT;
	}

	return (ret);
}

STATIC uint8_t
qla2x00_release_ioctl_access(adapter_state_t *ha)
{
	wait_q_t	*next_thread = NULL;

	DEBUG9(printk("qla2x00_release_ioctl_access:(%ld): entered.\n",
	    ha->host_no);)

	clear_bit(IOCTL_ACTIVE, (void *)&(ha->ioctl->access_bits));

	/* Wake up one pending ioctl thread in wait_q */
	qla2x00_wait_q_remove(ha, &next_thread);
	if (next_thread) {
		DEBUG9(printk("qla2x00_release_ioctl_access: found wait_q. "
		    "Waking up waitq %p.\n", &next_thread);)
		up(&next_thread->wait_q_sem);
	}

	DEBUG9(printk("qla2x00_release_ioctl_access:(%ld): exiting.\n",
	    ha->host_no);)

	return (QL_STATUS_SUCCESS);
}

/* Find a free wait_q member from the array. Must already got the
 * wait_q_lock spinlock.
 */
STATIC void
qla2x00_wait_q_memb_alloc(adapter_state_t *ha, wait_q_t **ret_wait_q_memb)
{
	uint8_t		i;
	wait_q_t	*ptmp = NULL;

	DEBUG9(printk("qla2x00_wait_q_memb_alloc: entered. "
	    "Inst=%d.\n", apiHBAInstance);)

	for (i = 0; i < MAX_IOCTL_WAIT_THREADS; i++) {
		if (!(ha->ioctl->wait_q_arr[i].flags & WQ_IN_USE)) {
			ha->ioctl->wait_q_arr[i].flags |= WQ_IN_USE;
			ptmp = &ha->ioctl->wait_q_arr[i];
			break;
		}
	}

	*ret_wait_q_memb = ptmp;

	DEBUG9(printk("qla2x00_wait_q_memb_alloc: return waitq_memb=%p. "
	    "Inst=%d.\n", *ret_wait_q_memb, apiHBAInstance);)
}

/* Free the specified wait_q member. Must already got the wait_q_lock
 * spinlock.
 */
STATIC void
qla2x00_wait_q_memb_free(adapter_state_t *ha, wait_q_t *pfree_wait_q_memb)
{
	DEBUG9(printk("qla2x00_wait_q_memb_free: entered. "
	    "Inst=%d.\n", apiHBAInstance);)

	if (pfree_wait_q_memb != NULL)
		pfree_wait_q_memb->flags &= ~WQ_IN_USE;

	DEBUG9(printk("qla2x00_wait_q_memb_free: exiting. "
	    "Inst=%d.\n", apiHBAInstance);)
}

/* Allocates a wait_q_t struct and add to the wait_q list. */
STATIC uint8_t
qla2x00_wait_q_add(adapter_state_t *ha, wait_q_t **ret_wq)
{
	uint8_t		ret;
	unsigned long	cpu_flags;
	wait_q_t	*ptmp = NULL;

	spin_lock_irqsave(&ha->ioctl->wait_q_lock, cpu_flags);

	DEBUG9(printk("qla2x00_wait_q_add: got wait_q spinlock. "
	    "Inst=%d.\n", apiHBAInstance);)

	qla2x00_wait_q_memb_alloc(ha, &ptmp);
	if (ptmp == NULL) {
		/* can't add any more threads */
		DEBUG9_10(printk("qla2x00_wait_q_add: ERROR no more "
		    "ioctl threads allowed. Inst=%d.\n", apiHBAInstance);)

		ret = QL_STATUS_RESOURCE_ERROR;
	} else {
		if (ha->ioctl->wait_q_tail == NULL) {
			/* First thread to queue. */
			set_bit(IOCTL_WANT, (void *)&ha->ioctl->access_bits);

			ha->ioctl->wait_q_head = ptmp;
		} else {
			ha->ioctl->wait_q_tail->pnext = ptmp;
		}
		ha->ioctl->wait_q_tail = ptmp;

		*ret_wq = ptmp;

		/* Now init the semaphore */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
		ptmp->wait_q_sem = MUTEX_LOCKED;
#else
		init_MUTEX_LOCKED(&ptmp->wait_q_sem);
#endif
		ret = QL_STATUS_SUCCESS;
	}

	DEBUG9(printk("qla2x00_wait_q_add: going to release spinlock. "
	    "ret_wq=%p, ret=%d. Inst=%d.\n", *ret_wq, ret, apiHBAInstance);)

	spin_unlock_irqrestore(&ha->ioctl->wait_q_lock, cpu_flags);

	return (ret);
}

/* Just remove one member from wait_q.  Don't free anything. */
STATIC void
qla2x00_wait_q_remove(adapter_state_t *ha, wait_q_t **ret_wq)
{
	unsigned long	cpu_flags;

	spin_lock_irqsave(&ha->ioctl->wait_q_lock, cpu_flags);

	DEBUG9(printk("qla2x00_wait_q_remove: got wait_q spinlock. "
	    "Inst=%d.\n", apiHBAInstance);)

	/* Remove from head */
	*ret_wq = ha->ioctl->wait_q_head;
	if (ha->ioctl->wait_q_head != NULL) {
		ha->ioctl->wait_q_head = ha->ioctl->wait_q_head->pnext;
		if (ha->ioctl->wait_q_head == NULL) {
			/* That's the last one in queue. */
			ha->ioctl->wait_q_tail = NULL;
		}
		(*ret_wq)->pnext = NULL;
	}

	DEBUG9(printk("qla2x00_wait_q_remove: return ret_wq=%p. Going to "
	    "release spinlock. Inst=%d.\n", *ret_wq, apiHBAInstance);)

	spin_unlock_irqrestore(&ha->ioctl->wait_q_lock, cpu_flags);
}


