/******************************************************************************
 *                  QLOGIC LINUX SOFTWARE
 *
 * QLogic ISP2x00 device driver for Linux 2.2.x and 2.4.x
 * Copyright (C) 2001 Qlogic Corporation
 * (www.qlogic.com)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 ******************************************************************************/

/****************************************************************************
              Please see revision.notes for revision history.
*****************************************************************************/


/*
 * Compile time Options:
 *     0 - Disable and 1 - Enable
 */
#define  LOOP_ID_FROM_ONE              0   /* loop ID start from 1 when P2P */
#define  QL2100_TARGET_MODE_SUPPORT    0   /* Target mode support */
#define  MEMORY_MAPPED_IO              0
#define  DEBUG_QLA2100_INTR            0
#define  USE_NVRAM_DEFAULTS	       0
#define  DEBUG_PRINT_NVRAM             0
#define  LOADING_RISC_ACTIVITY         0
#define  AUTO_ESCALATE_RESET           0   /* Automatically escalate resets */
#define  AUTO_ESCALATE_ABORT           0   /* Automatically escalate aborts */
#define  STOP_ON_ERROR                 0   /* Stop on aborts and resets  */
#define  STOP_ON_RESET                 0
#define  STOP_ON_ABORT                 0
#define  QLA2100_COMTRACE              0    /* One char tracing  */
#define  WATCH_THREADS_SIZ             0    /* watch size of pending queue */
#define  USE_PORTNAME                  1    /* option to use port names for targets */
#define  LUN_MASKING                   0
#define  USE_FLASH_DATABASE            0 /* Save persistent data to flash */
#define  QLA2100_PROFILE               1
#define  QLA_SCSI_VENDOR_DIR           0 /* Decode vendor specific opcodes for direction */
#define QLA2100_LIPTEST    	       0
#define REQ_TRACE    		       1

#undef   TRACECODE                       /* include tracing code in watchdog routines */
#define  CHECK_BINDING
#define  DUMP_INQ_DATA                 0  /* DEBUG_QLA2100 */

#define  DEBUG_QLA2100                 0  /* For Debug of qla2x00 */
#define  DEBUG_GET_FW_DUMP             0  /* also set DEBUG_QLA2100 and
                                             use COM1 and capture it */
#define  NO_LONG_DELAYS			0
#define  QL_TRACE_MEMORY		0

/* The following WORD_FW_LOAD is defined in Makefile for ia-64 builds
   and can also be decommented here for Word by Word confirmation of
   RISC code download operation */
/* #define  WORD_FW_LOAD               0  */

#define MPIO_SUPPORT			1  /* Multipath I/O */
#define VSA				1  /* Volume Set Addressing */

/*
 * Driver version
 */
#if DEBUG_QLA2100
#define QLA2100_VERSION      "5.38.1-debug"
#else
#define QLA2100_VERSION      "5.38.1"
#endif


/*************** 64 BIT PCI DMA ******************************************/
#define  FORCE_64BIT_PCI_DMA           0 /* set to one for testing only  */
/* Applicable to 64 version of the Linux 2.4.x and above only            */
/* NVRAM bit nv->cntr_flags_1.enable_64bit_addressing should be used for */
/* administrator control of PCI DMA width size per system configuration  */
/*************************************************************************/

/*
* String arrays
*/
#define LINESIZE    256
#define MAXARGS      26

/*
* Include files
*/
#include <linux/config.h>
#ifdef MODULE
#include <linux/module.h>
#endif

#ifndef LINUX_VERSION_CODE
#include <linux/version.h>
#endif  /* LINUX_VERSION_CODE not defined */

static int num_hosts;           /* ioctl related  */
static int apiHBAInstance = 0;  /* ioctl related keeps track of API HBA Instance */

#if QL_TRACE_MEMORY
static unsigned long mem_trace[1000];
static unsigned long mem_id[1000];
#endif


#include <stdarg.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/byteorder.h>
#include <asm/pgtable.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/blk.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>


#ifndef KERNEL_VERSION
#  define KERNEL_VERSION(x,y,z) (((x)<<16)+((y)<<8)+(z))
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
#define  APIDEV        1
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
#include <linux/tasks.h>
# include <linux/bios32.h>
#else
/*#include <asm/spinlock.h>*/

#define __KERNEL_SYSCALLS__

#include <linux/unistd.h>
#include <linux/smp_lock.h>

#include <asm/system.h>
/*
 * We must always allow SHUTDOWN_SIGS.  Even if we are not a module,
 * the host drivers that we are using may be loaded as modules, and
 * when we unload these,  we need to ensure that the error handler thread
 * can be shut down.
 *
 * Note - when we unload a module, we send a SIGHUP.  We mustn't
 * enable SIGTERM, as this is how the init shuts things down when you
 * go to single-user mode.  For that matter, init also sends SIGKILL,
 * so we mustn't enable that one either.  We use SIGHUP instead.  Other
 * options would be SIGPWR, I suppose.
 */
/*
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#define SHUTDOWN_SIGS	(sigmask(SIGHUP))
#else
#define SHUTDOWN_SIGS	(sigmask(SIGKILL)|sigmask(SIGINT)|sigmask(SIGTERM))
#endif
#define SHUTDOWN_SIGS	(sigmask(SIGKILL)|sigmask(SIGHUP))
*/
#define SHUTDOWN_SIGS	(sigmask(SIGHUP))
#endif
#include "sd.h"
#include "scsi.h"
#include "hosts.h"

#ifdef FC_IP_SUPPORT
#include <linux/skbuff.h>
#include "qlcommon.h"
#endif

#ifdef FC_SCTP_SUPPORT
#endif

#ifdef FC_VI_SUPPORT
#endif

#if BITS_PER_LONG > 32
#define	DMA_ADDR   uint64_t
#else
#define	DMA_ADDR   uint32_t
#endif


#include "exioct.h"
#include "qla2x00.h"
#define UNIQUE_FW_NAME                 /* unique F/W array names */
#ifdef ISP2200
#ifdef FC_IP_SUPPORT
#include "ql2200ip_fw.h"                   /* ISP RISC 2200 IP code */
#else
#include "ql2200_fw.h"                     /* ISP RISC 2200 TP code */
#endif
#endif
#ifdef ISP2300
#ifdef FC_IP_SUPPORT
#include "ql2300ip_fw.h"                   /* ISP RISC 2300 IP code */
#else
#include "ql2300_fw.h"                     /* ISP RISC 2300 TP code */
#endif
#endif
#include <linux/stat.h>
#include <linux/slab.h>

#include "qla_cfg.h"
#include "qla_gbl.h"
#include "qla_dbg.h"


#define  BZERO(ptr, amt)	memset((void *)(ptr), 0, amt)
#define  BCOPY(src, dst, amt)	memcpy((void *)(dst), (void *)(src), amt)
#define  KMALLOC_DMA(siz)		kmalloc((siz), GFP_ATOMIC | GFP_DMA)
#define  KMALLOC(siz)		kmalloc((siz), GFP_ATOMIC)
#define  KMFREE(ip,siz)		kfree((ip))
#if NO_LONG_DELAYS
#define  SYS_DELAY(x)		qla2x00_sleep(x)
#define  QLA2100_DELAY(sec)  qla2x00_sleep(sec * HZ)
#define NVRAM_DELAY() qla2x00_sleep(10) /* 10 microsecond delay */
#define  UDELAY(x)		qla2x00_sleep(x)
#else
#define  SYS_DELAY(x)		udelay(x);barrier()
#define  QLA2100_DELAY(sec)  mdelay(sec * HZ)
#define NVRAM_DELAY() udelay(10) /* 10 microsecond delay */
#define  UDELAY(x)		udelay(x)
#endif

/* 4.10 */
#if  BITS_PER_LONG > 32
#define pci_dma_lo32(a) (a & 0xffffffff)
#define pci_dma_hi32(a) ((a >> 32) & 0xffffffff)
#else
#define pci_dma_lo32(a) (a & 0xffffffff)
#define pci_dma_hi32(a) 0
#endif

#define  VIRT_TO_BUS(a) virt_to_bus((a))

#if  BITS_PER_LONG <= 32
#define  VIRT_TO_BUS_LOW(a) (uint32_t)virt_to_bus(((void *)a))
#define  VIRT_TO_BUS_HIGH(a) (uint32_t)(0x0)
#else
#define  VIRT_TO_BUS_LOW(a) (uint32_t)(0xffffffff & virt_to_bus((void *)(a)))
#define  VIRT_TO_BUS_HIGH(a) (uint32_t)(0xffffffff & (virt_to_bus((void *)(a))>>32))
#endif

#define  CACHE_FLUSH(a) (RD_REG_WORD(a))
#define  INVALID_HANDLE    (MAX_OUTSTANDING_COMMANDS+1)

#define  ABORTS_ACTIVE    ( (ha->dpc_flags & SRB_ABORT_NEEDED) || \
			(ha->dpc_flags & LOOP_RESET_NEEDED) || \
			(ha->dpc_flags & DEVICE_RESET_NEEDED) || \
			(ha->dpc_flags & DEVICE_ABORT_NEEDED) || \
			(ha->dpc_flags & ISP_ABORT_NEEDED))

#define  STATIC static

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
#define  OFFSET(w)   (((uint32_t) &w) & 0xFFFF)  /* 256 byte offsets */
#else
#define  OFFSET(w)   (((u_long) &w) & 0xFFFF)  /* 256 byte offsets */
#endif

#define  SCSI_BUS_32(scp)   ((scp)->channel)
#define  SCSI_TCN_32(scp)    ((scp)->target)
#define  SCSI_LUN_32(scp)    ((scp)->lun)

/*
 * LOCK MACROS
 */
#define QLA_RETRY_Q_LOCK(ap, cpu_flags)   \
    spin_lock_irqsave(&(ap)->retry_lock, cpu_flags);
#define QLA_RETRY_Q_UNLOCK(ap, cpu_flags) \
    spin_unlock_irqrestore(&(ap)->retry_lock, cpu_flags);
#define QLA_FAILBACK_LOCK(ap, cpu_flags)   \
    spin_lock_irqsave(&(ap)->failback_lock, cpu_flags);
#define QLA_FAILBACK_UNLOCK(ap, cpu_flags) \
    spin_unlock_irqrestore(&(ap)->failback_lock, cpu_flags);

#define QLA_DONE_Q_LOCK(ap, cpu_flags)   \
    spin_lock_irqsave(&(ap)->done_lock, cpu_flags);
#define QLA_DONE_Q_UNLOCK(ap, cpu_flags) \
    spin_unlock_irqrestore(&(ap)->done_lock, cpu_flags);

#define QLA_ABORT_Q_LOCK(ap, cpu_flags)   \
    spin_lock_irqsave(&(ap)->abort_q_lock, cpu_flags);
#define QLA_ABORT_Q_UNLOCK(ap, cpu_flags) \
    spin_unlock_irqrestore(&(ap)->abort_q_lock, cpu_flags);


#define QLA_MBX_REG_LOCK(ha)	\
    spin_lock_irqsave(&(ha)->mbx_reg_lock, mbx_flags);
#define QLA_MBX_REG_UNLOCK(ha)	\
    spin_unlock_irqrestore(&(ha)->mbx_reg_lock, mbx_flags);


#define  QLA2100_DPC_LOCK(ap)
#define  QLA2100_DPC_UNLOCK(ap)

#define	WATCH_INTERVAL		1       /* number of seconds */
#define	START_TIMER(f, h, w)	\
{ \
init_timer(&(h)->timer); \
(h)->timer.expires = jiffies + w * HZ;\
(h)->timer.data = (unsigned long) h; \
(h)->timer.function = (void (*)(unsigned long))f; \
(h)->flags.start_timer = FALSE; \
add_timer(&(h)->timer); \
(h)->timer_active = 1;\
}

#define	RESTART_TIMER(f, h, w)	\
{ \
(h)->timer.expires = jiffies + w * HZ;\
(h)->flags.start_timer = FALSE; \
add_timer(&(h)->timer); \
}

#define	STOP_TIMER(f, h)	\
{ \
del_timer(&(h)->timer); \
(h)->timer_active = 0;\
}

#ifdef ISP2200
#define DRIVER_NAME "qla2200"
#endif
#ifdef ISP2300
#define DRIVER_NAME "qla2300"
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,9)
MODULE_LICENSE("GPL");
#endif

typedef unsigned long paddr32_t;

/*
*  Qlogic Driver support Function Prototypes.
*/
STATIC uint8_t  qla2100_register_with_Linux(scsi_qla_host_t *ha, uint8_t maxchannels);
STATIC int   qla2100_done(scsi_qla_host_t *, srb_t **, srb_t **);
STATIC void qla2100_select_queue_depth(struct Scsi_Host *, Scsi_Device *);

STATIC void
qla2x00_abort_q_put(scsi_qla_host_t *, srb_t *, srb_t **, srb_t **),
qla2x00_abort_q_get(scsi_qla_host_t *, srb_t **);

STATIC void      qla2100_done_q_put(scsi_qla_host_t *, srb_t *, srb_t **, srb_t **),

qla2100_timer(scsi_qla_host_t *),
qla2100_timeout_insert(scsi_qla_host_t *, srb_t *),
qla2100_timeout_remove(scsi_qla_host_t *, srb_t *);

STATIC uint8_t   qla2100_mem_alloc(scsi_qla_host_t *);

static void qla2100_dump_regs(struct Scsi_Host *host);
#if  STOP_ON_ERROR
static void qla2100_panic(char *, struct Scsi_Host *host);
#endif
void qla2100_print_scsi_cmd(Scsi_Cmnd *cmd);
STATIC void qla2100_abort_lun_queue(scsi_qla_host_t *,uint32_t,uint32_t,uint32_t,uint32_t);

STATIC int qla2100_return_status(scsi_qla_host_t *ha, sts_entry_t *sts, Scsi_Cmnd *cp);
STATIC void qla2100_mem_free(scsi_qla_host_t *ha);
void qla2100_do_dpc(void *p);

static inline void qla2100_callback(scsi_qla_host_t *ha, Scsi_Cmnd *cmd);

static inline void qla2100_enable_intrs(scsi_qla_host_t *);
static inline void qla2100_disable_intrs(scsi_qla_host_t *);

#if 0
static void qla2100_extend_timeout(Scsi_Cmnd *cmd, int timeout);
#endif

static int  qla2100_get_tokens(char *line, char **argv, int maxargs );

/*
*  QLogic ISP2100 Hardware Support Function Prototypes.
*/
STATIC void
qla2x00_cfg_persistent_binding(adapter_state_t *ha);
STATIC  uint8_t  qla2100_initialize_adapter(scsi_qla_host_t *),
qla2100_isp_firmware(scsi_qla_host_t *),
qla2100_pci_config(scsi_qla_host_t *),
qla2100_set_cache_line(scsi_qla_host_t *),
qla2100_chip_diag(scsi_qla_host_t *),
qla2x00_setup_chip(scsi_qla_host_t *ha),
qla2x00_init_rings(scsi_qla_host_t *ha),
qla2x00_fw_ready(scsi_qla_host_t *ha),
qla2200_nvram_config(scsi_qla_host_t *),
qla2x00_get_link_status(scsi_qla_host_t *, uint8_t, void *, uint16_t *),

#if USE_FLASH_DATABASE
qla2100_get_database(scsi_qla_host_t *),
qla2100_save_database(scsi_qla_host_t *),
qla2100_program_flash_address(scsi_qla_host_t *, uint32_t, u_char),
qla2100_erase_flash_sector(scsi_qla_host_t *, uint32_t),
qla2100_poll_flash(scsi_qla_host_t *, uint32_t, u_char),
#endif

qla2x00_loop_reset(scsi_qla_host_t *ha),

qla2100_abort_isp(scsi_qla_host_t *),
qla2100_loop_resync(scsi_qla_host_t *);
STATIC void qla2100_cmd_wait(scsi_qla_host_t *ha);

STATIC void qla2100_nv_write(scsi_qla_host_t *, uint16_t),
qla2x00_nv_deselect(adapter_state_t *ha),
qla2100_poll(scsi_qla_host_t *),
qla2100_init_fc_db(scsi_qla_host_t *),
qla2100_init_tgt_map(scsi_qla_host_t *),
#if USE_FLASH_DATABASE
qla2100_flash_enable_database(scsi_qla_host_t *),
qla2100_flash_disable_database(scsi_qla_host_t *),
qla2x00_write_flash_byte(scsi_qla_host_t *, uint32_t, u_char),
#endif
qla2100_reset_adapter(scsi_qla_host_t *),
qla2100_enable_lun(scsi_qla_host_t *),
qla2100_isp_cmd(scsi_qla_host_t *),
qla2100_isr(scsi_qla_host_t *, uint16_t, srb_t **, srb_t **, uint8_t *),
qla2100_rst_aen(scsi_qla_host_t *),

qla2x00_response_pkt(scsi_qla_host_t *, uint16_t, srb_t **, srb_t **),
qla2x00_status_entry(scsi_qla_host_t *, sts_entry_t *, srb_t **, srb_t **),
qla2x00_error_entry(scsi_qla_host_t *, response_t *, srb_t **, srb_t **),
qla2x00_ms_entry(scsi_qla_host_t *, ms_iocb_entry_t *, srb_t **, srb_t **),

qla2x00_restart_queues(scsi_qla_host_t *, uint8_t),
qla2x00_abort_queues(scsi_qla_host_t *, uint8_t);

STATIC void qla2x00_flash_enable(adapter_state_t *);
STATIC void qla2x00_flash_disable(adapter_state_t *);
STATIC uint8_t qla2x00_read_flash_byte(adapter_state_t *, uint32_t);
STATIC uint16_t qla2x00_get_flash_version(adapter_state_t *);

STATIC uint16_t  qla2100_get_nvram_word(scsi_qla_host_t *, uint32_t),
qla2100_nvram_request(scsi_qla_host_t *, uint32_t),
qla2100_debounce_register(volatile uint16_t *);

STATIC request_t *qla2100_req_pkt(scsi_qla_host_t *);
STATIC request_t *qla2100_ms_req_pkt(scsi_qla_host_t *, srb_t *);
STATIC uint8_t qla2x00_configure_hba(scsi_qla_host_t *ha);
STATIC void qla2100_reset_chip(scsi_qla_host_t *ha);
#if QL2100_TARGET_MODE_SUPPORT
qla2100_enable_lun(scsi_qla_host_t *, uint8_t, uint32_t),
qla2100_notify_ack(scsi_qla_host_t *, notify_entry_t *),
qla2100_immed_notify(scsi_qla_host_t *, notify_entry_t *),
qla2100_accept_io(scsi_qla_host_t *, ctio_ret_entry_t *),
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
qla2100_64bit_continue_io(scsi_qla_host_t *, atio_entry_t *, uint32_t,
    u_long *),
#endif
qla2100_32bit_continue_io(scsi_qla_host_t *, atio_entry_t *, uint32_t,
    u_long *),
qla2100_atio_entry(scsi_qla_host_t *, atio_entry_t *),
qla2100_notify_entry(scsi_qla_host_t *, notify_entry_t *),

#endif  /* QLA2100_TARGET_MODE_SUPPORT */
STATIC void qla2100_display_fc_names(scsi_qla_host_t *ha);
void ql2100_dump_requests(scsi_qla_host_t *ha);
static void qla2100_get_properties(scsi_qla_host_t *ha, char *string);
STATIC uint8_t qla2100_find_propname(scsi_qla_host_t *ha, char *propname, char *propstr, char *db, int siz);
static int qla2100_get_prop_16chars(scsi_qla_host_t *ha, char *propname, char *propval, char *cmdline);
static char	*qla2100_get_line(char *str, char *line);
void qla2100_check_fabric_devices(scsi_qla_host_t *ha);

#ifdef FC_IP_SUPPORT

/* Entry points for IP network driver */
int  qla2x00_ip_inquiry(uint16_t wAdapterNumber, BD_INQUIRY_DATA *pInquiryData);
int  qla2x00_ip_enable(scsi_qla_host_t *ha, BD_ENABLE_DATA *pEnableData);
void qla2x00_ip_disable(scsi_qla_host_t *ha);
void qla2x00_add_buffers(scsi_qla_host_t *ha, uint16_t wBufferCount);
int  qla2x00_send_packet(scsi_qla_host_t *ha, SEND_CB *pSendCB);

static int qla2x00_ip_initialize(scsi_qla_host_t *ha);
static int qla2x00_add_new_ip_device(scsi_qla_host_t *ha,
uint16_t wLoopId,
uint8_t *pPortId,
uint8_t *pPortName,
int bForceAdd);
static int qla2x00_convert_to_arp(scsi_qla_host_t *ha, SEND_CB *pSendCB);
static void qla2x00_free_ip_block(scsi_qla_host_t *ha,
IP_DEVICE_BLOCK *pIpDevice);
static int qla2x00_get_ip_loopid(scsi_qla_host_t *ha,
uint8_t *pNodeName,
uint8_t *pLoopId);
static int qla2x00_send_farp_request(scsi_qla_host_t *ha,
uint8_t *pPortName);
static int qla2x00_register_ip_device(scsi_qla_host_t *ha);
static int qla2x00_reserve_ip_block(scsi_qla_host_t *ha,
PIP_DEVICE_BLOCK *pIpDevBlk);
static int qla2x00_update_ip_device_data(scsi_qla_host_t *ha,
device_data_t *pDeviceData);
static int qla2x00_reserve_loopid(scsi_qla_host_t *ha, uint16_t *pLoopId);
static void qla2x00_free_loopid(scsi_qla_host_t *ha, uint16_t wLoopId);
static int qla2x00_login_public_device(scsi_qla_host_t *ha,
uint16_t *pLoopId,
uint8_t  *pPortID,
uint16_t wOptions);
static int qla2x00_logout_public_device(scsi_qla_host_t *ha,
uint16_t wLoopId,
uint16_t wOptions);
#endif

static void qla2100_device_resync(scsi_qla_host_t *);
STATIC uint8_t qla2100_update_fc_database(scsi_qla_host_t *, fcdev_t *,
    uint8_t);

STATIC uint8_t qla2100_configure_fabric(scsi_qla_host_t *, uint8_t );
static uint8_t qla2100_find_all_fabric_devs(scsi_qla_host_t *, sns_data_t *,
    DMA_ADDR, struct new_dev *, uint16_t *, uint8_t *);
static uint8_t qla2x00_register_fc4(scsi_qla_host_t *, sns_data_t *, DMA_ADDR);
static uint8_t qla2100_gan(scsi_qla_host_t *, sns_data_t *, DMA_ADDR,
    fcdev_t *);
static uint8_t qla2100_fabric_login(scsi_qla_host_t *, fcdev_t *);

STATIC uint8_t qla2100_configure_loop(scsi_qla_host_t *);
static uint8_t qla2100_configure_local_loop(scsi_qla_host_t *, uint8_t );

void qla2100_free_sp( Scsi_Cmnd *, srb_t *);
srb_t * qla2100_allocate_sp( scsi_qla_host_t *,Scsi_Cmnd *);
STATIC uint8_t qla2x00_32bit_start_scsi(srb_t *sp);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
STATIC uint8_t qla2x00_64bit_start_scsi(srb_t *sp);
#endif

/* Routines for Failover */
os_tgt_t * qla2x00_tgt_alloc(scsi_qla_host_t *ha, uint16_t t, uint8_t *name);
#if  APIDEV
static int apidev_init(struct Scsi_Host*);
static int apidev_cleanup(void);
#endif
void qla2x00_tgt_free(scsi_qla_host_t *ha, uint16_t t);
os_lun_t * qla2x00_lun_alloc(scsi_qla_host_t *ha, uint16_t t, uint16_t l);

static void qla2x00_lun_free(scsi_qla_host_t *ha, uint16_t t, uint16_t l);
void qla2x00_add_link_b(ql_list_head_t *head, ql_list_link_t *link);
void qla2x00_add_link_t(ql_list_head_t *head, ql_list_link_t *link);
void qla2x00_next(os_tgt_t *tq, os_lun_t *lq);
static void qla2100_removeq(os_lun_t *lq, srb_t *sp);
static void qla2100_putq_t(os_lun_t *lq, srb_t *sp);
static void qla2100_putq_b(os_lun_t *lq, srb_t *sp);
void qla2x00_remove_link(ql_list_head_t *head, ql_list_link_t *link);
static int qla2x00_build_fcport_list(scsi_qla_host_t *ha);
static void qla2x00_config_os(scsi_qla_host_t *ha);
static uint16_t qla2x00_fcport_bind(scsi_qla_host_t *ha, fc_port_t *fcport);
static int qla2x00_update_fcport(scsi_qla_host_t *ha, fc_port_t *fcport, int);
static int qla2x00_lun_discovery(scsi_qla_host_t *ha, fc_port_t *fcport, int);

#if MPIO_SUPPORT
STATIC void qla2x00_process_failover(scsi_qla_host_t *ha) ;
#endif

void qla2x00_srb_abort(void *arg, int);
STATIC int qla2x00_device_reset(adapter_state_t *, uint16_t);

static int qla2x00_is_wwn_zero(uint8_t *nn);
STATIC int qla2x00_marker(scsi_qla_host_t *, uint16_t , uint16_t ,
    uint8_t );
void qla2100_get_lun_mask_from_config(scsi_qla_host_t *ha, fc_port_t *port,
	uint16_t tgt, uint16_t dev_no);
void qla2100_print_q_info(os_lun_t *q);
srb_t * qla2100_allocate_sp_from_ha( scsi_qla_host_t *, Scsi_Cmnd *);

#if MPIO_SUPPORT
STATIC void qla2x00_failover_cleanup(scsi_qla_host_t *, os_tgt_t *, os_lun_t	*, srb_t *);
#endif

void qla2x00_chg_endian(uint8_t buf[], size_t size);
STATIC void qla2x00_check_sense(Scsi_Cmnd       *cp, os_lun_t *);
STATIC void qla2x00_suspend_lun(scsi_qla_host_t *, os_lun_t *, int, int);
void qla2x00_flush_failover_q(scsi_qla_host_t *, os_lun_t *);



#if  DEBUG_QLA2100
#ifndef QL_DEBUG_ROUTINES
#define QL_DEBUG_ROUTINES
#endif
#endif

#ifdef QL_DEBUG_ROUTINES
/*
*  Driver Debug Function Prototypes.
*/
STATIC uint8_t  qla2100_getbyte(uint8_t *);
STATIC uint16_t qla2100_getword(uint16_t *);
STATIC uint32_t qla2100_getdword(uint32_t *);
STATIC void     qla2100_putbyte(uint8_t *, uint8_t),
    qla2100_putword(uint16_t *, uint16_t),
    qla2100_putdword(uint32_t *, uint32_t),
    qla2100_print(int8_t *),
    qla2100_output_number(u_long, uint8_t),
    qla2100_putc(int8_t),
    qla2100_dump_buffer(uint8_t *, uint32_t);
STATIC uint8_t ql2x_debug_print = 1;
#endif

/* ra 01/03/02 */
#if QLA2100_LIPTEST
STATIC int  mbxtimeout = 0;
#endif

#if DEBUG_GET_FW_DUMP
STATIC void qla2300_dump_isp(scsi_qla_host_t *ha),
qla2100_dump_word(uint8_t *, uint32_t, uint32_t);
#endif
#if  NO_LONG_DELAYS
STATIC void qla2x00_sleep_done (struct semaphore * sem);
#endif
void qla2x00_sleep (scsi_qla_host_t *ha, int timeout);

/* Debug print buffer */
char          debug_buff[LINESIZE*3];



/*
* insmod needs to find the variable and make it point to something
*/
static char *ql2xdevconf = NULL;
static int ql2xretrycount = 10;
static int qla2xenbinq = 1;
#ifdef MODULE
static char *ql2xopts = NULL;
static int ql2xmaxqdepth = 0;

#if MPIO_SUPPORT
static int ConfigRequired = 0;
static int recoveryTime = 10;
static int failbackTime = 60;
#endif	/* end of MPIO_SUPPORT */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,18)
/* insmod qla2100 ql2xopts=verbose" */
MODULE_PARM(ql2xopts, "s");
MODULE_PARM(ql2xmaxqdepth, "i");
#if MPIO_SUPPORT
MODULE_PARM(ConfigRequired, "i");
MODULE_PARM(recoveryTime, "i");
MODULE_PARM(failbackTime, "i");
#endif
#endif

#include "qla_fo.cfg"


/*
* Just in case someone uses commas to separate items on the insmod
* command line, we define a dummy buffer here to avoid having insmod
* write wild stuff into our code segment
*/
static char dummy_buffer[60] = "Please don't add commas in your insmod command!!\n";

#endif

/*
* Macros to change names of OS remapping routines.
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
#define ioremap vremap
#define iounmap vfree
#endif


/*
* This is the pointer to the /proc/scsi/qla2100 code.
* access the driver.
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
#if CONFIG_PROC_FS
/* this definition is normally in proc_fs.h     PROC_SCSI_QL2100 */
#define PROC_SCSI_QL2100  PROC_SCSI_QLOGICISP

struct proc_dir_entry proc_scsi_qla2100 = {
PROC_SCSI_QL2100, 7, DRIVER_NAME,
S_IFDIR | S_IRUGO | S_IXUGO, 2,
0, 0, 0, NULL, NULL, NULL, NULL, NULL, NULL, NULL
};
#endif
#endif
#if QLA2100_LIPTEST
static int qla2100_lip = 0;
#endif

#include <linux/ioctl.h>
#include <scsi/scsi_ioctl.h>

/* multi-OS QLOGIC IOCTL definition file */
#include "exioct.h"


#if QLA_SCSI_VENDOR_DIR
/* Include routine to set direction for vendor specific commands */
#include "qlavendor.c"
#endif
/***********************************************************************
* We use the Scsi_Pointer structure that's included with each command
* SCSI_Cmnd as a scratchpad for our SRB. This allows us to accept
* an unlimited number of commands.
*
* SCp will always point to the SRB structure (defined in qla2100.h).
* It is defined as follows:
*  - SCp.ptr  -- > pointer back to the cmd
*  - SCp.this_residual --> used as forward pointer to next srb
*  - SCp.buffer --> used as backward pointer to next srb
*  - SCp.buffers_residual --> used as flags field
*  - SCp.have_data_in --> not used
*  - SCp.sent_command --> not used
*  - SCp.phase --> not used
***********************************************************************/
#define NEW_SP
#ifdef NEW_SP
#define	CMD_SP(Cmnd)		((Cmnd)->SCp.ptr)
#else
#define	CMD_SP(Cmnd)		(&(Cmnd)->SCp)
#endif

#define	CMD_XFRLEN(Cmnd)	(Cmnd)->request_bufflen
#define	CMD_CDBLEN(Cmnd)	(Cmnd)->cmd_len
#define	CMD_CDBP(Cmnd)		(Cmnd)->cmnd
#define	CMD_SNSP(Cmnd)		(Cmnd)->sense_buffer
#define	CMD_SNSLEN(Cmnd)	(sizeof (Cmnd)->sense_buffer)
#define	CMD_RESULT(Cmnd)	((Cmnd)->result)
#define	CMD_HANDLE(Cmnd)	((Cmnd)->host_scribble)
#define	CMD_TIMEOUT(Cmnd)	((Cmnd)->timeout_per_command)

#define DID_RETRY		DID_ERROR


/*
* Macros use for debugging the driver.
*/
#ifdef QL_DEBUG_LEVEL_3
#define ENTER(x)	sprintf(debug_buff,"qla2100 : Entering %s()\n\r", x); \
qla2100_print(debug_buff);
#define LEAVE(x)	sprintf(debug_buff,"qla2100 : Leaving %s()\n\r", x); \
qla2100_print(debug_buff);
#define ENTER_INTR(x)	sprintf(debug_buff,"qla2100 : Entering %s()\n\r", x); \
qla2100_print(debug_buff);
#define LEAVE_INTR(x)	sprintf(debug_buff,"qla2100 : Leaving %s()\n\r", x); \
qla2100_print(debug_buff);
#define DEBUG3(x)	x
#define DEBUG2_3(x)     x
#define DEBUG2_3_11(x)	x
#define DEBUG3_11(x)	x
#else
#define ENTER(x)
#define LEAVE(x)
#define ENTER_INTR(x)
#define LEAVE_INTR(x)
#define DEBUG3(x)
#endif

#if  QLA2100_COMTRACE
#define COMTRACE(x)     qla2100_putc(x);
#else
#define COMTRACE(x)
#endif

#if  DEBUG_QLA2100
#define DEBUG(x)	x
#define DEBUG4(x)
#else
#define DEBUG(x)
#define DEBUG4(x)
#endif

#ifdef QL_DEBUG_LEVEL_2
#define DEBUG2(x)	x
#define DEBUG2_3(x)	x
#define DEBUG2_3_11(x)	x
#define DEBUG2_11(x)	x
#else
#define DEBUG2(x)
  #ifndef QL_DEBUG_LEVEL_3
  #define DEBUG2_3(x)
  #endif
#endif

#ifdef QL_DEBUG_LEVEL_5
#define DEBUG5(x)	x
#else
#define DEBUG5(x)
#endif

#ifdef QL_DEBUG_LEVEL_7
#define DEBUG7(x)          x
#else
#define DEBUG7(x)
#endif

#ifdef QL_DEBUG_LEVEL_9
#define DEBUG9(x)	x
#define DEBUG9_10(x)	x
#else
#define DEBUG9(x)
#endif

#ifdef QL_DEBUG_LEVEL_10
#define DEBUG10(x)	x
#define DEBUG9_10(x)	x
#else
#define DEBUG10(x)
  #ifndef QL_DEBUG_LEVEL_9
  #define DEBUG9_10(x)
  #endif
#endif

#ifdef QL_DEBUG_LEVEL_11
#define DEBUG11(x)	x
#define DEBUG2_11(x)	x
#define DEBUG2_3_11(x)	x
#define DEBUG3_11(x)	x
#else
#define DEBUG11(x)
  #ifndef QL_DEBUG_LEVEL_2
  #define DEBUG2_11(x)
    #ifndef QL_DEBUG_LEVEL_3
    #define DEBUG2_3_11(x)
    #endif
  #endif
  #ifndef QL_DEBUG_LEVEL_3
  #define DEBUG3_11(x)
  #endif
#endif


uint8_t copyright[48] = "Copyright 1999-2001, Qlogic Corporation";

/****************************************************************************/
/*  LINUX -  Loadable Module Functions.                                     */
/****************************************************************************/

/*****************************************/
/*   ISP Boards supported by this driver */
/*****************************************/
#define QLA2X00_VENDOR_ID   0x1077
#define QLA2200_DEVICE_ID   0x2200
#define QLA2200A_DEVICE_ID  0x2200A
#define QLA2300_DEVICE_ID   0x2300
#define QLA2312_DEVICE_ID   0x2312
#define QLA2200A_RISC_ROM_VER  4
#define FPM_2300            6
#define FPM_2310            7

#if defined(ISP2100) || defined(ISP2200)
#define NUM_OF_ISP_DEVICES  2
#endif
#ifdef ISP2300
#define NUM_OF_ISP_DEVICES  3
#endif

typedef struct _qlaboards
{
unsigned char   bdName[9];       /* Board ID String             */
unsigned long   device_id;       /* Device ID                   */
int   numPorts;                  /* number of loops on adapter  */
unsigned short   *fwcode;        /* pointer to FW array         */
unsigned long    *fwlen;         /* number of words in array    */
unsigned short   *fwstart;       /* start address for F/W       */
unsigned char   *fwver;          /* Ptr to F/W version array    */
} qla_boards_t;


static struct _qlaboards   QLBoardTbl_fc[NUM_OF_ISP_DEVICES] =
{
/* Name ,  Board PCI Device ID,         Number of ports */
#ifdef ISP2300
{"QLA2312 ", QLA2312_DEVICE_ID,           MAX_BUSES,
#ifdef FC_IP_SUPPORT
&fw2300ip_code01[0],  (unsigned long *)&fw2300ip_length01, &fw2300ip_addr01, &fw2300ip_version_str[0] },
#else
&fw2300tp_code01[0],  (unsigned long *)&fw2300tp_length01, &fw2300tp_addr01, &fw2300tp_version_str[0] },
#endif
{"QLA2300 ", QLA2300_DEVICE_ID,           MAX_BUSES,
#ifdef FC_IP_SUPPORT
&fw2300ip_code01[0],  (unsigned long *)&fw2300ip_length01, &fw2300ip_addr01, &fw2300ip_version_str[0] },
#else
&fw2300tp_code01[0],  (unsigned long *)&fw2300tp_length01, &fw2300tp_addr01, &fw2300tp_version_str[0] },
#endif
#endif
#ifdef ISP2200
{"QLA2200 ", QLA2200_DEVICE_ID,           MAX_BUSES,
#ifdef FC_IP_SUPPORT
&fw2200ip_code01[0],  (unsigned long *)&fw2200ip_length01, &fw2200ip_addr01, &fw2200ip_version_str[0] },
#else
&fw2200tp_code01[0],  (unsigned long *)&fw2200tp_length01, &fw2200tp_addr01, &fw2200tp_version_str[0] },
#endif
#endif
{"        ",                 0,           0}
};

/*
* Stat info for all adpaters
*/
static struct _qla2100stats  {
unsigned long   mboxtout;            /* mailbox timeouts */
unsigned long   mboxerr;             /* mailbox errors */
unsigned long   ispAbort;            /* ISP aborts */
unsigned long   debugNo;
unsigned long   loop_resync;
unsigned long   outarray_full;
unsigned long   retry_q_cnt;
scsi_qla_host_t *irqhba;
} qla2100_stats;

/*
 * Declare our global semaphores
 */
#ifdef ISP2200
DECLARE_MUTEX_LOCKED(qla2200_detect_sem);
#endif
#ifdef ISP2300
DECLARE_MUTEX_LOCKED(qla2300_detect_sem);
#endif

#ifdef	KERNEL_SEM_BUG
static atomic_t 	dpc_needed;         /* DPC routine is needed */
#endif

/*
* Command line options
*/
static unsigned long qla2100_verbose = 1L;
static unsigned long qla2100_quiet   = 0L;
static unsigned long qla2100_reinit = 0L;
static unsigned long qla2100_req_dmp = 0L;

#if QL_TRACE_MEMORY
extern unsigned long mem_trace[1000];
extern unsigned long mem_id[1000];
int	mem_trace_ptr = 0;
#endif

/* Enable for failover */
#if MPIO_SUPPORT
static unsigned long qla2100_failover = 1L;
#else
static unsigned long qla2100_failover = 0L;
#endif	/* end of MPIO_SUPPORT */

/*
 * List of host adapters
 */
static scsi_qla_host_t *qla2100_hostlist = NULL;

#ifdef QLA2100_PROFILE
static int qla2100_buffer_size = 0;
static char *qla2100_buffer = NULL;
#endif

#include <linux/ioctl.h>
#include <scsi/scsi_ioctl.h>
#include <asm/uaccess.h>

/*************************************************************************
*   qla2100_set_info
*
* Description:
*   Set parameters for the driver from the /proc filesystem.
*
* Returns:
*************************************************************************/
int
qla2100_set_info(char *buffer, int length, struct Scsi_Host *HBAptr)
{
    return (-ENOSYS);  /* Currently this is a no-op */
}

#include "qla_mbx.c"
#include "qla2x00_ioctl.c"
#ifdef INTAPI
#include "qla_inioct.c"
#endif


/*************************************************************************
* qla2100_proc_info
*
* Description:
*   Return information to handle /proc support for the driver.
*
* inout : decides on the direction of the dataflow and the meaning of the
*         variables
* buffer: If inout==FALSE data is being written to it else read from it
*         (ptrs to a page buffer)
* *start: If inout==FALSE start of the valid data in the buffer
* offset: If inout==FALSE offset from the beginning of the imaginary file
*         from which we start writing into the buffer
* length: If inout==FALSE max number of bytes to be written into the buffer
*         else number of bytes in the buffer
* Returns:
*************************************************************************/
#define	PROC_BUF	(&qla2100_buffer[len])
int
qla2100_proc_info( char *buffer, char **start, off_t offset, int length,
    int hostno, int inout)
{
#if QLA2100_PROFILE
	struct Scsi_Host *host;
	int             i;
	int             len = 0;
	int             size = 0;
	os_lun_t	*up;
	qla_boards_t    *bdp;
	scsi_qla_host_t *ha;
	uint32_t        t, l;
	uint32_t        tmp_sn;
	unsigned long   *flags;

#if REQ_TRACE
	Scsi_Cmnd       *cp;
	srb_t           *sp;
#endif

	DEBUG3(printk(KERN_INFO "Entering proc_info buff_in=%p, offset=0x%lx, "
	    "length=0x%x\n", buffer, offset, length);)

	host = NULL;

	/* Find the host that was specified */
	for (ha=qla2100_hostlist; (ha != NULL) && ha->host->host_no != hostno;
	    ha=ha->next) {
		continue;
	}

	/* if host wasn't found then exit */
	if( !ha ) {
		size = sprintf(buffer,
		    "Can't find adapter for host number %d\n", hostno);

		if( size > length ) {
			return (size);
		} else {
			return (0);
		}
	}

	host = ha->host;

	if( inout == TRUE ) /* Has data been written to the file? */ {
		printk("qla2100_proc: has data been written to the file. \n");
		return (qla2100_set_info(buffer, length, host));
	}

	/*
	 * if our old buffer is the right size use it otherwise
	 * allocate a new one.
	 */
	size = 4096 * 4;  /* get a page */
	if( qla2100_buffer_size != size ) {
		/* deallocate this buffer and get a new one */
		if( qla2100_buffer != NULL ) {
			kfree(qla2100_buffer);
			qla2100_buffer_size = 0;
		}
		qla2100_buffer = kmalloc(size, GFP_KERNEL);
	}

	if( qla2100_buffer == NULL ) {
		size = sprintf(buffer, "qla2100 - kmalloc error at line %d\n",
		    __LINE__);
		return size;
	}

	/* save the size of our buffer */
	qla2100_buffer_size = size;

	/* start building the print buffer */
	bdp = &QLBoardTbl_fc[ha->devnum];
#if MPIO_SUPPORT
	size =  sprintf(PROC_BUF,
	    "QLogic PCI to Fibre Channel Host Adapter for ISP21xx/ISP22xx/"
	    "ISP23xx:\n"
	    "        Firmware version: %2d.%02d.%02d, Driver version %s-fo\n",
	    bdp->fwver[0], bdp->fwver[1], bdp->fwver[2], QLA2100_VERSION);
#else
	size =  sprintf(PROC_BUF,
	    "QLogic PCI to Fibre Channel Host Adapter for ISP21xx/ISP22xx/"
	    "ISP23xx:\n"
	    "        Firmware version: %2d.%02d.%02d, Driver version %s\n",
	    bdp->fwver[0], bdp->fwver[1], bdp->fwver[2], QLA2100_VERSION);
#endif

	len += size;
	size = sprintf(PROC_BUF, "Entry address = %p\n",qla2100_set_info);
	len += size;

	tmp_sn = ((ha->serial0 & 0x1f) << 16) | (ha->serial2 << 8) | ha->serial1;
	size = sprintf(PROC_BUF, "HBA: %s, Serial# %c%05d\n",
	    bdp->bdName, ('A' + tmp_sn/100000), (tmp_sn%100000));
	len += size;

	size = sprintf(PROC_BUF,
	    "Request Queue = 0x%lx, Response Queue = 0x%lx\n",
	    (long unsigned int)ha->request_dma,
	    (long unsigned int)ha->response_dma);
	len += size;

	size = sprintf(PROC_BUF,
	    "Request Queue count= %ld, Response Queue count= %ld\n",
	    (long)REQUEST_ENTRY_CNT, (long)RESPONSE_ENTRY_CNT);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of pending commands = 0x%lx\n", ha->actthreads);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of queued commands = 0x%lx\n", ha->qthreads);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of free request entries = %d\n",ha->req_q_cnt);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of mailbox timeouts = %ld\n",qla2100_stats.mboxtout);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of ISP aborts = %ld\n",qla2100_stats.ispAbort);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of loop resyncs = %ld\n",qla2100_stats.ispAbort);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of retries for empty slots = %ld\n",
	    qla2100_stats.outarray_full);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of reqs in retry_q = %ld\n",qla2100_stats.retry_q_cnt);
	len += size;

	size = sprintf(PROC_BUF,
	    "Number of reqs in done_q = %ld\n",ha->done_q_cnt);
	len += size;

	flags = (unsigned long *) &ha->flags;

	size = sprintf(PROC_BUF,
	    "Host adapter: state = %s, flags= 0x%lx\n",
	    (ha->loop_state == LOOP_DOWN)? "DOWN": "UP", *flags);
	len += size;

	size = sprintf(PROC_BUF, "Dpc flags = 0x%lx\n",(unsigned long)ha->dpc_flags);
	len += size;

	size = sprintf(PROC_BUF, "\n");
	len += size;

#if REQ_TRACE
	if ( qla2100_req_dmp ) {
		size  = sprintf(PROC_BUF,
		    "Outstanding Commands on controller:\n");
		len += size;

		for( i=0; i < MAX_OUTSTANDING_COMMANDS; i++ ) {
			if( (sp = ha->outstanding_cmds[i]) == NULL )
				continue;

			if( (cp = sp->cmd) == NULL )
				continue;

			size = sprintf(PROC_BUF, "(%d): Pid=%d, sp flags=0x%lx,"
			    " cmd=0x%p, state=%d\n", i,
			    (int)sp->cmd->serial_number, (long)sp->flags,
			    CMD_SP(sp->cmd),(int)sp->state);
			len += size;

			if( len >= qla2100_buffer_size )
				goto profile_stop;
		}
	}
#endif /* REQ_TRACE */

	/* 2.25 node/port display to proc */
	/* Display the node name for adapter */
	size = sprintf(PROC_BUF, "\nSCSI Device Information:\n");
	len += size;
	size = sprintf(PROC_BUF,
	    "scsi-qla%d-adapter-node=%02x%02x%02x%02x%02x%02x%02x%02x;\n",
	    (int)ha->instance,
	    ha->init_cb->node_name[0],
	    ha->init_cb->node_name[1],
	    ha->init_cb->node_name[2],
	    ha->init_cb->node_name[3],
	    ha->init_cb->node_name[4],
	    ha->init_cb->node_name[5],
	    ha->init_cb->node_name[6],
	    ha->init_cb->node_name[7]);
	len += size;

	/* display the port name for adapter */
	size = sprintf(PROC_BUF,
	    "scsi-qla%d-adapter-port=%02x%02x%02x%02x%02x%02x%02x%02x;\n",
	    (int)ha->instance,
	    ha->init_cb->port_name[0],
	    ha->init_cb->port_name[1],
	    ha->init_cb->port_name[2],
	    ha->init_cb->port_name[3],
	    ha->init_cb->port_name[4],
	    ha->init_cb->port_name[5],
	    ha->init_cb->port_name[6],
	    ha->init_cb->port_name[7]);
	len += size;

	/* Print out device port names */
	for (i = 0; i < MAX_FIBRE_DEVICES; i++) {
		if (ha->fc_db[i].loop_id == PORT_UNUSED)
			continue;

		if (ha->flags.failover_enabled) {
			size = sprintf(PROC_BUF,
			    "scsi-qla%d-port-%d=%02x%02x%02x%02x%02x%02x%02x%"
			    "02x:%02x%02x%02x%02x%02x%02x%02x%02x;\n",
			    (int)ha->instance, i,
			    ha->fc_db[i].name[0], ha->fc_db[i].name[1],
			    ha->fc_db[i].name[2], ha->fc_db[i].name[3],
			    ha->fc_db[i].name[4], ha->fc_db[i].name[5],
			    ha->fc_db[i].name[6], ha->fc_db[i].name[7],
			    ha->fc_db[i].wwn[0], ha->fc_db[i].wwn[1],
			    ha->fc_db[i].wwn[2], ha->fc_db[i].wwn[3],
			    ha->fc_db[i].wwn[4], ha->fc_db[i].wwn[5],
			    ha->fc_db[i].wwn[6], ha->fc_db[i].wwn[7]);
		} else {
#if USE_PORTNAME  /* updated to use conditional compile */
			size = sprintf(PROC_BUF,
			    "scsi-qla%d-target-%d=%02x%02x%02x%02x%02x%02x%02x"
			    "%02x;\n",
			    (int)ha->instance, i,
			    ha->fc_db[i].wwn[0],
			    ha->fc_db[i].wwn[1],
			    ha->fc_db[i].wwn[2],
			    ha->fc_db[i].wwn[3],
			    ha->fc_db[i].wwn[4],
			    ha->fc_db[i].wwn[5],
			    ha->fc_db[i].wwn[6],
			    ha->fc_db[i].wwn[7]);
#else
			size = sprintf(PROC_BUF,
			    "scsi-qla%d-target-%d=%02x%02x%02x%02x%02x%02x%02x"
			    "%02x;\n",
			    (int)ha->instance, i,
			    ha->fc_db[i].name[0],
			    ha->fc_db[i].name[1],
			    ha->fc_db[i].name[2],
			    ha->fc_db[i].name[3],
			    ha->fc_db[i].name[4],
			    ha->fc_db[i].name[5],
			    ha->fc_db[i].name[6],
			    ha->fc_db[i].name[7]);
#endif /* USE_PORT_NAME */
		}

		len += size;
	} /* 2.25 node/port display to proc */

	size = sprintf(PROC_BUF, "\nSCSI LUN Information:\n");
	len += size;

	size = sprintf(PROC_BUF, "(Id:Lun)\n");
	len += size;

	/* scan for all equipment stats */
	for (t = 0; t < MAX_FIBRE_DEVICES; t++) {
		/* valid target */
		if (ha->fc_db[t].loop_id == PORT_UNUSED)
			continue;

		/* scan all luns */
		for( l = 0; l < ha->max_luns; l++ ) {

			up = (os_lun_t *) GET_LU_Q(ha, t, l);

			if( up == NULL )
				continue;

			if (up->io_cnt == 0 || up->io_cnt < 3 ||
			    up->fclun == NULL)
				continue;

			/* total reads since boot */
			/* total writes since boot */
			/* total requests since boot  */
			size = sprintf(PROC_BUF, "(%2d:%2d): Total reqs %ld,",
			    t,l,up->io_cnt);
			len += size;

			/* current number of pending requests */
			size = sprintf(PROC_BUF, " Pending %d,",
			    (int)up->q_outcnt);
			len += size;

			/* current number of pending requests */
			size = sprintf(PROC_BUF, " Queued %d,",
			    (int)up->q_incnt);
			len += size;

			/* current number of busy */
			size = sprintf(PROC_BUF, " full %d,",(int)up->q_bsycnt);
			len += size;

			size = sprintf(PROC_BUF, " flags 0x%x,",
			    (int)up->q_flag);
			len += size;

			size = sprintf(PROC_BUF, " %ld:%d:%02x,",
			    up->fclun->fcport->ha->instance,
			    up->fclun->fcport->cur_path,
			    up->fclun->fcport->loop_id);
			len += size;

			size = sprintf(PROC_BUF, "\n");
			len += size;

			if( len >= qla2100_buffer_size )
				goto profile_stop;
		}

		if( len >= qla2100_buffer_size )
			break;
	}

profile_stop:
	if( len >= qla2100_buffer_size ) {
		printk(KERN_WARNING "qla2x00: Overflow buffer in "
		    "qla2100_proc.c\n");
	}

	if( offset > len - 1 ) {
		kfree(qla2100_buffer);
		qla2100_buffer = NULL;
		qla2100_buffer_size = length = 0;
		*start = NULL;
	} else {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
		/* Start of wanted data */
		*start = &qla2100_buffer[offset];

		if( len - offset < length ) {
			length = len - offset;
		}
#else
		*start = buffer;
		if( len - offset < length ) {
			length = len - offset;
		}
		memcpy(buffer, &qla2100_buffer[offset], length);
#endif
	}

	DEBUG3(printk(KERN_INFO "Exiting proc_info: qlabuff=%p, offset=0x%lx, "
	    "length=0x%x\n", qla2100_buffer, offset, length);)

#if QLA2100_LIPTEST
	qla2100_lip = 1;
#endif

#else
	return (0);
#endif /* QLA2100_PROFILE */

	return (length);
}

/**************************************************************************
* qla2100_detect
*
* Description:
*    This routine will probe for Qlogic FC SCSI host adapters.
*    It returns the number of host adapters of a particular
*    type that were found.	 It also initialize all data necessary for
*    the driver.  It is passed-in the host number, so that it
*    knows where its first entry is in the scsi_hosts[] array.
*
* Input:
*     template - pointer to SCSI template
*
* Returns:
*  num - number of host adapters found.
**************************************************************************/
int
qla2100_detect(Scsi_Host_Template *template) {
	device_reg_t	*reg;
	int		i;
        uint16_t        subsystem_vendor, subsystem_device;
    struct Scsi_Host *host;
    scsi_qla_host_t *ha = NULL, *cur_ha;
    struct _qlaboards  *bdp;
	unsigned long		wait_switch = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
    unsigned short subsys;
#endif
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,1,95)
    unsigned int piobase;
    unsigned char pci_bus, pci_devfn, pci_irq;
    config_reg_t   *cfgp = 0;
#endif
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
    struct pci_dev *pdev = NULL;
#else
    int index;
#endif
#if 0
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
    struct semaphore sem = MUTEX_LOCKED;
#else
    DECLARE_MUTEX_LOCKED(sem);
#endif
#endif
    ENTER("qla2100_detect");

#ifndef NEW_SP
    if( sizeof(srb_t) > sizeof(Scsi_Pointer) )
		printk(KERN_WARNING "qla2x00: srb_t must be re-defined "
		    "- it's too big.\n");

#ifdef CODECHECK
    if( sizeof(srb_t) > sizeof(Scsi_Pointer) ) {
		printk("Redefine srb_t - it's too big.\n");
		return 0;
	}
#endif
#endif

#ifdef MODULE
	DEBUG2(sprintf(debug_buff,
	    "DEBUG: qla2100_set_info starts at address = %x\n",
	    (uint32_t)qla2100_set_info);)
	DEBUG2(qla2100_print(debug_buff);)
	DEBUG(printk(
	    "DEBUG: qla2100_set_info starts at address = %x\n",
	    (uint32_t)qla2100_set_info);)

    /*
    * If we are called as a module, the qla2100 pointer may not be null
    * and it would point to our bootup string, just like on the lilo
    * command line.  IF not NULL, then process this config string with
    * qla2100_setup
    *
    * Boot time Options
    * To add options at boot time add a line to your lilo.conf file like:
    * append="qla2100=verbose,tag_info:{{32,32,32,32},{32,32,32,32}}"
    * which will result in the first four devices on the first two
    * controllers being set to a tagged queue depth of 32.
    */
    if( ql2xopts )
        qla2100_setup(ql2xopts, NULL);
    if( dummy_buffer[0] != 'P' )
		printk(KERN_WARNING "qla2100: Please read the file "
		    "/usr/src/linux/drivers/scsi/README.qla2x00\n"
		    "qla2x00: to see the proper way to specify options to "
		    "the qla2x00 module\n"
		    "qla2x00: Specifically, don't use any commas when "
		    "passing arguments to\n"
		    "qla2x00: insmod or else it might trash certain memory "
		    "areas.\n");
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
    if( (int) !pcibios_present() )
#else
    if (!pci_present())
#endif
    {
		printk("scsi: PCI not present\n");
        return 0;
    } /* end of IF */

    bdp = &QLBoardTbl_fc[0];
    qla2100_hostlist = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
    template->proc_dir = &proc_scsi_qla2100;

#else
    template->proc_name = DRIVER_NAME;
#endif

    /* Try and find each different type of adapter we support */
    for( i=0; bdp->device_id != 0 && i < NUM_OF_ISP_DEVICES; i++, bdp++ ) {
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
                /* PCI_SUBSYSTEM_IDS supported */
		while ((pdev = pci_find_subsys(QLA2X00_VENDOR_ID,
			bdp->device_id, PCI_ANY_ID, PCI_ANY_ID, pdev) )) {
			if (pci_enable_device(pdev))
				continue;
#else
		while ((pdev = pci_find_device(QLA2X00_VENDOR_ID,
			bdp->device_id, pdev ) ))  {
#endif /* 2,3,18 */
#else  /* less than 2,1,95 */
		while (!(pcibios_find_device(QLA2X00_VENDOR_ID, bdp->device_id,
			index++, &pci_bus, &pci_devfn)) )  {
#endif /* 2,1,95 */

                /* found a adapter */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
			printk(
			    "qla2x00: Found  VID=%x DID=%x SSVID=%x SSDID=%x\n",
			pdev->vendor, pdev->device,
			pdev->subsystem_vendor, pdev->subsystem_device);
                        subsystem_vendor = pdev->subsystem_vendor;
                        subsystem_device = pdev->subsystem_device;

		  /* If it's an XXX SubSys Vendor ID adapter, skip it. */
                 /* if (pdev->subsystem_vendor == PCI_VENDOR_ID_XXX)
                    {
				printk(
				    "qla2x00: Skip XXX SubSys Vendor ID "
				    "Controller\n");
				continue;
			}
			*/
#else
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
		  pci_read_config_word(pdev, PCI_SUBSYSTEM_VENDOR_ID,
                                       &subsys);
                  subsystem_vendor = subsys;
		  pci_read_config_word(pdev, PCI_SUBSYSTEM_ID,
                                       &subsys);
                  subsystem_device = subsys;

		  /* Bypass all XXX SUBSYS VENDOR IDs */
                 /* if (subsys == PCI_VENDOR_ID_XXX)
                    {
				printk(
				    "qla2x00: Skip XXX SubSys Vendor ID "
				    "Controller\n");
				continue;
		        }
			*/
#endif /* 2,1,95 */
#endif /* 2,3,18 */

                if( (host = scsi_register(template, sizeof(scsi_qla_host_t))) == NULL )
                {
			printk("qla2x00: couldn't register with scsi layer\n");
			return( 0 );
		}

                ha = (scsi_qla_host_t *) host->hostdata;

                /* Clear our data area */
		memset((void *)(ha), 0, sizeof(scsi_qla_host_t));

                /* Sanitize the information from PCI BIOS.  */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
                host->irq = pdev->irq;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                host->io_port = (unsigned long) pdev->base_address[0];
#else
                host->io_port = pdev->resource[0].start;
#endif
                ha->pci_bus = pdev->bus->number;
                ha->pci_device_fn = pdev->devfn;
                ha->subsystem_vendor = subsystem_vendor;
                ha->subsystem_device = subsystem_device;
                ha->pdev = pdev;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,4)
			scsi_set_pci_device(host, pdev);
#endif
#else /* kern version <= 2,1,95 */
			pcibios_read_config_byte(pci_bus, pci_devfn,
			    OFFSET(cfgp->interrupt_line), &pci_irq);
			pcibios_read_config_dword(pci_bus, pci_devfn,
			    OFFSET(cfgp->base_port), &piobase);
                host->irq = pci_irq;
                host->io_port = (unsigned int) piobase;
                ha->pci_bus = pci_bus;
                ha->pci_device_fn = pci_devfn;
#endif
                ha->device_id = bdp->device_id;
                host->io_port &= PCI_BASE_ADDRESS_IO_MASK;
                ha->devnum = i;
                if( qla2100_verbose ) {
				printk(
				    "scsi%d: Found a %s @ bus %d, device 0x%x, "
				    "irq %d, iobase 0x%lx\n",
					host->host_no,
				    bdp->bdName, ha->pci_bus,
				    (ha->pci_device_fn & 0xf8) >> 3,
				    host->irq, (unsigned long)host->io_port);
                }

                ha->iobase = (device_reg_t *) host->io_port;
                ha->host = host;

                /* 4.23 Initialize /proc/scsi/qla2x00 counters */
                ha->actthreads = 0;
                ha->qthreads   = 0;
                ha->dump_done  = 0;
		ha->total_isr_cnt = 0;
		ha->total_isp_aborts = 0;
		ha->total_lip_cnt = 0;
		ha->total_dev_errs = 0;
		ha->total_ios = 0;
		ha->total_bytes = 0;

		if( qla2100_mem_alloc(ha) ) {
			printk(KERN_WARNING
			    "scsi%d: [ERROR] Failed to allocate "
			    "memory for adapter\n",host->host_no);
			qla2100_mem_free(ha);
			continue;
		}

                ha->prev_topology = 0;
                ha->ports = bdp->numPorts;
                ha->host_no = host->host_no;

                ha->max_targets = MAX_TARGETS_2200;

                /* load the F/W, read paramaters, and init the H/W */
                ha->instance = num_hosts;

#if MPIO_SUPPORT
		if( qla2100_failover )
                	ha->flags.failover_enabled = 1;
#else
               	ha->flags.failover_enabled = 0;
#endif

 		/*
		 * These locks are used to prevent more than one CPU
		 * from modifying the queue at the same time. The
		 * higher level "io_request_lock" will reduce most
		 * contention for these locks.
		 */
		ha->mbx_bits_lock = SPIN_LOCK_UNLOCKED;
                ha->mbx_reg_lock = SPIN_LOCK_UNLOCKED;
                ha->mbx_q_lock = SPIN_LOCK_UNLOCKED;
                ha->retry_lock = SPIN_LOCK_UNLOCKED;
                ha->failback_lock = SPIN_LOCK_UNLOCKED;
                ha->done_lock = SPIN_LOCK_UNLOCKED;
                ha->abort_q_lock = SPIN_LOCK_UNLOCKED;
                ha->dpc_lock = SPIN_LOCK_UNLOCKED;

                if (qla2100_initialize_adapter(ha) && !(ha->device_flags &
		    DFLG_NO_CABLE)) {

			printk(KERN_WARNING
			    "qla2x00: Failed to initialize adapter\n");
			DEBUG2(printk(
			    "scsi%ld: Failed to initialized adapter"
			    " - Adapter flags %d\n",ha->host_no,ha->device_flags);)
			qla2100_mem_free(ha);
			scsi_unregister(host);
			continue;
		}

                ha->next = NULL;
                /*  Mark preallocated Loop IDs in use. */
                ha->fabricid[SNS_FL_PORT].in_use = TRUE;
                ha->fabricid[FABRIC_CONTROLLER].in_use = TRUE;
                ha->fabricid[SIMPLE_NAME_SERVER].in_use = TRUE;


                /* Register our resources with Linux */
                if( qla2100_register_with_Linux(ha, bdp->numPorts-1) ) {
				printk("scsi%ld: Failed to register our "
				    "resources\n",ha->host_no);
				qla2100_mem_free(ha);
				scsi_unregister(host);
				continue;
			}

			DEBUG2(sprintf(debug_buff,
			    "DEBUG: detect hba %ld at address = %p\n",
			    ha->host_no,ha);)
			DEBUG2(qla2100_print(debug_buff);)

			reg = ha->iobase;
			/* Disable ISP interrupts. */
			qla2100_disable_intrs(ha);

		/*
                * Startup the kernel thread for this host adapter
                */
#ifdef ISP2200
                ha->dpc_notify = &qla2200_detect_sem;
#endif
#ifdef ISP2300
                ha->dpc_notify = &qla2300_detect_sem;
#endif

                kernel_thread((int (*)(void *))qla2100_do_dpc,
                (void *) ha, 0);

			/*
			 * Now wait for the kernel dpc thread to initialize
			 * and go to sleep.
 			 */
#ifdef ISP2200
			down(&qla2200_detect_sem);
#endif
#ifdef ISP2300
			down(&qla2300_detect_sem);
#endif
			ha->dpc_notify = NULL;

			/* Insure mailbox registers are free. */
			WRT_REG_WORD(&reg->semaphore, 0);
			WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
			WRT_REG_WORD(&reg->host_cmd, HC_CLR_HOST_INT);


#if MPIO_SUPPORT
		/* if failover is enabled read the user configuration */
                if(ha->flags.failover_enabled) {
		 	if( ConfigRequired > 0 )
				mp_config_required = 1;
			else
				mp_config_required = 0;

			DEBUG(printk(
			    "qla2x00_detect: qla2x00_cfg_init for hba %ld \n",
			    ha->instance);)

			qla2x00_cfg_init(ha);
		}
#endif

                /* Enable chip interrupts. */
                qla2100_enable_intrs(ha);

                /* Insert new entry into the list of adapters */
                ha->next = NULL;

                if( qla2100_hostlist == NULL ) {
                    qla2100_hostlist = ha;
                } else {
                    cur_ha = qla2100_hostlist;

                    while( cur_ha->next != NULL )
                        cur_ha = cur_ha->next;

                    cur_ha->next = ha;
                }

                /* v2.19.5b6                                              */
		/*
		 * Wait around max loop_reset_delay secs for the devices to come
		 * on-line. We don't want Linux scanning before we
		 * are ready.
		 */
	        for (wait_switch = jiffies + (ha->loop_reset_delay * HZ);
                        wait_switch > jiffies  &&
                        !(ha->device_flags & DFLG_NO_CABLE) &&
                        !(ha->device_flags & SWITCH_FOUND) ;) {
                     qla2100_check_fabric_devices(ha);
                }

                /* List the target we have found */
                if( !ha->flags.failover_enabled )
                	qla2100_display_fc_names(ha);

                ha->init_done = 1;
                num_hosts++;
            }
        } /* end of FOR */

#if MPIO_SUPPORT
        qla2x00_cfg_display_devices();
#endif

        LEAVE("qla2100_detect");
        return num_hosts;
}

/**************************************************************************
*   qla2100_register_with_Linux
*
* Description:
*   Free the passed in Scsi_Host memory structures prior to unloading the
*   module.
*
* Input:
*     ha - pointer to host adapter structure
*     maxchannels - MAX number of channels.
*
* Returns:
*  0 - Sucessfully reserved resources.
*  1 - Failed to reserved a resource.
**************************************************************************/
STATIC uint8_t  qla2100_register_with_Linux(scsi_qla_host_t *ha,
    uint8_t maxchannels)
{

        struct Scsi_Host *host = ha->host;
        char drvname[9];

        host->can_queue = 0xfffff;  /* unlimited  */
        host->cmd_per_lun = 1;
        host->select_queue_depths = qla2100_select_queue_depth;
        host->n_io_port = 0xFF;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
        host->base = (unsigned char *) ha->mmpbase;
#else
        host->base = (u_long) ha->mmpbase;
#endif
        host->max_channel = maxchannels;
        /* fix: 07/31 host->max_lun = MAX_LUNS-1; */
        host->max_lun = ha->max_luns;
        host->unique_id = ha->instance;
        host->max_id = ha->max_targets;

        /* set our host ID  (need to do something about our two IDs) */
        host->this_id = 255;

        /* Register the I/O space with Linux */
        if( check_region(host->io_port, 0xff) ) {
            printk(KERN_WARNING
              "qla2x00 : Failed to reserved i/o base region 0x%04lx-0x%04lx already in use\n",
            host->io_port, host->io_port + 0xff);
            /* 6/15/01 - free_irq(host->irq, NULL); */
            return 1;
        }

        request_region(host->io_port, 0xff, drvname);

        /* ER# 4368 */
        sprintf(drvname,"qla2x00#%02d",host->unique_id);
        /* Register the IRQ with Linux (sharable) */
        if( request_irq(host->irq, qla2100_intr_handler, SA_INTERRUPT| SA_SHIRQ, DRIVER_NAME, ha) ) {
            printk(KERN_WARNING
              "qla2x00 : Failed to reserved interrupt %d already in use\n", host->irq);
            return 1;
        }

        /* Initialized the timer */
        START_TIMER(qla2100_timer,ha,WATCH_INTERVAL);

        return 0;
}


/**************************************************************************
*   qla2100_release
*
* Description:
*   Free the passed in Scsi_Host memory structures prior to unloading the
*   module.
*
* Input:
*     ha - pointer to host adapter structure
*
* Returns:
*  0 - Always returns good status
**************************************************************************/
int
qla2100_release(struct Scsi_Host *host)
{
        scsi_qla_host_t *ha = (scsi_qla_host_t *) host->hostdata;
#if  QL_TRACE_MEMORY
	int t;
#endif

        ENTER("qla2100_release");

        /* turn-off interrupts on the card */
	if (ha->interrupts_on)
		qla2100_disable_intrs(ha);

        /* Detach interrupts */
        if( host->irq )
            free_irq(host->irq, ha);

        /* release io space registers  */
        if( host->io_port )
            release_region(host->io_port, 0xff);

        /* Disable timer */
        if( ha->timer_active )
            STOP_TIMER(qla2100_timer,ha)

            /* Kill the kernel thread for this host */
            if( ha->dpc_handler != NULL ) {
#ifdef ISP2200
                ha->dpc_notify = &qla2200_detect_sem;
#endif
#ifdef ISP2300
                ha->dpc_notify = &qla2300_detect_sem;
#endif
#if 0
                send_sig(SIGKILL, ha->dpc_handler, 1);
#else
                send_sig(SIGHUP, ha->dpc_handler, 1);
#endif
#ifdef ISP2200
                down(&qla2200_detect_sem);
#endif
#ifdef ISP2300
                down(&qla2300_detect_sem);
#endif
                ha->dpc_notify = NULL;
            }
#if USE_FLASH_DATABASE
            /* Move driver database to flash, if enabled. */
            if( ha->flags.enable_flash_db_update && ha->flags.updated_fc_db ) {
                ha->flags.updated_fc_db = FALSE;
                qla2100_save_database(ha);
            }
#endif
#if MEMORY_MAPPED_IO
            if( ha->mmpbase ) {
                iounmap((void *) (((unsigned long) ha->mmpbase) & PAGE_MASK));
            }
#endif /* MEMORY_MAPPED_IO */

#if  APIDEV
            apidev_cleanup();
#endif
            qla2100_mem_free(ha);
#if MPIO_SUPPORT
		if (qla2100_failover)
            		qla2x00_cfg_mem_free(ha);
#endif

        if( qla2100_buffer != NULL ) {
            kfree(qla2100_buffer);
            qla2100_buffer_size = 0;
        }

#if QL_TRACE_MEMORY
	for( t = 0; t < 1000; t++ ) {
		if ( mem_trace[t] == 0L )
			continue;
         printk("mem_trace[%d]=%lx, %lx\n",
		t, mem_trace[t],mem_id[t]);
	}
#endif
            ha->flags.online = FALSE;

            LEAVE("qla2100_release");
            return(0);
}

/**************************************************************************
*   qla2100_info
*
* Description:
*
* Input:
*     host - pointer to Scsi host adapter structure
*
* Returns:
*     Return a text string describing the driver.
**************************************************************************/
const char *
qla2100_info(struct Scsi_Host *host)
{
	static char qla2100_buffer[255];
	char *bp;
	scsi_qla_host_t *ha;
	qla_boards_t   *bdp;

#if  APIDEV
	/* We must create the api node here instead of qla2100_detect
	 * since we want the api node to be subdirectory of /proc/scsi/qla2x00
	 * which will not have been created when qla2100_detect exits,
	 * but which will have been created by this point.
	 */

	apidev_init(host);
#endif

	bp = &qla2100_buffer[0];
	ha = (scsi_qla_host_t *)host->hostdata;
	bdp = &QLBoardTbl_fc[ha->devnum];
	memset(bp, 0, sizeof(qla2100_buffer));

#if MPIO_SUPPORT
	sprintf(bp, "QLogic %sPCI to Fibre Channel Host Adapter: bus %d "
	    "device %d irq %d\n"
	    "        Firmware version: %2d.%02d.%02d, Driver version %s-fo",
	    (char *)&bdp->bdName[0], ha->pci_bus,
	    (ha->pci_device_fn & 0xf8) >> 3, host->irq,
	    bdp->fwver[0], bdp->fwver[1], bdp->fwver[2], QLA2100_VERSION);
#else
	sprintf(bp, "QLogic %sPCI to Fibre Channel Host Adapter: bus %d "
	    "device %d irq %d\n"
	    "        Firmware version: %2d.%02d.%02d, Driver version %s",
	    (char *)&bdp->bdName[0], ha->pci_bus,
	    (ha->pci_device_fn & 0xf8) >> 3, host->irq,
	    bdp->fwver[0], bdp->fwver[1], bdp->fwver[2], QLA2100_VERSION);
#endif

	return(bp);
}

/**************************************************************************
*   qla1200_queuecommand
*
* Description:
*     Queue a command to the controller.
*
* Input:
*     cmd - pointer to Scsi cmd structure
*     fn - pointer to Scsi done function
*
* Returns:
*   0 - Always
*
* Note:
* The mid-level driver tries to ensures that queuecommand never gets invoked
* concurrently with itself or the interrupt handler (although the
* interrupt handler may call this routine as part of request-completion
* handling).
**************************************************************************/
int
qla2100_queuecommand(Scsi_Cmnd *cmd, void (*fn)(Scsi_Cmnd *))
{
	fc_port_t	*fcport;
	os_lun_t	*lq;
	os_tgt_t	*tq;
	scsi_qla_host_t	*ha, *ha2;
	srb_t		*sp;
	struct Scsi_Host	*host;

	uint32_t	b, t, l;
#if  BITS_PER_LONG <= 32
	uint32_t	handle;
#else
	u_long		handle;
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	unsigned long	cpu_flags = 0;
#endif

	ENTER("qla2100_queuecommand");
	COMTRACE('C')

	host = cmd->host;
	ha = (scsi_qla_host_t *) host->hostdata;

	/* Get our previous SCSI request pointer */
	sp = (srb_t *) CMD_SP(cmd);
	cmd->scsi_done = fn;


	DEBUG5(qla2100_print_scsi_cmd(cmd);)
	DEBUG5(printk("qla2100_queuecmd: pid=%ld, opcode=%d, timeout=%d\n",
	    cmd->serial_number,cmd->cmnd[0],CMD_TIMEOUT(cmd));)

	/* Generate LU queue on bus, target, LUN */
	b = SCSI_BUS_32(cmd);
	t = SCSI_TCN_32(cmd);
	l = SCSI_LUN_32(cmd);

	if ((tq = (os_tgt_t *) TGT_Q(ha, t)) != NULL &&
	    (lq = (os_lun_t *) LUN_Q(ha, t, l)) != NULL ) {
		fcport = lq->fclun->fcport;
		ha2 = fcport->ha;
	} else {
		lq = NULL;
		fcport = NULL;
		ha2 = ha;
	}

	/* Set an invalid handle until we issue the command to ISP */
	/* then we will set the real handle value.                 */
	handle = INVALID_HANDLE;
	CMD_HANDLE(cmd) = (unsigned char *)handle;

	/*
	 * Allocate a command packet if necessary. If we can't
	 * get one from the lun pool or the hba pool then
	 * let the scsi layer come back later.
	 */
	if(sp == NULL){
		if( (sp = qla2100_allocate_sp(ha,cmd)) == NULL){
			CMD_RESULT(cmd) = DID_RETRY << 16;
			printk(KERN_WARNING "qla2100_command: Couldn't "
			    "allocate memory for sp - retried.\n");
			    (*(cmd)->scsi_done)(cmd);
			LEAVE("qla2100_queuecommand");
			return(0);
		}
	}
	sp->cmd =  cmd;

	DEBUG5(printk("scsi(%2d:%2d:%2d): (queuecmd) queue sp = %p, "
	    "flags=0x%x fo retry=%d, pid=%ld, cmd flags= 0x%x\n",
	    (int)ha->host_no,t,l,sp,sp->flags,sp->fo_retry_cnt,
	    cmd->serial_number,cmd->flags);)

	/* Bookkeeping information */
	sp->r_start = jiffies;       /* time the request was recieved */
	sp->u_start = 0;

	/* Setup device queue pointers. */
	sp->tgt_queue = tq;
	sp->lun_queue = lq;

	/*
	 * NOTE : q is NULL
	 * 1. When device is added from persistent binding
	 *    but has not been discovered yet.The state of
	 *    loopid == PORT_AVAIL.
	 * 2. When device is never found on the bus.(loopid == UNUSED)
	 *
	 * IF Device Queue is not created, or device is not in
	 * a valid state and link down error reporting is enabled,
	 * reject IO.
	 */
	if (fcport == NULL) {
		DEBUG3(printk("scsi(%2d:%2d:%2d): port unavailable\n",
		    (int)ha->host_no,t,l);)

		CMD_RESULT(cmd) = DID_NO_CONNECT << 16;
		qla2100_callback(ha,cmd);

		/* LEAVE("qla2100_queuecommand"); */
		return(0);
	}

	if (l >= ha->max_luns) {
		CMD_RESULT(cmd) = DID_NO_CONNECT << 16;
		qla2100_callback(ha,cmd);
		LEAVE("qla2100_queuecommand");
		return(0);
	}

	/* reset flags if new command */
	if( cmd->flags == 0 &&
	    !(sp->flags & SRB_BUSY) &&	/* 10/20/00 */
	    cmd->retries == 0 ) {
		sp->flags = 0;
	}
	sp->flags &= ~SRB_ISP_COMPLETED;

	sp->fclun = lq->fclun;
	sp->ha = ha2;

	/*  Check destination HBA and port on backend */
	if ((ha2->loop_down_timer == 0 && ha2->flags.link_down_error_enable &&
	    ha2->loop_state == LOOP_DOWN) ||
	    (fcport->loop_id == FC_NO_LOOP_ID) ||
	    (ha2->flags.link_down_error_enable && fcport->state != FC_ONLINE)){

		CMD_RESULT(cmd) = DID_NO_CONNECT << 16;

		if( fcport->loop_id == FC_NO_LOOP_ID)  {
			sp->err_id = 1;

			DEBUG(printk("scsi(%2d:%2d:%2d): port is down - "
			    "pid=%ld, sp=%p loopid=0x%x queued to scsi%d "
			    "done_q\n", (int)ha2->host_no,t,l,
			    cmd->serial_number,sp,fcport->loop_id,
			    (int)ha->host_no);)
		} else {
			sp->err_id = 2;
			if( fcport->state != FC_ONLINE )
				sp->err_id = 3;
			DEBUG(printk("scsi(%2d:%2d:%2d): loop is down - "
			    "pid=%ld, sp=%p loop_id=0x%x queued to scsi%d "
			    "done_q \n", (int)ha2->host_no,t,l,
			    cmd->serial_number,sp,fcport->loop_id,
			    (int)ha->host_no);)
			DEBUG(printk("scsi(%2d:%2d:%2d): port state=%d, loop "
			    "state = %d, lq flags = 0x%x\n", (int)ha2->host_no,
			    t,l,fcport->state,ha2->loop_state,lq->q_flag);)
		}

		ha->flags.done_requests_needed = TRUE;
		qla2100_done_q_put(ha, sp, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last);

		LEAVE("qla2100_queuecommand");
		return (0);
	}

	/*
	* SCSI Kluge
	* ========
	* Whenever, we need to wait for an event such as loop down
	* (i.e. loop_down_timer ) or port down (i.e. LUN request qeueue is
	* suspended) then we will recycle new commands back to the SCSI layer.
	* We do this because this is normally a temporary condition and we don't
	* want the mid-level scsi.c driver to get upset and start aborting
	* commands.
	* The timeout value is extracted from the command minus 1-second
	* and put on a retry queue (watchdog). Once the command timeout it
	* is returned to the mid-level with a BUSY status, so the mid-level
	* will retry it. This process continues until the LOOP DOWN time
	* expires or the condition goes away.
	*/
	if( ha2->loop_down_timer || ha->cfg_active ||
	    ha2->loop_state == LOOP_DOWN || PORT_DOWN(fcport) > 0 ||
	    fcport->login_retry  > 0 ||
	    ABORTS_ACTIVE ||
	    /* ((ha->device_flags & LOGIN_RETRY_NEEDED)  &&
	    ha2->loop_state != LOOP_READY) || */
	    (ha2->loop_state != LOOP_READY) ||
	    (sp->flags & SRB_FAILOVER) ||
	    (lq->q_flag & QLA2100_QSUSP) ) {

		/* Insert command into watchdog queue */
		qla2100_timeout_insert(ha, sp);
		sp->ha = ha;

		LEAVE("qla2100_queuecommand");
		return (0);
	} else {
	  sp->flags &= ~SRB_BUSY;   /* v5.21b16 */
	}


	/* Set retry count if this is a new command */
	if( sp->flags == 0 && !(lq->q_flag & QLA2100_QSUSP) ) {
		sp->retry_count = ha->retry_count;

		if ( PORT_DOWN_RETRY(fcport) == 0 )
			sp->port_down_retry_count = ha->port_down_retry_count;
	}

	/* No timeout necessary, because the upper layer is doing it for us */
	sp->wdg_time = 0;

	DEBUG3(printk("pid=%ld ",cmd->serial_number);)

	/* add the command to the lun queue */
	ha->qthreads++;
	qla2100_putq_b(lq, sp);

	/* First start cmds for this lun if possible */
	qla2x00_next(tq, lq);

	/* start any other cmds for loop down conditions */
	qla2x00_restart_queues(ha,FALSE);

	COMTRACE('c')
	LEAVE("qla2100_queuecommand");
	return (0);
}

/**************************************************************************
*   qla1200_abort
*
* Description:
*     Abort the specified SCSI command.
*
* Input:
*     cmd - pointer to Scsi cmd structure
*
* Returns:
**************************************************************************/
int
qla2100_abort(Scsi_Cmnd *cmd)
{
	int		found = 0;
	int		i;
	int		return_status = SCSI_ABORT_SUCCESS;
	os_lun_t	*q;
	ql_list_link_t	*link;
	ql_list_link_t	*next_link;
	scsi_qla_host_t *ha;
	scsi_qla_host_t *vis_ha, *orig_ha;
	srb_t		*sp;
	srb_t		*rp, *rp_next;
	struct Scsi_Host *host;
	u_long		handle;
	uint8_t		discard;
	uint16_t	int_bit;
	uint16_t	data;
	uint32_t	b, t, l;
	unsigned long	cpu_flags = 0;

	ENTER("qla2100_abort");
	COMTRACE('A')

	vis_ha = (scsi_qla_host_t *) cmd->host->hostdata;
	/* Get the SCSI request ptr */
	sp = (srb_t *) CMD_SP(cmd);
	ha = (scsi_qla_host_t *)sp->ha;
	host = ha->host;

#if DEBUG_GET_FW_DUMP
	if (ha->device_id == QLA2300_DEVICE_ID ||
	    ha->device_id == QLA2312_DEVICE_ID) {
		if (ha->dump_done != 1) {
			DEBUG(printk("\nqla2100_abort handle=%x: >>>> DUMP "
			    "2300 FW <<<<\n",CMD_HANDLE(cmd));)

			qla2300_dump_isp(ha);
			ha->dump_done = 1;
		}
	}
#endif

	DRIVER_LOCK

	/*
	 * if the handle is NULL then we already completed the command.
	 * We always give the handle a value of "INVALID_HANDLE" when
	 * we received it.
	 */
	if( (u_long)   CMD_HANDLE(cmd) == 0L ) {
		DRIVER_UNLOCK
#if  STOP_ON_ABORT
		qla2100_panic("qla2100_abort",ha->host);
#endif
        printk("qla2x00_abort: INVALID HANDLE FOUND!!!\n");
        /* no action - we don't have command */
		return(SCSI_ABORT_NOT_RUNNING);
	}

	/* Check for a pending interrupt. */
#if defined(ISP2100) || defined(ISP2200)
	data = RD_REG_WORD(&ha->iobase->istatus);
	int_bit = RISC_INT;
#else
	data = RD_REG_WORD(&ha->iobase->host_status_lo);
	int_bit = HOST_STATUS_INT;
#endif

	if (!(ha->flags.in_isr) && (data & int_bit))
		qla2100_isr(ha, data, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last, &discard);

	/*
	 * if no LUN queue then something is very wrong!!!
	 */
	handle = (u_long) CMD_HANDLE(cmd);

	/* Generate LU queue on bus, target, LUN */
	b = SCSI_BUS_32(cmd);
	t = SCSI_TCN_32(cmd);
	l = SCSI_LUN_32(cmd);

	if( (q = GET_LU_Q(vis_ha, t, l)) == NULL ) {
		COMTRACE('a')

		DRIVER_UNLOCK
        printk("qla2x00 (%d:%d:%d): No LUN queue for the "
		    "specified device\n",(int)b,(int)t,(int)l);

		/* no action - we don't have command */
		return(SCSI_ABORT_NOT_RUNNING);
	}

#if AUTO_ESCALATE_ABORT
	if( (sp->flags & SRB_ABORTED) ) {
		DRIVER_UNLOCK
		DEBUG2(printk("qla2100_abort: Abort escalayted - returning "
		    "SCSI_ABORT_SNOOZE.\n\r"));
		return(SCSI_ABORT_SNOOZE);
	}
#endif

	/*
	 * if the command ahs a abort pending then tell the upper layer
	 */
	if( (sp->flags & SRB_ABORT_PENDING) ) {
		COMTRACE('a')
		DRIVER_UNLOCK

		if( qla2100_verbose )
			printk("scsi(): Command has a pending abort message - "
			    "ABORT_PENDING.\n");

        printk("qla2100: Command has a pending abort message - "
            "ABORT_PENDING.\n\r");

		return(SCSI_ABORT_PENDING);
	}

        printk("scsi(%ld): ABORTing command= 0x%x, "
	    "jiffies = 0x%lx, timeout=%x, dpc_flags=%x, vis_ha->dpc_flags=%x\n",
	    ha->host_no,(int)cmd,jiffies,CMD_TIMEOUT(cmd)/HZ, ha->dpc_flags,
        vis_ha->dpc_flags);
        DEBUG2(qla2100_print_scsi_cmd(cmd);)
        DEBUG2(qla2100_print_q_info(q);)

	/*
	 * Clear retry queue
	 */
	QLA_RETRY_Q_LOCK(vis_ha, cpu_flags);
	for( rp = vis_ha->retry_q_first; (rp); rp = rp_next ) {
		rp_next = rp->s_next;

		if( sp != rp )
			continue;

        printk("ABORTing from retry queue - SP = %p\n",sp);

		qla2100_timeout_remove(vis_ha, sp);

		CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
		CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;

		qla2100_done_q_put(vis_ha, sp, (srb_t **)&vis_ha->done_q_first,
		  (srb_t **)&vis_ha->done_q_last);

		found++;
		break;
	}
	QLA_RETRY_Q_UNLOCK(vis_ha, cpu_flags);

	/*
	 * Clear failover queue
	 */
	if (!found) {
		orig_ha = sp->tgt_queue->ha;
		for (link = orig_ha->failover_queue.first; link != NULL;
		    link = next_link) {
			next_link = link->next;
			rp = link->base_address;

			if( sp != rp )
				continue;

			printk("ABORTing from failover queue - SP = %p\n",sp);

			/* Remove srb from failover queue. */
			qla2x00_remove_link(&orig_ha->failover_queue,
			    &rp->s_cmd);

			CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
			CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;

			qla2100_done_q_put(orig_ha, sp,
			    (srb_t **)&orig_ha->done_q_first,
			    (srb_t **)&orig_ha->done_q_last);

			found++;
			break;
		}
	}

	/*
	 * Our SP pointer points at the command we want to remove from the
	 * LUN queue providing we haven't already sent it to the adapter.
	 */
	if( found ) {
		return_status = SCSI_ABORT_PENDING;

	} else if( !(sp->flags & SRB_DMA_VALID) ) {
		found++;

        printk("qla2100: Cmd in LUN queue aborted serial_number "
            "%ld.\n\r", sp->cmd->serial_number);

		/* Remove srb from SCSI LU queue. */
		qla2100_removeq(q, sp);

		sp->flags |=  SRB_ABORTED;
		CMD_RESULT(cmd) = DID_ABORT << 16;
		qla2100_done_q_put(ha, sp, &ha->done_q_first, &ha->done_q_last);

		return_status = SCSI_ABORT_PENDING;

	} else {  /* find the command in our active list */
		for( i = 0; i < MAX_OUTSTANDING_COMMANDS; i++ ) {
			if (sp != ha->outstanding_cmds[i])
				continue;

			found++;
            printk("qla2100(%ld): RISC aborting pid "
			    "%ld sp->state=%x\n",
                ha->host_no,sp->cmd->pid,sp->state);

			qla2x00_abort_q_put(ha, sp,
			    (srb_t **)&ha->abort_q_first,
			    (srb_t **)&ha->abort_q_last);

			ha->dpc_flags |= SRB_ABORT_NEEDED;
			return_status = SCSI_ABORT_PENDING;

			break;
		}
	}

#if  STOP_ON_ABORT
	qla2100_panic("qla2100_abort",ha->host);
#endif
	if( found == 0 )
		return_status = SCSI_ABORT_NOT_RUNNING;  /* no action */

    printk("qla2100_abort: Aborted status returned = 0x%x.\n\r",
        return_status);

	/*
	 * Complete any commands
	 */
	if (ha->done_q_first)
		qla2100_done(ha, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last);

	if ((q->q_flag & LUN_QUEUE_SUSPENDED))
		q->q_flag = q->q_flag & ~LUN_QUEUE_SUSPENDED;

	if (found) {
		qla2x00_restart_queues(vis_ha, TRUE);
	}

	DRIVER_UNLOCK

	LEAVE("qla2100_abort");
	COMTRACE('a')
	return(return_status);
}

/**************************************************************************
* qla1200_reset
*
* Description:
*    The reset function will reset the SCSI bus and abort any executing
*    commands.
*
* Input:
*      cmd = Linux SCSI command packet of the command that cause the
*            bus reset.
*      flags = SCSI bus reset option flags (see scsi.h)
*
* Returns:
*      DID_RESET in cmd.host_byte of aborted command(s)
*
* Note:
*      Resetting the bus always succeeds - is has to, otherwise the
*      kernel will panic! Try a surgical technique - sending a BUS
*      DEVICE RESET message - on the offending target before pulling
*      the SCSI bus reset line.
**************************************************************************/
int
qla2100_reset(Scsi_Cmnd *cmd, unsigned int flags)
{
	typedef enum {
	    ABORT_DEVICE = 1,
	    DEVICE_RESET = 2,
	    BUS_RESET    = 3,
	    ADAPTER_RESET= 4,
	    RESET_DELAYED= 5,
	    FAIL         = 6
	} action_t;
	action_t	action = ADAPTER_RESET;
	int		result;
	os_lun_t	*q;
	scsi_qla_host_t	*ha;
	srb_t		*sp;
	uint16_t	data;
	uint16_t	int_bit;
	uint32_t        b, t, l;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	unsigned long	cpu_flags = 0;
#endif
	uint8_t		discard;
	fc_port_t	*fcport;


	COMTRACE('R')

	if (cmd == NULL) {
		printk(KERN_WARNING
		    "(scsi?:?:?:?) Reset called with NULL Scsi_Cmnd "
		    "pointer, failing.\n");
		return(SCSI_RESET_SNOOZE);
	}

	ha = (scsi_qla_host_t *) cmd->host->hostdata;
	DEBUG3(printk("qla2100_reset(%ld): entered.\n", ha->host_no);)

#if  STOP_ON_RESET
	sprintf(debug_buff,"ABORTing command= 0x%x\n",(int)cmd);
	qla2100_print(debug_buff);
	qla2100_print_scsi_cmd(cmd);
	qla2100_panic("qla2100_reset",ha->host);
#endif

	DRIVER_LOCK

	/* Check for pending interrupts. */
#if defined(ISP2300)
	data = RD_REG_WORD(&ha->iobase->host_status_lo);
	int_bit = HOST_STATUS_INT;
#else
	data = RD_REG_WORD(&ha->iobase->istatus);
	int_bit = RISC_INT;
#endif
	if (!(ha->flags.in_isr) && (data & int_bit)) {
		qla2100_isr(ha, data, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last, &discard);
	}

	DRIVER_UNLOCK

	/*
	 * Determine the suggested action that the mid-level driver wants
	 * us to perform.
	 */
	if( CMD_HANDLE(cmd) == (unsigned char *) 0 ) {
		/*
		 * if mid-level driver called reset with a orphan SCSI_Cmnd
		 * (i.e. a command that's not pending ), so perform the
		 * function specified.
		 */
		/* 4.23 */
		if( (flags & SCSI_RESET_SUGGEST_HOST_RESET) )
			action = ADAPTER_RESET;
		else if( flags & SCSI_RESET_SUGGEST_BUS_RESET )
			action = BUS_RESET;
		else
			action = DEVICE_RESET;
	} else {
		/*
		 * Mid-level driver has called reset with this SCSI_Cmnd and
		 * its pending.
		 */
		if( flags & SCSI_RESET_SUGGEST_HOST_RESET )
			action = ADAPTER_RESET;
		else if( flags & SCSI_RESET_SUGGEST_BUS_RESET )
			action = BUS_RESET;
		else
			action = DEVICE_RESET;
	}

	b = SCSI_BUS_32(cmd);
	t = SCSI_TCN_32(cmd);
	l = SCSI_LUN_32(cmd);
	DEBUG2(printk("qla2x00 (%d:%d:%d): Reset() device\n",
	    (int)b, (int)t, (int)l);)

	sp = (srb_t *) CMD_SP(cmd);
	q = GET_LU_Q(ha, t, l);

	/*
	 *  By this point, we want to already know what we are going to do,
	 *  so we only need to perform the course of action.
	 */

	DRIVER_LOCK
	result = SCSI_RESET_ERROR;

	switch( action ) {
	case FAIL:
		break;

	case RESET_DELAYED:
		break;

	case ABORT_DEVICE:
		if( qla2100_verbose )
			printk(KERN_INFO
			    "scsi(%d:%d:%d:%d): going to issue ABORT DEVICE.\n",
			    (int)ha->host_no,(int)b,(int)t,(int)l);

		DEBUG2(printk(
		    "scsi(%d:%d:%d:%d): going to issue ABORT DEVICE.\n",
		    (int)ha->host_no,(int)b,(int)t,(int)l);)

		qla2100_abort_lun_queue(ha,b,t,l,DID_ABORT);

		if (!ha->loop_down_timer) {
			ha->dpc_flags |= DEVICE_ABORT_NEEDED;
			ha->reset_bus_id = b;
			ha->reset_tgt_id = t;
			ha->reset_lun = l;

			DEBUG3(printk("device abort: going to wake up dpc.\n");)

			result = SCSI_RESET_PENDING;
			fcport = q->fclun->fcport;
		}
		break;

	case DEVICE_RESET:
		if( qla2100_verbose )
			printk(KERN_INFO
			    "scsi(%d:%d:%d:%d): going to issue DEVICE RESET.\n",
			    (int) ha->host_no,(int)b,(int)t,(int)l);

		DEBUG2(printk(
		    "scsi(%d:%d:%d:%d): going to issue DEVICE RESET.\n",
		    (int) ha->host_no,(int)b,(int)t,(int)l);)

		for (l = 0; l < ha->max_luns; l++)
			qla2100_abort_lun_queue(ha,b,t,l,DID_ABORT);

		if( !ha->loop_down_timer ) {
			ha->dpc_flags |= DEVICE_RESET_NEEDED;
			ha->reset_bus_id = b;
			ha->reset_tgt_id = t;

			DEBUG3(printk("device reset: going to wake up dpc.\n");)

			result = SCSI_RESET_PENDING;
		}
		q->q_flag |= QLA2100_QRESET;
		break;

	case BUS_RESET:
		if( qla2100_verbose )
			printk(KERN_INFO
			    "scsi(%d:%d:%d:%d): going to issue LOOP RESET.\n",
			    (int) ha->host_no,(int)b,(int)t,(int)l);

		DEBUG2(printk(
		    "scsi(%d:%d:%d:%d): going to issue LOOP RESET.\n",
		    (int) ha->host_no,(int)b,(int)t,(int)l);)

		for( t = 0; t < ha->max_targets; t++ )
			for( l = 0; l <  ha->max_luns; l++ )
				qla2100_abort_lun_queue(ha,b,t,l,DID_RESET);

		if( !ha->loop_down_timer ) {
			ha->dpc_flags |= LOOP_RESET_NEEDED;
			DEBUG3(printk("reset: going to wake up dpc.\n");)

			/* if (!ha->dpc_active && ha->dpc_wait)
				up(ha->dpc_wait); */

			result = SCSI_RESET_SUCCESS | SCSI_RESET_BUS_RESET;
		}

		/*
		 * The reset loop routine returns all the outstanding
		 * commands back with "DID_RESET" in the status field.
		 */
		if( flags & SCSI_RESET_SYNCHRONOUS ) {
			CMD_RESULT(cmd) = (int) (DID_BUS_BUSY << 16);
			(*(cmd)->scsi_done)(cmd);
		}

		break;

	case ADAPTER_RESET:
	default:
		if( qla2100_verbose ) {
			printk(KERN_INFO
			    "scsi(%d:%d:%d:%d): will issue ADAPTER RESET.\n",
			    (int) ha->host_no,(int)b,(int)t,(int)l);
		}

                DEBUG2(printk(
                    "scsi(%d:%d:%d:%d): going to issue ADAPTER RESET.\n",
                    (int) ha->host_no,(int)b,(int)t,(int)l);)

		/*
		 * We restarted all of the commands automatically, so the
		 * mid-level code can expect completions momentitarily.
		 */
		ha->dpc_flags |= ISP_ABORT_NEEDED;

		result = SCSI_RESET_SUCCESS | SCSI_RESET_HOST_RESET;

		break;
	}

	if( ha->done_q_first ) {
		qla2100_done(ha, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last);
	}

	DEBUG3(printk("qla2100_reset: going to call restart_queues. "
	    "jiffies=%lx.\n", jiffies);)

	qla2x00_restart_queues(ha,TRUE);
	DRIVER_UNLOCK

	COMTRACE('r')

	DEBUG3(printk("qla2100_reset: exiting %d. jiffies=%lx.\n",
	    result, jiffies);)

	return(result);
}

/**************************************************************************
* qla1200_biosparam
*
* Description:
*   Return the disk geometry for the given SCSI device.
**************************************************************************/
int
qla2100_biosparam(Disk *disk, kdev_t dev, int geom[])
{
        int heads, sectors, cylinders;

        heads = 64;
        sectors = 32;
        cylinders = disk->capacity / (heads * sectors);
        if( cylinders > 1024 ) {
            heads = 255;
            sectors = 63;
            cylinders = disk->capacity / (heads * sectors);
        }

        geom[0] = heads;
        geom[1] = sectors;
        geom[2] = cylinders;

        return (0);
}

/**************************************************************************
* qla2100_intr_handler
*
* Description:
*   Handles the actual interrupt from the adapter.
*
* Context: Interrupt
**************************************************************************/
void
qla2100_intr_handler(int irq, void *dev_id, struct pt_regs *regs)
{
        unsigned long mbx_flags = 0;
        scsi_qla_host_t *ha;
        uint16_t    data;
        uint8_t     got_mbx = 0;
        device_reg_t *reg;

        ENTER_INTR("qla2100_intr_handler");
        COMTRACE('I')
        ha = (scsi_qla_host_t *) dev_id;
        if( !ha ) {
            printk(KERN_INFO "qla2100_intr_handler: NULL host ptr\n");
            COMTRACE('X')
            return;
        }
        qla2100_stats.irqhba = ha;

        /* Prevent concurrent access to adapters register */
        spin_lock_irqsave(&io_request_lock, ha->cpu_flags);

        reg = ha->iobase;

        /* Check for pending interrupts. */
#if defined(ISP2100) || defined(ISP2200)
        while ((data = RD_REG_WORD(&reg->istatus)) & RISC_INT)
#else
        while ((data = RD_REG_WORD(&reg->host_status_lo)) & HOST_STATUS_INT)
#endif
	{
		ha->total_isr_cnt++;
        	qla2100_isr(ha, data, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last, &got_mbx);
        }

	if (got_mbx && ha->flags.mbox_int) {
		/* There was a mailbox completion */
		DEBUG3(printk("qla2x00_intr_handler: going to get mbx reg "
		    "lock.\n");)

		QLA_MBX_REG_LOCK(ha);
		got_mbx = 0;

		if (ha->mcp == NULL) {
			DEBUG3(printk("qla2x00_intr_handler: error mbx "
			    "pointer.\n");)
		} else {
			DEBUG3(printk("qla2x00_intr_handler: going to set mbx "
			    "intr flags. cmd=%x.\n",
			    ha->mcp->mb[0]);)
		}

		set_bit(MBX_INTERRUPT, (void *)&(ha->mbx_cmd_flags));

		DEBUG3(printk("qla2x00_intr_handler(%ld): going to wake up mbx "
		    "function for completion.\n", ha->host_no);)

		up(&ha->mbx_intr_sem);

		DEBUG3(printk("qla2x00_intr_handler: going to unlock mbx "
		    "reg.\n");)
		QLA_MBX_REG_UNLOCK(ha);
	}

        if( ha->done_q_first )
        	qla2100_done(ha, (srb_t **)&ha->done_q_first,
		    (srb_t **)&ha->done_q_last);

        spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

/* timer will wake up dpc. */
#if 0
        /* Wakeup the DPC routine */
        if( (!ha->flags.mbox_busy &&
		    ( ha->dpc_flags & (ISP_ABORT_NEEDED |
			   RESET_MARKER_NEEDED | LOOP_RESYNC_NEEDED) ) )
        && ha->dpc_wait && !ha->dpc_active ) {  /* v2.19.4 */
#ifdef	KERNEL_SEM_BUG
            atomic_inc(&dpc_needed);
#else
            up(ha->dpc_wait);
#endif
        }

        if( !ha->dpc_wait  ){
            DEBUG(printk("qla2x00 %d: DPC handler died.\n",(int)ha->host_no));
	}
#endif

        COMTRACE('i')
        LEAVE_INTR("qla2100_intr_handler");
}

/**************************************************************************
* qla2100_do_dpc
*   This kernel thread is a task that is schedule by the interrupt handler
*   to perform the background processing for interrupts.
*
* Notes:
* This task always run in the context of a kernel thread.  It
* is kick-off by the driver's detect code and starts up
* up one per adapter. It immediately goes to sleep and waits for
* some fibre event.  When either the interrupt handler or
* the timer routine detects a event it will one of the task
* bits then wake us up.
**************************************************************************/
void
qla2100_do_dpc(void *p)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
        struct semaphore sem = MUTEX_LOCKED;
#else
        DECLARE_MUTEX_LOCKED(sem);
#endif

        fcdev_t         dev;
	fc_port_t	*fcport;
    	os_lun_t        *q;
        scsi_qla_host_t *ha = (scsi_qla_host_t *) p;
        srb_t           *sp, *sp_next;
        srb_t           *tmp_srb;
        uint8_t         found = 0;
	uint8_t		status;
        uint32_t        t;
        unsigned long	cpu_flags = 0;
        unsigned long	done_flags = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
#ifdef MODULE
        siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
#else
        siginitsetinv(&current->blocked, 0);
#endif
#endif
        lock_kernel();

        /*
        * If we were started as result of loading a module, close all of the
        * user space pages.  We don't need them, and if we didn't close them
        * they would be locked into memory.
        */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
        exit_mm(current);
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)

	/* Flush resources */
        daemonize();

#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
        current->session = 1;
        current->pgrp = 1;
#endif
        /*
        * FIXME(dg) this is still a child process of the one that did
        * the insmod.  This needs to be attached to task[0] instead.
        */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
        siginitsetinv(&current->blocked, SHUTDOWN_SIGS);
#else
        siginitsetinv(&current->blocked, sigmask(SIGKILL));  
#endif
        current->fs->umask = 0;
#endif


#ifdef ISP2200
        /*
         * Set the name of this process.
         */
        sprintf(current->comm, "qla2200_dpc%ld", ha->host_no);
        ha->dpc_wait = &sem;
#endif
#ifdef ISP2300
        sprintf(current->comm, "qla2300_dpc%ld", ha->host_no);
        ha->dpc_wait = &sem;
#endif
        ha->dpc_handler = current;

        unlock_kernel();

        /*
         * Wake up the thread that created us.
         */
        DEBUG(printk("qla2100_dpc: Wake up parent %d\n",
            ha->dpc_notify->count.counter);)

        up(ha->dpc_notify);

#ifdef	KERNEL_SEM_BUG
	atomic_set(&dpc_needed, 0); 	
#endif

        while( 1 ) {
            /*
            * If we get a signal, it means we are supposed to go
            * away and die.  This typically happens if the user is
            * trying to unload a module.
            */
            DEBUG3(printk("qla2x00: DPC handler sleeping\n");)

            LEAVE("qla2100: DPC");
#ifdef	KERNEL_SEM_BUG
		do {
            		if( signal_pending(current) )
                		break;   /* get out */
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(2*HZ);
		} while( atomic_read(&dpc_needed) == 0 );
		atomic_dec(&dpc_needed);

#else
#ifdef ISP2200
            down_interruptible(&sem);
#endif
#ifdef ISP2300
            down_interruptible(&sem);
#endif
#endif

            if( signal_pending(current) )
                break;   /* get out */

            ENTER("qla2100: DPC");
            DEBUG3(printk("qla2x00: DPC handler waking up\n");)

            if (!ha->init_done)
		/* Initialization not yet finished. Don't do anything yet. */
		continue;

            DEBUG3(printk("scsi%d: DPC handler\n",(int)ha->host_no);)
            QLA2100_DPC_LOCK(ha);

            spin_lock_irqsave(&io_request_lock, ha->cpu_flags);
            ha->dpc_active = 1;

                if( ha->done_q_first ) {
                    qla2100_done(ha, (srb_t **)&ha->done_q_first,
                      (srb_t **)&ha->done_q_last);
                }

		if (ha->dpc_flags & COMMAND_WAIT_NEEDED) {
			ha->dpc_flags &= ~COMMAND_WAIT_NEEDED;

			if (!(ha->dpc_flags & COMMAND_WAIT_ACTIVE)) {
				ha->dpc_flags |= COMMAND_WAIT_ACTIVE;

				QLA2100_DPC_UNLOCK(ha);
        			DEBUG(printk("qla%d: waiting on commands\n",
				    (int)ha->host_no);)

				qla2100_cmd_wait(ha);

				DEBUG(printk("qla%d: cmd_wait - done\n",
				    (int)ha->host_no);)
				QLA2100_DPC_LOCK(ha);

				ha->dpc_flags &= ~COMMAND_WAIT_ACTIVE;

			}
		}

            /* Determine what action is necessary */

            /* Flush all commands in watchdog queue */
            if( ha->flags.port_restart_needed ) {

                DEBUG(printk("qla%ld: DPC port restarting - flushing "
                    "all cmds in watchdog queue.\n", ha->instance);)

                ha->flags.port_restart_needed = FALSE;
		QLA_RETRY_Q_LOCK(ha, cpu_flags);

                for( sp = ha->retry_q_first; (sp != NULL); sp =  sp_next ) {
                    sp_next = sp->s_next;
		    q = sp->lun_queue;

                    DEBUG3(printk("qla2100_retry_q: pid=%ld sp=%p, "
                        "spflags=0x%x,q_flag= 0x%x\n",sp->cmd->serial_number,
                        sp,sp->flags,q->q_flag);)

                    if( q == NULL )
                        continue;

                    q->q_flag &= ~QLA2100_QSUSP;
                    qla2100_timeout_remove(ha, sp);

                    DEBUG3(printk("qla2100_retry_q: remove sp - pid=%ld "
                        "spfalgs=0x%x q_flag= 0x%x\n",sp->cmd->serial_number,
                        sp->flags,q->q_flag);)

                    CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
                    CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;
                    qla2100_done_q_put(ha, sp, (srb_t **)&ha->done_q_first,
                        (srb_t **)&ha->done_q_last);
                }
		QLA_RETRY_Q_UNLOCK(ha, cpu_flags);

                /* v2.19 -  We want to wait until the end to
                 * return all requests back to OS.
                 */
                /* deleted: qla2x00_restart_queues(ha, FALSE); */
                ha->flags.restart_queues_needed = TRUE;
            }

            /* Process any pending mailbox commands */

            if (!ha->flags.mbox_busy) {

	    if ( ha->dpc_flags & ISP_ABORT_NEEDED ) {
                    DEBUG(printk("dpc: qla2100_abort_isp ha = %p\n", ha);)
	            ha->dpc_flags &= ~ISP_ABORT_NEEDED;
		    if (!(ha->dpc_flags & ABORT_ISP_ACTIVE)) {
			ha->dpc_flags |= ABORT_ISP_ACTIVE;
			if (qla2100_abort_isp(ha)) {
				/* failed. retry later */
				ha->dpc_flags |= ISP_ABORT_NEEDED;
			}
			ha->dpc_flags &= ~ABORT_ISP_ACTIVE;
		    }
            }

            if (ha->dpc_flags & LOOP_RESET_NEEDED) {

                ha->dpc_flags &= ~LOOP_RESET_NEEDED;

                DEBUG(printk("dpc: loop_reset_needed(%ld) calling "
                  "loop_reset.\n", ha->host_no);)

                qla2x00_loop_reset(ha);
            }

            if (ha->dpc_flags & DEVICE_RESET_NEEDED) {

                ha->dpc_flags &= ~DEVICE_RESET_NEEDED;

                DEBUG(printk("dpc: device_reset_needed(%ld) calling "
                  "device_reset.\n", ha->host_no);)

                qla2x00_device_reset(ha, ha->reset_tgt_id);
            }

            if (ha->dpc_flags & DEVICE_ABORT_NEEDED) {

                ha->dpc_flags &= ~DEVICE_ABORT_NEEDED;

                DEBUG(printk("dpc: device_abort_needed(%ld) calling "
                  "device_abort.\n", ha->host_no);)

                t = ha->reset_tgt_id;
                if (ha->otgt[t] && ha->otgt[t]->vis_port)
                    qla2x00_abort_device(ha, ha->otgt[t]->vis_port->loop_id,
                      ha->reset_lun);
            }

            if (ha->dpc_flags & SRB_ABORT_NEEDED) {
                ha->dpc_flags &= ~SRB_ABORT_NEEDED;

                DEBUG(printk("dpc: srb_abort_needed(%ld) processing "
                  "abort_q.\n", ha->host_no);)

                /* Process all SRBs in abort_q */
                while (ha->abort_q_first) {
                    qla2x00_abort_q_get(ha, &sp);
                    if (sp == NULL) {
                        /* We should not get here. But in case we do,
                         * don't panic the system.
                         */
                        DEBUG2(printk("dpc: No more srb in abort_q.\n");)
                        break;
                    }

                    /* search done queue in case it's finished. */
                    QLA_DONE_Q_LOCK(ha, done_flags);
                    tmp_srb = ha->done_q_first;
                    while (tmp_srb) {
                        if (sp == tmp_srb) {
                            found = 1;
                            break;
                        }
                        tmp_srb = tmp_srb->s_next;
                    }
                    QLA_DONE_Q_UNLOCK(ha, done_flags);

                    if (!found && qla2x00_abort_command(ha, sp)) {
                        DEBUG(printk("dpc: Failed to abort pid %ld.\n",
                          sp->cmd->pid);)
                    }
                }
            }

            if( ha->dpc_flags & RESET_MARKER_NEEDED ) {
		    ha->dpc_flags &= ~RESET_MARKER_NEEDED;
		    if (!(ha->dpc_flags & RESET_ACTIVE)) {
		        ha->dpc_flags |= RESET_ACTIVE;
                        DEBUG(printk("dpc: qla2100_reset_marker \n");)
                        qla2100_rst_aen(ha);
		        ha->dpc_flags &= ~RESET_ACTIVE;
		    }
            }

                /* v2.19.8 Retry each device up to login retry count */
                if ((ha->device_flags & RELOGIN_NEEDED) &&
                  ha->loop_state != LOOP_DOWN) { /* v2.19.5 */

                    DEBUG(printk("dpc: qla2100_login\n");)
                    ha->device_flags &= ~RELOGIN_NEEDED;

	            for (fcport = ha->fcport; fcport != NULL;
			fcport = fcport->next) {

		          if ( fcport->login_retry ) {
		              fcport->login_retry--;
                              dev.d_id.b24 = fcport->d_id.b24;

	                      if ((status = qla2100_fabric_login(ha, &dev)) ==
                                  0) {
                                  DEBUG(printk("dpc: logged in ID %x\n",
                                      fcport->loop_id);)
		                  fcport->login_retry = 0;
	                      } else if ( status == 1 ) {
	                            /* dg 07/05/01  
					ha->dpc_flags |= ISP_ABORT_NEEDED;
					*/
		                        fcport->login_retry = 0;
					/* 07/05 */
				        fcport->state = FC_LOGIN_NEEDED;
	                      } else {
                                  ha->device_flags |= RELOGIN_NEEDED;
				  /* retry the login again */
                                DEBUG(printk("dpc: Retry %d logged in ID %x\n",
                                    fcport->login_retry, fcport->loop_id);)
	                      }
		        }
	            }
                }

                /* v2.19.5 */
                if ( (ha->device_flags & LOGIN_RETRY_NEEDED) &&
                     ha->loop_state != LOOP_DOWN ) { /* v2.19.5 */
                    ha->device_flags &= ~LOGIN_RETRY_NEEDED;
                    DEBUG(printk("dpc: qla2100_login_retry\n");)
		    ha->dpc_flags |= LOOP_RESYNC_NEEDED;
                    DEBUG(printk("dpc: qla2100_login_retry end.\n");)
                }

                /* v2.19.5b5 */
                if( ha->dpc_flags & LOOP_RESYNC_NEEDED ) {

                    DEBUG(printk("dpc: qla2100_LOOP_RESYNC\n");)
		    ha->dpc_flags &= ~LOOP_RESYNC_NEEDED;

		    if (!(ha->dpc_flags & LOOP_RESYNC_ACTIVE)) {
		            ha->dpc_flags |=  LOOP_RESYNC_ACTIVE;
                	    QLA2100_DPC_UNLOCK(ha);
                            qla2100_loop_resync(ha);
                 	    QLA2100_DPC_LOCK(ha);
			    ha->dpc_flags &= ~ LOOP_RESYNC_ACTIVE;

	            }
                    DEBUG(printk("dpc: qla2100_LOOP_RESYNC done\n");)
                }

#if	MPIO_SUPPORT
		if ( !ha->cfg_active ) {
                   QLA_FAILBACK_LOCK(ha, cpu_flags);
		   if ( (ha->failback_flags & FAILOVER_EVENT) ) {
		      ha->failback_flags &= ~FAILOVER_EVENT;
                      QLA_FAILBACK_UNLOCK(ha, cpu_flags);
                      DEBUG(printk("dpc: qla2100_cfg_event_notify\n");)
		      if( ha->flags.online) { /* 11/21/00 */
		  	   qla2x00_cfg_event_notify(ha, ha->failover_type);
			   /* v5.21b16 */
/* Following block of code is moved to qla_cfg.c file.
                	   for( t = 0; t < MAX_TARGETS; t++ ) {
			      for (l = 0; l < MAX_LUNS; l++) {
				if( (q = GET_LU_Q(ha, t, l)) != NULL ){
                                   if (q->q_flag & LUN_FAILBACK) {
    			              qla2x00_suspend_lun(ha, q, recoveryTime, 1);
	                              for (link = ha->failover_queue.first;
	                                 link != NULL; link = next_link) {
		                         next_link = link->next;
		                         rp = link->base_address;
                                         if (q == rp->lun_queue) {
                                            CMD_RESULT(rp->cmd) = DID_BUS_BUSY << 16;
                                            CMD_HANDLE(rp->cmd) = (unsigned char *) NULL;
                                            qla2100_done_q_put(ha, rp, (srb_t **)&ha->done_q_first, (srb_t **)&ha->done_q_last);

	                                 }
				      }
                                      q->q_flag &= ~LUN_FAILBACK;
				   }
			        }
                              }
                           }
*/
                       }
                       DEBUG(printk("\ndpc: qla2100_cfg_event_notify - done\n");)
		    }
                    else
                       QLA_FAILBACK_UNLOCK(ha, cpu_flags);

                    if( (ha->dpc_flags & FAILOVER_NEEDED) ) {
	    	       /* get any requests from failover queue */
               	       ha->dpc_flags &= ~FAILOVER_NEEDED;
                       DEBUG(printk("dpc: qla2100_process failover\n");)
	               qla2x00_process_failover(ha);
                       DEBUG(printk("dpc: qla2100_process failover - done\n");)
                    }
		}
#endif

                if( ha->flags.restart_queues_needed ) {
                    DEBUG(printk("dpc: qla2100_restart_queues\n");)
                    qla2x00_restart_queues(ha,FALSE);
                    DEBUG(printk("dpc: qla2100_restart_queues - done\n");)
                }

                if( ha->flags.abort_queue_needed ) {
                    DEBUG(printk("dpc: qla2100_abort_queues\n");)
                    qla2x00_abort_queues(ha, FALSE);
                }
                if( !ha->interrupts_on )
                    qla2100_enable_intrs(ha);
            }

            if( ha->flags.done_requests_needed )
                ha->flags.done_requests_needed = FALSE;

            if( ha->done_q_first ) {
                qla2100_done(ha, (srb_t **)&ha->done_q_first,
                  (srb_t **)&ha->done_q_last);
            }
            spin_unlock_irqrestore(&io_request_lock, ha->cpu_flags);

            ha->dpc_active = 0;
            QLA2100_DPC_UNLOCK(ha);

            /* The spinlock is really needed up to this point. (DB) */
        }
        DEBUG(qla2100_print("dpc: DPC handler exiting\n");)

        /*
        * Make sure that nobody tries to wake us up again.
        */
        ha->dpc_wait = NULL;
        ha->dpc_handler = NULL;
        ha->dpc_active = 0;

        /*
        * If anyone is waiting for us to exit (i.e. someone trying to unload
        * a driver), then wake up that process to let them know we are on
        * the way out the door.  This may be overkill - I *think* that we
        * could probably just unload the driver and send the signal, and when
        * the error handling thread wakes up that it would just exit without
        * needing to touch any memory associated with the driver itself.
        */
        if( ha->dpc_notify != NULL )
            up(ha->dpc_notify);


}

/**************************************************************************
* qla2100_device_queue_depth
*   Determines the queue depth for a given device.  There are two ways
*   a queue depth can be obtained for a tagged queueing device.  One
*   way is the default queue depth which is determined by whether
*   If it is defined, then it is used
*   as the default queue depth.  Otherwise, we use either 4 or 8 as the
*   default queue depth (dependent on the number of hardware SCBs).
**************************************************************************/
void qla2100_device_queue_depth(scsi_qla_host_t *p, Scsi_Device *device)
{
        int default_depth = 16;

        device->queue_depth = default_depth;
        if( device->tagged_supported ) {
            device->tagged_queue = 1;
            device->current_tag = 0;
#ifdef MODULE
            if( !(ql2xmaxqdepth == 0 || ql2xmaxqdepth > 256) )
              device->queue_depth = ql2xmaxqdepth;
#endif

            printk(KERN_INFO "scsi(%d:%d:%d:%d): Enabled tagged queuing, queue depth %d.\n",
            (int)p->host_no, device->channel, device->id,
            device->lun, device->queue_depth);
        }

}

/**************************************************************************
*   qla2100_select_queue_depth
*
* Description:
*   Sets the queue depth for each SCSI device hanging off the input
*   host adapter.  We use a queue depth of 2 for devices that do not
*   support tagged queueing.
**************************************************************************/
STATIC void
qla2100_select_queue_depth(struct Scsi_Host *host, Scsi_Device *scsi_devs)
{
        Scsi_Device *device;
        scsi_qla_host_t  *p = (scsi_qla_host_t *) host->hostdata;

        ENTER("qla2100_select_queue_depth");
        for( device = scsi_devs; device != NULL; device = device->next ) {
            if( device->host == host )
                qla2100_device_queue_depth(p, device);
        }
        LEAVE("qla2100_select_queue_depth");
}

/**************************************************************************
* ** Driver Support Routines **
*
* qla2100_enable_intrs
* qla2100_disable_intrs
**************************************************************************/
static inline void qla2100_enable_intrs(scsi_qla_host_t *ha)
{
        device_reg_t *reg;

        reg = ha->iobase;
        ha->interrupts_on = 1;
        /* enable risc and host interrupts */
        WRT_REG_WORD(&reg->ictrl, (ISP_EN_INT+ ISP_EN_RISC));
}

static inline void qla2100_disable_intrs(scsi_qla_host_t *ha)
{
        device_reg_t *reg;

        reg = ha->iobase;
        ha->interrupts_on = 0;
        /* disable risc and host interrupts */
        WRT_REG_WORD(&reg->ictrl, 0);
}

/*
 * Copy out of kernel space
 */
int os_copyout(void *src,void *dst,size_t len,int mode)
{
	copy_to_user(dst,src,len);
	return(0);
}

/*
 * Copy into kernel space
 */
int os_copyin(void *src,void *dst, size_t len,int mode)
{
 	copy_from_user(dst,src,len);
	return(0);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)
/**************************************************************************
*  mdelay
*     Delay in milliseconds
*
**************************************************************************/
STATIC inline void
mdelay(int milliseconds)
{
        int i;

        for( i=0; i<milliseconds; i++ ) {
            SYS_DELAY(HZ);
        }
}
#endif

/**************************************************************************
* qla2100_done
*      Process completed commands.
*
* Input:
*      ha           = adapter block pointer.
*      done_q_first = done queue first pointer.
*      done_q_last  = done queue last pointer.
*
* Returns:
*     None
**************************************************************************/
STATIC int
qla2100_done(scsi_qla_host_t *old_ha, srb_t **done_q_first, srb_t **done_q_last)
{
        srb_t           *sp;
	os_tgt_t	*tq;
	os_lun_t	*lq;
        Scsi_Cmnd  *cmd;
        unsigned long done_flags = 0;
    	scsi_qla_host_t *ha;
	int     cnt;
	int     send_marker_once = 0;

        ENTER("qla2100_done");
        COMTRACE('D')

        cnt = 0;

        while( *done_q_first ) {
            /* remove command from done list */

            QLA_DONE_Q_LOCK(old_ha, done_flags);
            if (*done_q_first == NULL) {
                QLA_DONE_Q_UNLOCK(old_ha, done_flags);
                break;
            }

            sp = *done_q_first;
            if( !(*done_q_first = sp->s_next) )
                *done_q_last = NULL;
            else
                (*done_q_first)->s_prev = NULL;

            if (old_ha->done_q_cnt)
                old_ha->done_q_cnt--;

            DEBUG(sp->state = 5;)
            DEBUG3(printk("D%ld ",sp->cmd->serial_number);)
            DEBUG4(printk("qla2100_done: callback pid %d\n",sp->cmd->serial_number);)
            cnt++;
            QLA_DONE_Q_UNLOCK(old_ha, done_flags);

            cmd = sp->cmd;
		tq = sp->tgt_queue;
		lq = sp->lun_queue;
		ha = lq->fclun->fcport->ha;

            DEBUG3(printk("qla2100_done: cmd=%p tq=%p lq=%p ha=%p\n",
              cmd,tq,lq,ha);)

            if (sp->flags & SRB_DMA_VALID) {
                sp->flags &= ~SRB_DMA_VALID;
                /* Decrement outstanding commands on device. */
                if( lq->q_outcnt )
                    lq->q_outcnt--;
                if( lq->q_outcnt < ha->hiwat ) {
	        	lq->q_flag = lq->q_flag & ~LUN_BUSY;
                   	if( lq->q_bsycnt )
               	        	lq->q_bsycnt--;
                }  
            /* 4.10   64 and 32 bit */
            /* Release memory used for this I/O */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,18)
                if (cmd->use_sg) {
                    pci_unmap_sg(ha->pdev, cmd->request_buffer, cmd->use_sg,
                     scsi_to_pci_dma_dir(cmd->sc_data_direction));
                } else if (cmd->request_bufflen) {
                    pci_unmap_single(ha->pdev,sp->saved_dma_handle,
                     cmd->request_bufflen,
                     scsi_to_pci_dma_dir(cmd->sc_data_direction));
               }
#endif
           }

#if MPIO_SUPPORT
            if(ha->flags.failover_enabled) {
            	DEBUG(sp->state = 6;)
	    	qla2x00_fo_check(ha,sp);
		/* we failover this device */
		if( ( sp->flags & SRB_FAILOVER) ) {
            		DEBUG(sp->state = 10;)
			continue;
		}
	    }
#endif

            switch( (CMD_RESULT(cmd)>>16) ) {
                case DID_RESET:
                    lq->q_flag &= ~QLA2100_QRESET;
		    /*
                     * set marker needed, so we don't have to
                     * send multiple markers
                     */
                    /* ra 01/10/02 */
                    if (!send_marker_once) {
                        ha->marker_needed = 1;
                        send_marker_once++;
                    }

                    break;
                case DID_ABORT:
                    sp->flags &= ~SRB_ABORT_PENDING;
                    sp->flags |= SRB_ABORTED;
                    if( sp->flags & SRB_TIMEOUT )
                        CMD_RESULT(cmd)= DID_TIME_OUT << 16;
                    break;
                default:
                    break;
            }

            /* Call the mid-level driver bh handler */
            qla2100_callback(ha,cmd);
	    qla2x00_next(tq, lq);

        }

        COMTRACE('d')
        LEAVE("qla2100_done");
		return( cnt );
}

/*
 *  qla2x00_suspend_lun
 *	Suspend lun and start port down timer
 *
 * Input:
 *	ha = adapter block pointer.
 *  lq = lun queue
 *  time = time in seconds
 *  count = number of times to let time expire
 *
 * Context:
 *	Interrupt context.
 */
STATIC void 
qla2x00_suspend_lun(scsi_qla_host_t *ha, os_lun_t *lq, int time, int count)
{
	fc_port_t *fcport;
	srb_t *sp1;
	ql_list_link_t	*link;
	ql_list_link_t	*next_link = NULL;
	
	/* suspend lun queue */	
	lq->q_flag = lq->q_flag | LUN_QUEUE_SUSPENDED;
	
	DEBUG3(printk("SUSPEND_LUN: lq=%p ha instance=%d count=%d\n",
	    lq,(int)ha->instance,count);)

	/* Remove all pending commands from request queue */
	/* and return them back to OS.                    */
	for (link = lq->cmd.first; link != NULL; link = next_link) {
		next_link = link->next;
		sp1 = link->base_address;

		/* Remove srb from request LUN queue. */
		qla2x00_remove_link(&lq->cmd, &sp1->s_cmd);
		lq->q_incnt--;

		/* Set ending status. */
		CMD_RESULT(sp1->cmd) = DID_BUS_BUSY << 16;
		CMD_HANDLE(sp1->cmd) = (unsigned char *) NULL;
		qla2100_callback(ha,sp1->cmd);
	}

	/* suspend port */	
	if(lq->fclun )  {
		fcport = lq->fclun->fcport;
		if( fcport ) {
			/* Save visible ha for timer to restart this port */
			fcport->vis_ha = ha;
			if( PORT_DOWN(fcport) == 0 )
				PORT_DOWN(fcport) = time;

			if( PORT_DOWN_RETRY(fcport) == 0 ) 
				PORT_DOWN_RETRY(fcport) = count * time;
		}
	}
}

/*
 *  qla2x00_flush_failover_queue
 *	Return cmds of a specific LUN from the failover queue with
 *      DID_BUS_BUSY status.
 *
 * Input:
 *	ha = adapter block pointer.
 *      q  = lun queue.
 *
 * Context:
 *	Interrupt context.
 */
void 
qla2x00_flush_failover_q(scsi_qla_host_t *ha, os_lun_t *q)
{
	ql_list_link_t  *link, *next_link;
	srb_t  *rp;

	for (link = ha->failover_queue.first; link != NULL; link = next_link) {
		next_link = link->next;
		rp = link->base_address;
		if (q == rp->lun_queue) {
			/* Remove srb from failover queue. */
			qla2x00_remove_link(&ha->failover_queue,
			    &rp->s_cmd);
			CMD_RESULT(rp->cmd) = DID_BUS_BUSY << 16;
			CMD_HANDLE(rp->cmd) = (unsigned char *) NULL;
			qla2100_callback(ha, rp->cmd);
		}
	}

}

/*
 *  qla2x00_check_sense
 *
 * Input:
 * cp = SCSI command structure
 * lq = lun queue
 *
 * Context:
 *	Interrupt context.
 */
STATIC void qla2x00_check_sense(Scsi_Cmnd *cp, os_lun_t *lq) 
{
	scsi_qla_host_t *ha, *vis_ha;
	srb_t *sp = (srb_t *) CMD_SP(cp);

	if (((cp->sense_buffer[0] & 0x70) >> 4) != 7) {
		return;
	}

	switch (cp->sense_buffer[2] & 0xf) {
	case 1:
		/* recoverable errors or like OK */
		CMD_RESULT(cp)  = DID_OK << 16;
		break;

	case 2:	 /* need to retry until device becomes ready */
		if( sp->retry_count > 0 ) {
			sp->retry_count--;
			CMD_RESULT(cp) = (int) DID_BUS_BUSY << 16;
			ha = lq->fclun->fcport->ha;

			DEBUG2(printk("CHECK_SENSE: cmd=%p lq=%p "
			    "ha instance=%d retries=%d\n",
			    cp,lq,(int)ha->instance,sp->retry_count);)

		        vis_ha = (scsi_qla_host_t *)cp->host->hostdata;
			if (!(lq->q_flag & LUN_QUEUE_SUSPENDED)) {
				qla2x00_suspend_lun(vis_ha, lq, 6, 1); 
			}
		}
		break;
	}
}

/**************************************************************************
* qla2100_return_status
* 	Translates a ISP error to a Linux SCSI error
**************************************************************************/
STATIC int qla2100_return_status(scsi_qla_host_t *ha, sts_entry_t *sts,
    Scsi_Cmnd *cp)
{
        int host_status = DID_ERROR;
        int scsi_status;
        int comp_status;
        unsigned resid;
        srb_t  *sp;
        uint8_t  *strp;
		os_tgt_t	*tq;
		os_lun_t	*lq;
	    fc_port_t	*fcport;
#if DEBUG_QLA2100_INTR
        STATIC char *reason[] = {
            "DID_OK",
            "DID_NO_CONNECT",
            "DID_BUS_BUSY",
            "DID_TIME_OUT",
            "DID_BAD_TARGET",
            "DID_ABORT",
            "DID_PARITY",
            "DID_ERROR",
            "DID_RESET",
            "DID_BAD_INTR"
        };
#endif /* DEBUG_QLA2100_INTR */

        ENTER("qla2100_return_status");

#if DEBUG_QLA2100_INTR
        /*
        DEBUG(printk("qla2x00: compl status = 0x%04x\n", sts->comp_status));
        */
#endif
        scsi_status = sts->scsi_status;
        comp_status = sts->comp_status;
        sp = (srb_t *) CMD_SP(cp);
	tq = sp->tgt_queue;
	lq = sp->lun_queue;

        if( (scsi_status & SS_RESIDUAL_OVER) )
            comp_status = CS_DATA_OVERRUN;
        else if( (scsi_status & SS_RESPONSE_INFO_LEN_VALID) &&
        sts->rsp_info[3] == (uint8_t) 0 )
            comp_status = CS_COMPLETE;

	/* If scanning and missing lun then let the scsi layer skip it */
	/* 4.24   dg 01/18/2001 */
#if  DUMP_INQ_DATA 
        if( cp->cmnd[0] == 0x12  ) {
		strp = (uint8_t *)cp->request_buffer;
                DEBUG(printk("scsi%d: Inquiry data\n",ha->instance); )
                DEBUG(qla2100_dump_buffer((uint8_t *)strp, 16 );)
	}
#endif
        if( cp->cmnd[0] == 0x12 && 
	    /* 7/05/01 - dg */
	    qla2xenbinq &&
	    cp->lun == 0 ) { /* inquiry */

		strp = (uint8_t *)cp->request_buffer;
		if (*strp == 0x7f && lq->io_cnt < 5) { 
			/* Report lun offline */
			*strp = 0x20;
		} else if (tq->vis_port &&
		    (EXT_IS_LUN_BIT_SET(&(tq->vis_port->lun_mask), 0))) {
			/* Also report lun offline */
			*strp = 0x20;
		}
	}
        switch( comp_status ) {
            case CS_COMPLETE:
                host_status = DID_OK;
		        /* v2.19.5b2 Reset port down retry on success. */
			    sp->port_down_retry_count = ha->port_down_retry_count;
	    		sp->fo_retry_cnt = 0;
                break;
            case CS_PORT_UNAVAILABLE:
                /* release target data structure */
                host_status = DID_NO_CONNECT;
		sp->err_id = 4;
                DEBUG3(printk("scsi: Unavail port detected 0x%x-0x%x.\n",
                    sts->comp_status, sts->scsi_status);)
                break;
            case CS_PORT_LOGGED_OUT:
		sp->err_id = 5;
                host_status = DID_NO_CONNECT;
                break;
            case CS_PORT_CONFIG_CHG:
		sp->err_id = 6;
                host_status = DID_NO_CONNECT;
                break;
            case CS_PORT_BUSY:
		sp->err_id = 7;
                host_status = DID_NO_CONNECT;
                break;
            case CS_INCOMPLETE:
		sp->err_id = 8;
                host_status = DID_NO_CONNECT;
                DEBUG(printk("hba%ld: Port Error detected 0x%x-0x%x.\n",
                    ha->instance, sts->comp_status, sts->scsi_status); )
                break;
            case CS_RESET:
                host_status = DID_RESET;
                break;
            case CS_ABORTED:
                host_status = DID_ABORT;
                break;
            case CS_TIMEOUT:
                  host_status = DID_ERROR;
                  /* v2.19.8 if timeout then check to see if logout occurred*/
				   fcport = lq->fclun->fcport;
		        if ( (fcport->flags & FC_FABRIC_DEVICE)  &&
                       (sts->status_flags & IOCBSTAT_SF_LOGO) ) {
				        /* fcport->state = FC_LOGIN_NEEDED; */
                    DEBUG(printk(
					"scsi(%d): Timeout occurred with Logo, status flag (%x) with public device loop id (%x)\n",
                    (int)ha->instance,sts->status_flags, fcport->loop_id);)
					/* Suspend port */
		            fcport->login_retry = ha->port_down_retry_count;
                    ha->device_flags |= RELOGIN_NEEDED;
                }
                break;
            case CS_DATA_UNDERRUN:
                resid = sts->residual_length;
                /* if RISC reports underrun and target does not report it
                * then we must have a lost frame, so tell upper layer
                * to retry it by reporting an error.
                */
                if( !(sts->scsi_status & SS_RESIDUAL_UNDER) ) {
                    resid = (unsigned)(CMD_XFRLEN(cp));
                }

                if( (unsigned)(CMD_XFRLEN(cp) - resid) < cp->underflow ) {
                        host_status = DID_ERROR;
                        DEBUG3(printk("scsi: Underflow detected - retrying command.\n");)
                } else {
		            /* v2.19.5b2 Reset port down retry on success. */
			        sp->port_down_retry_count = ha->port_down_retry_count;
                    host_status = DID_OK;
	    		sp->fo_retry_cnt = 0;
                }
                break;

            default:
                DEBUG3(printk("scsi: Error detected 0x%x-0x%x.\n",
                sts->comp_status, sts->scsi_status);)
                host_status = DID_ERROR;
                break;
        }

#if DEBUG_QLA2100_INTR
	if ( host_status == DID_ERROR) {
        printk("status: host status (%s) scsi status %x\n", reason[host_status], scsi_status);
	}
#endif

        LEAVE("qla2100_return_status");

        return ((scsi_status & STATUS_MASK) | (host_status << 16) );
}

/*
* qla2100_done_q_put
*      Place SRB command on done queue.
*
* Input:
*      ha           = host pointer
*      sp           = srb pointer.
*      done_q_first = done queue first pointer.
*      done_q_last  = done queue last pointer.
*/
STATIC void
qla2100_done_q_put(scsi_qla_host_t *ha, srb_t *sp, srb_t **done_q_first,
    srb_t **done_q_last)
{
	unsigned long	done_flags;

	/* We must already got the io_request_lock */

	DEBUG3(printk("qla2100_done_q_put: entered.\n");)

	QLA_DONE_Q_LOCK(ha, done_flags);

        /* Place block on done queue */
        sp->s_next = NULL;
        sp->s_prev = *done_q_last;
        if( !*done_q_first )
            *done_q_first = sp;
        else
            (*done_q_last)->s_next = sp;
        *done_q_last = sp;

        ha->done_q_cnt++;
        DEBUG(sp->state = 3;)

	QLA_DONE_Q_UNLOCK(ha, done_flags);

	DEBUG3(printk("qla2100_done_q_put: exiting.\n");)
}

/*
 * qla2x00_abort_q_put
 *      Place SRB command on abort_q.
 *
 * Input:
 *      ha           = host pointer
 *      sp           = srb pointer.
 *      abort_q_first = abort_q first pointer.
 *      abort_q_last  = abort_q last pointer.
 */
STATIC void
qla2x00_abort_q_put(scsi_qla_host_t *ha, srb_t *sp, srb_t **abort_q_first,
    srb_t **abort_q_last)
{
	unsigned long cpu_flags;

	DEBUG3(printk("qla2100_abort_q_put: entered.\n");)

	if (sp == NULL) {
		/* Error */
		return;
	}

	QLA_ABORT_Q_LOCK(ha, cpu_flags);

	/* Place block on abort_q */

	sp->s_next = NULL;
	sp->s_prev = *abort_q_last;

	if (*abort_q_first == NULL)
		*abort_q_first = sp;
	else
		(*abort_q_last)->s_next = sp;

	*abort_q_last = sp;

	DEBUG(sp->state = 7;)

	QLA_ABORT_Q_UNLOCK(ha, cpu_flags);

	DEBUG3(printk("qla2x00_abort_q_put: exiting.\n");)
}

/*
 * qla2x00_abort_q_get
 *      Return one SRB command from abort_q_first.
 *
 * Input:
 *      ha           = host pointer
 *      ret_sp       = srb pointer.
 */
STATIC void
qla2x00_abort_q_get(scsi_qla_host_t *ha, srb_t **ret_sp)
{
	srb_t *sp;
	unsigned long cpu_flags;

	QLA_ABORT_Q_LOCK(ha, cpu_flags);

	if ((*ret_sp = ha->abort_q_first) == NULL) {
		QLA_ABORT_Q_UNLOCK(ha, cpu_flags);
		return;
	}

	sp = *ret_sp;
	if ((ha->abort_q_first = sp->s_next) == NULL) {
		ha->abort_q_last = NULL;
	} else {
		(ha->abort_q_first)->s_prev = NULL;
	}
	sp->s_next = NULL;

	QLA_ABORT_Q_UNLOCK(ha, cpu_flags);
}

/**************************************************************************
*   qla2100_timer
*
* Description:
*   One second timer
*
* Context: Interrupt
***************************************************************************/
STATIC void
qla2100_timer(scsi_qla_host_t *ha)
{
	int		stop_timer, kick_off = 0;
	int		cnt;
	int		t;
	srb_t		*sp, *sp_next;
	unsigned long	cpu_flags = 0;
	fc_port_t	*fcport;


	stop_timer = 0;

	for (t=0, fcport = ha->fcport; fcport != NULL; fcport = fcport->next,
	    t++) {

		if( PORT_DOWN_RETRY(fcport) != 0 ){
			DEBUG(printk("qla%ld: %d fcport(%p) - port retry count %d "
			    "remainning\n", ha->instance, t, fcport,
			    PORT_DOWN_RETRY(fcport));)

			PORT_DOWN_RETRY(fcport)--;
		}

		/* Port Down Handler. */
		if ( PORT_DOWN(fcport) ) {
			PORT_DOWN(fcport)--;

			DEBUG(printk("qla%ld: Port down time in secs %d\n",
			    ha->instance, PORT_DOWN(fcport));)

			if ( PORT_DOWN(fcport) == 0) {
				if (fcport->vis_ha) {
					fcport->vis_ha->flags.port_restart_needed = TRUE;
				} else {
					ha->flags.port_restart_needed = TRUE;
				}
				DEBUG(printk("qla%ld: Port Down complete - "
				    "restarting commands in the queues\n",
				    ha->instance);)
				stop_timer++;
			}
		}
	}

	/* Port Down Handler. */
	if( ha->queue_restart_timer > 0 ) {
		ha->queue_restart_timer--;
		DEBUG3(printk("qla%ld: Port down in secs %d\n",
		    ha->instance, ha->queue_restart_timer);)
		/*
		 * When a port goes DOWN, we suspend the queue and wait 1 second
		 * (one timer tick) before trying to kick off the commands
		 * again. We will do this for "port_down_retry_count" times per
		 * command before giving up on the command altogether.
		 */
		if( !ha->queue_restart_timer ) {
			ha->flags.port_restart_needed = TRUE;
			DEBUG(printk("qla%ld: Port Down complete - restarting "
			    "commands in the queues\n", ha->instance);)
			stop_timer++;
		}
	}

	/* Loop down handler. */
	if (ha->loop_down_timer > 0 && !(ha->dpc_flags & ABORT_ISP_ACTIVE) &&
	    ha->flags.online) {
		if( ha->loop_down_timer == LOOP_DOWN_TIME ) {
			DEBUG(printk("qla%ld: Loop Down - aborting the queues "
			    "before time expire\n", ha->instance);)
			QLA2100_DPC_LOCK(ha);
			ha->flags.abort_queue_needed = TRUE;
			QLA2100_DPC_UNLOCK(ha);
		}

		ha->loop_down_timer--;
		DEBUG3(printk("qla%ld: Loop Down - seconds remainning %d\n",
		    ha->instance, ha->loop_down_timer);)

		/* if the loop has been down for 4 minutes, reinit adapter */
		if( !ha->loop_down_timer ) {
			DEBUG(printk("qla%ld: Loop down exceed 4 mins - "
			    "restarting queues and abort ISP.\n",ha->instance);)

			ha->flags.restart_queues_needed = TRUE;

			if ( qla2100_reinit  && !ha->flags.failover_enabled ) {
				QLA2100_DPC_LOCK(ha);
				ha->dpc_flags |= ISP_ABORT_NEEDED;
				QLA2100_DPC_UNLOCK(ha);
			}
			stop_timer++;
		}
	}

	/*
         * Retry Handler -- This handler will recycle queued requests until the
         * temporary loop down condition terminates.
         */
	QLA_RETRY_Q_LOCK(ha, cpu_flags);

	for( cnt=0, sp = ha->retry_q_first; (sp); sp =  sp_next ) {
		if( sp->wdg_time )
			sp->wdg_time--;

		sp_next = sp->s_next;
		if( sp->wdg_time ==  0 ) {
			kick_off++;

			DEBUG4(printk("qla%ld timer: CMD timeout %p, pid %d\n",
			    ha->instance, sp, sp->cmd->serial_number);)

			cnt++;
			qla2100_timeout_remove(ha, sp);

			CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
			CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;

			qla2100_done_q_put(ha, sp,
			    (srb_t **)&ha->done_q_first,
			    (srb_t **)&ha->done_q_last);
		}
	}
	QLA_RETRY_Q_UNLOCK(ha, cpu_flags);

	if( cnt > 0 ) {
		DEBUG2(printk("qla%ld: timer - found %d requests\n",
		    ha->instance, cnt);)
	}

	if( ha->done_q_first )
		ha->flags.done_requests_needed = TRUE;

#if QLA2100_LIPTEST
	/* This block is used to periodically schedule isp abort after
	 * qla2100_lip flag is set. 
	 */
/*
	if (qla2100_lip && (ha->forceLip++) == (60*2)) {
		printk("timer: schedule isp abort.\n");
		ha->dpc_flags |= ISP_ABORT_NEEDED ;
		ha->forceLip = 0;
	}
*/

	/* This block is used to periodically schedule mailbox cmd timeout
	 * simulation
	 */
	if (qla2100_lip && (ha->forceLip++) == (60*6)) {
		printk("qla2100_timer: Going to force mbx timeout\n");

		ha->forceLip = 0;
                mbxtimeout = 1;
	}
#endif

#if MPIO_SUPPORT 
         QLA_FAILBACK_LOCK(ha, cpu_flags);
	 if( (ha->failback_flags & FAILOVER_EVENT_NEEDED)  ){
	    	if( ha->failback_delay )  {
	    		ha->failback_delay--;
	    		if( ha->failback_delay == 0 )  {
	    			ha->failback_flags |= FAILOVER_EVENT;
		        	ha->failback_flags &= ~FAILOVER_EVENT_NEEDED;
			}
		}
                else
                {
	    		ha->failback_flags |= FAILOVER_EVENT;
	        	ha->failback_flags &= ~FAILOVER_EVENT_NEEDED;
		}
	} 
        QLA_FAILBACK_UNLOCK(ha, cpu_flags);
#endif

	/* Schedule the DPC routine if needed */
	if (((ha->dpc_flags & (ISP_ABORT_NEEDED|LOOP_RESYNC_NEEDED|
	    COMMAND_WAIT_NEEDED)) ||
	    ha->flags.restart_queues_needed ||
	    ha->flags.port_restart_needed ||
	    ha->flags.done_requests_needed ||
	    (ha->device_flags & LOGIN_RETRY_NEEDED) ||
#if MPIO_SUPPORT
	    (ha->failback_flags & FAILOVER_EVENT) ||
	    (ha->dpc_flags & FAILOVER_NEEDED) ||
#endif
	    (ha->dpc_flags & MAILBOX_CMD_NEEDED) ||
	    kick_off > 0  ||
	    ha->flags.abort_queue_needed ) &&
	    ha->dpc_wait && !ha->dpc_active ) {  /* v2.19.4 */
#ifdef	KERNEL_SEM_BUG
		atomic_inc(&dpc_needed);
#else
		up(ha->dpc_wait);
#endif
	}

	RESTART_TIMER(qla2100_timer,ha,WATCH_INTERVAL);
}

/*
* qla2100_timeout_insert
*      Function used to insert a command block onto the
*      watchdog timer queue.
*
*      Note: Must insure that sc_time is not zero
*            before calling qla2100_timeout_insert.
*
* Input:
*      ha = adapter block pointer.
*      sp = srb pointer.
*/
STATIC void
qla2100_timeout_insert(scsi_qla_host_t *ha, srb_t *sp)
{
            uint8_t timeoutcnt;
        unsigned long cpu_flags = 0;

#ifdef QL_DEBUG_LEVEL_5
            ENTER("qla2100_timeout_insert");
#endif

            /* Compute number of time intervals */
            timeoutcnt = (uint8_t) (CMD_TIMEOUT(sp->cmd)/HZ);
            if( timeoutcnt > 10 )
		/* If more than 10, it's something that may take more time
		 * so just leave ourselves 10 in case we need even more we'd
		 * have enough time to return to kernel to request more time.
		 */
                sp->wdg_time = timeoutcnt - 10;
            else if( timeoutcnt >  1 )
                sp->wdg_time = timeoutcnt;
	    else
                sp->wdg_time = 1;

            DEBUG5(printk("Watchdog (insert) - pid=%ld, tmo=%d, start=0x%lx\n",
		sp->cmd->serial_number,sp->wdg_time,sp->r_start);)

            /* Acquire watchdoq queue specific lock */
            QLA_RETRY_Q_LOCK(ha, cpu_flags);
            /* Add command to queue */
            if (ha->retry_q_first == NULL) {
                ha->retry_q_first = sp;
                ha->retry_q_last = sp;
            } else {
                ha->retry_q_last->s_next = sp;
                ha->retry_q_last = sp;
            }
            sp->s_next = NULL;
            qla2100_stats.retry_q_cnt++;
            DEBUG(sp->state = 2;)

            sp->flags |= SRB_WATCHDOG;
            ha->flags.watchdog_enabled = TRUE;

            /* Release watchdog queue specific lock */
            QLA_RETRY_Q_UNLOCK(ha, cpu_flags);

#ifdef QL_DEBUG_LEVEL_5
            LEAVE("qla2100_timeout_insert");
#endif
}

/*
* qla2100_timeout_remove
*      Function used to remove a command block from the
*      watchdog timer queue.
*
* Note: Must insure that command is on watchdog
*       list before calling qla2100_timeout_remove.
*       if (sp->flags & SRB_WATCHDOG)
*       Must have gotten RETRY_Q_LOCK before calling
*       this function.
*
* Input:
*      ha = adapter block pointer.
*      sp = srb pointer.
*/
STATIC void
qla2100_timeout_remove(scsi_qla_host_t *ha, srb_t *sp)
{
            srb_t *nextsp;

#ifdef QL_DEBUG_LEVEL_3
            /* ENTER("qla2100_timeout_remove"); */
#endif

            DEBUG4(printk("Watchdog (remove) pid=%d time%d\n",
		sp->cmd->serial_number,sp->wdg_time);)

            /* Acquire watchdoq queue specific lock */
            /* Remove command block from watchdog queue. */
            if (ha->retry_q_first != NULL) {
                if (ha->retry_q_first == sp) {
                    /* Remove from top of queue */
                    ha->retry_q_first = sp->s_next;
                    if (ha->retry_q_first == NULL) {

                        ha->retry_q_last = NULL;
                        ha->flags.watchdog_enabled = FALSE;
                    }
                } else {
                    /* Remove from middle of queue or bottom of queue */
                    for (nextsp = ha->retry_q_first; nextsp->s_next != NULL;
                    nextsp = nextsp->s_next) {
                        if (nextsp->s_next == sp) {
                            nextsp->s_next = sp->s_next;
                            if (nextsp->s_next == NULL)
                                ha->retry_q_last = nextsp;
                            break;
                        }
                    }
                }
                sp->wdg_time = 0;
                sp->flags &= ~(SRB_WATCHDOG|SRB_BUSY);
                if (qla2100_stats.retry_q_cnt)
                    qla2100_stats.retry_q_cnt--;
            }

            /* Release watchdog queue specific lock */

#ifdef QL_DEBUG_LEVEL_3
            /* LEAVE("qla2100_timeout_remove"); */
#endif
}

#if  NO_LONG_DELAYS
/*
 * This would normally need to get the IO request lock,
 * but as it doesn't actually touch anything that needs
 * to be locked we can avoid the lock here..
 */
STATIC
void qla2x00_sleep_done (struct semaphore * sem)
{
    if( sem != NULL )
    {
        up(sem);
    }
}

void qla2x00_sleep (scsi_qla_host_t *ha, int timeout)
{
    struct semaphore sem = MUTEX_LOCKED;
    struct timer_list timer;

    init_timer(&timer);
    timer.data = (unsigned long) &sem;
    timer.expires = jiffies + timeout;
    timer.function = (void (*)(unsigned long))scsi_sleep_done;
    
    DEBUG(printk("Sleeping for timer tics %d\n", timeout));
    
    add_timer(&timer);

    if( ha->dpc_active  ) 
	{
    spin_unlock_irq(&io_request_lock);
    down(&sem);
    spin_lock_irq(&io_request_lock);
	}
	else
    	down(&sem);
    
    del_timer(&timer);
}
#endif

/*
* qla2100_callback
*      Returns the completed SCSI command to LINUX.
*
* Input:
*	ha -- Host adapter structure
*	cmd -- SCSI mid-level command structure.
* Returns:
*      None
*/
static inline void
qla2100_callback(scsi_qla_host_t *ha, Scsi_Cmnd *cmd)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	unsigned long cpu_flags = 0;
#endif
	srb_t *sp = (srb_t *) CMD_SP(cmd);

#ifdef QL_DEBUG_LEVEL_3
	ENTER("qla2100_callback");
#endif
	CMD_HANDLE(cmd) = (unsigned char *) NULL;
	/*
	 * if command status is not DID_BUS_BUSY then go ahead
	 * and freed sp.
	 */
    	if( sp != NULL &&
	   (CMD_RESULT(cmd) != (DID_BUS_BUSY << 16))  ) {
		/* Never free this resource since we can reuse it */
		if (sp->used == 0x81)
			qla2100_free_sp( cmd, sp);

		sp->flags &= ~SRB_BUSY;

    	} else {
		sp->flags |= SRB_BUSY;
	}

	if ( sp != NULL  ) {
            	DEBUG(sp->state = 12;)

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
            DRIVER_LOCK
#endif
	        /* v2.19.14 
  	        * Perform internal retries, if needed.
	        */
#if 0
		if ( (CMD_RESULT(cmd)>>16) == DID_ERROR ) {
			if( sp->retry_count > 0 ) {
				sp->retry_count--;
				DEBUG3(printk(
				    "qla2x00: RETRY - os retry %d, drv retry "
				    "%d, port retry %d\n\r",
				    cmd->retries,sp->retry_count,
				    sp->port_down_retry_count));
                       	CMD_RESULT(cmd) = (int) DID_BUS_BUSY << 16;
		    } else {
              		sp->retry_count = ha->retry_count;
			    /* all resetted commands must return with RESET */
			    if(cmd->flags & IS_RESETTING ) {
                       		CMD_RESULT(cmd) = (int) DID_RESET << 16;
            			DEBUG3(printk("qla2x00: RESET cmd %p\n",cmd));
			    }
            		DEBUG3(printk("qla2x00: OSerr = %p\n\r",cmd));
		    }
            }
#endif
	}

	if ( (CMD_RESULT(cmd)>>16) == DID_OK ) {
		/* device ok */
		ha->total_bytes += cmd->bufflen;
	} else if ( (CMD_RESULT(cmd)>>16) == DID_ERROR ) {
		/* device error */
		ha->total_dev_errs++;
	}

	DEBUG3(printk("(%ld) ",cmd->serial_number);)

	/* Call the mid-level driver interrupt handler */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	sti();
	(*(cmd)->scsi_done)(cmd);
	cli();
#else
	(*(cmd)->scsi_done)(cmd);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	DRIVER_UNLOCK
#endif
}

/*
* qla2100_mem_alloc
*      Allocates adapter memory.
*
* Returns:
*      0  = success.
*      1  = failure.
*/
STATIC uint8_t
qla2100_mem_alloc(scsi_qla_host_t *ha)
{
	uint8_t   status = 1;
	uint8_t   i;
	int	retry= 10;
	mbx_cmdq_t	*ptmp;
	mbx_cmdq_t	*tmp_q_head;
	mbx_cmdq_t	*tmp_q_tail;

	DEBUG3(printk("qla2100_mem_alloc: entered.\n");)

	do {
		/*
		 * This will loop only once if everything goes well, else
		 * some number of retries will be performed to get around
		 * a kernel bug where available mem is not allocated until
		 * after a little delay and a retry.
		 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
		/* get consistent memory allocated for request/response rings */
		ha->request_ring = KMALLOC_DMA(REQUEST_ENTRY_SIZE *
		    REQUEST_ENTRY_CNT + 1);
		if (ha->request_ring != NULL) {
			ha->request_dma = VIRT_TO_BUS(ha->request_ring);
			BZERO((caddr_t)ha->request_ring,
			    REQUEST_ENTRY_SIZE * REQUEST_ENTRY_CNT+1);
		} else {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - request_ring",(int)ha->host_no);
			UDELAY(HZ/2);
			continue;
		}

		ha->response_ring = KMALLOC_DMA(RESPONSE_ENTRY_SIZE *
		    RESPONSE_ENTRY_CNT+1);
		if (ha->response_ring) {
			ha->response_dma = VIRT_TO_BUS(ha->response_ring);
			BZERO((caddr_t)ha->response_ring,
			    RESPONSE_ENTRY_SIZE * RESPONSE_ENTRY_CNT+1);
		} else {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - response_ring",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}

#ifdef FC_IP_SUPPORT
		ha->pIpBufferQueue      = &ha->asIpBuffers[0];
		ha->ppIpBufferQueueLow  = VIRT_TO_BUS_LOW(&ha->asIpBuffers[0]);
		ha->ppIpBufferQueueHigh = VIRT_TO_BUS_HIGH(&ha->asIpBuffers[0]);
#endif /* FC_IP_SUPPORT */

		/* get consistent memory allocated for init control block */
		ha->init_cb = KMALLOC_DMA(sizeof(init_cb_t));
		if (ha->init_cb != NULL) {
			ha->init_cb_dma = VIRT_TO_BUS(ha->init_cb);
			BZERO((caddr_t)ha->init_cb, sizeof(init_cb_t));
		} else {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - init_cb",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}

#ifdef FC_IP_SUPPORT
		/* not ported yet to new pci_alloc_consistent */
		ha->pIpBufferQueue      = &ha->asIpBuffers[0];
		ha->ppIpBufferQueueLow  = VIRT_TO_BUS_LOW(&ha->asIpBuffers[0]);
		ha->ppIpBufferQueueHigh = VIRT_TO_BUS_HIGH(&ha->asIpBuffers[0]);
#endif
#else  /* KERNEL >= 2.3.18 */ 
		ha->request_ring  = pci_alloc_consistent(ha->pdev,
		    ((REQUEST_ENTRY_CNT + 1) * (sizeof(request_t))),
		    &ha->request_dma);
		if (ha->request_ring == NULL) {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - request_ring",(int)ha->host_no);
			UDELAY(HZ/2);
			continue;
		}

		ha->response_ring  = pci_alloc_consistent(ha->pdev,
		    ((RESPONSE_ENTRY_CNT + 1) * (sizeof(response_t))),
		    &ha->response_dma);

		if (ha->response_ring == NULL) {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - response_ring",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}

		/* get consistent memory allocated for init control block */
		ha->init_cb = pci_alloc_consistent(ha->pdev, sizeof(init_cb_t),
		    &ha->init_cb_dma);
		if (ha->init_cb == NULL) {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - init_cb",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}
		BZERO((caddr_t)ha->init_cb, sizeof(init_cb_t));
#endif

		/* Allocate ioctl related memory. */
		if (qla2x00_alloc_ioctl_mem(ha)) {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - ioctl_mem",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}

		/*
		 * Allocate an initial list of mailbox semaphore queue
		 * to be used for serialization of the mailbox commands.
		 */
		tmp_q_head = (void *)KMEM_ZALLOC(sizeof(mbx_cmdq_t), 20);
		if (tmp_q_head == NULL) {
			/* error */
			printk(KERN_WARNING "scsi(%d): Memory Allocation "
			    "failed - mbx_cmd_q",(int)ha->host_no);
			qla2100_mem_free(ha);
			UDELAY(HZ/2);
			continue;
		}
		ha->mbx_sem_pool_head = tmp_q_head;
		tmp_q_tail = tmp_q_head;
		/* Now try to allocate more */
		for (i = 1; i < MBQ_INIT_LEN; i++) {
			ptmp = (void *)KMEM_ZALLOC(sizeof(mbx_cmdq_t), 20+i);
			if (ptmp == NULL) {
				/*
				 * Error. Just exit. If more is needed later
				 * they will be allocated at that time.
				 */
				break;
			}
			tmp_q_tail->pnext = ptmp;
			tmp_q_tail = ptmp;
		}
		ha->mbx_sem_pool_tail = tmp_q_tail;

		/* Done all allocations without any error. */
		status = 0;

	} while (retry-- && status != 0);

	if (status) {
		printk(KERN_WARNING "qla2100_mem_alloc: **** FAILED ****\n");
	}

	DEBUG3(printk("qla2100_mem_alloc: exiting.\n");)

	return(status);
}

/*
* qla2100_mem_free
*      Frees all adapter allocated memory.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_mem_free(scsi_qla_host_t *ha)
{
	uint32_t	t;
	fc_lun_t	*fclun, *fclun_next;
	fc_port_t	*fcport, *fcport_next;
	mbx_cmdq_t	*ptmp;
	mbx_cmdq_t	*tmp_q_head;
	unsigned long	wtime;/* max wait time if mbx cmd is busy. */

	DEBUG3(printk("qla2100_mem_free: entered.\n");)

	if (ha == NULL) {
		/* error */
		DEBUG2(printk("qla2100_mem_free: ERROR invalid ha pointer.\n");)
		return;
	}

	/* Free the target queues */
	for (t = 0; t < MAX_TARGETS; t++) {
		qla2x00_tgt_free(ha, t);
	}

	/* Make sure all other threads are stopped. */
	wtime = 60 * HZ;
	while ((ha->dpc_wait != NULL || ha->mbx_q_head != NULL) && wtime) {
		set_current_state(TASK_INTERRUPTIBLE);
		wtime = schedule_timeout(wtime);
	}
	/* Now free the mbx sem pool */
	tmp_q_head = ha->mbx_sem_pool_head;
	while (tmp_q_head != NULL) {
		ptmp = tmp_q_head->pnext;
		KMEM_FREE(tmp_q_head, sizeof(mbx_cmdq_t));
		tmp_q_head = ptmp;
	}
	ha->mbx_sem_pool_head = NULL;

	/* free ioctl memory */
	qla2x00_free_ioctl_mem(ha);

	/* 4.10 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	/* free memory allocated for init_cb */
	if (ha->init_cb)
		KMFREE(ha->init_cb, sizeof(init_cb_t));

	if (ha->request_ring)
		KMFREE(ha->request_ring,
		    REQUEST_ENTRY_SIZE * REQUEST_ENTRY_CNT + 1);

	if (ha->response_ring)
		KMFREE(ha->response_ring,
		    RESPONSE_ENTRY_SIZE * RESPONSE_ENTRY_CNT + 1);
#else
	/* free memory allocated for init_cb */
	if (ha->init_cb)
	pci_free_consistent(ha->pdev, sizeof(init_cb_t),
	    ha->init_cb, ha->init_cb_dma);

	if (ha->request_ring)
	pci_free_consistent(ha->pdev,
	    ((REQUEST_ENTRY_CNT + 1) * (sizeof(request_t))),
	    ha->request_ring, ha->request_dma);

	if (ha->response_ring)
	pci_free_consistent(ha->pdev,
	    ((RESPONSE_ENTRY_CNT + 1) * (sizeof(response_t))),
	    ha->response_ring, ha->response_dma);

#endif
	ha->init_cb = NULL;
	ha->request_ring = NULL;
	ha->request_dma = 0;
	ha->response_ring = NULL;
	ha->response_dma = 0;

	/* fc ports */
	for (fcport = ha->fcport; fcport != NULL;
	    fcport = fcport_next) {

		fcport_next = fcport->next;

		/* fc luns */
		for (fclun = fcport->fclun; fclun != NULL;
		    fclun = fclun_next) {
			fclun_next = fclun->next;
			KMFREE(fclun, sizeof(fc_lun_t));
		}
		KMFREE(fcport, sizeof(fc_port_t));
	}

	DEBUG3(printk("qla2100_mem_free: exiting.\n");)
}

/*
*  qla2100_abort_lun_queue
*      Abort all commands on a device queues.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void qla2100_abort_lun_queue(scsi_qla_host_t *ha,
    uint32_t b,uint32_t t,uint32_t l,uint32_t stat)
{
            os_lun_t *q;
            srb_t     *sp;
			ql_list_link_t	*link;

            ENTER("qla2100_abort_lun_queue");
            DEBUG5(sprintf(debug_buff,"Abort queue single %2ld:%2d:%2d\n",
            ha->host_no,t,l);)
            DEBUG5(qla2100_print(debug_buff);)

            if ( (q = (os_lun_t  *)GET_LU_Q(ha, t, l)) ) {
				/* abort all commands on LUN queue. */
				for (link = q->cmd.first; link != NULL; link = link->next) {
					sp = link->base_address;
					qla2100_removeq(q,sp);
                	CMD_RESULT(sp->cmd) = stat << 16;
                	qla2100_done_q_put(ha, sp, (srb_t **)&ha->done_q_first, (srb_t **)&ha->done_q_last);
            	}
			}
            LEAVE("qla2100_abort_lun_queue");
}

/****************************************************************************/
/*                QLogic ISP2100 Hardware Support Functions.                */
/****************************************************************************/

/*
* qla2100_initialize_adapter
*      Initialize board.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success
*/
uint8_t
qla2100_initialize_adapter(scsi_qla_host_t *ha)
{
            device_reg_t *reg;
            uint8_t      status;
            uint8_t      isp_init = 0;
            uint8_t      restart_risc = 0;
            uint8_t      retry;
            unsigned long	wait_device = 0;

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2100_initialize_adapter");
#endif

            /* Clear adapter flags. */
            ha->forceLip = 0;
            ha->flags.online = FALSE;
            ha->flags.disable_host_adapter = FALSE;
            ha->flags.reset_active = FALSE;
            ha->flags.watchdog_enabled = FALSE;
            ha->loop_down_timer = LOOP_DOWN_TIME;
            ha->loop_state = LOOP_DOWN;
            ha->flags.start_timer  = FALSE;
            ha->flags.done_requests_needed = FALSE;
            ha->device_flags = 0;
            ha->sns_retry_cnt = 0;
            ha->device_flags = 0;
            ha->dpc_flags = 0;
            ha->sns_retry_cnt = 0;
	    ha->failback_delay = 0;
	    ha->failback_flags = 0;
	    ha->marker_needed = 0;
            /* 4.11 */
            ha->flags.management_server_logged_in = 0;
            /* ra 11/27/01 */
            ha->marker_needed = 0;

            DEBUG(printk("Configure PCI space for adapter...\n"));

            if( !(status = qla2100_pci_config(ha)) ) {
                reg = ha->iobase;

                qla2100_reset_chip(ha);

                /* Initialize Fibre Channel database. */
                qla2100_init_fc_db(ha);

                /* Initialize target map database. */
                qla2100_init_tgt_map(ha);

		/* Get Flash Version */
		qla2x00_get_flash_version(ha);

                if( qla2100_verbose )
                    printk("scsi(%d): Configure NVRAM parameters...\n",(int)ha->host_no);

                qla2200_nvram_config(ha);
				ha->retry_count = ql2xretrycount;
#if USE_PORTNAME
				ha->flags.port_name_used =1;
#else
				ha->flags.port_name_used =0;
#endif

                if( qla2100_verbose )
                    printk("scsi(%d): Verifying loaded RISC code...\n",(int)ha->host_no);

                qla2100_set_cache_line(ha);

                /* If the user specified a device configuration on
                * the command line then use it as the configuration.
                * Otherwise, we scan for all devices.
                */
                if ( ql2xdevconf ) {
		    ha->cmdline = ql2xdevconf;
                    if ( !ha->flags.failover_enabled )
                    	qla2100_get_properties(ha, ql2xdevconf);
                }

                retry = 10;
                /*
                * Try an configure the loop.
                */
                do {
                    restart_risc = 0;
		    isp_init = 0;
                    DEBUG(printk("qla2100_initialize_adapter: check if "
			"firmware needs to be loaded\n");)

                    /* If firmware needs to be loaded */
                    if( qla2100_isp_firmware(ha) ) {
                        if( qla2100_verbose )
                            printk("scsi(%d): Verifying chip...\n",(int)ha->host_no);

                        if( !(status = qla2100_chip_diag(ha)) )
                            status = qla2x00_setup_chip(ha);

                        if( !status ) {
                            DEBUG(printk("scsi(%d): Chip verified and RISC "
				"loaded...\n", (int)ha->host_no));
			}
                    }
                    if( !status && !(status = qla2x00_init_rings(ha)) ) {
                        /* dg - 7/3/1999
                         *
                         * Wait for a successful LIP up to a maximum of (in
			 * seconds): RISC login timeout value, RISC retry
			 * count value, and port down retry value OR a minimum
			 * of 4 seconds OR If no cable, only 5 seconds.
                         */
                        DEBUG(printk("qla2x00_init_rings OK, call "
			    "qla2x00_fw_ready...\n");)

                        if( !qla2x00_fw_ready(ha) ) {

                            ha->dpc_flags &=
                                 ~(RESET_MARKER_NEEDED | COMMAND_WAIT_NEEDED);

                            /* Go setup flash database devices with proper Loop ID's. */
                            do {
                                ha->dpc_flags &= ~LOOP_RESYNC_NEEDED;
                                status = qla2100_configure_loop(ha);

				/*
				 * Temp code: delay a while for certain
				 * slower devices to become ready.
				 */
				for ((wait_device = jiffies + HZ);
				    !time_after_eq(jiffies,wait_device); ) {
					qla2100_check_fabric_devices(ha);
				}

                            } while (!ha->loop_down_timer &&
				!(ha->dpc_flags & ISP_ABORT_NEEDED) &&
				(ha->dpc_flags & LOOP_RESYNC_NEEDED) );
			}

                        if( ha->flags.update_config_needed ) {
                            ha->init_cb->additional_firmware_options.
				connection_options = ha->operating_mode;
                            restart_risc = 1;
                        }

			if( ha->mem_err )
			{
                            restart_risc = 1;
			}
                        isp_init = 1;

                    }
                } while( restart_risc && retry-- );

                if( isp_init ) {
                    ha->dpc_flags &= ~RESET_MARKER_NEEDED;
                    qla2x00_marker(ha, 0, 0, MK_SYNC_ALL);

                    ha->flags.online = TRUE;

                    /* Enable target response to SCSI bus. */
                    if( ha->flags.enable_target_mode )
                        qla2100_enable_lun(ha);
                }

            }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
            if( status )
                qla2100_print("qla2100_initialize_adapter: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
            else
                LEAVE("qla2100_initialize_adapter");
#endif
            return(status);
}

/*
* ISP Firmware Test
*      Checks if present version of RISC firmware is older than
*      driver firmware.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = firmware does not need to be loaded.
*/
STATIC uint8_t
qla2100_isp_firmware(scsi_qla_host_t *ha)
{
            uint8_t  status = 1; /* assume loading risc code */

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2x00_isp_firmware");
#endif

            if( ha->flags.disable_risc_code_load ) {
                /* Verify checksum of loaded RISC code. */
		status = qla2x00_verify_checksum(ha);
            }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
            if( status )
                qla2100_print(
                "qla2100_isp_firmware: **** Load RISC code ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
            else
                LEAVE("qla2x00_isp_firmware");
#endif
            return(status);
}

/*
* (08/05/99)
*
* PCI configuration
*      Setup device PCI configuration registers.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
STATIC uint8_t
qla2100_pci_config(scsi_qla_host_t *ha)
{
            uint8_t      status = 1;
            uint32_t     command;
#if MEMORY_MAPPED_IO
            uint32_t  page_offset, base;
            uint32_t   mmapbase;
#endif
            config_reg_t *creg = 0;
            uint16_t     buf_wd;

            ENTER("qla2100_pci_config");
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,1,95)
/* 4.12 */
           /* turn on PCI master; for system BIOSes that don't turn
              it on by default */
            pci_set_master(ha->pdev);
#endif
            pci_read_config_word(ha->pdev,OFFSET(creg->revision_id), &buf_wd);
            ha->revision = buf_wd;
            if( !ha->iobase ) {
                /* Get command register. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                if( pci_read_config_word(ha->pdev,OFFSET(creg->command), &buf_wd) == PCIBIOS_SUCCESSFUL ) {
#else
                if( pci_read_config_word(ha->pdev,OFFSET(creg->command), &buf_wd) == PCIBIOS_SUCCESSFUL ) {
#endif
                    command = buf_wd;

                    /*
                    * Set Bus Master Enable (bit-2), Memory Address Space Enable and
                    * reset any error bits.
                    */
                    buf_wd &= ~0x7;
#if MEMORY_MAPPED_IO
                    DEBUG(printk("qla2x00: I/O SPACE and MEMORY MAPPED IO is enabled.\n"));
                    buf_wd |= BIT_2 + BIT_1 + BIT_0;
#else
                    DEBUG(printk("qla2x00: I/O SPACE Enabled and MEMORY MAPPED IO is disabled.\n"));
                    buf_wd |= BIT_2 + BIT_0;
#endif
                    if( pci_write_config_word(ha->pdev,OFFSET(creg->command), buf_wd) ) {
                        printk(KERN_WARNING "qla2x00: Could not write config word.\n");
                    }
                    /* Get expansion ROM address. */
                    if( pci_read_config_word(ha->pdev, OFFSET(creg->expansion_rom), &buf_wd) == PCIBIOS_SUCCESSFUL ) {
                        /* Reset expansion ROM address decode enable. */
                        buf_wd &= ~BIT_0;
                        if( pci_write_config_word(ha->pdev, OFFSET(creg->expansion_rom), buf_wd) == PCIBIOS_SUCCESSFUL ) {
#if MEMORY_MAPPED_IO
                            /* Get memory mapped I/O address. */
                            pcibios_read_config_dword(ha->pci_bus, ha->pci_device_fn,OFFSET(cfgp->mem_base_addr), &mmapbase);
                            mmapbase &= PCI_BASE_ADDRESS_MEM_MASK;

                            /* Find proper memory chunk for memory map I/O reg. */
                            base = mmapbase & PAGE_MASK;
                            page_offset = mmapbase - base;
                            /* Get virtual address for I/O registers. */
                            ha->mmpbase = ioremap(base,page_offset + 256);
                            if( ha->mmpbase ) {
                                ha->mmpbase += page_offset;
                                /* ha->iobase = ha->mmpbase; */
                                status = 0;
                            }
#else /* MEMORY_MAPPED_IO */
                            status = 0;
#endif /* MEMORY_MAPPED_IO */
                        }
                    }
                }
            } else
                status = 0;


            LEAVE("qla2100_pci_config");
            return(status);
}
#else
STATIC uint8_t
qla2100_pci_config(scsi_qla_host_t *ha)
{
            uint8_t      status = 1;
            uint32_t     command;
#if MEMORY_MAPPED_IO
            uint32_t  page_offset, base;
            uint32_t   mmapbase;
#endif
            config_reg_t *creg = 0;
            uint16_t     buf_wd;

            ENTER("qla2100_pci_config");

            pcibios_read_config_word(ha->pci_bus,ha->pci_device_fn, OFFSET(creg->revision_id), &buf_wd);
            ha->revision = buf_wd;
            if( !ha->iobase ) {
                /* Get command register. */
                if( pcibios_read_config_word(ha->pci_bus,ha->pci_device_fn, OFFSET(creg->command), &buf_wd) == PCIBIOS_SUCCESSFUL ) {
                    command = buf_wd;

                    /*
                    * Set Bus Master Enable (bit-2), Memory Address Space Enable and
                    * reset any error bits.
                    */
                    buf_wd &= ~0x7;
#if MEMORY_MAPPED_IO
                    DEBUG(printk("qla2x00: I/O SPACE and MEMORY MAPPED IO is enabled.\n"));
                    buf_wd |= BIT_2 + BIT_1 + BIT_0;
#else
                    DEBUG(printk("qla2x00: I/O SPACE Enabled and MEMORY MAPPED IO is disabled.\n"));
                    buf_wd |= BIT_2 + BIT_0;
#endif
                    if( pcibios_write_config_word(ha->pci_bus,ha->pci_device_fn, OFFSET(creg->command), buf_wd) ) {
                        printk(KERN_WARNING "qla2x00: Could not write config word.\n");
                    }
                    /* Get expansion ROM address. */
                    if( pcibios_read_config_word(ha->pci_bus,ha->pci_device_fn, OFFSET(creg->expansion_rom), &buf_wd) == PCIBIOS_SUCCESSFUL ) {
                        /* Reset expansion ROM address decode enable. */
                        buf_wd &= ~BIT_0;
                        if( pcibios_write_config_word(ha->pci_bus,ha->pci_device_fn, OFFSET(creg->expansion_rom), buf_wd) == PCIBIOS_SUCCESSFUL ) {
#if MEMORY_MAPPED_IO
                            /* Get memory mapped I/O address. */
                            pcibios_read_config_dword(ha->pci_bus, ha->pci_device_fn,OFFSET(cfgp->mem_base_addr), &mmapbase);
                            mmapbase &= PCI_BASE_ADDRESS_MEM_MASK;

                            /* Find proper memory chunk for memory map I/O reg. */
                            base = mmapbase & PAGE_MASK;
                            page_offset = mmapbase - base;
                            /* Get virtual address for I/O registers. */
                            ha->mmpbase = ioremap(base,page_offset + 256);
                            if( ha->mmpbase ) {
                                ha->mmpbase += page_offset;
                                /* ha->iobase = ha->mmpbase; */
                                status = 0;
                            }
#else /* MEMORY_MAPPED_IO */
                            status = 0;
#endif /* MEMORY_MAPPED_IO */
                        }
                    }
                }
            } else
                status = 0;


            LEAVE("qla2100_pci_config");
            return(status);
}
#endif


/*
* qla2100_set_cache_line
*      Sets PCI cache line parameter.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_set_cache_line(scsi_qla_host_t *ha)
{
            int          status = 0;
            uint8_t      buf;
            config_reg_t *creg = 0;

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2100_set_cache_line");
#endif
            /* Set the cache line. */
            if( ha->flags.set_cache_line_size_1 ) {
                buf = 1;
                if( pcibios_write_config_byte(ha->pci_bus,
                    ha->pci_device_fn,
                    (u_long)&creg->cache_line,buf) != PCIBIOS_SUCCESSFUL )
                    status = 1;
            }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
            if( status )
                qla2100_print("qla2100_set_cache_line: **** FAILED ****\n\r");
#endif
#ifdef QL_DEBUG_LEVEL_3
            else
                LEAVE("qla2100_set_cache_line");
#endif
            return(status);
}

/*
* Chip diagnostics
*      Test chip for proper operation.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_chip_diag(scsi_qla_host_t *ha)
{
            device_reg_t *reg   = ha->iobase;
#ifdef ISP2300
            config_reg_t *creg = 0;
            uint16_t     buf_wd;
#endif
            uint16_t     data;
            uint32_t     cnt;
            uint8_t      status = 0;

            ENTER("qla2100_chip_diag");
#ifdef QL_DEBUG_LEVEL_3
            qla2100_print("qla2100_chip_diag: testing device at ");
            qla2100_output_number((u_long)&reg->flash_address, 16);
            qla2100_print("\n");
#endif

            /* Reset ISP chip. */
            WRT_REG_WORD(&reg->ctrl_status, ISP_RESET);
            data = qla2100_debounce_register(&reg->ctrl_status);
            for( cnt = 6000000 ; cnt && (data & ISP_RESET); cnt-- ) {
                SYS_DELAY(5);
                data = RD_REG_WORD(&reg->ctrl_status);
            }

            if( cnt ) {
#if defined(QL_DEBUG_LEVEL_3)
                qla2100_print("qla2100_chip_diag: reset register cleared by chip reset\n\r");
#endif
#ifdef ISP2300
                pci_read_config_word(ha->pdev,OFFSET(creg->command), &buf_wd);
                buf_wd |= BIT_1 | BIT_2;
                data = RD_REG_WORD(&reg->mailbox6);

                if ((ha->device_id == QLA2312_DEVICE_ID) || 
                   ((data & 0xff) == FPM_2310))
                   /* Enable Memory Write and Invalidate. */
                   buf_wd |= BIT_4;
                else
                   buf_wd &= ~BIT_4;
                pci_write_config_word(ha->pdev,OFFSET(creg->command), buf_wd);
#endif
                /* Reset RISC processor. */
                WRT_REG_WORD(&reg->host_cmd, HC_RESET_RISC);
                WRT_REG_WORD(&reg->host_cmd, HC_RELEASE_RISC);
                /* Workaround for Qla2312 PCI parity error */
                if (ha->device_id == QLA2312_DEVICE_ID)
                {
                   SYS_DELAY(10);
                }
                else
                {
                   data = qla2100_debounce_register(&reg->mailbox0);

                   for( cnt = 6000000; cnt && (data == MBS_BUSY); cnt-- ) {
                       SYS_DELAY(5);
                       data = RD_REG_WORD(&reg->mailbox0);
                   }
                }

                if( cnt ) {
                    /* Check product ID of chip */
#if defined(QL_DEBUG_LEVEL_3)
                    qla2100_print("{{{qla2100_chip_diag: Checking product ID of chip}}}\n\r");
#endif
                    if( RD_REG_WORD(&reg->mailbox1) != PROD_ID_1 ||
                    (RD_REG_WORD(&reg->mailbox2) != PROD_ID_2 &&
                    RD_REG_WORD(&reg->mailbox2) != PROD_ID_2a) ||
                    RD_REG_WORD(&reg->mailbox3) != PROD_ID_3 ||
                    (qla2100_debounce_register(&reg->mailbox4) != PROD_ID_4 &&
                    qla2100_debounce_register(&reg->mailbox4) != PROD_ID_4a) ) {
                        printk(KERN_WARNING "qla2x00: Wrong product ID = 0x%x,0x%x,0x%x,0x%x\n",
                        RD_REG_WORD(&reg->mailbox1),
                        RD_REG_WORD(&reg->mailbox2),
                        RD_REG_WORD(&reg->mailbox3),
                        RD_REG_WORD(&reg->mailbox4));
                        status = 1;
                    } else {
                        /* Now determine if we have a 2200A board */
                        if( ( ha->device_id == QLA2200_DEVICE_ID   ||
                        ha->device_id == QLA2200A_DEVICE_ID ) &&
                        RD_REG_WORD(&reg->mailbox7) == QLA2200A_RISC_ROM_VER ) {
                            ha->device_id = QLA2200A_DEVICE_ID;
#if defined(QL_DEBUG_LEVEL_3)
                            qla2100_print("qla2100_chip_diag: Found QLA2200A chip.\n\r");
#endif
                        }

                        DEBUG(printk("qla2100_chip_diag: Checking mailboxes.\n"));
#if defined(QL_DEBUG_LEVEL_3)
                        qla2100_print("qla2100_chip_diag: Checking mailboxes.\n\r");
#endif
                        /* Wrap Incoming Mailboxes Test. */
			status = qla2x00_mbx_reg_test(ha);
                        if (status) {
                                printk(KERN_WARNING "qla2100_chip_diag: failed mailbox send register test\n");
                                DEBUG(qla2100_print("qla2100_chip_diag: Failed mailbox send register test\n\r");)
                        }
                    }
                } else
                    status = 1;
            } else
                status = 1;

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
            if( status )
                qla2100_print("qla2100_chip_diag: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
            else
                LEAVE("qla2100_chip_diag");
#endif
            return(status);
}

/*
* Setup chip
*      Load and start RISC firmware.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2x00_setup_chip(scsi_qla_host_t *ha)
{
	uint16_t cnt;
	uint16_t risc_address;
	uint16_t *risc_code_address;
	long risc_code_size;
	uint8_t  status = 0;
	int num;
#ifdef WORD_FW_LOAD
	uint16_t data;
	uint16_t *ql21_risc_code_addr01;
	uint16_t  ql21_risc_code_length01;
	uint8_t   dump_status;
#endif

	DEBUG3(printk("qla2x00_setup_chip(%ld): entered.\n",
	    ha->host_no);)

	/* Load RISC code. */
	risc_address = *QLBoardTbl_fc[ha->devnum].fwstart;
	risc_code_address = QLBoardTbl_fc[ha->devnum].fwcode;
	risc_code_size    = (long)(*QLBoardTbl_fc[ha->devnum].fwlen & 0xffff);

	DEBUG(printk("qla2x00_setup_chip: Loading RISC code size =(0x%lx)\n",
	    risc_code_size);)
	DEBUG(printk("    req virt=%p phys=%x\n", ha->request_ring,
	    ha->request_dma);)

	/*
	 * Save active FC4 type depending on firmware support. This
	 * info is needed by ioctl cmd.
	 */
	ha->active_fc4_types = EXT_DEF_FC4_TYPE_SCSI;
#ifdef FC_IP_SUPPORT
	if (risc_address == fw2300ip_code01)
		ha->active_fc4_types |= EXT_DEF_FC4_TYPE_IP;
#endif
#ifdef FC_SCTP_SUPPORT
	if (risc_address == fw2300sctp_code01)
		ha->active_fc4_types |= EXT_DEF_FC4_TYPE_SCTP;
#endif
#ifdef FC_VI_SUPPORT
	if (risc_address == fw2300vi_code01)
		ha->active_fc4_types |= EXT_DEF_FC4_TYPE_VI;
#endif

	num = 0;

	while (risc_code_size > 0 && !status) {

		/* for 2200A set transfer size to 128 bytes */
		if( ha->device_id == QLA2200A_DEVICE_ID )
			cnt = 128 >> 1;
		else
			cnt = REQUEST_ENTRY_SIZE * REQUEST_ENTRY_CNT >> 1;

		if( cnt > risc_code_size )
			cnt = risc_code_size;

		DEBUG7(printk("qla2x00_setup_chip: loading risc segment@ "
		    "addr %p, number of bytes 0x%x, offset 0x%x.\n",
		    risc_code_address,cnt,risc_address);)

		BCOPY((caddr_t)risc_code_address, (caddr_t)ha->request_ring,
		    (cnt <<1));

		/* flush written firmware to the
		 * ha->request_ring buffer before DMA
		 */
		flush_cache_all();

		status = qla2x00_load_ram(ha, ha->request_dma, risc_address,
		    cnt);

		if (status) {
			qla2100_dump_regs(ha->host);
			printk(KERN_WARNING "qla2x00: [ERROR] Failed to load "
			    "segment %d of FW\n", num);
			DEBUG(printk("qla2x00_setup_chip: Failed to load "
			    "segment %d of FW\n", num);)
			break;
		}

		risc_address += cnt;
		risc_code_size -= cnt;
		risc_code_address += cnt;
		num++;
	}

#ifdef WORD_FW_LOAD
	{
		int i;
		uint8_t temp;

		risc_address = *QLBoardTbl_fc[ha->devnum].fwstart;
		ql21_risc_code_addr01  = QLBoardTbl_fc[ha->devnum].fwcode;
		ql21_risc_code_length01 =
		    (long)(*QLBoardTbl_fc[ha->devnum].fwlen & 0xffff);

		for( i = 0; i < ql21_risc_code_length01 ; i++ ) {

			dump_status = qla2x00_write_ram_word(ha,
			    risc_address + i, *(ql21_risc_code_addr01 + i));

			if( dump_status ) {
				printk(KERN_WARNING "qla2x00: [ERROR] firmware "
				    "load failure\n");
				break;
			}

			dump_status = qla2x00_read_ram_word(ha,
			    risc_address + i, &data);

			if( dump_status ) {
				printk(KERN_WARNING "qla2x00: [ERROR] RISC FW "
				    "Read Failure\n");
				break;
			}

			if( data != *(ql21_risc_code_addr01 + i) ) {
				printk(KERN_WARNING "qla2x00: [ERROR] RISC FW "
				    "Compare ERROR @ (0x%p)\n",
				    (void *) (ql21_risc_code_addr01+i));
			}
		}
		printk(KERN_INFO "qla2x00: RISC FW download confirmed... \n");
	}
#endif /* WORD_FW_LOAD */

	/* Verify checksum of loaded RISC code. */
	if( !status ) {
		DEBUG(printk("qla2x00_setup_chip: Verifying Check Sum of "
		    "loaded RISC code.\n");)

		status = (uint8_t)qla2x00_verify_checksum(ha);

		if (status == QL_STATUS_SUCCESS) {
			/* Start firmware execution. */
			DEBUG(printk("qla2x00_setup_chip: CS Ok, Start "
			    "firmware running\n");)
			status = qla2x00_execute_fw(ha);
		}
#if defined(QL_DEBUG_LEVEL_2)
		else {
			printk("qla2x00_setup_chip: ISP FW Failed Check Sum\n");
		}
#endif
	}

	if (status) {
		DEBUG2_3(printk("qla2x00_setup_chip: **** FAILED ****\n");)
	} else {
		DEBUG3(printk("qla2x00_setup_chip: Returning Good Status\n");)
	}

	return(status);
}

/*
* qla2x00_init_rings
*      Initializes firmware.
*
*      Beginning of request ring has initialization control block
*      already built by nvram config routine.
*
* Input:
*      ha                = adapter block pointer.
*      ha->request_ring  = request ring virtual address
*      ha->response_ring = response ring virtual address
*      ha->request_dma   = request ring physical address
*      ha->response_dma  = response ring physical address
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2x00_init_rings(scsi_qla_host_t *ha)
{
	uint8_t  status;
	int cnt;

	DEBUG3(printk("qla2x00_init_rings(%ld): entered.\n",
	    ha->host_no);)

	/* Clear outstanding commands array. */
	for (cnt = 0; cnt < MAX_OUTSTANDING_COMMANDS; cnt++)
		ha->outstanding_cmds[cnt] = 0;

#ifdef RSCN
	/* Clear RSCN queue. */
	ha->rscn_in_ptr = 0;
	ha->rscn_out_ptr = 0;
#endif

	/* Initialize firmware. */
	ha->request_ring_ptr  = ha->request_ring;
	ha->req_ring_index    = 0;
	ha->req_q_cnt         = REQUEST_ENTRY_CNT;
	ha->response_ring_ptr = ha->response_ring;
	ha->rsp_ring_index    = 0;

	DEBUG(printk("qla2x00_init_rings(%ld): issue init firmware.\n",
	    ha->host_no);)
	status = qla2x00_init_firmware(ha, sizeof(init_cb_t));

	if (status) {
		/* Empty */
		DEBUG2_3(printk("qla2x00_init_rings(%ld): **** FAILED ****.\n",
		    ha->host_no);)
	} else {
		/* Empty */
		DEBUG3(printk("qla2x00_init_rings(%ld): exiting normally.\n",
		    ha->host_no);)
	}

	return(status);
}

/*
* qla2x00_fw_ready
*      Waits for firmware ready.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2x00_fw_ready(scsi_qla_host_t *ha)
{
	uint8_t  status = 0;
	uint8_t  loop_forever = 1;
	unsigned long wtime, mtime;
	uint16_t min_wait; /* minimum wait time if loop is down */
	uint16_t wait_time;/* wait time if loop is becoming ready */
	uint16_t pause_time;
	uint16_t fw_state;

	DEBUG3(printk("qla2x00_fw_ready(%ld): entered.\n",
	    ha->host_no);)

	min_wait = 20; /* 20 seconds for loop down. */
	ha->device_flags &= ~DFLG_NO_CABLE;

	/* Firmware should take at most one RATOV to login, plus 5 seconds
	 * for our own processing.
	 */
	if ((wait_time = (ha->retry_count * ha->login_timeout) + 5) <
	    min_wait) {
		wait_time = min_wait;
	}

	pause_time = 1000; /* 1000 usec */

	/* min wait time if loop down */
	mtime = jiffies + (min_wait * HZ);

	/* wait time before firmware ready */
	wtime = jiffies + (wait_time * HZ);

	/* Wait for ISP to finish LIP */
	if (!qla2100_quiet)
		printk(KERN_INFO "scsi(%d): Waiting for LIP to complete...\n",
		    (int)ha->host->host_no);

	DEBUG3(printk("scsi(%d): Waiting for LIP to complete...\n",
	    (int)ha->host->host_no);)

        do {

		status = qla2x00_get_firmware_state(ha, &fw_state);

		if (status == QL_STATUS_SUCCESS) {

			if (fw_state == FSTATE_READY) {
				qla2x00_get_retry_cnt(ha, &ha->retry_count,
				    &ha->login_timeout);
				status = QL_STATUS_SUCCESS;

				DEBUG(printk(
				  "qla2x00_fw_ready(%ld): F/W Ready - OK \n",
				  ha->host_no);)

				break;
			}

			status = QL_STATUS_ERROR;

			if (ha->loop_down_timer ||
			    fw_state == FSTATE_LOSS_OF_SYNC) {
				/* Loop down. Timeout on min_wait */
				if (time_after_eq(jiffies,mtime) ) {
					printk(KERN_INFO "scsi(%d): Cable is "
					    "unplugged...\n",
		    			    (int)ha->host->host_no);
					ha->device_flags |= DFLG_NO_CABLE;
					break;
				}
			}
		} else {
			/* Mailbox cmd failed. Timeout on min_wait. */
			if (time_after_eq(jiffies,mtime))
				break;
		}

		if (time_after_eq(jiffies,wtime))
			break;

		/* Delay for a while */
		UDELAY(pause_time);

		DEBUG3(printk("qla2x00_fw_ready: fw_state=%x curr time=%lx.\n",
		    fw_state, jiffies);)

	} while(loop_forever);

	DEBUG(printk("qla2x00_fw_ready(%d): fw_state=%x curr time=%lx.\n",
	    (int)ha->host_no,fw_state, jiffies);)

	if (status) {
		DEBUG2_3(printk("qla2x00_fw_ready(%ld): **** FAILED ****.\n",
		    ha->host_no);)
	} else {
		DEBUG3(printk("qla2x00_fw_ready(%ld): exiting normally.\n",
		    ha->host_no);)
	}


	return(status);
}

/*
*  qla2x00_configure_hba
*      Setup adapter context.
*
* Input:
*      ha = adapter state pointer.
*
* Returns:
*      0 = success
*
* Context:
*      Kernel context.
*/
STATIC uint8_t
qla2x00_configure_hba(scsi_qla_host_t *ha)
{
	uint8_t       rval;
	uint8_t       connect_type[22];
	uint16_t      loop_id;
	uint16_t      topo;
	uint8_t       al_pa;
	uint8_t       area;
	uint8_t       domain;

	DEBUG3(printk("qla2x00_configure_hba(%ld): entered.\n",
	    ha->host_no);)

	/* Get host addresses. */
	rval = qla2x00_get_adapter_id(ha, &loop_id, &al_pa, &area, &domain,
	    &topo);
	if (rval != QL_STATUS_SUCCESS) {
		printk(KERN_WARNING
		    "qla2x00_configure_hba(%ld): ERROR Get host loop ID.\n",
		    ha->host_no);
		return (rval);
	}

	if (topo == 4) {
		printk(KERN_INFO "scsi%d: Cannot get topology - "
		    "retrying.\n", (int)ha->host_no);
		return (QL_STATUS_ERROR);
	}

	ha->loop_id = loop_id;

#ifdef ISP2100
	/* Make sure 2100 only has loop, in case of any firmware bug. */
	topo = 0;
#endif

	/* initialize */
	ha->min_external_loopid = SNS_FIRST_LOOP_ID;
	ha->operating_mode = LOOP;

	switch (topo) {
	case 0:
		DEBUG3(printk("qla2x00(%ld): HBA in NL topology.\n",
		    ha->host_no);)
		ha->current_topology = ISP_CFG_NL;
		strcpy((char *)&connect_type[0],"(Loop)");
		break;

	case 1:
		DEBUG3(printk("qla2x00(%ld): HBA in FL topology.\n",
		    ha->host_no);)
		ha->current_topology = ISP_CFG_FL;
		strcpy((char *)&connect_type[0],"(FL_Port)");
		break;

	case 2:
		DEBUG3(printk( "qla2x00(%ld): HBA in N P2P topology.\n",
		    ha->host_no);)
		ha->operating_mode = P2P;
		ha->current_topology = ISP_CFG_N;
		strcpy((char *)&connect_type[0],"(N_Port-to-N_Port)");
		break;

	case 3:
		DEBUG3(printk(
		    "qla2x00(%ld): HBA in F P2P topology.\n",
		    ha->host_no);)
		ha->operating_mode = P2P;
		ha->current_topology = ISP_CFG_F;
		strcpy((char *)&connect_type[0],"(F_Port)");
		break;

	default:
		DEBUG3(printk(
		    "qla2x00(%ld): HBA in unknown topology %x. "
		    "Using NL.\n", ha->host_no, topo);)
		ha->current_topology = ISP_CFG_NL;
		strcpy((char *)&connect_type[0],"(Loop)");
		break;
	}

	/* Save Host port and loop ID. */
	/* byte order - Big Endian */
	ha->d_id.b.domain = domain;
	ha->d_id.b.area = area;
	ha->d_id.b.al_pa = al_pa;

	if (!qla2100_quiet)
		printk(KERN_INFO
		    "scsi%d: Topology - %s, Host Loop address 0x%x\n",
		    (int)ha->host_no, connect_type, ha->loop_id);

	if (rval != 0) {
		/* Empty */
		DEBUG2_3(printk("qla2x00_configure_hba(%ld): FAILED.\n",
		    ha->host_no);)
	} else {
		/* Empty */
		DEBUG3(printk("qla2x00_configure_hba(%ld): exiting normally.\n",
		    ha->host_no);)
	}

	return(rval);
}


/*
* NVRAM configuration for the 2200.
*
* Input:
*      ha                = adapter block pointer.
*      ha->request_ring  = request ring virtual address
*      ha->response_ring = response ring virtual address
*      ha->request_dma   = request ring physical address
*      ha->response_dma  = response ring physical address
*
* Output:
*      initialization control block in response_ring
*      host adapters parameters in host adapter block
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2200_nvram_config(scsi_qla_host_t *ha)
{
#ifdef ISP2300
            device_reg_t *reg = ha->iobase;
            uint16_t  data;
#endif
            uint8_t   status = 0;
            uint8_t   chksum = 0;
            uint16_t  cnt, base;
            caddr_t   dptr1, dptr2;
            init_cb_t *icb   = ha->init_cb;
            nvram22_t *nv    = (nvram22_t *)ha->request_ring;
            uint16_t  *wptr  = (uint16_t *)ha->request_ring;

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2200_nvram_config");
#endif

            if( !ha->flags.nvram_config_done ) {
#ifdef ISP2300
                if (ha->device_id == QLA2312_DEVICE_ID)
                {
                   data = RD_REG_WORD(&reg->ctrl_status);
                   if ((data >> 14) == 1)
                      base = 0x80;
                   else
                      base = 0;
                   data = RD_REG_WORD(&reg->nvram);
                   while (data & NV_BUSY)
                   {
                      UDELAY(100);
                      data = RD_REG_WORD(&reg->nvram);
                   }
                   /* Lock resource */
                   WRT_REG_WORD(&reg->host_semaphore, 0x1);
                   UDELAY(5);
                   data = RD_REG_WORD(&reg->host_semaphore);
                   while ((data & BIT_0) == 0)
                   {
                      /* Lock failed */
                      UDELAY(100);
                      WRT_REG_WORD(&reg->host_semaphore, 0x1);
                      UDELAY(5);
                      data = RD_REG_WORD(&reg->host_semaphore);
                   }
                }
                else
                   base = 0;
#else
                base = 0;
#endif
                /* Verify valid NVRAM checksum. */
                for( cnt = 0; cnt < sizeof(nvram22_t)/2; cnt++ ) {
                    *wptr = qla2100_get_nvram_word(ha, (cnt+base));
                    chksum += (uint8_t)*wptr;
                    chksum += (uint8_t)(*wptr >> 8);
                    wptr++;
                }
#ifdef ISP2300
                if (ha->device_id == QLA2312_DEVICE_ID)
                {
                    /* Unlock resource */
                    WRT_REG_WORD(&reg->host_semaphore, 0);
                }
#endif

#if  DEBUG_PRINT_NVRAM
                qla2100_print(
                "qla2200_nvram_config: Contents of NVRAM ");
                qla2100_print("\n\r");
                qla2100_dump_buffer((uint8_t *)ha->request_ring, sizeof(nvram22_t));
#endif

                /* Bad NVRAM data, set defaults parameters. */
                if( chksum || nv->id[0] != 'I' || nv->id[1] != 'S' || nv->id[2] != 'P' ||
                nv->id[3] != ' ' || nv->nvram_version < 1 ) {
                    /* Reset NVRAM data. */
                    DEBUG(printk("Using defaults for NVRAM: \n"));
                    DEBUG(printk("checksum=0x%x, Id=%c, version=0x%x\n",chksum,nv->id[0],nv->nvram_version));
                    wptr = (uint16_t *)nv;
                    for( cnt = 0; cnt < sizeof(nvram21_t)/2; cnt++ )
                        *wptr++ = 0;

                    /*
                    * Set default initialization control block.
                    */
                    nv->parameter_block_version = ICB_VERSION;
                    nv->firmware_options.enable_fairness = 1;
                    nv->firmware_options.enable_fast_posting = 1;
                    nv->firmware_options.enable_full_login_on_lip = 1;
                    nv->firmware_options.expanded_ifwcb = 1;

                    nv->frame_payload_size  = 1024;
                    nv->max_iocb_allocation = 256;
                    nv->execution_throttle  = 16;
                    nv->retry_count         = 8;
                    nv->retry_delay         = 1;
                    nv->port_name[0]        = 32;
                    nv->port_name[3]        = 224;
                    nv->port_name[4]        = 139;
                    nv->login_timeout       = 4;
#ifdef ISP2200
                    nv->additional_firmware_options.connection_options = P2P_LOOP;
#else
                    nv->additional_firmware_options.connection_options = LOOP_P2P;
#endif
                    /*
                    * Set default host adapter parameters
                    */
                    nv->host_p.enable_lip_full_login = 1;
                    nv->reset_delay = 5;
                    nv->port_down_retry_count = 8;
                    nv->maximum_luns_per_target = 8;
                    status = 1;
                }

                /* Reset NVRAM data. */
                BZERO((caddr_t)icb, sizeof(init_cb_t));
                /*
                * Copy over NVRAM RISC parameter block
                * to initialization control block.
                */
                dptr1 = (caddr_t)icb;
                dptr2 = (caddr_t)&nv->parameter_block_version;
                cnt = (caddr_t)&nv->additional_firmware_options - (caddr_t)&nv->parameter_block_version;
                while( cnt-- )
                    *dptr1++ = *dptr2++;

                dptr1 += (caddr_t)&icb->additional_firmware_options - (caddr_t)&icb->request_q_outpointer;
                cnt = (caddr_t)&nv->host_p - (caddr_t)&nv->additional_firmware_options;
                while( cnt-- )
                    *dptr1++ = *dptr2++;

                /* HBA node name 0 correction */
                for (cnt=0 ; cnt<8 ; cnt++) {
                    if (icb->node_name[cnt] != 0)
                       break;
                }
                if (cnt == 8) {
                   for ( cnt= 0 ; cnt < 8 ; cnt++)
                       icb->node_name[cnt] = icb->port_name[cnt];
                   icb->node_name[0] = icb->node_name[0] & ~BIT_0;
                   icb->port_name[0] = icb->port_name[0] |  BIT_0;
                }

                /*
                * Setup driver firmware options.
                */
                icb->firmware_options.enable_full_duplex       = 0;
#if  QL2100_TARGET_MODE_SUPPORT
                icb->firmware_options.enable_target_mode       = 1;
#else
                icb->firmware_options.enable_target_mode       = 0;
#endif
                icb->firmware_options.disable_initiator_mode   = 0;
                icb->firmware_options.enable_port_update_event = 1;
                icb->firmware_options.enable_full_login_on_lip = 1;
#ifdef ISP2300
                icb->firmware_options.enable_fast_posting = 0;
#endif
#ifndef FC_IP_SUPPORT
                /* Enable FC-Tape support */
                icb->firmware_options.node_name_option = 1;
                icb->firmware_options.expanded_ifwcb = 1;
                icb->additional_firmware_options.enable_fc_tape = 1;
                icb->additional_firmware_options.enable_fc_confirm = 1;
#endif
                /*
                * Set host adapter parameters
                */
                ha->flags.enable_target_mode = icb->firmware_options.enable_target_mode;
                ha->flags.disable_luns            = nv->host_p.disable_luns;
                ha->flags.disable_risc_code_load  = nv->host_p.disable_risc_code_load;
                ha->flags.set_cache_line_size_1   = nv->host_p.set_cache_line_size_1;
                ha->flags.enable_64bit_addressing = nv->host_p.enable_64bit_addressing;
#if BITS_PER_LONG > 32
    /* Enable 64bit addressing for OS/System combination supporting it   */
    /* actual NVRAM bit is: nv->cntr_flags_1.enable_64bit_addressing     */ 
    /* but we will ignore it and use BITS_PER_LONG macro to setup for    */
    /* 64 or 32 bit access of host memory in all x86/ia-64/Alpha systems */
    ha->flags.enable_64bit_addressing = 1;
#else
    ha->flags.enable_64bit_addressing = 0;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,18)
        if (ha->flags.enable_64bit_addressing)
           printk(KERN_INFO "qla2x00: 64 Bit PCI Addressing Enabled\n");

#if BITS_PER_LONG > 32
           /* Update our PCI device dma_mask for full 64 bit mask */
           ha->pdev->dma_mask = 0xffffffffffffffff;
#endif
#endif
                ha->flags.enable_lip_reset        = nv->host_p.enable_lip_reset;
                ha->flags.enable_lip_full_login   = nv->host_p.enable_lip_full_login;
                ha->flags.enable_target_reset     = nv->host_p.enable_target_reset;
                ha->flags.enable_flash_db_update  = nv->host_p.enable_database_storage;
                ha->operating_mode = icb->additional_firmware_options.connection_options;

            /* new for IOCTL support of APIs */
            ha->node_name[0] = icb->node_name[0];
            ha->node_name[1] = icb->node_name[1];
            ha->node_name[2] = icb->node_name[2];
            ha->node_name[3] = icb->node_name[3];
            ha->node_name[4] = icb->node_name[4];
            ha->node_name[5] = icb->node_name[5];
            ha->node_name[6] = icb->node_name[6];
            ha->node_name[7] = icb->node_name[7];
            ha->nvram_version = nv->nvram_version;


                ha->hiwat                 = icb->iocb_allocation;
                ha->execution_throttle  = nv->execution_throttle;

                ha->retry_count           = nv->retry_count;
                /* Set minimum login_timeout to 4 seconds. */
                if( nv->login_timeout < 4 )
                    nv->login_timeout = 4;
                ha->login_timeout         = nv->login_timeout;
                icb->login_timeout         = nv->login_timeout;
                ha->port_down_retry_count = nv->port_down_retry_count;
                ha->minimum_timeout       = (ha->login_timeout * ha->retry_count)
                + ha->port_down_retry_count;
                ha->loop_reset_delay      = nv->reset_delay;
                /* Will get the value from nvram. */
                ha->loop_down_timeout     = LOOP_DOWN_TIMEOUT;
                ha->loop_down_abort_time  = LOOP_DOWN_TIME - ha->loop_down_timeout;

                /* save HBA serial number */
                ha->serial0 = nv->port_name[5];
                ha->serial1 = nv->port_name[6];
                ha->serial2 = nv->port_name[7];
	    	ha->flags.link_down_error_enable  = 1;
                /* save OEM related items for QLA2200s and QLA2300s */
                ha->oem_id     = nv->oem_id;
                ha->oem_spare0 = nv->oem_spare0;
                for ( cnt= 2 ; cnt < 8 ; cnt++ ) 
                  ha->oem_string[cnt] = nv->oem_string[cnt]; 

                for ( cnt= 0 ; cnt < 8 ; cnt++ ) {
                  ha->oem_part[cnt]   = nv->oem_part[cnt]; 
                  ha->oem_fru[cnt]    = nv->oem_fru[cnt]; 
                  ha->oem_ec[cnt]     = nv->oem_ec[cnt]; 
                }

#ifdef FC_IP_SUPPORT
                for (cnt = 0; cnt < 8; cnt++)
                    ha->acPortName[cnt] = nv->port_name[cnt];
#endif

#if  USE_BIOS_MAX_LUNS
                if( !nv->maximum_luns_per_target )
                    ha->max_luns = MAX_LUNS;
                else if( nv->maximum_luns_per_target < MAX_LUNS )
                    ha->max_luns = nv->maximum_luns_per_target;
                else
                    ha->max_luns = MAX_LUNS;
#else
                ha->max_luns = MAX_LUNS;
#endif

                /*
                * Setup ring parameters in initialization control block
                */
                icb->request_q_outpointer  = 0;
                icb->response_q_inpointer  = 0;
                icb->request_q_length      = REQUEST_ENTRY_CNT;
                icb->response_q_length     = RESPONSE_ENTRY_CNT;
                icb->request_q_address[0]  = LS_64BITS(ha->request_dma);
                icb->request_q_address[1]  = MS_64BITS(ha->request_dma);
                icb->response_q_address[0] = LS_64BITS(ha->response_dma);
                icb->response_q_address[1] = MS_64BITS(ha->response_dma);

                icb->lun_enables = 0;
                icb->command_resource_count = 0;
                icb->immediate_notify_resource_count = 0;
                icb->timeout = 0;
                icb->reserved_2 = 0;

                ha->flags.nvram_config_done = 1;
            }


#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
            if( status )
                qla2100_print("qla2200_nvram_config: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
            else
                LEAVE("qla2200_nvram_config");
#endif
            return(status);
}

/*
* Get NVRAM data word
*      Calculates word position in NVRAM and calls request routine to
*      get the word from NVRAM.
*
* Input:
*      ha      = adapter block pointer.
*      address = NVRAM word address.
*
* Returns:
*      data word.
*/
STATIC uint16_t
qla2100_get_nvram_word(scsi_qla_host_t *ha, uint32_t address)
{
            uint32_t nv_cmd;
            uint16_t data;

#ifdef QL_DEBUG_ROUTINES
            uint8_t  saved_print_status = ql2x_debug_print;
#endif
#ifdef QL_DEBUG_LEVEL_4
            qla2100_print("qla2100_get_nvram_word: entered\n");
#endif

            nv_cmd = address << 16;
            nv_cmd |= NV_READ_OP;

#ifdef QL_DEBUG_ROUTINES
            ql2x_debug_print = FALSE;
#endif
            data = qla2100_nvram_request(ha, nv_cmd);
#ifdef QL_DEBUG_ROUTINES
            ql2x_debug_print = saved_print_status;
#endif

#ifdef QL_DEBUG_LEVEL_4
            qla2100_print("qla2100_get_nvram_word: exiting normally NVRAM data = ");
        qla2100_output_number((u_long)data, 16);
            qla2100_print("\n");
#endif
            return(data);
}

/*
* NVRAM request
*      Sends read command to NVRAM and gets data from NVRAM.
*
* Input:
*      ha     = adapter block pointer.
*      nv_cmd = Bit 26     = start bit
*               Bit 25, 24 = opcode
*               Bit 23-16  = address
*               Bit 15-0   = write data
*
* Returns:
*      data word.
*/
STATIC uint16_t
qla2100_nvram_request(scsi_qla_host_t *ha, uint32_t nv_cmd)
{
            uint8_t      cnt;
            device_reg_t *reg = ha->iobase;
            uint16_t     data = 0;
            uint16_t     reg_data;

            /* Send command to NVRAM. */

            nv_cmd <<= 5;
            for( cnt = 0; cnt < 11; cnt++ ) {
                if( nv_cmd & BIT_31 )
                    qla2100_nv_write(ha, NV_DATA_OUT);
                else
                    qla2100_nv_write(ha, 0);
                nv_cmd <<= 1;
            }

            /* Read data from NVRAM. */

            for( cnt = 0; cnt < 16; cnt++ ) {
                WRT_REG_WORD(&reg->nvram, NV_SELECT+NV_CLOCK);
                /* qla2100_nv_delay(ha); */
                NVRAM_DELAY();
                data <<= 1;
                reg_data = RD_REG_WORD(&reg->nvram);
                if( reg_data & NV_DATA_IN )
                    data |= BIT_0;
                WRT_REG_WORD(&reg->nvram, NV_SELECT);
                /* qla2100_nv_delay(ha); */
                NVRAM_DELAY();
            }

            /* Deselect chip. */

            WRT_REG_WORD(&reg->nvram, NV_DESELECT);
            /* qla2100_nv_delay(ha); */
            NVRAM_DELAY();

            return(data);
}

STATIC void
qla2100_nv_write(scsi_qla_host_t *ha, uint16_t data)
{
            device_reg_t *reg = ha->iobase;

            WRT_REG_WORD(&reg->nvram, data | NV_SELECT);
            NVRAM_DELAY();
            /* qla2100_nv_delay(ha); */
            WRT_REG_WORD(&reg->nvram, data | NV_SELECT | NV_CLOCK);
            /* qla2100_nv_delay(ha); */
            NVRAM_DELAY();
            WRT_REG_WORD(&reg->nvram, data | NV_SELECT);
            /* qla2100_nv_delay(ha); */
            NVRAM_DELAY();
}

STATIC void
qla2x00_nv_deselect(adapter_state_t *ha)
{
	device_reg_t *reg = ha->iobase;

	WRT_REG_WORD(&reg->nvram, NV_DESELECT);
	NVRAM_DELAY();
}

/*
* qla2100_poll
*      Polls ISP for interrupts.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_poll(scsi_qla_host_t *ha)
{
            device_reg_t *reg   = ha->iobase;
            uint8_t     discard;
            uint16_t     data;
            srb_t        *done_q_first = 0;
            srb_t        *done_q_last = 0;

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2100_poll");
#endif

            /* Acquire interrupt specific lock */
            QLA2100_INTR_LOCK(ha);

            /* Check for pending interrupts. */
#if defined(ISP2100) || defined(ISP2200)
            data = RD_REG_WORD(&reg->istatus);
            if( data & RISC_INT )
                qla2100_isr(ha, data, (srb_t **)&done_q_first, (srb_t **)&done_q_last, &discard);
#else
            if (ha->device_id == QLA2312_DEVICE_ID)
            {
                data = RD_REG_WORD(&reg->istatus);
                if( data & RISC_INT )
                {
                    data = RD_REG_WORD(&reg->host_status_lo);
                    qla2100_isr(ha, data, (srb_t **)&done_q_first, (srb_t **)&done_q_last, &discard);
                }

            }
            else
            {
                data = RD_REG_WORD(&reg->host_status_lo);
                if( data & HOST_STATUS_INT )
                    qla2100_isr(ha, data, (srb_t **)&done_q_first, (srb_t **)&done_q_last, &discard);
            }
#endif

            /* Release interrupt specific lock */
            QLA2100_INTR_UNLOCK(ha);


            if( done_q_first )
                qla2100_done(ha, (srb_t **)&done_q_first, (srb_t **)&done_q_last);

#ifdef QL_DEBUG_LEVEL_3
            LEAVE("qla2100_poll");
#endif
}

/*
*  qla2100_restart_isp
*      restarts the ISP after a reset
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success
*/
int
qla2x00_restart_isp(scsi_qla_host_t *ha)
{
       uint8_t status = 0;

        /* If firmware needs to be loaded */
        if( qla2100_isp_firmware(ha) ) {
                ha->flags.online = FALSE;
                if( !(status = qla2100_chip_diag(ha)) )
                    status = qla2x00_setup_chip(ha);
        }
       	if (!status && !(status = qla2x00_init_rings(ha))) {
       	   ha->dpc_flags &= ~(RESET_MARKER_NEEDED | COMMAND_WAIT_NEEDED);
               if( !(status = qla2x00_fw_ready(ha)) ) {
		DEBUG(printk("restart_isp: Start configure loop ,status = %d\n",status);)
                ha->flags.online = TRUE;
       		do {
       			ha->dpc_flags &= ~LOOP_RESYNC_NEEDED;
       			qla2100_configure_loop(ha);
       		} while (!ha->loop_down_timer && 
			!(ha->dpc_flags & ISP_ABORT_NEEDED) &&
			(ha->dpc_flags & LOOP_RESYNC_NEEDED) );
	       }
	  DEBUG(printk("restart_isp: Configure loop done, status = 0x%x\n",status);)
	}
	return( status );
}

/*
*  qla2100_abort_isp
*      Resets ISP and aborts all outstanding commands.
*
* Input:
*      ha           = adapter block pointer.
*
* Returns:
*      0 = success
*/
STATIC uint8_t
qla2100_abort_isp(scsi_qla_host_t *ha)
{
	uint16_t       cnt;
	srb_t          *sp;
	os_lun_t      *q;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
	unsigned long cpu_flags = 0;
#endif
	uint8_t        status = 0;

	ENTER("qla2100_abort_isp");

	DRIVER_LOCK
	if( ha->flags.online ) {
	    ha->flags.online = FALSE;
	    ha->dpc_flags &= ~COMMAND_WAIT_NEEDED;
	    ha->dpc_flags &= ~COMMAND_WAIT_ACTIVE;
	    qla2100_stats.ispAbort++;
            ha->total_isp_aborts++;  /* used by ioctl */
	    ha->sns_retry_cnt = 0;

	    printk(KERN_INFO
	      "qla2x00: Performing ISP error recovery - ha= %p.\n", ha);
	    qla2100_reset_chip(ha);

	    if( ha->loop_state != LOOP_DOWN ) {
	        ha->loop_state = LOOP_DOWN;
	        ha->loop_down_timer = LOOP_DOWN_TIME;
	    }

#ifdef FC_IP_SUPPORT
	    /* Return all IP send packets */
	    for (cnt = 0; cnt < MAX_SEND_PACKETS; cnt++) {
	        if (ha->apActiveIpQueue[cnt] != NULL) {
	            (*ha->pSendCompletionRoutine)(ha->apActiveIpQueue[cnt]);

	            ha->apActiveIpQueue[cnt] = NULL;
	        }
	    }
#endif

	    /* Requeue all commands in outstanding command list. */
	    for( cnt = 0; cnt < MAX_OUTSTANDING_COMMANDS; cnt++ ) {
	        sp = ha->outstanding_cmds[cnt];
	        if( sp ) {
	            ha->outstanding_cmds[cnt] = 0;
                    ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads--;
		    q = sp->lun_queue;
	            q->q_outcnt = 0;
	            q->q_flag &= ~QLA2100_QBUSY;
	            sp->flags = 0;
	            /* we need to send the command back to OS */
	            CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
	            CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;
	            qla2100_done_q_put(ha, sp, (srb_t **)&ha->done_q_first, (srb_t **)&ha->done_q_last);
	        }
	    }
		/* fixme(dg): Retrun the ones in the retry queue as well */

	    qla2200_nvram_config(ha);

	    if( !qla2x00_restart_isp(ha) ) {
	        ha->dpc_flags &= ~RESET_MARKER_NEEDED;

	        if( !ha->loop_down_timer )
	            qla2x00_marker(ha, 0, 0, MK_SYNC_ALL);

	        ha->flags.online = TRUE;

	        /* Enable target response to SCSI bus. */
	        if( ha->flags.enable_target_mode )
	            qla2100_enable_lun(ha);

#ifdef FC_IP_SUPPORT
	        /* Reenable IP support */
	        if (ha->flags.enable_ip)
	            qla2x00_ip_initialize(ha);
#endif
	        /* Enable ISP interrupts. */
	        qla2100_enable_intrs(ha);

	        /* v2.19.5b6 Return all commands */
	        qla2x00_abort_queues(ha, TRUE);

	        /* Restart queues that may have been stopped. */
	        /* 6/9 if( !ha->loop_down_timer ) */
	            qla2x00_restart_queues(ha,TRUE);
	    } else {
	        printk(KERN_WARNING
	        "qla2100: ISP error recovery failed - board disabled\n");
	        qla2100_reset_adapter(ha);
	        qla2x00_abort_queues(ha, FALSE);
	    	ha->flags.online = TRUE;
	    }
	}

	if( status ) {
		printk(KERN_INFO "qla2100_abort_isp(%ld): **** FAILED ****\n",
		    ha->host_no);
	} else {
		DEBUG(printk(KERN_INFO "qla2100_abort_isp(%ld): exiting.\n",
		    ha->host_no);)
	}

	return(status);
}

/*
* qla2100_init_fc_db
*      Initializes Fibre Channel Device Database.
*
* Input:
*      ha = adapter block pointer.
*
* Output:
*      ha->fc_db = initialized
*/
STATIC void
qla2100_init_fc_db(scsi_qla_host_t *ha)
{
            uint16_t cnt;

#ifdef QL_DEBUG_LEVEL_3
            ENTER("qla2100_init_fc_db");
#endif

            /* Initialize fc database if it is not initialized. */
            if( !ha->fc_db[0].loop_id && !ha->fc_db[1].loop_id ) {
                ha->flags.updated_fc_db = FALSE;

                /* Initialize target database. */
                for( cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++ ) {
                    ha->fc_db[cnt].name[0] = 0L;
                    ha->fc_db[cnt].name[1] = 0L;
                    ha->fc_db[cnt].loop_id = PORT_UNUSED;
                    ha->fc_db[cnt].port_login_retry_count = 8;
                    ha->fc_db[cnt].flag = 0;   /* v2.19.5b3 */
                }

#if USE_FLASH_DATABASE
                /* Move flash database to driver database. */
                qla2100_get_database(ha);
#endif
            }

#ifdef QL_DEBUG_LEVEL_3
            LEAVE("qla2100_init_fc_db");
#endif
}


/*
* qla2100_init_tgt_map
*      Initializes target map.
*
* Input:
*      ha = adapter block pointer.
*
* Output:
*      TGT_Q initialized
*/
STATIC void
qla2100_init_tgt_map(scsi_qla_host_t *ha)
{
                uint32_t t;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_init_tgt_map");
#endif
                    for( t = 0; t < MAX_TARGETS; t++ )
                        TGT_Q(ha, t) = (os_tgt_t *) NULL;
#ifdef QL_DEBUG_LEVEL_3
                    LEAVE("qla2100_init_tgt_map");
#endif
}

#if USE_FLASH_DATABASE
/*
* qla2100_get_database
*      Copies and converts flash database to driver database.
*      (may sleep)
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_get_database(scsi_qla_host_t *ha)
{
                flash_database_t *fptr;
                uint8_t          status = 1;
                uint32_t         addr;
                uint16_t         cnt;
                uint8_t          *bptr;
                uint8_t          checksum;
                uint32_t         b, t;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_get_database");
#endif

                /* Default setup. */
                ha->flash_db = FLASH_DATABASE_0;
                ha->flash_seq = 0;

                if( (fptr = (flash_database_t *)KMALLOC(sizeof(flash_database_t))) ) {
                    /* Enable Flash Read/Write. */
                    qla2100_flash_enable_database(ha);

                    /* Start with flash database with the highest sequence number. */
                    b = qla2100_read_flash_byte(ha, FLASH_DATABASE_0);
                    b |= qla2100_read_flash_byte(ha, FLASH_DATABASE_0 + 1) << 8;
                    b |= qla2100_read_flash_byte(ha, FLASH_DATABASE_0 + 1) << 16;
                    b |= qla2100_read_flash_byte(ha, FLASH_DATABASE_0 + 1) << 24;
                    t = qla2100_read_flash_byte(ha, FLASH_DATABASE_1);
                    t |= qla2100_read_flash_byte(ha, FLASH_DATABASE_1 + 1) << 8;
                    t |= qla2100_read_flash_byte(ha, FLASH_DATABASE_1 + 1) << 16;
                    t |= qla2100_read_flash_byte(ha, FLASH_DATABASE_1 + 1) << 24;
                    if( t > b ) {
                        ha->flash_db = FLASH_DATABASE_1;
                    }

                    /* Select the flash database with the good checksum. */
                    for( t = 0; t < 2; t++ ) {
                        checksum = 0;
                        addr = ha->flash_db;
                        bptr = (uint8_t *)fptr;
                        fptr->hdr.size = sizeof(flash_database_t);

                        /* Read flash database to driver. */
                        for( cnt = 0; cnt < fptr->hdr.size; cnt++ ) {
                            *bptr = (uint8_t)qla2100_read_flash_byte(ha, addr++);
                            checksum += *bptr++;
                            if( bptr == &fptr->hdr.spares[0] &&
                            (fptr->hdr.size > sizeof(flash_database_t) ||
                            fptr->hdr.size < sizeof(flash_hdr_t) ||
                            !fptr->hdr.version) ) {
                                checksum = 1;
                                break;
                            }
                        }

                        if( !checksum ) {
                            status = 0;
                            break;
                        }
                        /* trying other database */
                        if( ha->flash_db == FLASH_DATABASE_0 ) {
                            ha->flash_db = FLASH_DATABASE_1;
                        } else {
                            ha->flash_db = FLASH_DATABASE_0;
                        }
                    }

                    if( !status ) {
                        ha->flash_seq = fptr->hdr.seq;

                        /* Convert flash database to driver database format. */
                        if( fptr->hdr.size -= sizeof(flash_hdr_t) ) {
                            for( cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++ ) {
                                ha->fc_db[cnt].name[0] = fptr->node[cnt].name[0];
                                ha->fc_db[cnt].name[1] = fptr->node[cnt].name[1];
                                cnt,
                                ha->fc_db[cnt].name[1],
                                ha->fc_db[cnt].name[0]);

                                ha->fc_db[cnt].loop_id = PORT_AVAILABLE;
                                ha->fc_db[cnt].flag = 0;  /* v2.19.05b3 */
                                if( !(fptr->hdr.size -= sizeof(flash_node_t)) )
                                    break;
                            }
                        }
                    }

                    qla2100_flash_disable_database(ha);

                    KMFREE(fptr, sizeof(flash_database_t));
                }
				else
				{
					printk(KERN_WARNING "scsi(%d): Memory Allocation failed - flash mem",(int)ha->host_no);
					ha->mem_err++;
				}

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( status )
                    qla2100_print("qla2100_get_database: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_get_database");
#endif
                return(status);
}

/*
* qla2100_save_database
*      Copies and converts driver database to flash database.
*      (may sleep)
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_save_database(scsi_qla_host_t *ha)
{
                flash_database_t *fptr;
                uint8_t          status = 1;
                uint32_t         addr;
                uint16_t         cnt;
                uint8_t          *bptr;
                uint8_t          checksum;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_save_database");
#endif

                if( (fptr = (flash_database_t *)KMALLOC(sizeof(flash_database_t)) ) ) {
                    /* Enable Flash Read/Write. */
                    qla2100_flash_enable_database(ha);

                    fptr->hdr.seq     = ++ha->flash_seq;
                    fptr->hdr.version = FLASH_DATABASE_VERSION;
                    fptr->hdr.size    = sizeof(flash_hdr_t);

                    /* Copy and convert driver database to flash database. */
                    for( cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++ ) {
                        if( ha->fc_db[cnt].loop_id == PORT_UNUSED )
                            break;
                        else {
                            fptr->node[cnt].name[0] = ha->fc_db[cnt].name[0];
                            fptr->node[cnt].name[1] = ha->fc_db[cnt].name[1];
                            fptr->hdr.size += sizeof(flash_node_t);
                        }
                    }

                    /* Calculate checksum. */
                    checksum = 0;
                    bptr = (uint8_t *)fptr;
                    for( cnt = 0; cnt < fptr->hdr.size; cnt++ )
                        checksum += *bptr++;
                    fptr->hdr.checksum = ~checksum + 1;

                    /* Setup next sector address for flash */
                    if( ha->flash_db == FLASH_DATABASE_0 )
                        addr = FLASH_DATABASE_1;
                    else
                        addr = FLASH_DATABASE_0;
                    ha->flash_db = addr;

                    /* Erase flash sector prior to write. */
                    status = qla2100_erase_flash_sector(ha, addr);

                    /* Write database to flash. */
                    bptr = (uint8_t *)fptr;
                    for( cnt = 0; cnt < fptr->hdr.size && !status; cnt++ )
                        status = qla2100_program_flash_address(ha, addr++, *bptr++);

                    qla2100_flash_disable_database(ha);

                    KMFREE(fptr, sizeof(flash_database_t));
                }
				else
				{
					printk(KERN_WARNING "scsi(%d): Memory Allocation failed - flash",(int)ha->host_no);
					ha->mem_err++;
				}

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( status )
                    qla2100_print("qla2100_save_database: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_save_database");
#endif
                return(status);
}

/*
* qla2100_program_flash_address
*      Program flash address.
*
* Input:
*      ha   = adapter block pointer.
*      addr = flash byte address.
*      data = data to be written to flash.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_program_flash_address(scsi_qla_host_t *ha, uint32_t addr, uint8_t data)
{
                uint8_t status;

                /* Write Program Command Sequence */
                qla2100_write_flash_byte(ha, 0x5555, 0xaa);
                qla2100_write_flash_byte(ha, 0x2aaa, 0x55);
                qla2100_write_flash_byte(ha, 0x5555, 0xa0);
                qla2100_write_flash_byte(ha, addr, data);

                /* Wait for write to complete. */
                status = qla2100_poll_flash(ha, addr, data);

#ifdef QL_DEBUG_LEVEL_2
                if( status )
                    qla2100_print("qla2100_program_flash_address: **** FAILED ****\n");
#endif
                return(status);
}

/*
* qla2100_erase_flash_sector
*      Erases flash sector.
*
* Input:
*      ha   = adapter block pointer.
*      addr = sector address.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_erase_flash_sector(scsi_qla_host_t *ha, uint32_t addr)
{
                uint8_t status;

                addr &= 0x1c000;

                /* Individual Sector Erase Command Sequence */
                qla2100_write_flash_byte(ha, 0x5555, 0xaa);
                qla2100_write_flash_byte(ha, 0x2aaa, 0x55);
                qla2100_write_flash_byte(ha, 0x5555, 0x80);
                qla2100_write_flash_byte(ha, 0x5555, 0xaa);
                qla2100_write_flash_byte(ha, 0x2aaa, 0x55);
                qla2100_write_flash_byte(ha, addr, 0x30);

                SYS_DELAY(150);

                /* Wait for erase to complete. */
                status = qla2100_poll_flash(ha, addr, 0x80);

#ifdef QL_DEBUG_LEVEL_2
                if( status )
                    qla2100_print("qla2100_erase_flash_sector: **** FAILED ****\n");
#endif
                return(status);
}

/*
* qla2100_poll_flash
*      Polls flash for completion.
*
* Input:
*      ha   = adapter block pointer.
*      addr = flash byte address.
*      data = data to be polled.
*
* Returns:
*      0 = success.
*/
STATIC uint8_t
qla2100_poll_flash(scsi_qla_host_t *ha, uint32_t addr, uint8_t poll_data)
{
                uint8_t  status = 1;
                uint8_t  flash_data;
                uint32_t cnt;

                poll_data &= BIT_7;

                /* Wait for 30 seconds for command to finish. */
                for( cnt = 3000000; cnt; cnt-- ) {
                    flash_data = (uint8_t)qla2100_read_flash_byte(ha, addr);

                    if( (flash_data & BIT_7) == poll_data ) {
                        status = 0;
                        break;
                    }
                    if( flash_data & BIT_5 && cnt > 2 )
                        cnt = 2;
                    SYS_DELAY(10);
                }

                return(status);
}

/*
* qla2100_flash_enable_database
*      Setup flash for reading/writing.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_flash_enable_database(scsi_qla_host_t *ha)
{
                device_reg_t *reg = ha->iobase;

                /* Setup bit 16 of flash address. */
                WRT_REG_WORD(&reg->nvram, NV_SELECT);

                /* Enable Flash Read/Write. */
                WRT_REG_WORD(&reg->ctrl_status, ISP_FLASH_ENABLE);

                /* Read/Reset Command Sequence */
                qla2x00_write_flash_byte(ha, 0x5555, 0xaa);
                qla2x00_write_flash_byte(ha, 0x2aaa, 0x55);
                qla2x00_write_flash_byte(ha, 0x5555, 0xf0);
                qla2x00_read_flash_byte(ha, FLASH_DATABASE_0);
}

/*
* qla2100_flash_disable_database
*      Disable flash and allow RISC to run.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_flash_disable_database(scsi_qla_host_t *ha)
{
                device_reg_t *reg = ha->iobase;

                /* Restore chip registers. */
                WRT_REG_WORD(&reg->ctrl_status, 0);
                WRT_REG_WORD(&reg->nvram, 0);
}

/*
* qla2x00_write_flash_byte
*      Write byte to flash.
*
* Input:
*      ha   = adapter block pointer.
*      addr = flash byte address.
*      data = data to be written.
*/
STATIC void
qla2x00_write_flash_byte(scsi_qla_host_t *ha, uint32_t addr, uint8_t data)
{
	device_reg_t *reg = ha->iobase;

	WRT_REG_WORD(&reg->flash_address, (uint16_t)addr);
	WRT_REG_WORD(&reg->flash_data, (uint16_t)data);
}

#endif

/*
 * qla2x00_flash_enable
 *      Setup flash for reading/writing.
 *
 * Input:
 *      ha = adapter block pointer.
 */
STATIC void
qla2x00_flash_enable(adapter_state_t *ha)
{
	device_reg_t *reg = ha->iobase;

	/* Enable Flash Read/Write. */
	WRT_REG_WORD(&reg->ctrl_status, ISP_FLASH_ENABLE);

}

/*
 * qla2x00_flash_disable
 *      Disable flash and allow RISC to run.
 *
 * Input:
 *      ha = adapter block pointer.
 */
STATIC void
qla2x00_flash_disable(adapter_state_t *ha)
{
	device_reg_t *reg = ha->iobase;

	/* Restore chip registers. */
	WRT_REG_WORD(&reg->ctrl_status, 0);
}

/*
 * qla2x00_read_flash_byte
 *      Reads byte from flash, but must read a word from chip.
 *
 * Input:
 *      ha   = adapter block pointer.
 *      addr = flash byte address.
 *
 * Returns:
 *      byte from flash.
 */
STATIC uint8_t
qla2x00_read_flash_byte(scsi_qla_host_t *ha, uint32_t addr)
{
	device_reg_t *reg = ha->iobase;
	uint16_t     data;

	WRT_REG_WORD(&reg->flash_address, (uint16_t)addr);
	data = qla2100_debounce_register(&reg->flash_data);

	return((uint8_t)data);
}

/*
 * qla2x00_get_flash_version
 *      Reads version info from flash.
 *
 * Input:
 *      ha   = adapter struct pointer.
 *
 * Returns:
 *      byte from flash.
 */
STATIC uint16_t
qla2x00_get_flash_version(adapter_state_t *ha)
{
	uint16_t	ret = QL_STATUS_SUCCESS;
	uint32_t	loop_cnt = 1;  /* this is for error exit only */
	uint32_t	pcir_adr;

	DEBUG3(printk("qla2x00_get_flash_version: entered.\n");)

	qla2x00_flash_enable(ha);

	do {	/* Loop once to provide quick error exit */

		/* Match signature */
		if (!(qla2x00_read_flash_byte(ha, 0) == 0x55 &&
		    qla2x00_read_flash_byte(ha, 1) == 0xaa)) {
			/* No signature */
			DEBUG2(printk("qla2x00_get_flash_version: No matching "
			    "signature.\n");)
			ret = QL_STATUS_ERROR;
			break;
		}

		pcir_adr = qla2x00_read_flash_byte(ha, 0x18) & 0xff;

		/* validate signature of PCI data structure */
		if ((qla2x00_read_flash_byte(ha, pcir_adr)) == 'P' &&
		    (qla2x00_read_flash_byte(ha, pcir_adr + 1)) == 'C' &&
		    (qla2x00_read_flash_byte(ha, pcir_adr + 2)) == 'I' &&
		    (qla2x00_read_flash_byte(ha, pcir_adr + 3)) == 'R') {

			/* Read version */
			ha->optrom_minor = qla2x00_read_flash_byte(ha,
			    pcir_adr + 0x12);
			ha->optrom_major = qla2x00_read_flash_byte(ha,
			    pcir_adr + 0x13);
			DEBUG3(printk("qla2x00_get_flash_version: got %d.%d.\n",
			    ha->optrom_major, ha->optrom_minor);)

		} else {
			/* error */
			DEBUG2(printk("qla2x00_get_flash_version: PCI data "
			    "struct not found. pcir_adr=%x.\n", pcir_adr);)
			ret = QL_STATUS_ERROR;
			break;
		}

	} while (--loop_cnt);

	qla2x00_flash_disable(ha);

	DEBUG3(printk("qla2x00_get_flash_version: exiting normally.\n");)
	return (ret);
}

/*
* qla2100_reset_adapter
*      Reset adapter.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_reset_adapter(scsi_qla_host_t *ha)
{
                device_reg_t *reg = ha->iobase;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_reset_adapter");
#endif

                ha->flags.online = FALSE;
                qla2100_disable_intrs(ha);
                /* WRT_REG_WORD(&reg->ictrl, 0); */
                /* Reset RISC processor. */
                WRT_REG_WORD(&reg->host_cmd, HC_RESET_RISC);
                WRT_REG_WORD(&reg->host_cmd, HC_RELEASE_RISC);
#ifdef QL_DEBUG_LEVEL_3
                LEAVE("qla2100_reset_adapter");
#endif
}

/*
* qla2x00_loop_reset
*      Issue loop reset.
*
* Input:
*      ha = adapter block pointer.
*
* Returns:
*      0 = success
*/
STATIC uint8_t
qla2x00_loop_reset(scsi_qla_host_t *ha)
{
	uint8_t  status = QL_STATUS_SUCCESS;
	uint16_t t;

	DEBUG3(printk("qla2x00_loop_reset(%ld): entered.\n",
	    ha->host_no);)

	if (ha->flags.enable_lip_reset) {
		status = qla2x00_lip_reset(ha);
	}

	if (status == QL_STATUS_SUCCESS && ha->flags.enable_target_reset) {
		for (t = 0; t < MAX_FIBRE_DEVICES; t++) {
			if (ha->otgt[t] && ha->otgt[t]->vis_port == NULL)
				continue; 

			status = qla2x00_target_reset(ha, 0, t);
			if (status != QL_STATUS_SUCCESS) {
				break;
			}
		}
	}

	if (status == QL_STATUS_SUCCESS &&
	    ((!ha->flags.enable_target_reset && !ha->flags.enable_lip_reset) ||
	    ha->flags.enable_lip_full_login)) {
		status = qla2x00_full_login_lip(ha);
	}

	/* Issue marker command. */
	qla2x00_marker(ha, 0, 0, MK_SYNC_ALL);

	if (status) {
		/* Empty */
		DEBUG2_3(printk("qla2x00_loop_reset(%ld): **** FAILED ****\n",
		    ha->host_no);)
	} else {
		/* Empty */
		DEBUG3(printk("qla2x00_loop_reset(%ld): exiting normally.\n",
		    ha->host_no);)
	}

	return(status);
}

/*
 * qla2x00_device_reset
 *	Issue bus device reset message to the target.
 *
 * Input:
 *	ha = adapter block pointer.
 *	t = SCSI ID.
 *	TARGET_QUEUE_LOCK must be released.
 *	ADAPTER_STATE_LOCK must be released.
 *
 * Context:
 *	Kernel context.
 */
STATIC  int
qla2x00_device_reset(scsi_qla_host_t *ha, uint16_t t)
{
	uint16_t	l;
	os_lun_t	*lq;
	fc_port_t	*fcport;
	uint8_t  status = 0;

	ENTER("qla2x00_device_reset:");

	for (l = 0; l < MAX_LUNS; l++) {
		lq = LUN_Q(ha, t, l);
		if (lq == NULL)
			continue;
		fcport = lq->fclun->fcport;
		if (LOOP_RDY(fcport->ha)) {
			qla2x00_abort_device(fcport->ha, fcport->loop_id,
			    lq->fclun->lun);
		}
	}
	LEAVE("qla2x00_device_reset: exiting normally");
	return( status );
}

/*
 *  Issue marker command.
 *	Function issues marker IOCB.
 *
 * Input:
 *	ha = adapter block pointer.
 *	loop_id = loop ID
 *	lun = LUN
 *	type = marker modifier
 *
 * Returns:
 *	qla2x00 local function return status code.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
STATIC int
qla2x00_marker(scsi_qla_host_t *ha, uint16_t loop_id, uint16_t lun,
    uint8_t type)
{
	mrk_entry_t	*pkt;

	ENTER("qla2x00_marker: entered");

	pkt = (mrk_entry_t *)qla2100_req_pkt(ha);
	if (pkt == NULL) {
#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
        qla2100_print("qla2x00_marker: **** FAILED ****\n");
#endif
		return (QLA2X00_FAILED);
	}

	pkt->entry_type = MARKER_TYPE;
	pkt->modifier = type;

	if (type != MK_SYNC_ALL) {
		pkt->lun = lun;
		pkt->target = (uint8_t)loop_id;
	}

	/* Issue command to ISP */
	qla2100_isp_cmd(ha);

	LEAVE("qla2x00_marker: exiting normally");

	return (QLA2X00_SUCCESS);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,3,18)
/*
* qla2100_64bit_start_scsi
*      The start SCSI is responsible for building request packets on
*      request ring and modifying ISP input pointer.
*
* Input:
*      ha = adapter block pointer.
*      sp = SB structure pointer.
*
* Returns:
*      0 = success, was able to issue command.
*/
STATIC uint8_t
qla2x00_64bit_start_scsi(srb_t *sp)
{
                device_reg_t    *reg;
                uint8_t         status = 0;
                uint8_t         found = 0;
                Scsi_Cmnd       *cmd = sp->cmd;
                uint32_t        cnt, index;
                cmd_a64_entry_t     *pkt;
                uint16_t        req_cnt;
                uint16_t        seg_cnt;
                uint16_t        cdb_len,temp;
                struct scatterlist    *sg = (struct scatterlist *) NULL;
                uint32_t        timeout;
                caddr_t         data_ptr;
                uint32_t        *dword_ptr;
                uint64_t        dma_handle;
            	scsi_qla_host_t *ha;
		fc_lun_t	*fclun;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_64bit_start_scsi:");
#endif
                DEBUG(sprintf(debug_buff,
                 "64bit_start: cmd=%p sp=%p CDB=%x\n\r",cmd,sp,cmd->cmnd[0]);)
                DEBUG(qla2100_print(debug_buff));

				/* Setup device pointers. */
				fclun = sp->lun_queue->fclun;
				ha = fclun->fcport->ha;
				reg = ha->iobase;

                /* Calculate number of entries and segments required. */
                seg_cnt = 0;
                req_cnt = 1;
                if ( cmd->use_sg ) {
                    /* 4.10 64 bit S/G Data Transfer */
                    sg = (struct scatterlist *) cmd->request_buffer;
                    seg_cnt = pci_map_sg(ha->pdev,
                               sg,cmd->use_sg,
                               scsi_to_pci_dma_dir(cmd->sc_data_direction));

                    if( seg_cnt > 2 ) {
                        req_cnt += (uint16_t)(seg_cnt - 2) / 5;
                        if( (uint16_t)(seg_cnt - 2) % 5 )
                            req_cnt++;
                    }
                } else if( cmd->request_bufflen )
                {   /* no S/G Data Transfer */
                    DEBUG5(printk("Single data transfer (0x%x)\n",
                           cmd->request_bufflen));
                    seg_cnt = 1;
                }

                /* Acquire ring specific lock */
                QLA2100_RING_LOCK(ha);

                if( (uint16_t)(req_cnt + 2) >= ha->req_q_cnt ) {
                    /* Calculate number of free request entries. */
#if defined(ISP2100) || defined(ISP2200)
                    cnt = RD_REG_WORD(&reg->mailbox4);
#else
                    cnt = RD_REG_WORD(&reg->req_q_out);
#endif
                    if( ha->req_ring_index < cnt )
                        ha->req_q_cnt = cnt - ha->req_ring_index;
                    else
                        ha->req_q_cnt = REQUEST_ENTRY_CNT - (ha->req_ring_index - cnt);
                }

                /* If room for request in request ring. */
                if( (uint16_t)(req_cnt + 2) < ha->req_q_cnt ) {
                    /* Check for room in outstanding command list. */
                    cnt = ha->current_outstanding_cmd;
                    for( index = 0; index < MAX_OUTSTANDING_COMMANDS; index++)
                    {
                        cnt++;
                        if (cnt == MAX_OUTSTANDING_COMMANDS)
                           cnt = 0;
                        if (ha->outstanding_cmds[cnt] == 0)
                        {
                           found = 1;
                           ha->current_outstanding_cmd = cnt;
                           break;
                        }  
                    }
                    if( found ) {
                        ha->outstanding_cmds[cnt] = sp;
                        ha->req_q_cnt -= req_cnt;
                        CMD_HANDLE(sp->cmd) = (unsigned char *) (u_long) cnt;

                        /*
                        * Build command packet.
                        */
                        pkt = ha->request_ring_ptr;

                        pkt->entry_type = COMMAND_A64_TYPE;
                        pkt->entry_count = (uint8_t)req_cnt;
                        pkt->sys_define = (uint8_t)ha->req_ring_index;
                        pkt->control_flags= 0;
                        pkt->entry_status = 0;
                        pkt->handle = (uint32_t)cnt;

                        /* Zero out remaining portion of packet. */
                        dword_ptr = (uint32_t *)pkt + 2;
                        for( cnt = 2; cnt < REQUEST_ENTRY_SIZE/4; cnt++ )
                            *dword_ptr++ = 0;

                        /*
                        * We subtract 5 sec. from the timeout value to insure
                        * the ISP time-out before the mid-level or the driver.
                        */
                        timeout = (uint32_t) CMD_TIMEOUT(cmd)/HZ;
                        if( timeout > 5 )
                            pkt->timeout = (uint16_t) timeout - 5;
                        else
                            pkt->timeout = (uint16_t) timeout;
#if VSA
			if( (cmd->data_cmnd[0] == 0x26) ||
			    (cmd->data_cmnd[0] == 0xA0) ||
			    (cmd->data_cmnd[0] == 0xCB) ) {
				pkt->lun = fclun->lun;
			}
                        else if( (fclun->fcport->flags & FC_VSA) ) 
				pkt->lun = (fclun->lun | 0x4000);
			else
				pkt->lun = fclun->lun;
#else
			/* Set LUN number*/
			pkt->lun = fclun->lun;
#endif

			/* Set target ID */
			pkt->target = (uint8_t)fclun->fcport->loop_id;

                        /* Enable simple tag queuing if device supports it. */
                        if ( cmd->device->tagged_queue ) {
                            switch (cmd->tag) {
                                case SIMPLE_QUEUE_TAG:
                                    pkt->control_flags = CF_SIMPLE_TAG;
                                    break;
                                case HEAD_OF_QUEUE_TAG:
                                    pkt->control_flags = CF_HEAD_TAG;
                                    break;
                                case ORDERED_QUEUE_TAG:
                                    pkt->control_flags = CF_ORDERED_TAG;
                                    break;
                                default:
                                    pkt->control_flags = CF_SIMPLE_TAG;
                            }
                        } else
                            pkt->control_flags = CF_SIMPLE_TAG;

                        /* Load SCSI command packet. */
                        cdb_len = (uint16_t)CMD_CDBLEN(cmd);
                        if (cdb_len > MAX_CMDSZ) cdb_len = MAX_CMDSZ;
                        data_ptr = (uint8_t *) &(CMD_CDBP(cmd));
                        for( cnt = 0; cnt < cdb_len; cnt++ )
                            pkt->scsi_cdb[cnt] = *data_ptr++;
                        pkt->byte_count = (uint32_t) CMD_XFRLEN(cmd);

                        /*
                        * Load data segments.
                        */
                        if( seg_cnt )                /* If data transfer. */
                        {
                            switch( cmd->data_cmnd[0] ) {
                                case FORMAT_UNIT:
                                case WRITE_6:
                                case MODE_SELECT:
                                case SEND_DIAGNOSTIC:
                                case WRITE_10:
                                case WRITE_BUFFER:
                                case WRITE_LONG:
                                case WRITE_SAME:
                                case MODE_SELECT_10:
                                case WRITE_12:
                                case WRITE_VERIFY:
                                case WRITE_VERIFY_12:
                                case SEND_VOLUME_TAG:
                                    pkt->control_flags |= BIT_6;
                                    break;
                                default:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#if QLA_SCSI_VENDOR_DIR
    				    qla2100_set_vend_direction(ha, cmd, pkt);
#else
                                    pkt->control_flags |= BIT_5; /* READ */
#endif
#else  /* kernel version is 2.4.0 or higher */
                                    if (cmd->sc_data_direction == SCSI_DATA_WRITE)
                                        pkt->control_flags |= BIT_6;/*WRITE*/
                                    else 
                                        pkt->control_flags |= BIT_5;/*READ*/
#endif /* kernel version 2.4.0 */
                                    break;
                            }
                            sp->dir = pkt->control_flags & (BIT_5|BIT_6);

                            /* Set total data segment count. */
                            pkt->dseg_count = seg_cnt;

                            /* Setup packet address segment pointer. */
                            dword_ptr = (uint32_t *)&pkt->dseg_0_address;

                            if( cmd->use_sg )              /* If scatter gather */
                            {
                                /* Load command entry data segments. */
                                for( cnt = 0; cnt < 2 && seg_cnt; cnt++, seg_cnt-- ) {
                                    /* 4.10 64 bit */
                                    *dword_ptr++ = cpu_to_le32(pci_dma_lo32(sg_dma_address(sg)));
                                    *dword_ptr++ = cpu_to_le32(pci_dma_hi32(sg_dma_address(sg)));
                                    *dword_ptr++ = cpu_to_le32(sg_dma_len(sg));
                                    sg++;
                                      /* DEBUG(sprintf(debug_buff,
                                      "S/G Segment phys_addr=%x %x, len=0x%x\n\r",
                                      cpu_to_le32(pci_dma_hi32(sg_dma_address(sg))),
                                      cpu_to_le32(pci_dma_lo32(sg_dma_address(sg))),
                                      cpu_to_le32(sg_dma_len(sg)));)
                                      DEBUG(qla2100_print(debug_buff)); */
                                }
#ifdef QL_DEBUG_LEVEL_5
                                qla2100_print(
                                "qla2x00_64bit_start_scsi: Scatter/gather command packet data - ");
                                qla2100_print("b");
                                qla2100_output_number((uint32_t)SCSI_BUS_32(cmd), 10);
                                qla2100_print("t");
                                qla2100_output_number((uint32_t)SCSI_TCN_32(cmd), 10);
                                qla2100_print("d");
                                qla2100_output_number((uint32_t)SCSI_LUN_32(cmd), 10);
                                qla2100_print("\n\r");
                                qla2100_dump_buffer((caddr_t)pkt, REQUEST_ENTRY_SIZE);
#endif
                                /*
                                * Build continuation packets.
                                */
                                while( seg_cnt > 0 ) {
                                    /* Adjust ring index. */
                                    ha->req_ring_index++;
                                    if( ha->req_ring_index == REQUEST_ENTRY_CNT ) {
                                        ha->req_ring_index = 0;
                                        ha->request_ring_ptr = ha->request_ring;
                                    } else
                                        ha->request_ring_ptr++;

                                    pkt = (cmd_a64_entry_t *)ha->request_ring_ptr;

                                    /* Zero out packet. */
                                    dword_ptr = (uint32_t *)pkt;
                                    for( cnt = 0;cnt < REQUEST_ENTRY_SIZE/4; cnt++ )
                                        *dword_ptr++ = 0;

                                    /* Load packet defaults. */
                                    ((cont_a64_entry_t *)pkt)->entry_type =
                                    CONTINUE_A64_TYPE;
                                    ((cont_a64_entry_t *)pkt)->entry_count = 1;
                                    ((cont_a64_entry_t *)pkt)->sys_define = (uint8_t)
                                    ha->req_ring_index;

                                    /* Setup packet address segment pointer. */
                                    dword_ptr = (uint32_t *)
                                    &((cont_a64_entry_t *)pkt)->dseg_0_address;

                                    /* Load continuation entry data segments. */
                                    for( cnt = 0; cnt < 5 && seg_cnt; cnt++, seg_cnt-- ) {
                                       /* 4.10 64 bit */
                                       *dword_ptr++ = cpu_to_le32(pci_dma_lo32(sg_dma_address(sg)));
                                       *dword_ptr++ = cpu_to_le32(pci_dma_hi32(sg_dma_address(sg)));
                                       *dword_ptr++ = cpu_to_le32(sg_dma_len(sg));
                                       sg++;
                                      /* DEBUG(sprintf(debug_buff,
                                      "S/G Segment Cont. phys_addr=%x %x, len=0x%x\n\r",
                                       cpu_to_le32(pci_dma_hi32(sg_dma_address(sg))),
                                       cpu_to_le32(pci_dma_lo32(sg_dma_address(sg))),
                                       cpu_to_le32(sg_dma_len(sg)));)
                                       DEBUG(qla2100_print(debug_buff)); */
                                    }
#ifdef QL_DEBUG_LEVEL_5
                                    qla2100_print(
                                    "qla2x00_64bit_start_scsi: continuation packet data - c");
                                    qla2100_print("b");
                            qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);

                            qla2100_print("t");
                            qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                            qla2100_print("d");
                            qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                            qla2100_print("\n\r");
                                    qla2100_dump_buffer((caddr_t)pkt, REQUEST_ENTRY_SIZE);
#endif
                                }
                            } else   /* No scatter gather data transfer */
                            {   /* 4.10 64 bit */
                                dma_handle = pci_map_single(ha->pdev,
                                                         cmd->request_buffer,
                                                         cmd->request_bufflen,
                                                         scsi_to_pci_dma_dir(cmd->sc_data_direction));
                                /* save dma_handle for pci_unmap_single */
                                sp->saved_dma_handle = dma_handle;

                                *dword_ptr++ = cpu_to_le32(pci_dma_lo32(dma_handle));
                                *dword_ptr++ = cpu_to_le32(pci_dma_hi32(dma_handle));
                                *dword_ptr   = (uint32_t) cmd->request_bufflen;
                                /* DEBUG(sprintf(debug_buff,
                                "64_bit: No S/G map_single saved_dma_handle=%lx len=%x \n\r",dma_handle, cmd->request_bufflen));
                                DEBUG(qla2100_print(debug_buff)); */
#ifdef QL_DEBUG_LEVEL_5
                                qla2100_print(
                                "qla2x00_64bit_start_scsi: No scatter/gather command packet data - c");
                                qla2100_print("b");
                                qla2x00_output_number((uint32_t)SCSI_BUS_32(cmd), 10);
                                qla2100_print("t");
                                qla2x00_output_number((uint32_t)SCSI_TCN_32(cmd), 10);
                                qla2100_print("d");
                                qla2x00_output_number((uint32_t)SCSI_LUN_32(cmd), 10);
                                qla2100_print("\n\r");
                                qla2x00_dump_buffer((caddr_t)pkt, REQUEST_ENTRY_SIZE);
#endif
                            }
                        }
                        else                            /* No data transfer */
                        {
                            *dword_ptr++ = (uint32_t) 0;
                            *dword_ptr++ = (uint32_t) 0;
                            *dword_ptr = (uint32_t)  0;
#ifdef QL_DEBUG_LEVEL_5
                         qla2100_print(
                         "qla2x00_64bit_start_scsi: No data, command packet data - c");
                         qla2100_print("b");
                         qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);
                         qla2100_print("t");
                         qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                         qla2100_print("d");
                         qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                         qla2100_print("\n\r");
                         qla2x00_dump_buffer((caddr_t)pkt, REQUEST_ENTRY_SIZE);
#endif
                        }
                        DEBUG4(qla2100_print("\nqla2100_64bit_start_scsi: Wakeup RISC for pending command\n\r"));
                        /* Adjust ring index. */
                        ha->req_ring_index++;
                        if( ha->req_ring_index == REQUEST_ENTRY_CNT ) {
                            ha->req_ring_index = 0;
                            ha->request_ring_ptr = ha->request_ring;
                        } else
                            ha->request_ring_ptr++;

                        ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->qthreads--;
                        ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads++;
                        ha->total_ios++;
                        sp->lun_queue->io_cnt++;
			sp->ha = ha;
                        sp->flags |= SRB_DMA_VALID;
        		sp->u_start = jiffies;   

                        /* Set chip new ring index. */
#if WATCH_THREADS_SIZE
                        DEBUG3(printk("qla2100_64bit_start_scsi: actthreads=%d.\n", (uint32_t)((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads);)
#endif
#if defined(ISP2100) || defined(ISP2200)
                        temp = CACHE_FLUSH(&reg->mailbox4);
                        WRT_REG_WORD(&reg->mailbox4, ha->req_ring_index);
#else
                        temp = CACHE_FLUSH(&reg->req_q_in);
                        WRT_REG_WORD(&reg->req_q_in, ha->req_ring_index);
#endif
                    } else {
                        status = 1;
#ifdef QL_DEBUG_LEVEL_2
                        qla2100_print(
                        "qla2x00_64bit_start_scsi: NO ROOM IN OUTSTANDING ARRAY\n\r");
                        qla2100_print(" req_q_cnt=");
                qla2100_output_number((u_long)ha->req_q_cnt, 16);
#endif
                    }
                } else {
                    status = 1;
#ifdef QL_DEBUG_LEVEL_2
                    qla2100_print("qla2x00_64bit_start_scsi: in-ptr=");
            qla2100_output_number((u_long)ha->req_ring_index, 16);
            qla2100_print(" req_q_cnt=");
            qla2100_output_number((u_long)ha->req_q_cnt, 16);
            qla2100_print(" req_cnt=");
            qla2100_output_number((u_long)req_cnt, 16);
                    qla2100_print("\n\r");
#endif
                }

                /* Release ring specific lock */
                QLA2100_RING_UNLOCK(ha);

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( status )
                    qla2100_print("qla2x00_64bit_start_scsi: **** FAILED ****\n\r");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    qla2100_print("qla2x00_64bit_start_scsi: exiting normally\n\r");
#endif
                return(status);
}
#endif

/*
* qla2x00_32bit_start_scsi
*      The start SCSI is responsible for building request packets on
*      request ring and modifying ISP input pointer.
*
*      The Qlogic firmware interface allows every queue slot to have a SCSI
*      command and up to 4 scatter/gather (SG) entries.  If we need more
*      than 4 SG entries, then continuation entries are used that can
*      hold another 7 entries each.  The start routine determines if there
*      is eought empty slots then build the combination of requests to
*      fulfill the OS request.
*
* Input:
*      ha = adapter block pointer.
*      sp = SCSI Request Block structure pointer.
*
* Returns:
*      0 = success, was able to issue command.
*/
STATIC uint8_t
qla2x00_32bit_start_scsi(srb_t *sp)
{
                device_reg_t    *reg;
                uint8_t         status = 0;
                uint8_t         found = 0;
                Scsi_Cmnd       *cmd = sp->cmd;
                uint16_t        cdb_len, temp;
                uint32_t        cnt, index;
                cmd_entry_t     *pkt;
                uint16_t        req_cnt;
                uint16_t        seg_cnt;
                struct scatterlist    *sg = (struct scatterlist *) NULL;
                caddr_t         data_ptr;
                uint32_t        *dword_ptr;
                uint32_t        timeout;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,18)
                uint64_t       dma_handle;
#endif
            	scsi_qla_host_t *ha;
		fc_lun_t	*fclun;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_32bit_start_scsi");
#endif

		/* Setup device pointers. */
		fclun = sp->lun_queue->fclun;
		ha = fclun->fcport->ha;
		reg = ha->iobase;

                /* ra 11/27/01
                * Send marker if required.
                */
                if (ha->marker_needed != 0) {
                    if (qla2x00_marker(ha, 0, 0, MK_SYNC_ALL) != 0) {
                        return (1);
                    }
                    ha->marker_needed = 0;
                }
                COMTRACE('S')
DEBUG5(printk( "32bit_start: ha(%d) cmd=%p sp=%p CDB=%x\n\r",(int)ha->instance,cmd,sp,cmd->cmnd[0]);)
                /* Calculate number of entries and segments required. */
                seg_cnt = 0;
                req_cnt = 1;
                if( cmd->use_sg ) {
                    sg = (struct scatterlist *) cmd->request_buffer;
                    /* 4.10 32 bit S/G Data Transfer */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                    seg_cnt =  cmd->use_sg;
#else
                    seg_cnt = pci_map_sg(ha->pdev,sg,cmd->use_sg,
                              scsi_to_pci_dma_dir(cmd->sc_data_direction));
#endif
                    /*
                    * if greater than four sg entries then we need to allocate
                    * continuation entries
                    */
                    if( seg_cnt > 2 ) {
                        req_cnt += (uint16_t)(seg_cnt - 3) / 7;
                        if( (uint16_t)(seg_cnt - 3) % 7 )
                            req_cnt++;
                    }
                    DEBUG5(sprintf(debug_buff,
                    "S/G for data transfer -num segs(%d), req blk cnt(%d)\n\r",seg_cnt,req_cnt));
                    DEBUG5(qla2100_print(debug_buff));
                } else if( CMD_XFRLEN(cmd) )  /* If data transfer. */
                {  /* no S/G Data Transfer */
                    DEBUG5(printk("Single data transfer (0x%x)\n",
                           cmd->request_bufflen));
                    seg_cnt = 1;
                }

                /* Acquire ring specific lock */
                QLA2100_RING_LOCK(ha);

                if( (uint16_t)(req_cnt + 2) >= ha->req_q_cnt ) {
                    /* Calculate number of free request entries. */
#if defined(ISP2100) || defined(ISP2200)
                    cnt = qla2100_debounce_register(&reg->mailbox4);
#else
                    cnt = qla2100_debounce_register(&reg->req_q_out);
#endif
                    if( ha->req_ring_index < cnt )
                        ha->req_q_cnt = cnt - ha->req_ring_index;
                    else
                        ha->req_q_cnt = REQUEST_ENTRY_CNT - (ha->req_ring_index - cnt);
                }

                DEBUG5(sprintf(debug_buff,
                 "Number of free entries = (%d)\n\r",ha->req_q_cnt));
                DEBUG5(qla2100_print(debug_buff));
                /* If room for request in request ring. */
                if( (uint16_t)(req_cnt + 2) < ha->req_q_cnt ) {
                    /* Check for room in outstanding command list. */
                    cnt = ha->current_outstanding_cmd;
                    for( index = 0; index < MAX_OUTSTANDING_COMMANDS; index++)
                    {
                        cnt++;
                        if (cnt == MAX_OUTSTANDING_COMMANDS)
                           cnt = 0;
                        if (ha->outstanding_cmds[cnt] == 0)
                        {
                           found = 1;
                           ha->current_outstanding_cmd = cnt;
                           break;
                        }  
                    }
                    if( found ) {
                        ha->outstanding_cmds[cnt] = sp;
                        ha->req_q_cnt -= req_cnt;
                        /* save the handle -- it help if we have to abort it */
                        CMD_HANDLE(sp->cmd) = (unsigned char *) (u_long) cnt;

                        /*
                        * Build command packet.
                        */
                        pkt = (cmd_entry_t *)ha->request_ring_ptr;

                        pkt->entry_type = COMMAND_TYPE;
                        pkt->entry_count = (uint8_t)req_cnt;
                        pkt->sys_define = (uint8_t)ha->req_ring_index;
                        pkt->entry_status = 0;
                        pkt->control_flags= 0;
                        pkt->handle = (uint32_t)cnt;

                        /* Zero out remaining portion of packet. */
                        dword_ptr = (uint32_t *)pkt + 2;
                        for( cnt = 2; cnt < REQUEST_ENTRY_SIZE/4; cnt++ )
                            *dword_ptr++ = 0;

                        /*
                        * We subtract 5 sec. from the timeout value to insure
                        * the ISP time-out before the mid-level or the driver.
                        */
                        timeout = (uint32_t) CMD_TIMEOUT(cmd)/HZ;
                        if( timeout > 5 )
                            pkt->timeout = (uint16_t) timeout - 5;
                        else
                            pkt->timeout = (uint16_t) timeout;

#if VSA
			if( (cmd->data_cmnd[0] == 0x26) ||
			    (cmd->data_cmnd[0] == 0xA0) ||
			    (cmd->data_cmnd[0] == 0xCB) ) {
				pkt->lun = fclun->lun;
			}
                        else if( (fclun->fcport->flags & FC_VSA) ) 
				pkt->lun = (fclun->lun | 0x4000);
			else
				pkt->lun = fclun->lun;
#else
			/* Set LUN number*/
			pkt->lun = fclun->lun;
#endif

			/* Set target ID */
			pkt->target = (uint8_t)fclun->fcport->loop_id;

                        /* Enable simple tag queuing if device supports it. */
                        if ( cmd->device->tagged_queue ) {
                            switch (cmd->tag) {
                                case SIMPLE_QUEUE_TAG:
                                    pkt->control_flags = CF_SIMPLE_TAG;
                                    break;
                                case HEAD_OF_QUEUE_TAG:
                                    pkt->control_flags = CF_HEAD_TAG;
                                    break;
                                case ORDERED_QUEUE_TAG:
                                    pkt->control_flags = CF_ORDERED_TAG;
                                    break;
                                default:
                                    pkt->control_flags = CF_SIMPLE_TAG;
                            }
                        } else
                            pkt->control_flags = CF_SIMPLE_TAG;

                        /* Load SCSI command packet. */
                        cdb_len = (uint16_t)CMD_CDBLEN(cmd);
                        if( cdb_len > MAX_CMDSZ )
                            cdb_len = MAX_CMDSZ;
                        data_ptr = (uint8_t *) &(CMD_CDBP(cmd));
                        for( cnt = 0; cnt < cdb_len; cnt++ )
                            pkt->scsi_cdb[cnt] = *data_ptr++;
          DEBUG4(sprintf(debug_buff,
          "32bit_start: Hba=%02d, lun=%02d, scsi Cdb[0]=0x%x, handle=0x%x, loop id =0x%x, sp=%p\n",ha->instance,pkt->lun,pkt->scsi_cdb[0],pkt->handle,pkt->target,sp);)
                        DEBUG4(qla2100_print(debug_buff);)
                        pkt->byte_count = (uint32_t) CMD_XFRLEN(cmd);

                        /*
                        * Load data segments.
                        */
                        if( seg_cnt ) {
                            /* Set transfer direction (READ and WRITE) */
                            /* Linux doesn't tell us                   */

                            switch( cmd->data_cmnd[0] ) {
                                case FORMAT_UNIT:
                                case WRITE_6:
                                case MODE_SELECT:
                                case SEND_DIAGNOSTIC:
                                case WRITE_10:
                                case WRITE_BUFFER:
                                case WRITE_LONG:
                                case WRITE_SAME:
                                case MODE_SELECT_10:
                                case WRITE_12:
                                case WRITE_VERIFY:
                                case WRITE_VERIFY_12:
                                case SEND_VOLUME_TAG:
                                    pkt->control_flags |= BIT_6;
                                    break;
                                default:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#if QLA_SCSI_VENDOR_DIR
    				    qla2100_set_vend_direction(ha, cmd, pkt);
#else
                                    pkt->control_flags |= BIT_5; /* READ */
#endif
#else  /* kernel version is 2.4.0 or higher */
                                    if (cmd->sc_data_direction == SCSI_DATA_WRITE)
                                        pkt->control_flags |= BIT_6;/*WRITE*/
                                    else 
                                        pkt->control_flags |= BIT_5;/*READ*/
#endif /* kernel version 2.4.0 */
                                    break;
                            }
                            sp->dir = pkt->control_flags & (BIT_5|BIT_6);

                            /* Set total data segment count. */
                            pkt->dseg_count = seg_cnt;

                            /* Setup packet address segment pointer. */
                            dword_ptr = (uint32_t *)&pkt->dseg_0_address;

                            if( cmd->use_sg )     /* If scatter gather */
                            {
                                DEBUG5(qla2100_print("Building S/G data segments..\n\r"));
                                DEBUG5(qla2100_dump_buffer((caddr_t)sg, 4*16));
                                /* Load command entry data segments. */
                                for( cnt = 0; cnt < 3 && seg_cnt; cnt++, seg_cnt-- ) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                                   *dword_ptr++ = (uint32_t) cpu_to_le32(VIRT_TO_BUS(sg->address));
                                   *dword_ptr++ = sg->length;
                                   /* DEBUG(sprintf(debug_buff,
                                      "S/G Segment phys_addr=0x%x, len=0x%x\n\r",
                                       cpu_to_le32(VIRT_TO_BUS(sg->address)),sg->length));
                                       DEBUG(qla2100_print(debug_buff)); */
#else
                                   *dword_ptr++ = cpu_to_le32(pci_dma_lo32(sg_dma_address(sg)));
                                   *dword_ptr++ = cpu_to_le32(sg_dma_len(sg));
                                      /* DEBUG(sprintf(debug_buff,
                                      "S/G Segment phys_addr=0x%x, len=0x%x\n\r",
                                       cpu_to_le32(pci_dma_lo32(sg_dma_address(sg))),
                                       cpu_to_le32(sg_dma_len(sg)));)
                                       DEBUG(qla2100_print(debug_buff)); */
#endif
                                   sg++;
                                }
#ifdef QL_DEBUG_LEVEL_5
                                qla2100_print(
                                "qla2100_32bit_start_scsi: Scatter/gather command packet data - ");
                        qla2100_output_number((u_long)ha->host_no, 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                                qla2100_print("\n");
                                qla2100_dump_buffer((uint8_t *)pkt, REQUEST_ENTRY_SIZE);
#endif
                                /*
                                * Build continuation packets.
                                */
                                while( seg_cnt > 0 ) {
                                    /* Adjust ring index. */
                                    ha->req_ring_index++;
                                    if( ha->req_ring_index == REQUEST_ENTRY_CNT ) {
                                        ha->req_ring_index = 0;
                                        ha->request_ring_ptr = ha->request_ring;
                                    } else
                                        ha->request_ring_ptr++;

                                    pkt = (cmd_entry_t *)ha->request_ring_ptr;

                                    /* Zero out packet. */
                                    dword_ptr = (uint32_t *)pkt;
                                    for( cnt = 0;cnt < REQUEST_ENTRY_SIZE/4; cnt++ )
                                        *dword_ptr++ = 0;

                                    /* Load packet defaults. */
                                    ((cont_entry_t *)pkt)->entry_type = CONTINUE_TYPE;
                                    ((cont_entry_t *)pkt)->entry_count = 1;
                                    ((cont_entry_t *)pkt)->sys_define = (uint8_t)
                                    ha->req_ring_index;

                                    /* Setup packet address segment pointer. */
                                    dword_ptr = (uint32_t *)
                                    &((cont_entry_t *)pkt)->dseg_0_address;

                                    /* Load continuation entry data segments. */
                                    for( cnt = 0; cnt < 7 && seg_cnt; cnt++, seg_cnt-- ) {
                                       /* 4.10 32 bit */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                                       *dword_ptr++ = (u_int) cpu_to_le32(VIRT_TO_BUS(sg->address));
                                       *dword_ptr++ = sg->length;
                                       /* DEBUG(sprintf(debug_buff,
                                       "S/G Segment Cont. phys_addr=0x%x, len=0x%x\n\r",
                                        cpu_to_le32(pci_dma_lo32(VIRT_TO_BUS(sg->address))),sg->length);)
                                        DEBUG(qla2100_print(debug_buff)); */
#else
                                       *dword_ptr++ = cpu_to_le32(pci_dma_lo32(sg_dma_address(sg)));
                                       *dword_ptr++ = cpu_to_le32(sg_dma_len(sg));
                                       /* DEBUG(sprintf(debug_buff,
                                       "S/G Segment Cont. phys_addr=0x%x, len=0x%x\n\r",
                                       cpu_to_le32(pci_dma_lo32(sg_dma_address(sg))),
                                       cpu_to_le32(sg_dma_len(sg)));)
                                       DEBUG(qla2100_print(debug_buff)); */
#endif
                                       sg++;
                                    }
#ifdef QL_DEBUG_LEVEL_5
                                    qla2100_print(
                                    "qla2100_32bit_start_scsi: continuation packet data - ");
                            qla2100_output_number((u_long)ha->host_no, 10);
                            qla2100_print(":");
                            qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);
                            qla2100_print(":");
                            qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                            qla2100_print(":");
                            qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                                    qla2100_print("\n");
                                    qla2100_dump_buffer((uint8_t *)pkt,
                                    REQUEST_ENTRY_SIZE);
#endif
                                }
                            } else  /* No scatter gather data transfer */
                            {
                              /* 4.10 32 bit */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                              *dword_ptr++ = (uint32_t) cpu_to_le32(VIRT_TO_BUS(cmd->request_buffer));
                              *dword_ptr = (uint32_t) cmd->request_bufflen;
#else
                              dma_handle = pci_map_single(ha->pdev,
                                                          cmd->request_buffer,
                                                          cmd->request_bufflen,
                                                          scsi_to_pci_dma_dir(cmd->sc_data_direction));
                              sp->saved_dma_handle = dma_handle;

                              *dword_ptr++ = cpu_to_le32(pci_dma_lo32(dma_handle));
                              *dword_ptr   = (uint32_t) cmd->request_bufflen;
                              /* DEBUG(sprintf(debug_buff,
                              "32_bit: No S/G map_single dma_handle=%lx len=%x\n\r",dma_handle,cmd->request_bufflen));
                              DEBUG(qla2100_print(debug_buff)); */
#endif
                              DEBUG5(printk("Single Segment ap=0x%p, len=0x%x\n",cmd->request_buffer,cmd->request_bufflen));

#ifdef QL_DEBUG_LEVEL_5
                                qla2100_print(
                                "qla2100_32bit_start_scsi: No scatter/gather command packet data - ");
                        qla2100_output_number((u_long)ha->host_no, 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                                qla2100_print("\n");
                                qla2100_dump_buffer((uint8_t *)pkt, REQUEST_ENTRY_SIZE);
#endif
                            }
                        }
                        else     /* No data transfer */
                        {
                            *dword_ptr++ = (uint32_t) 0;
                            *dword_ptr = (uint32_t)  0;
#ifdef QL_DEBUG_LEVEL_5
                    qla2100_print(
                    "qla2100_32bit_start_scsi: No data, command packet data - ");
                    qla2100_output_number((u_long)ha->host_no, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)SCSI_BUS_32(cmd), 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)SCSI_TCN_32(cmd), 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)SCSI_LUN_32(cmd), 10);
                    qla2100_print("\n");
                    qla2100_dump_buffer((uint8_t *)pkt, REQUEST_ENTRY_SIZE);
#endif
                        }
                        /* Adjust ring index. */
                        ha->req_ring_index++;
                        if( ha->req_ring_index == REQUEST_ENTRY_CNT ) {
                            ha->req_ring_index = 0;
                            ha->request_ring_ptr = ha->request_ring;
                        } else
                            ha->request_ring_ptr++;

                        /* Set chip new ring index. */
                        DEBUG5(qla2100_print("qla2100_32bit_start_scsi: Wakeup RISC for pending command\n\r"));
                        DEBUG5(qla2100_dump_buffer((caddr_t)pkt, 64));

                        ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->qthreads--;
                        ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads++;
                        ha->total_ios++;
                        sp->lun_queue->io_cnt++;
			sp->ha = ha;
                        sp->flags |= SRB_DMA_VALID;
        		sp->u_start = jiffies;   

			DEBUG5(printk("%ld ",cmd->serial_number); )

#if WATCH_THREADS_SIZE
                	DEBUG3(printk("qla2100_32bit_start_scsi: actthreads=%d.\n", (u_long)((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads);)
#endif
#if defined(ISP2100) || defined(ISP2200)
                        temp = CACHE_FLUSH(&reg->mailbox4);
                        WRT_REG_WORD(&reg->mailbox4, ha->req_ring_index);
#else
                        temp = CACHE_FLUSH(&reg->req_q_in);
                        WRT_REG_WORD(&reg->req_q_in, ha->req_ring_index);
#endif
                    } else {
                        status = 1;
                        qla2100_stats.outarray_full++;
#ifdef QL_DEBUG_LEVEL_2
                        qla2100_print(
                        "qla2100_32bit_start_scsi: NO ROOM IN OUTSTANDING ARRAY\n");
#endif
                    }
                } else {
                    status = 1;

#ifdef QL_DEBUG_LEVEL_8
                    qla2100_print("qla2100_32bit_start_scsi: in-ptr=");
            qla2100_output_number((u_long)ha->req_ring_index, 16);
            qla2100_print(" req_q_cnt=");
            qla2100_output_number((u_long)ha->req_q_cnt, 16);
            qla2100_print(" req_cnt=");
            qla2100_output_number((u_long)req_cnt, 16);
                    qla2100_print("\n");
#endif
                }

                /* Release ring specific lock */
                QLA2100_RING_UNLOCK(ha);

#if defined(QL_DEBUG_LEVEL_3)
                if( status )
                    printk("qla2100_32bit_start_scsi: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                LEAVE("qla2100_32bit_start_scsi");
#endif
                COMTRACE('s')
                return(status);
}

/*
* qla2100_ms_req_pkt
*      Function is responsible for locking ring and
*      getting a zeroed out Managment Server request packet.
*
* Input:
*      ha  = adapter block pointer.
*      sp  = srb_t pointer to handle post function call
* Returns:
*      0 = failed to get slot.
*/
STATIC request_t *
qla2100_ms_req_pkt(scsi_qla_host_t *ha, srb_t  *sp)
{
	device_reg_t *reg = ha->iobase;
	request_t    *pkt = 0;
	uint16_t     cnt, i, index;
	uint32_t     *dword_ptr;
	uint32_t     timer;
	uint8_t      found = 0;
	uint16_t     req_cnt = 1;

#ifdef QL_DEBUG_LEVEL_3
	ENTER("qla2100_ms_req_pkt");
#endif

	/* Wait 1 second for slot. */
	for( timer = HZ; timer; timer-- ) {
		/* Acquire ring specific lock */
		QLA2100_RING_LOCK(ha);

		if ((uint16_t)(req_cnt + 2) >= ha->req_q_cnt) {
			/* Calculate number of free request entries. */

#if defined(ISP2100) || defined(ISP2200)
			cnt = qla2100_debounce_register(&reg->mailbox4);
#else
			cnt = qla2100_debounce_register(&reg->req_q_out);
#endif

			if (ha->req_ring_index < cnt) {
				ha->req_q_cnt = cnt - ha->req_ring_index;
			} else {
				ha->req_q_cnt = REQUEST_ENTRY_CNT -
				    (ha->req_ring_index - cnt);
			}
		}

		/* Check for room in outstanding command list. */
		cnt = ha->current_outstanding_cmd;
		for (index = 0; index < MAX_OUTSTANDING_COMMANDS; index++) {
			cnt++;
			if (cnt == MAX_OUTSTANDING_COMMANDS)
				cnt = 0;

			if (ha->outstanding_cmds[cnt] == 0) {
				found = 1;
				ha->current_outstanding_cmd = cnt;
				break;
			}  
		}

		/* If room for request in request ring. */  
		if (found && (uint16_t)(req_cnt + 2) < ha->req_q_cnt) {

			pkt = ha->request_ring_ptr;

			/* Zero out packet. */
			dword_ptr = (uint32_t *)pkt;
			for( i = 0; i < REQUEST_ENTRY_SIZE/4; i++ )
			     *dword_ptr++ = 0;

			DEBUG5(printk("qla2100_ms_req: putting sp=%p in "
			    "outstanding_cmds[%x]\n",sp,cnt);)

			ha->outstanding_cmds[cnt] = sp;

			/* save the handle */
			CMD_HANDLE(sp->cmd) = (unsigned char *) (u_long) cnt;
			CMD_SP(sp->cmd) = (void *)sp;

			ha->req_q_cnt--;
			pkt->handle = (uint32_t)cnt;

			/* Set system defined field. */
			pkt->sys_define = (uint8_t)ha->req_ring_index;
			pkt->entry_status = 0;

			break;
		}
			

		/* Release ring specific lock */
		QLA2100_RING_UNLOCK(ha);

		SYS_DELAY(5);

		/* Check for pending interrupts. */
		qla2100_poll(ha);
	}

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
	if( !pkt )
		qla2100_print("qla2100_ms_req_pkt: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
	else
		LEAVE("qla2100_ms_req_pkt");
#endif
	return(pkt);
}

/*
* qla2100_req_pkt
*      Function is responsible for locking ring and
*      getting a zeroed out request packet.
*
* Input:
*      ha  = adapter block pointer.
*
* Returns:
*      0 = failed to get slot.
*/
STATIC request_t *
qla2100_req_pkt(scsi_qla_host_t *ha)
{
                device_reg_t *reg = ha->iobase;
                request_t    *pkt = 0;
                uint16_t     cnt;
                uint32_t     *dword_ptr;
                uint32_t     timer;
                /* ra  11/27/01 */ 
		uint16_t     req_cnt = 1;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_req_pkt");
#endif

                /* Wait 1 second for slot. */
                for( timer = HZ; timer; timer-- ) {

                    /* Acquire ring specific lock */
                    QLA2100_RING_LOCK(ha);

                    /* ra 11/27/01 */
   	            if( (uint16_t)(req_cnt + 2) >= ha->req_q_cnt ) {

                        /* Calculate number of free request entries. */
#if defined(ISP2100) || defined(ISP2200)
                        cnt = qla2100_debounce_register(&reg->mailbox4);
#else
                        cnt = qla2100_debounce_register(&reg->req_q_out);
#endif
                        if( ha->req_ring_index < cnt )
                            ha->req_q_cnt = cnt - ha->req_ring_index;
                        else
                            ha->req_q_cnt = REQUEST_ENTRY_CNT - (ha->req_ring_index - cnt);
                    }
                     /* ra 11/27/01 */
                    /* If room for request in request ring. */
                    if( (uint16_t)(req_cnt + 2) < ha->req_q_cnt ) {
                       /*if( ha->req_q_cnt ) {*/
                        ha->req_q_cnt--;
                        pkt = ha->request_ring_ptr;

                        /* Zero out packet. */
                        dword_ptr = (uint32_t *)pkt;
                        for( cnt = 0; cnt < REQUEST_ENTRY_SIZE/4; cnt++ )
                            *dword_ptr++ = 0;

                        /* Set system defined field. */
                        pkt->sys_define = (uint8_t)ha->req_ring_index;

                        /* Set entry count. */
                        pkt->entry_count = 1;

                        break;
                    }

                    /* Release ring specific lock */
                    QLA2100_RING_UNLOCK(ha);

                    SYS_DELAY(5);   /* 5 us */

                    /* Check for pending interrupts. */
                    qla2100_poll(ha);
                }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( !pkt )
                    printk("qla2100_req_pkt: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_req_pkt");
#endif
                return(pkt);
}

/*
* qla2100_isp_cmd
*      Function is responsible for modifying ISP input pointer.
*      Releases ring lock.
*
* Input:
*      ha  = adapter block pointer.
*/
STATIC void
qla2100_isp_cmd(scsi_qla_host_t *ha)
{
	device_reg_t *reg = ha->iobase;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_isp_cmd");
#endif

#ifdef QL_DEBUG_LEVEL_5
                qla2100_print("qla2100_isp_cmd: IOCB data:\n");
                qla2100_dump_buffer((uint8_t *)ha->request_ring_ptr, REQUEST_ENTRY_SIZE);
#endif

                /* Adjust ring index. */
                ha->req_ring_index++;
                if( ha->req_ring_index == REQUEST_ENTRY_CNT ) {
                    ha->req_ring_index = 0;
                    ha->request_ring_ptr = ha->request_ring;
                } else
                    ha->request_ring_ptr++;

                /* Set chip new ring index. */
#if defined(ISP2100) || defined(ISP2200)
                WRT_REG_WORD(&reg->mailbox4, ha->req_ring_index);
#else
                WRT_REG_WORD(&reg->req_q_in, ha->req_ring_index);
#endif

                /* Release ring specific lock */
                QLA2100_RING_UNLOCK(ha);

#ifdef QL_DEBUG_LEVEL_3
                LEAVE("qla2100_isp_cmd");
#endif
}

/*
* qla2100_enable_lun
*      Issue enable LUN entry IOCB.
*
* Input:
*      ha = adapter block pointer.
*/
STATIC void
qla2100_enable_lun(scsi_qla_host_t *ha)
{
                elun_entry_t *pkt;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_enable_lun");
#endif

                /* Get request packet. */
                if( (pkt = (elun_entry_t *)qla2100_req_pkt(ha)) != NULL ) {
                    pkt->entry_type = ENABLE_LUN_TYPE;
                    pkt->command_count = 32;
                    pkt->immed_notify_count = 1;
                    pkt->timeout = 0xffff;

                    /* Issue command to ISP */
                    qla2100_isp_cmd(ha);
                }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( !pkt )
                    qla2100_print("qla2100_enable_lun: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_enable_lun");
#endif
}


#if QL2100_TARGET_MODE_SUPPORT
/****************************************************************************/
/*                      Target Mode Support Functions.                      */
/****************************************************************************/


/*
* qla2100_notify_ack
*      Issue notify acknowledge IOCB.
*      If sequence ID is zero, acknowledgement of
*      SCSI bus reset or bus device reset is assumed.
*
* Input:
*      ha      = adapter block pointer.
*      inotify = immediate notify entry pointer.
*/
STATIC void
qla2100_notify_ack(scsi_qla_host_t *ha, notify_entry_t *inotify)
{
                nack_entry_t *pkt;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_notify_ack: entered\n");
#endif

                /* Get request packet. */
                if( pkt = (nack_entry_t *)qla2100_req_pkt(ha) ) {
                    pkt->entry_type = NOTIFY_ACK_TYPE;
                    pkt->initiator_id = inotify->initiator_id;
                    pkt->target_id = inotify->target_id;

                    if( (pkt->status = inotify->status) == 0xe )
                        /* Reset LIP occurred. */
                        pkt->flags = OF_RESET;
                    else
                        /* Increment Immediate Notify Resource Count. */
                        pkt->flags = OF_INC_RC;

                    pkt->task_flags = inotify->task_flags;
                    pkt->seq_id = inotify->seq_id;

                    /* Issue command to ISP */
                    qla2100_isp_cmd(ha);
                }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( !pkt )
                    qla2100_print("qla2100_notify_ack: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_notify_ack: exiting normally\n");
#endif
}

/*
* qla2100_64bit_continue_io
*      Issue continue target I/O IOCB.
*
* Input:
*      ha   = adapter block pointer.
*      atio = atio pointer.
*      len  = total bytecount.
*      addr = physical address pointer.
*/
STATIC void
qla2100_64bit_continue_io(scsi_qla_host_t *ha, atio_entry_t *atio, uint32_t len,
    u_long *addr)
{
                ctio_a64_entry_t *pkt;
                uint32_t         *dword_ptr;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_64bit_continue_io: entered\n");
#endif

                /* Get request packet. */
                if( pkt = (ctio_a64_entry_t *)qla2100_req_pkt(ha) ) {
                    pkt->entry_type = CTIO_A64_TYPE;
                    pkt->initiator_id = atio->initiator_id;
                    pkt->exchange_id = atio->exchange_id;
                    pkt->flags = atio->flags | OF_FAST_POST;
                    pkt->scsi_status = atio->scsi_status;

                    if( len ) {
                        pkt->dseg_count = 1;
                        pkt->transfer_length = len;
                        pkt->dseg_0_length = len;
                        dword_ptr = (uint32_t *)addr;
                        pkt->dseg_0_address[0] = *dword_ptr++;
                        pkt->dseg_0_address[1] = *dword_ptr;
                    }

                    /* Issue command to ISP */
                    qla2100_isp_cmd(ha);
                }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( !pkt )
                    qla2100_print("qla2100_64bit_continue_io: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_64bit_continue_io: exiting normally\n");
#endif
}

/*
* qla2100_32bit_continue_io
*      Issue continue target I/O IOCB.
*
* Input:
*      ha   = adapter block pointer.
*      atio = atio pointer.
*      len  = total bytecount.
*      addr = physical address pointer.
*/
STATIC void
qla2100_32bit_continue_io(scsi_qla_host_t *ha, atio_entry_t *atio, uint32_t len,
    u_long *addr)
{
                ctio_entry_t *pkt;
                uint32_t     *dword_ptr;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_continue_io: entered\n");
#endif

                /* Get request packet. */
                if( pkt = (ctio_entry_t *)qla2100_req_pkt(ha) ) {
                    pkt->entry_type = CONTINUE_TGT_IO_TYPE;
                    pkt->initiator_id = atio->initiator_id;
                    pkt->exchange_id = atio->exchange_id;
                    pkt->flags = atio->flags | OF_FAST_POST;
                    pkt->scsi_status = atio->scsi_status;

                    if( len ) {
                        pkt->dseg_count = 1;
                        pkt->transfer_length = len;
                        pkt->dseg_0_length = len;
                        dword_ptr = (uint32_t *)addr;
                        pkt->dseg_0_address = *dword_ptr;
                    }

                    /* Issue command to ISP */
                    qla2100_isp_cmd(ha);
                }

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                if( !pkt )
                    qla2100_print("qla2100_32bit_continue_io: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                else
                    LEAVE("qla2100_32bit_continue_io: exiting normally\n");
#endif
}
#endif /* QL2100_TARGET_MODE_SUPPORT */

/****************************************************************************/
/*                        Interrupt Service Routine.                        */
/****************************************************************************/

/*
*  qla2100_isr
*      Calls I/O done on command completion.
*
* Input:
*      ha           = adapter block pointer.
*      done_q_first = done queue first pointer.
*      done_q_last  = done queue last pointer.
*      INTR_LOCK must be already obtained.
*/
STATIC void
qla2100_isr(scsi_qla_host_t *ha, uint16_t data, srb_t **done_q_first,
    srb_t **done_q_last, uint8_t *got_mbx)
{
        device_reg_t *reg = ha->iobase;
        srb_t        *sp;
        uint32_t     index;
        unsigned long done_flags = 0;
        uint16_t     *iptr, *mptr;
        uint16_t     mailbox[MAILBOX_REGISTER_COUNT];
        uint16_t     cnt, temp1;
        uint16_t     response_index = RESPONSE_ENTRY_CNT;
#ifdef ISP2300
        uint16_t     temp2;
        uint8_t      mailbox_int;
#endif
        uint8_t      rscn_queue_index;
	unsigned long	mbx_flags = 0;

#ifdef QL_DEBUG_LEVEL_3
        ENTER("qla2100_isr: entered\n");
#endif

        /* Check for mailbox interrupt. */
#if defined(ISP2100) || defined(ISP2200)
        response_index = qla2100_debounce_register(&reg->mailbox5);
        temp1 = RD_REG_WORD(&reg->semaphore);
        if( temp1 & BIT_0 ) {
            temp1 = RD_REG_WORD(&reg->mailbox0);
#else
        temp2 = RD_REG_WORD(&reg->host_status_hi);
        mailbox_int = 0;
        switch (data & 0xFF)
        {
           case ROM_MB_CMD_COMP:
           case ROM_MB_CMD_ERROR:
           case MB_CMD_COMP:
           case MB_CMD_ERROR:
           case ASYNC_EVENT:
                mailbox_int = 1;
                temp1 = temp2;
                break;
           case FAST_SCSI_COMP:
                mailbox_int = 1;
                temp1 = MBA_SCSI_COMPLETION;
                break;
           case RESPONSE_QUEUE_INT:
                response_index = temp2;
                goto response_queue_int;
                break;
           default:
                WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
                goto isr_end;
                break;
        }
        if (mailbox_int)
        {
#endif
	   if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
	     QLA_MBX_REG_LOCK(ha);

           if (temp1 == MBA_SCSI_COMPLETION) 
           {
#if defined(ISP2100) || defined(ISP2200)
              mailbox[1] = RD_REG_WORD(&reg->mailbox1);
#else
              mailbox[1] = temp2;
#endif
              mailbox[2] = RD_REG_WORD(&reg->mailbox2);
           }
           else
           {
              mailbox[0] = temp1;
              DEBUG3(printk("qla2100_isr: Saving return mbx data\n");)
              /* Get mailbox data. */
              mptr = &mailbox[1];
              iptr = (uint16_t *)&reg->mailbox1;
              for (cnt = 1; cnt < MAILBOX_REGISTER_COUNT; cnt++) {
#ifdef ISP2200
                 if (cnt == 8)
                    iptr = (uint16_t *)&reg->mailbox8;
#endif
                 if (cnt == 4 || cnt == 5)
                    *mptr = qla2100_debounce_register(iptr);
                 else
                    *mptr = RD_REG_WORD(iptr);
                 mptr++;
                 iptr++;
              }
           }
	   if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
	     QLA_MBX_REG_UNLOCK(ha);

           /* Release mailbox registers. */
           WRT_REG_WORD(&reg->semaphore, 0);
           WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);

#ifdef QL_DEBUG_LEVEL_5
           qla2100_print("qla2100_isr: mailbox interrupt mailbox[0] = ");
           qla2100_output_number((u_long)temp1, 16);
           qla2100_print("\n");
#endif

           /* Handle asynchronous event */
           switch( temp1 ) {
                case MBA_SCSI_COMPLETION:   /* Response completion */
#ifdef QL_DEBUG_LEVEL_5
                    qla2100_print("qla2100_isr: mailbox response completion\n");
#endif
                    if( ha->flags.online ) {
                        /* Get outstanding command index. */
                        index = (uint32_t)(mailbox[2] << 16 | mailbox[1]);

                        /* Validate handle. */
                        if( index < MAX_OUTSTANDING_COMMANDS ) {
                            sp = ha->outstanding_cmds[index];
                        } else {
			    DEBUG2(printk("qla2100_isr(%ld): invalid scsi "
				"completion handle %d.\n", ha->host_no, index);)
                            sp = 0;
			}

                        if( sp ) {
                            /* Free outstanding command slot. */
                            ha->outstanding_cmds[index] = 0;
                            ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads--;
                            sp->flags |= SRB_ISP_COMPLETED;

                            /* Save ISP completion status */
                            CMD_RESULT(sp->cmd) = DID_OK;
		            /* v2.19.5b2 Reset port down retry on success. */
			    sp->port_down_retry_count = ha->port_down_retry_count;
	    		    sp->fo_retry_cnt = 0;

                            QLA_DONE_Q_LOCK(ha, done_flags);
                            /* Place block on done queue */
                            ha->done_q_cnt++;
                            DEBUG(sp->state = 3;)

                            sp->s_next = NULL;
                            sp->s_prev = *done_q_last;
                            if( !(*done_q_first) )
                                *done_q_first = sp;
                            else
                                (*done_q_last)->s_next = sp;
                            *done_q_last = sp;
                            QLA_DONE_Q_UNLOCK(ha, done_flags);
                        } 
                        else {
                            DEBUG2(printk("scsi(%d): qla2100_isr: ISP invalid "
				"handle\n",(int)ha->host_no);)
                            printk(KERN_WARNING "qla2x00: ISP invalid handle");

                            QLA2100_DPC_LOCK(ha);
                            ha->dpc_flags |= ISP_ABORT_NEEDED;
                            QLA2100_DPC_UNLOCK(ha);
                        }
                   }
                   break;

               case MBA_RESET:                 /* Reset */
#ifdef QL_DEBUG_LEVEL_2
                    printk("scsi(%d) qla2100_isr: asynchronous RESET\n",(int)ha->host_no);
#endif
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= RESET_MARKER_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    break;

                case MBA_SYSTEM_ERR:        /* System Error */
                    printk(KERN_WARNING
                     "qla2x00: ISP System Error - mbx1=%xh, mbx2=%xh, mbx3=%xh",
                     mailbox[1], mailbox[2], mailbox[3]);

                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    break;

                case MBA_REQ_TRANSFER_ERR:  /* Request Transfer Error */
#ifdef QL_DEBUG_LEVEL_2
                    printk("qla2100_isr: ISP Request Transfer Error\n");
#endif
                    printk(KERN_WARNING "qla2x00: ISP Request Transfer Error\n");
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    break;

                case MBA_RSP_TRANSFER_ERR:  /* Response Transfer Error */
#ifdef QL_DEBUG_LEVEL_2
                    printk("qla2100_isr: ISP Response Transfer Error\n");
#endif
                    printk(KERN_WARNING "qla2100: ISP Response Transfer Error\n");
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    break;

                case MBA_WAKEUP_THRES:      /* Request Queue Wake-up */
#ifdef QL_DEBUG_LEVEL_2
                    printk("qla2100_isr: asynchronous WAKEUP_THRES\n");
#endif
                    break;

                case MBA_LIP_OCCURRED:      /* Loop Initialization Procedure */
                    if(!qla2100_quiet) printk(KERN_INFO "scsi(%d): LIP occurred.\n",(int)ha->host_no);
                    /* occurred. */
#ifdef QL_DEBUG_LEVEL_2
                    printk(
                    "qla2100_isr: asynchronous MBA_LIP_OCCURRED\n");
#endif
		    ha->dpc_flags = ha->dpc_flags | COMMAND_WAIT_NEEDED;
                    /* Save LIP sequence. */
                    ha->lip_seq = mailbox[1];
                    if( ha->loop_state != LOOP_DOWN ) {
                        ha->loop_state = LOOP_DOWN;
                        ha->loop_down_timer = LOOP_DOWN_TIME;
                    }
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= COMMAND_WAIT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);

                    ha->flags.management_server_logged_in = 0;

		    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
			/* Update AEN queue. */
			qla2x00_enqueue_aen(ha, MBA_LIP_OCCURRED, NULL);
		    }    

                    ha->total_lip_cnt++;

                    break;

                case MBA_LOOP_UP:
                    printk(KERN_INFO "scsi(%d): LOOP UP detected\n",(int)ha->host_no);
#ifdef QL_DEBUG_LEVEL_2
                    printk("qla2100_isr: asynchronous MBA_LOOP_UP\n");
#endif
                    ha->flags.management_server_logged_in = 0;
		    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
			/* Update AEN queue. */
			qla2x00_enqueue_aen(ha, MBA_LOOP_UP, NULL);
	            }
                    ha->loop_state = LOOP_UP;
                    break;

                case MBA_LOOP_DOWN:
                    printk(KERN_INFO "scsi(%d): LOOP DOWN detected\n",(int)ha->host_no);
#ifdef QL_DEBUG_LEVEL_2
                    printk("scsi(%ld) qla2100_isr: asynchronous MBA_LOOP_DOWN\n",ha->host_no);
#endif
                    if( ha->loop_state != LOOP_DOWN ) {
                        ha->loop_state = LOOP_DOWN;
                        ha->loop_down_timer = LOOP_DOWN_TIME;
                    }

                    ha->flags.management_server_logged_in = 0;

		    /* no wait 10/19/2000 */
		    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
			/* Update AEN queue. */
			qla2x00_enqueue_aen(ha, MBA_LOOP_DOWN, NULL);
		    }
                    break;

                case MBA_LIP_RESET:         /* LIP reset occurred. */
                    if(!qla2100_quiet)
			printk(KERN_INFO "scsi(%d): LIP reset occurred\n",
			    (int)ha->host_no);
#ifdef QL_DEBUG_LEVEL_2
                    printk(
                    "scsi(%ld) qla2100_isr: asynchronous MBA_LIP_RESET\n",ha->host_no);
#endif
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= COMMAND_WAIT_NEEDED;
                    ha->dpc_flags |= RESET_MARKER_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);

                    ha->loop_down_timer = LOOP_DOWN_TIME;
                    ha->loop_state = LOOP_DOWN;
                    ha->operating_mode = LOOP;
                    ha->flags.management_server_logged_in = 0;

		    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
			/* Update AEN queue. */
			qla2x00_enqueue_aen(ha, MBA_LIP_RESET, NULL);
		    }

                    ha->total_lip_cnt++;

                    break;

                case MBA_LINK_MODE_UP:      /* Link mode up. */
                    DEBUG(printk("scsi(%ld): Link node is up\n",ha->host_no);)
                    DEBUG2(printk( 
                      "qla2100_isr(%ld): asynchronous MBA_LINK_MODE_UP\n",
			ha->host_no);)

		    /* Save the current speed for use by ioctl. */
#ifdef ISP2100
                    ha->current_speed = EXT_DEF_PORTSPEED_1GBIT;
#elif defined(ISP2200)
                    ha->current_speed = EXT_DEF_PORTSPEED_1GBIT;
#elif defined(ISP2300)
                    if (mailbox[1] == 1)
                        ha->current_speed = EXT_DEF_PORTSPEED_2GBIT;
                    else
                        ha->current_speed = EXT_DEF_PORTSPEED_1GBIT;
#endif

                    QLA2100_DPC_LOCK(ha);

		    /* Until there's a transition from loop down to loop up,
		     * treat this as loop down only.
		     */
		    if (!(ha->dpc_flags & ABORT_ISP_ACTIVE)) {
                        ha->dpc_flags |= COMMAND_WAIT_NEEDED;
                        ha->dpc_flags |= RESET_MARKER_NEEDED;
		    }

		    if (!ha->loop_down_timer)
	                ha->loop_down_timer = LOOP_DOWN_TIME;

                    ha->loop_state = LOOP_DOWN;

                    QLA2100_DPC_UNLOCK(ha);

                    break;

                case MBA_UPDATE_CONFIG:      /* Update Configuration. */
                    printk(KERN_INFO "scsi(%ld): Configuration change "
			"detected: value %d.\n",ha->host_no,mailbox[1]);
                    DEBUG2(printk("scsi(%ld) qla2100_isr: asynchronous "
			"MBA_UPDATE_CONFIG\n",ha->host_no);)

                    ha->loop_state = LOOP_DOWN;  /* dg - 03/30 */
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= LOOP_RESYNC_NEEDED | LOCAL_LOOP_UPDATE;
                    QLA2100_DPC_UNLOCK(ha);
                    break;

                case MBA_PORT_UPDATE:       /* Port database update occurred. */
                    DEBUG(printk("scsi(%ld): Port database changed\n",ha->host_no);)
                    DEBUG2(printk("scsi%ld qla2100_isr: asynchronous "
			"MBA_PORT_UPDATE\n", ha->host_no);)

		    /* 
		     * dg - 06/19/01 - Mark all devices as missing,
		     * so we will login again.
		     */
		    ha->flags.rscn_queue_overflow = 1;

                    ha->loop_down_timer = 0;
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= LOOP_RESYNC_NEEDED | LOCAL_LOOP_UPDATE;
                    QLA2100_DPC_UNLOCK(ha);
                    /* 9/23 ha->flags.loop_resync_needed = TRUE; */
                    ha->loop_state = LOOP_UPDATE;
	       	    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
		       /* Update AEN queue. */
		       qla2x00_enqueue_aen(ha, MBA_PORT_UPDATE, NULL);
	       	    }
                    break;

                case MBA_SCR_UPDATE:        /* State Change Registration. */
                    DEBUG2(printk("scsi%ld qla2100_isr: asynchronous "
			"MBA_RSCR_UPDATE\n",ha->host_no);)
                    DEBUG(printk("scsi(%ld): RSCN database changed - "
			"0x%x,0x%x\n",ha->host_no,mailbox[1],mailbox[2]);)

	            rscn_queue_index = ha->rscn_in_ptr + 1;
		    if (rscn_queue_index == MAX_RSCN_COUNT)
		       rscn_queue_index = 0;
		    if (rscn_queue_index != ha->rscn_out_ptr) {
		       ha->rscn_queue[ha->rscn_in_ptr].format = (uint8_t)(mailbox[1]>>8);
	               ha->rscn_queue[ha->rscn_in_ptr].d_id.b.domain = (uint8_t)mailbox[1];
		       ha->rscn_queue[ha->rscn_in_ptr].d_id.b.area = (uint8_t)(mailbox[2]>>8);
		       ha->rscn_queue[ha->rscn_in_ptr].d_id.b.al_pa = (uint8_t)mailbox[2];
		       ha->rscn_in_ptr = (uint8_t)rscn_queue_index;
		    } else {
		       ha->flags.rscn_queue_overflow = 1;
		    }

                    ha->loop_down_timer = 0;
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= LOOP_RESYNC_NEEDED | RSCN_UPDATE;
                    QLA2100_DPC_UNLOCK(ha);
                    ha->loop_down_timer = 0;
                    ha->flags.management_server_logged_in = 0;

                    ha->loop_state = LOOP_UPDATE;
		    if (ha->ioctl->flags & IOCTL_AEN_TRACKING_ENABLE) {
			/* Update AEN queue. */
			qla2x00_enqueue_aen(ha, MBA_RSCN_UPDATE, &mailbox[0]);
		    }
                    break;

                case MBA_CTIO_COMPLETION:
#ifdef QL_DEBUG_LEVEL_2
                    qla2100_print(
                    "qla2100_isr: asynchronous MBA_CTIO_COMPLETION\n");
#endif
                    break;

                default:
                    if( temp1 < MBA_ASYNC_EVENT ) {
			/* mailbox completion */
			*got_mbx = TRUE;
                        BCOPY(&mailbox[0], &ha->mailbox_out[0],
                        MAILBOX_REGISTER_COUNT);
				ha->flags.mbox_int = TRUE;
				if (ha->mcp) {
					DEBUG11(printk("qla2x00_isr: got "
					    "mailbox completion. cmd=%x.\n",
					    ha->mcp->mb[0]);)
				} else {
					DEBUG11(printk("qla2x00_isr: mbx "
					    "pointer ERROR.\n");)
				}
#ifdef QL_DEBUG_LEVEL_5
                        qla2100_print("qla2100_isr: Returning mailbox data\n");
#endif
                    }
                    break;
           }
        }
        else
#ifdef ISP2300
response_queue_int: 
#endif
        {
            WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
            /*
            * Response ring
            */
            if( ha->flags.online ) {
                if( response_index < RESPONSE_ENTRY_CNT ) {
		    qla2x00_response_pkt(ha, response_index,
		    done_q_first, done_q_last);
		} 
                else {
		    /* Invalid response pointer value. */
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
		    DEBUG(printk("qla2100_isr: Response Pointer "
		    "Error. mb5=%x.\n", response_index);)
		}
	    }
        }

#ifdef ISP2300
isr_end:
#endif

#ifdef QL_DEBUG_LEVEL_3
        /* LEAVE("qla2100_isr"); */
#endif
}


/*
*  qla2100_rst_aen
*      Processes asynchronous reset.
*
* Input:
*      ha  = adapter block pointer.
*/
STATIC void
qla2100_rst_aen(scsi_qla_host_t *ha)
{
#if QL2100_TARGET_MODE_SUPPORT
                notify_entry_t nentry;
#endif /* QL2100_TARGET_MODE_SUPPORT */

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_rst_aen");
#endif

                if ( ha->flags.online && !ha->flags.reset_active &&
                !ha->loop_down_timer && !(ha->dpc_flags & ABORT_ISP_ACTIVE) ) {
                    /* 10/15 ha->flags.reset_active = TRUE; */
                    do {
                        ha->dpc_flags &= ~RESET_MARKER_NEEDED;

                        /* Issue marker command. */
                        qla2x00_marker(ha, 0, 0, MK_SYNC_ALL);

#if QL2100_TARGET_MODE_SUPPORT
                        if( !ha->loop_down_timer && !(ha->dpc_flags & RESET_MARKER_NEEDED) ) {
                            /* Issue notify acknowledgement command. */
                            BZERO((caddr_t)&nentry, sizeof(notify_entry_t));
                            nentry.initiator_id = ha->id;
                            /* dg 7/3/99 nentry.target_id = ha->id; */
                            nentry.task_flags = BIT_13;
                            qla2100_notify_ack(ha, &nentry);
                        }
#endif /* QL2100_TARGET_MODE_SUPPORT */
                    }while( !ha->loop_down_timer &&
			(ha->dpc_flags & RESET_MARKER_NEEDED) );
                    /* 10/15 ha->flags.reset_active = FALSE; */
                }

#ifdef QL_DEBUG_LEVEL_3
                LEAVE("qla2100_rst_aen");
#endif
}

#if  QLA2100_TARGET_MODE_SUPPORT
/*
*  qla2100_atio_entry
*      Processes received ISP accept target I/O entry.
*
* Input:
*      ha  = adapter block pointer.
*      pkt = entry pointer.
*/
STATIC void
qla2100_atio_entry(scsi_qla_host_t *ha, atio_entry_t *pkt)
{
                uint64_t  *a64;
                uint64_t  *end_a64;
        u_long    phy_addr[2];
        u_long    end_addr[2];
                uint32_t  len;
                uint32_t  offset;
                uint8_t   t;
                uint8_t   *sense_ptr;

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_atio_entry: entered\n");
#endif

                t = pkt->initiator_id;
                sense_ptr = ha->tsense + t * TARGET_SENSE_SIZE;
                a64 = (uint64_t *)&phy_addr[0];
                end_a64 = (uint64_t *)&end_addr[0];

                switch( pkt->status & ~BIT_7 ) {
                    case 7:                         /* Path invalid */
#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                        qla2100_print("qla2100_atio_entry: Path invalid\n");
#endif
                        break;
                    case 0x16:                  /* Requested Capability Not Available */
#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                        qla2100_print(
                        "qla2100_atio_entry: Requested Capability Not Available\n");
#endif
                        break;
                    case 0x17:                  /* Bus Device Reset Message Received */
#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                        qla2100_print(
                        "qla2100_atio_entry: Bus Device Reset Message Received\n");
#endif
                        break;
                    case 0x3D:                  /* CDB Received */

                        /* Check for invalid LUN */
                        if( pkt->lun && pkt->cdb[0] != SS_INQUIR &&
                        pkt->cdb[0] != SS_REQSEN )
                            pkt->cdb[0] = SS_TEST;

                        switch( pkt->cdb[0] ) {
                        case SS_TEST:
#ifdef QL_DEBUG_LEVEL_3
                            qla2100_print("qla2100_atio_entry: SS_TEST\n");
#endif
                            BZERO(sense_ptr, TARGET_SENSE_SIZE);
                            len = 0;
                            if( pkt->lun == 0 )
                                pkt->scsi_status = S_GOOD;
                            else {
                                *sense_ptr = 0x70;
                                *(sense_ptr+2) = SD_ILLREQ;
                                *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                *(sense_ptr+12) = SC_INVLUN;
                                pkt->scsi_status = S_CKCON;
                            }
                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC | OF_NO_DATA);
                            break;
                        case SS_REQSEN:
#ifdef QL_DEBUG_LEVEL_3
                            qla2100_print("qla2100_atio_entry: SS_REQSEN\n");
#endif
                            phy_addr[0] = ha->tsense_dma;
                            phy_addr[1] = 0;
                            *a64 += t * TARGET_SENSE_SIZE;
                            if( pkt->cdb[4] > TARGET_SENSE_SIZE )
                                len = TARGET_SENSE_SIZE;
                            else
                                len = pkt->cdb[4];
                            pkt->scsi_status = S_GOOD;
                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC | OF_DATA_IN);
                            break;
                        case SS_INQUIR:
#ifdef QL_DEBUG_LEVEL_3
                            qla2100_print("qla2100_atio_entry: SS_INQUIR\n");
#endif
                            BZERO(sense_ptr, TARGET_SENSE_SIZE);
                            phy_addr[0] = ha->tbuf_dma;
                            phy_addr[1] = 0;
                            *a64 += TARGET_INQ_OFFSET;

                            if( pkt->lun == 0 ) {
                                    ha->tbuf->inq.id_type = ID_PROCESOR;
                                    ha->tbuf->inq.id_pqual = ID_QOK;
                            } else {
                                    ha->tbuf->inq.id_type = ID_NODEV;
                                    ha->tbuf->inq.id_pqual = ID_QNOLU;
                            }

                            if( pkt->cdb[4] > sizeof(struct ident) )
                                len = sizeof(struct ident);
                            else
                                len = pkt->cdb[4];
                            pkt->scsi_status = S_GOOD;
                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC | OF_DATA_IN);
                            break;
                        case SM_WRDB:
                            BZERO(sense_ptr, TARGET_SENSE_SIZE);
                            offset = pkt->cdb[5];
                            offset |= pkt->cdb[4] << 8;
                            offset |= pkt->cdb[3] << 16;
                            len = pkt->cdb[8];
                            len |= pkt->cdb[7] << 8;
                            len |= pkt->cdb[6] << 16;
                            end_addr[0] = phy_addr[0] = ha->tbuf_dma;
                            end_addr[1] = phy_addr[1] = 0;
                            *end_a64 += TARGET_DATA_OFFSET + TARGET_DATA_SIZE;
                            switch( pkt->cdb[1] & 7 ) {
                                case RW_BUF_HDATA:
#ifdef QL_DEBUG_LEVEL_3
                                    qla2100_print("qla2100_atio_entry: SM_WRDB, RW_BUF_HDATA\n");
#endif
                                    if( len > TARGET_DATA_SIZE + 4 ) {
#ifdef QL_DEBUG_LEVEL_2
                                        qla2100_print("qla2100_atio_entry: SM_WRDB, length > buffer size\n");
#endif
                                        *sense_ptr = 0x70;
                                        *(sense_ptr+2) = SD_ILLREQ;
                                        *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                        *(sense_ptr+12) = SC_ILLCDB;
                                        pkt->scsi_status = S_CKCON;
                                        pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                        OF_NO_DATA);
                                        len = 0;
                                    } else if( len ) {
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags =(uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_DATA_OUT);
                                    } else {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_WRDB, zero length\n");
#endif
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    }

                                    break;
                                case RW_BUF_DATA:
#ifdef QL_DEBUG_LEVEL_3
                                    qla2100_print("qla2100_atio_entry: SM_WRDB, RW_BUF_DATA\n");
#endif
                                    *a64 += offset + TARGET_DATA_OFFSET;
                                    if( pkt->cdb[2] != 0 || *a64 >= *end_a64 ||
                                    *a64 + len > *end_a64 ) {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_WRDB, RW_BUF_DATA BAD\n");
                                            qla2100_print("buf_id=");
                        qla2100_output_number((u_long)pkt->cdb[2], 16);
                        qla2100_print(", offset=");
                        qla2100_output_number((u_long)offset, 16);
                                            qla2100_print(", length=");
                        qla2100_output_number((u_long)len, 16);
                                            qla2100_print("\n");
#endif
                                            *sense_ptr = 0x70;
                                            *(sense_ptr+2) = SD_ILLREQ;
                                            *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                            *(sense_ptr+12) = SC_ILLCDB;
                                            len = 0;
                                            pkt->scsi_status = S_CKCON;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    } else if( len ) {
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags =(uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_DATA_OUT);
                                    } else {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_WRDB, zero length\n");
#endif
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    }
                                    break;
                                default:
#ifdef QL_DEBUG_LEVEL_2
                                    qla2100_print("qla2100_atio_entry: SM_WRDB unknown mode\n");
#endif
                                    *sense_ptr = 0x70;
                                    *(sense_ptr+2) = SD_ILLREQ;
                                    *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                    *(sense_ptr+12) = SC_ILLCDB;
                                    len = 0;
                                    pkt->scsi_status = S_CKCON;
                                    pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                    OF_NO_DATA);
                                    break;
                            }
                            break;
                        case SM_RDDB:
                            BZERO(sense_ptr, TARGET_SENSE_SIZE);
                            offset = pkt->cdb[5];
                            offset |= pkt->cdb[4] << 8;
                            offset |= pkt->cdb[3] << 16;
                            len = pkt->cdb[8];
                            len |= pkt->cdb[7] << 8;
                            len |= pkt->cdb[6] << 16;
                            end_addr[0] = phy_addr[0] = ha->tbuf_dma;
                            end_addr[1] = phy_addr[1] = 0;
                            *end_a64 += TARGET_DATA_OFFSET + TARGET_DATA_SIZE;
                            switch( pkt->cdb[1] & 7 ) {
                                case RW_BUF_HDATA:
#ifdef QL_DEBUG_LEVEL_3
                                    qla2100_print("qla2100_atio_entry: SM_RDDB, RW_BUF_HDATA\n");
#endif
                                    if( len ) {
                                        ha->tbuf->hdr[0] = 0;
                                        ha->tbuf->hdr[1] =
                                        (uint8_t)(TARGET_DATA_SIZE >> 16);
                                        ha->tbuf->hdr[2] =
                                        (uint8_t)(TARGET_DATA_SIZE >> 8);
                                        ha->tbuf->hdr[3] = (uint8_t)TARGET_DATA_SIZE;
                                        if( len > TARGET_DATA_SIZE + 4 )
                                            len = TARGET_DATA_SIZE + 4;
                                        pkt->scsi_status = S_GOOD;
                                        pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                        OF_DATA_IN);
                                    } else {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_RDDB, zero length\n");
#endif
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    }
                                    break;
                                case RW_BUF_DATA:
#ifdef QL_DEBUG_LEVEL_3
                                    qla2100_print("qla2100_atio_entry: SM_RDDB, RW_BUF_DATA\n");
#endif
                                    *a64 += offset + TARGET_DATA_OFFSET;
                                    if( pkt->cdb[2] != 0 || *a64 >= *end_a64 ) {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_RDDB, RW_BUF_DATA BAD\n");
                                            qla2100_print("buf_id=");
                        qla2100_output_number((u_long)pkt->cdb[2], 16);
                        qla2100_print(", offset=");
                        qla2100_output_number((u_long)offset, 16);
                                            qla2100_print("\n");
#endif
                                            *sense_ptr = 0x70;
                                            *(sense_ptr+2) = SD_ILLREQ;
                                            *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                            *(sense_ptr+12) = SC_ILLCDB;
                                            len = 0;
                                            pkt->scsi_status = S_CKCON;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    } else {
                                            if( *a64 + len > *end_a64 )
                                                len = *end_a64 - *a64;
                                            if( len ) {
                                                pkt->scsi_status = S_GOOD;
                                                pkt->flags = (uint16_t)(OF_SSTS |
                                                OF_INC_RC | OF_DATA_IN);
                                            } else {
#ifdef QL_DEBUG_LEVEL_2
                                                    qla2100_print("qla2100_atio_entry: SM_RDDB, zero length\n");
#endif
                                                    pkt->scsi_status = S_GOOD;
                                                    pkt->flags = (uint16_t)(OF_SSTS |
                                                    OF_INC_RC | OF_NO_DATA);
                                            }
                                    }
                                    break;
                                case RW_BUF_DESC:
#ifdef QL_DEBUG_LEVEL_3
                                    qla2100_print("qla2100_atio_entry: SM_RDDB, RW_BUF_DESC\n");
#endif
                                    if( len ) {
                                            if( len > 4 )
                                                len = 4;

                                            ha->tbuf->hdr[0] = 0;
                                            if( pkt->cdb[2] != 0 ) {
                                                ha->tbuf->hdr[1] = 0;
                                                ha->tbuf->hdr[2] = 0;
                                                ha->tbuf->hdr[3] = 0;
                                            } else {
                                                    ha->tbuf->hdr[1] =
                                                    (uint8_t)(TARGET_DATA_SIZE >> 16);
                                                    ha->tbuf->hdr[2] =
                                                    (uint8_t)(TARGET_DATA_SIZE >> 8);
                                                    ha->tbuf->hdr[3] =
                                                    (uint8_t)TARGET_DATA_SIZE;
                                            }
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_DATA_IN);
                                    } else {
#ifdef QL_DEBUG_LEVEL_2
                                            qla2100_print("qla2100_atio_entry: SM_RDDB, zero length\n");
#endif
                                            pkt->scsi_status = S_GOOD;
                                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                            OF_NO_DATA);
                                    }
                                    break;
                                default:
#ifdef QL_DEBUG_LEVEL_2
                                    qla2100_print("qla2100_atio_entry: SM_RDDB unknown mode\n");
#endif
                                    *sense_ptr = 0x70;
                                    *(sense_ptr+2) = SD_ILLREQ;
                                    *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                                    *(sense_ptr+12) = SC_ILLCDB;
                                    len = 0;
                                    pkt->scsi_status = S_CKCON;
                                    pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC |
                                    OF_NO_DATA);
                                    break;
                            }
                            break;
                        default:
#ifdef QL_DEBUG_LEVEL_2
                            qla2100_print("qla2100_atio_entry: Unknown SCSI command\n");
                            qla2100_dump_buffer((uint8_t *)&pkt->cdb[0], MAX_CMDSZ);
#endif
                            BZERO(sense_ptr, TARGET_SENSE_SIZE);
                            *sense_ptr = 0x70;
                            *(sense_ptr+2) = SD_ILLREQ;
                            *(sense_ptr+7) = TARGET_SENSE_SIZE-8;
                            *(sense_ptr+12) = SC_INVOPCODE;
                            len = 0;
                            pkt->scsi_status = S_CKCON;
                            pkt->flags = (uint16_t)(OF_SSTS | OF_INC_RC | OF_NO_DATA);
                            break;
                        }
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,18)
                        if ( ha->flags.enable_64bit_addressing )
                            qla2100_64bit_continue_io(ha, pkt, len, &phy_addr);
                        else
#endif
                        qla2100_32bit_continue_io(ha, pkt, len, &phy_addr);
                        break;
                    default:
                        break;
                }

#ifdef QL_DEBUG_LEVEL_3
                LEAVE("qla2100_atio_entry: exiting normally\n");
#endif
}
#endif  /* QLA2100_TARGET_MODE_SUPPORT */

STATIC void
qla2x00_response_pkt(scsi_qla_host_t *ha, uint16_t index, srb_t **done_q_first,
    srb_t **done_q_last)
{
	device_reg_t	*reg = ha->iobase;
	response_t	*pkt;

	DEBUG3(printk("qla2x00_response_pkt: entered.\n");)

	while (ha->rsp_ring_index != index) {
	       pkt = ha->response_ring_ptr;

	       DEBUG5(printk("qla2x00_response_pkt: ha->rsp_ring_index=%ld "
		    "index=%ld.\n",
		    (u_long)ha->rsp_ring_index, (u_long)index);)
	       DEBUG5(printk("qla2x00_response_pkt: response packet data:");)
	       DEBUG5(qla2100_dump_buffer((uint8_t *)pkt,
		    RESPONSE_ENTRY_SIZE);)

#ifdef FC_IP_SUPPORT
                /* Handle IP send completion */
               if (pkt->entry_type == ET_IP_COMMAND_64) {
                  uint32_t            lTagVal;
                  SEND_CB             *pSendCB;

                  /* Set packet pointer from queue entry handle */
                  if ((lTagVal = pkt->handle) < MAX_SEND_PACKETS) {
                      if ((pSendCB = (SEND_CB *)ha->apActiveIpQueue[lTagVal]) != NULL) {
                        ha->apActiveIpQueue[lTagVal] = NULL;

                        /* Return send packet to IP driver */
                        (*ha->pSendCompletionRoutine)(pSendCB);
                      } else {
                        /* Invalid handle from RISC, reset RISC firmware */
                        printk(KERN_WARNING "qla2x00: bad IP send handle %x\n", lTagVal);
                        QLA2100_DPC_LOCK(ha);
                        ha->dpc_flags |= ISP_ABORT_NEEDED;
                        QLA2100_DPC_UNLOCK(ha);
                     }
                 } else {
                     /* Invalid handle from RISC, reset RISC firmware */
                     printk(KERN_WARNING "qla2x00: bad IP send handle %x\n", lTagVal);
                     QLA2100_DPC_LOCK(ha);
                     ha->dpc_flags |= ISP_ABORT_NEEDED;
                     QLA2100_DPC_UNLOCK(ha);
                 }

                 /* Adjust ring index. */
                 ha->rsp_ring_index++;
                 if( ha->rsp_ring_index == RESPONSE_ENTRY_CNT ) {
                     ha->rsp_ring_index = 0;
                     ha->response_ring_ptr = ha->response_ring;
                 } else
                     ha->response_ring_ptr++;
#if defined(ISP2100) || defined(ISP2200)
                 WRT_REG_WORD(&reg->mailbox5, ha->rsp_ring_index);
#else
                 WRT_REG_WORD(&reg->rsp_q_out, ha->rsp_ring_index);
#endif

                 continue;
              }

              /* Handle IP receive packet */
              else if (pkt->entry_type == ET_IP_RECEIVE) {
                 PIP_RECEIVE_ENTRY   pIpReceiveEntry = (PIP_RECEIVE_ENTRY)pkt;
                 PBUFFER_CB          pBufferCB, pNextBufferCB;
                 uint32_t            lTagVal;
                 uint32_t            lPacketSize;
                 uint16_t            wBufferCount;
                 uint32_t            lReceiveBufferSize;

                 /* If split buffer, set header size for 1st buffer */
                 if (pIpReceiveEntry->wCompletionStatus & IP_REC_STATUS_SPLIT_BUFFER)
                    lReceiveBufferSize = ha->wHeaderSize;
                 else
                    lReceiveBufferSize = ha->lReceiveBufferSize;

                 if ((lTagVal = pIpReceiveEntry->waBufferHandle[0]) >=
                 ha->wReceiveBufferCount) {
                    /* Invalid handle from RISC, reset RISC firmware */
                    printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    goto InvalidIpHandle;
                 }
                 pBufferCB = &ha->pReceiveBufferCBs[lTagVal];

                 if (!(pBufferCB->lFlags &  BCB_FLAGS_RISC_OWNS_BUFFER)) {
                    /* Invalid handle from RISC, reset RISC firmware */
                    printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
                    QLA2100_DPC_LOCK(ha);
                    ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    goto InvalidIpHandle;
                 }

                 /* Set buffer belongs to driver now */
                 pBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER;

                 lPacketSize = pIpReceiveEntry->wSequenseLength;
                 pBufferCB->lPacketSize = lPacketSize;
                 pNextBufferCB = pBufferCB;

                 for (wBufferCount = 1; ; wBufferCount++) {
                    if (lPacketSize > lReceiveBufferSize) {
                       pNextBufferCB->lBufferSize = lReceiveBufferSize;
                       lPacketSize               -= lReceiveBufferSize;
                       /* If split buffer, only use header size on 1st buffer */
                       lReceiveBufferSize = ha->lReceiveBufferSize;

                       if ((lTagVal = pIpReceiveEntry->waBufferHandle[wBufferCount]) >=
                       ha->wReceiveBufferCount) {
                          /* Invalid handle from RISC, reset RISC firmware */
                          printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
                          QLA2100_DPC_LOCK(ha);
                          ha->dpc_flags |= ISP_ABORT_NEEDED;
                          QLA2100_DPC_UNLOCK(ha);
                          goto InvalidIpHandle;
                        }
                        pNextBufferCB->pNextBufferCB = &ha->pReceiveBufferCBs[lTagVal];;
                        pNextBufferCB = pNextBufferCB->pNextBufferCB;

                        if (!(pNextBufferCB->lFlags & BCB_FLAGS_RISC_OWNS_BUFFER)) {
                           /* Invalid handle from RISC, reset RISC firmware */
                           printk(KERN_WARNING "qla2x00: bad IP buffer handle %x\n", lTagVal);
                           QLA2100_DPC_LOCK(ha);
                           ha->dpc_flags |= ISP_ABORT_NEEDED;
                           QLA2100_DPC_UNLOCK(ha);
                           goto InvalidIpHandle;
                        }

                        /* Set buffer belongs to driver now */
                        pNextBufferCB->lFlags &= ~BCB_FLAGS_RISC_OWNS_BUFFER;
                     } else {
                        pNextBufferCB->lBufferSize = lPacketSize;
                        pNextBufferCB->pNextBufferCB = NULL;
                        break;
                     }
                  }
                  /* Check for incoming ARP packet with matching IP address */
                  if (pIpReceiveEntry->wServiceClass == 0) {
                     PPACKET_HEADER  pPacket = (PPACKET_HEADER)pBufferCB->pBuffer;
                     PIP_DEVICE_BLOCK pIpDevice;
                     uint8_t     acPortId[3];

                     /* Scan list of IP devices to see if login needed */
                     for (pIpDevice = ha->pIpDeviceTop; pIpDevice != NULL;
                     pIpDevice = pIpDevice->pNextIpDevice) {
                        if (*(uint16_t *)(&pIpDevice->acWorldWideName[2]) ==
                        pPacket->sNetworkHeader.wSourceAddrHigh &&
                        *(uint32_t *)(&pIpDevice->acWorldWideName[4]) ==
                        pPacket->sNetworkHeader.lSourceAddrLow) {
                           /* Device already in IP list, skip login */
                           goto SkipDeviceLogin;
                        }
                     }

                     /* Device not in list, need to do login */
                     acPortId[0] = pIpReceiveEntry->cS_IDHigh;
                     acPortId[1] = (uint8_t)(pIpReceiveEntry->wS_IDLow >> 8);
                     acPortId[2] = (uint8_t)pIpReceiveEntry->wS_IDLow;

                     /* Make sure its not a local device */
                     if (acPortId[0] == ha->port_id[0] &&
                     acPortId[1] == ha->port_id[1]) {
                        goto SkipDeviceLogin;
                     }

                     if (qla2x00_add_new_ip_device(ha, PUBLIC_LOOP_DEVICE,
                     acPortId,
                     (uint8_t *)&pPacket->sNetworkHeader.wSourceNAA,
                     TRUE) == QL_STATUS_FATAL_ERROR) {
                        /* Fatal error, reinitialize */
                        QLA2100_DPC_LOCK(ha);
                        ha->dpc_flags |= ISP_ABORT_NEEDED;
                        QLA2100_DPC_UNLOCK(ha);
                     }
                  }
                  SkipDeviceLogin:
                  /* Pass received packet to IP driver */
                  pBufferCB->wBufferCount = wBufferCount;
                  (*ha->pReturnReceivePacketsRoutine)
                  (ha->pReturnReceivePacketsContext,
                  pBufferCB);

                  /* Keep track of RISC buffer pointer (for IP reinit) */
                  ha->wIpBufferOut += wBufferCount;
                  if (ha->wIpBufferOut >= IP_BUFFER_QUEUE_DEPTH)
                     ha->wIpBufferOut -= IP_BUFFER_QUEUE_DEPTH;
                  InvalidIpHandle:
                  /* Adjust ring index. */
                  ha->rsp_ring_index++;
                  if( ha->rsp_ring_index == RESPONSE_ENTRY_CNT ) {
                     ha->rsp_ring_index = 0;
                     ha->response_ring_ptr = ha->response_ring;
                  } else
                     ha->response_ring_ptr++;
#if defined(ISP2100) || defined(ISP2200)
                  WRT_REG_WORD(&reg->mailbox5, ha->rsp_ring_index);
#else
                  WRT_REG_WORD(&reg->rsp_q_out, ha->rsp_ring_index);
#endif

                  continue;
               }

               /* Handle IP FARP request */
               else if (pkt->entry_type == ET_IP_FARP_REQUEST) {
                  PIP_FARP_REQUEST_ENTRY  pIpFarpRequestEntry;
                  uint8_t                 acPortId[3];
                  uint8_t                 acPortName[8];

                  pIpFarpRequestEntry = (PIP_FARP_REQUEST_ENTRY)pkt;
                  acPortId[0] = pIpFarpRequestEntry->cRequesterPortIdHigh;
                  acPortId[1] = (uint8_t)(pIpFarpRequestEntry->wRequesterPortIdLow >> 8);
                  acPortId[2] = (uint8_t)pIpFarpRequestEntry->wRequesterPortIdLow;
                  acPortName[0] = pIpFarpRequestEntry->acRequesterPortName[7];
                  acPortName[1] = pIpFarpRequestEntry->acRequesterPortName[6];
                  acPortName[2] = pIpFarpRequestEntry->acRequesterPortName[5];
                  acPortName[3] = pIpFarpRequestEntry->acRequesterPortName[4];
                  acPortName[4] = pIpFarpRequestEntry->acRequesterPortName[3];
                  acPortName[5] = pIpFarpRequestEntry->acRequesterPortName[2];
                  acPortName[6] = pIpFarpRequestEntry->acRequesterPortName[1];
                  acPortName[7] = pIpFarpRequestEntry->acRequesterPortName[0];
                  /* Login and add device to IP database */
                  if (qla2x00_add_new_ip_device(ha, PUBLIC_LOOP_DEVICE,
                  acPortId,
                  acPortName,
                  TRUE) == QL_STATUS_FATAL_ERROR) {
                     /* Fatal error, reinitialize */
                     QLA2100_DPC_LOCK(ha);
                     ha->dpc_flags |= ISP_ABORT_NEEDED;
                     QLA2100_DPC_UNLOCK(ha);
                  }

                  /* Adjust ring index. */
                  ha->rsp_ring_index++;
                  if( ha->rsp_ring_index == RESPONSE_ENTRY_CNT ) {
                     ha->rsp_ring_index = 0;
                     ha->response_ring_ptr = ha->response_ring;
                  } else
                     ha->response_ring_ptr++;
#if defined(ISP2100) || defined(ISP2200)
                  WRT_REG_WORD(&reg->mailbox5, ha->rsp_ring_index);
#else
                  WRT_REG_WORD(&reg->rsp_q_out, ha->rsp_ring_index);
#endif
                  continue;
               }
#endif /* FC_IP_SUPPORT */

	       /* Correct this line: ha->actthreads--; */
	       if (pkt->entry_status != 0) {
		  DEBUG3(printk(KERN_INFO
		    "qla2x00_response_pkt: process error entry.\n");)
		    qla2x00_error_entry(ha, pkt, done_q_first, done_q_last);
	       } else {
		  DEBUG3(printk(KERN_INFO
		  "qla2x00_response_pkt: process response entry.\n");)

		  switch (pkt->entry_type) {
		  case STATUS_TYPE:
			qla2x00_status_entry(ha, (sts_entry_t *)pkt,
			    done_q_first, done_q_last);
			break;

		  case MS_IOCB_TYPE:
			qla2x00_ms_entry(ha, (ms_iocb_entry_t *)pkt,
			    done_q_first, done_q_last);
			break;

		  default:
			/* Type Not Supported. */
			DEBUG3(printk(KERN_WARNING
			    "qla2x00_response_pkt: received unknown "
			    "response pkt type %x entry status=%x.\n",
			    pkt->entry_type, pkt->entry_status);)
			break;
		  }
	       }
	       /* Adjust ring index. */
	       ha->rsp_ring_index++;
	       if (ha->rsp_ring_index == RESPONSE_ENTRY_CNT) {
	          ha->rsp_ring_index = 0;
		  ha->response_ring_ptr = ha->response_ring;
	       } else {
		  ha->response_ring_ptr++;
	       }

#if defined(ISP2100) || defined(ISP2200)
               WRT_REG_WORD(&reg->mailbox5, ha->rsp_ring_index);
#else
               WRT_REG_WORD(&reg->rsp_q_out, ha->rsp_ring_index);
#endif
	} /* while (ha->rsp_ring_index != index) */

	DEBUG3(printk("qla2x00_response_pkt: exiting.\n");)
}

/*
*  qla2x00_status_entry
*      Processes received ISP status entry.
*
* Input:
*      ha           = adapter block pointer.
*      pkt          = entry pointer.
*      done_q_first = done queue first pointer.
*      done_q_last  = done queue last pointer.
*/
STATIC void
qla2x00_status_entry(scsi_qla_host_t *ha, sts_entry_t *pkt,
    srb_t **done_q_first, srb_t **done_q_last)
{
#ifdef QL_DEBUG_LEVEL_2
                uint32_t       b, t, l;
#endif
                uint8_t        sense_sz = 0;
                srb_t          *sp;
                os_lun_t       *lq;
		os_tgt_t		*tq;

                Scsi_Cmnd       *cp;
#ifdef DPC_LOCK
                unsigned long cpu_flags = 0;
#endif
#ifndef MPIO_SUPPORT
                srb_t          *sp1;
                os_lun_t       *lq1;
		ql_list_link_t	*link;
		ql_list_link_t	*next_link = NULL;
                fc_port_t       *fcport = NULL;
#endif

#ifdef QL_DEBUG_LEVEL_3
                ENTER("qla2100_status_entry");
#endif

                /* Validate handle. */
                if( pkt->handle < MAX_OUTSTANDING_COMMANDS )
                    sp = ha->outstanding_cmds[pkt->handle];
                else
                    sp = 0;

                if( sp ) {
                    /* Free outstanding command slot. */
                    ha->outstanding_cmds[pkt->handle] = 0;
                    cp = sp->cmd;
                    ((scsi_qla_host_t *)(cp->host->hostdata))->actthreads--;
                    sp->ccode = pkt->comp_status;
                    sp->scode = pkt->scsi_status;
           	    sp->flags |= SRB_ISP_COMPLETED;

                    /* Generate LU queue on cntrl, target, LUN */
#ifdef QL_DEBUG_LEVEL_2
                    b = SCSI_BUS_32(cp);
                    t = SCSI_TCN_32(cp);
                    l = SCSI_LUN_32(cp);
#endif
                    tq = sp->tgt_queue;
                    lq = sp->lun_queue;

                    /* Target busy */
                    if( pkt->scsi_status & SS_BUSY_CONDITION ) {
#ifdef QL_DEBUG_LEVEL_2
                        qla2100_print("qla2100_status_entry: SCSI busy status, scsi(");
                        qla2100_output_number((u_long)ha->host_no, 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)b, 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)t, 10);
                        qla2100_print(":");
                        qla2100_output_number((u_long)l, 10);
                        qla2100_print(")\n");
#endif
                        sp->retry_count--;
                        if ((uint8_t)pkt->scsi_status == SS_RESERVE_CONFLICT) {
                           CMD_RESULT(cp) = (int) (DID_OK << 16) |
                           SS_RESERVE_CONFLICT;
                        }
                        else {
                           CMD_RESULT(cp) = (int) (DID_BUS_BUSY << 16) |
                           (pkt->scsi_status & 0xff);
                        }
                    }
                    else if( ha->loop_down_timer ||
                             ha->loop_state != LOOP_READY ) {
#ifdef QL_DEBUG_LEVEL_2
                printk("scsi(%02d:%02d:%02d:%02d): Loop down - serial_number = %d, handle=0x%x\n",
		(int)ha->host_no,
                (int)b,
                (int)t,
                (int)l,
                (int)sp->cmd->serial_number,
                pkt->handle);
#endif
                            CMD_RESULT(cp) = (int) (DID_BUS_BUSY << 16);
#ifndef MPIO_SUPPORT
		} else if (sp->port_down_retry_count > 1 &&
		    (pkt->comp_status == CS_INCOMPLETE ||
		    pkt->comp_status == CS_ABORTED ||
		    pkt->comp_status == CS_PORT_UNAVAILABLE ||
		    pkt->comp_status == CS_PORT_LOGGED_OUT ||
		    pkt->comp_status == CS_PORT_CONFIG_CHG ||
		    pkt->comp_status == CS_PORT_BUSY)) {
                            /* if the port is unavaliable and we haven't exceeded the port down count */
                            /* then send command back to the mid-level. */
#ifdef QL_DEBUG_LEVEL_2
                            qla2100_print("scsi(");
                qla2100_output_number((u_long)ha->host_no, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)b, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)t, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)l, 10);
                            qla2100_print(")");
			    DEBUG(sprintf(debug_buff,
": Port Down Retry count=%d, pid=%ld, compl status=0x%x, q_flag=0x%x\n", 
			sp->port_down_retry_count,
                	sp->cmd->serial_number,
                	pkt->comp_status,lq->q_flag );)
                        DEBUG(qla2100_print(debug_buff));
#endif

                            sp->port_down_retry_count--;

                            /* dg 08/17/99
                            * Force the SCSI layer to keep retrying until our
                            * port_down_retry_count expire.
                            */
                            /* v2.19.14 */
                            CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;

			/* Acquire target queue lock */
			TGT_LOCK(tq);

			if (!(lq->q_flag & LUN_QUEUE_SUSPENDED)) {
				lq->q_flag = lq->q_flag | LUN_QUEUE_SUSPENDED;
				/* Decrement port down count. */
				if (ha->flags.link_down_error_enable) {
					if (tq->port_down_retry_count != 0)
						tq->port_down_retry_count--;
				}

                /* Decrement port down count on all pending commands. */
                /* and return them back to OS.                        */
				/* fixme(dg) */
				for (link = lq->cmd.first; link != NULL;
			    	link = next_link) {
					next_link = link->next;
					sp1 = link->base_address;

					lq1 = lq;
					/* Remove srb from LUN queue. */
					qla2x00_remove_link(&lq1->cmd, &sp1->s_cmd);
            				lq1->q_incnt--;

#ifdef QL_DEBUG_LEVEL_2
                            qla2100_print("scsi(");
                qla2100_output_number((u_long)ha->host_no, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)b, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)t, 10);
                qla2100_print(":");
                qla2100_output_number((u_long)l, 10);
                            qla2100_print(")");
			    sprintf(debug_buff,
": Q-Port Down Retry count=%d, pid=%ld, compl status=0x%x, q_flag=0x%x\n", 
			sp1->port_down_retry_count,
                	sp1->cmd->serial_number,
                	pkt->comp_status,lq->q_flag );
                        DEBUG(qla2100_print(debug_buff));
#endif

					/* Set ending status. */
                	CMD_RESULT(sp1->cmd) = DID_BUS_BUSY << 16;
					/* SP get freed */
                        /* qla2100_callback(ha,sp1->cmd); */
                        CMD_HANDLE(sp1->cmd) = (unsigned char *) NULL;
                        qla2100_done_q_put(ha, sp1, done_q_first, done_q_last);

                        }
          
			fcport = lq->fclun->fcport;
			if( PORT_DOWN(fcport) == 0 )
				PORT_DOWN(fcport) = PORT_RETRY_TIME;

                	if( PORT_DOWN_RETRY(fcport) == 0 ) {
               			PORT_DOWN_RETRY(fcport) =
				ha->port_down_retry_count * PORT_RETRY_TIME;
			}
			fcport->vis_ha = 
				(scsi_qla_host_t *)sp->cmd->host->hostdata;
		        }
			/* Release target queue lock */
			TGT_UNLOCK(tq);
#endif

                        } else {
#ifdef QL_DEBUG_LEVEL_2
                            if( pkt->comp_status ) {
                                qla2100_print(
                        "qla2100_status_entry: Compl error = ");
                    qla2100_output_number((u_long)pkt->comp_status, 16);
                                qla2100_print(", scsi(");
                    qla2100_output_number((u_long)ha->host_no, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)b, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)t, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)l, 10);
                                qla2100_print("), retry count= ");
                    qla2100_output_number((u_long)sp->port_down_retry_count, 10);
                                qla2100_print(", pid = ");
                    qla2100_output_number((u_long)cp->serial_number, 16);
                                qla2100_print("\n");
                            }
#endif
                            /* Set ISP completion status and arget status byte. */
                            CMD_RESULT(cp) = qla2100_return_status(ha, pkt, cp);

                            memset((caddr_t)cp->sense_buffer, 0,sizeof(cp->sense_buffer));

#ifdef QL_DEBUG_LEVEL_3
                    qla2100_print("qla2100_status_entry: pre-check: scsi_status (0x");
                    qla2100_output_number((u_long)pkt->scsi_status, 16);
                    qla2100_print(")\n");
#endif

                            if( pkt->scsi_status & SS_CHECK_CONDITION ) {
                                /* Mid-level always zero sense buffer before giving it to us */
                                if( pkt->scsi_status & SS_SENSE_LEN_VALID ) {
                                    if( pkt->req_sense_length < CMD_SNSLEN(cp) )
                                        sense_sz = pkt->req_sense_length;
                                    else
                                        sense_sz = CMD_SNSLEN(cp) - 1;

                                    BCOPY((caddr_t)&pkt->req_sense_data, cp->sense_buffer, sense_sz);
            						qla2x00_check_sense(cp, lq);
                                }

#ifdef QL_DEBUG_LEVEL_2
                                qla2100_print(
                                "qla2100_status_entry: Check condition Sense data, scsi(");
                    qla2100_output_number((u_long)ha->host_no, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)b, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)t, 10);
                    qla2100_print(":");
                    qla2100_output_number((u_long)l, 10);
                                qla2100_print(")\n");
#ifdef QL_DEBUG_LEVEL_3
                                if( sense_sz )
                                    qla2100_dump_buffer(cp->sense_buffer, sense_sz);
#endif
#endif
                            }
                        }
                        /* Place command on done queue. */
                        qla2100_done_q_put(ha, sp, done_q_first, done_q_last);
                    } else {
#ifdef QL_DEBUG_LEVEL_2
                        qla2100_print("qla2100_status_entry: ISP Invalid handle\n");
#endif
                        printk(KERN_WARNING "qla2x00: Status Entry invalid handle");

                    QLA2100_DPC_LOCK(ha);
	                ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    }
#ifdef QL_DEBUG_LEVEL_3
                    LEAVE("qla2100_status_entry");
#endif
}

/*
*  qla2x00_error_entry
*      Processes error entry.
*
* Input:
*      ha           = adapter block pointer.
*      pkt          = entry pointer.
*      done_q_first = done queue first pointer.
*      done_q_last  = done queue last pointer.
*/
STATIC void
qla2x00_error_entry(scsi_qla_host_t *ha, response_t *pkt, srb_t **done_q_first,
    srb_t **done_q_last)
{
                    srb_t *sp;
#ifdef DPC_LOCK
                    unsigned long cpu_flags = 0;
#endif

#ifdef QL_DEBUG_LEVEL_3
                    ENTER("qla2100_error_entry");
#endif

#ifdef QL_DEBUG_LEVEL_2
                    if( pkt->entry_status & BIT_5 )
                        qla2100_print("qla2100_error_entry: Invalid Entry Order\n");
                    else if( pkt->entry_status & BIT_4 )
                        qla2100_print("qla2100_error_entry: Invalid Entry Count\n");
                    else if( pkt->entry_status & BIT_3 )
                        qla2100_print("qla2100_error_entry: Invalid Entry Parameter\n");
                    else if( pkt->entry_status & BIT_2 )
                        qla2100_print("qla2100_error_entry: Invalid Entry Type\n");
                    else if( pkt->entry_status & BIT_1 )
                        qla2100_print("qla2100_error_entry: Busy\n");
                    else
                        qla2100_print("qla2100_error_entry: UNKNOWN flag error\n");
#endif

                    /* Validate handle. */
                    if( pkt->handle < MAX_OUTSTANDING_COMMANDS )
                        sp = ha->outstanding_cmds[pkt->handle];
                    else
                        sp = 0;

                    if( sp ) {
                        /* Free outstanding command slot. */
                        ha->outstanding_cmds[pkt->handle] = 0;
                        ((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads--;

           		sp->flags |= SRB_ISP_COMPLETED;

                        /* Bad payload or header */
                        if( pkt->entry_status & (BIT_5 + BIT_4 + BIT_3 + BIT_2) ) {
                            /* Bad payload or header, set error status. */
                            CMD_RESULT(sp->cmd) = (int) DID_ERROR << 16;

                        } else if( pkt->entry_status & BIT_1 && sp->retry_count ) /* FULL flag */
                        {
                            sp->retry_count--;
                            CMD_RESULT(sp->cmd) = (int) DID_BUS_BUSY << 16;
                        } else {
                            /* Set error status. */
                            CMD_RESULT(sp->cmd) =(int)  DID_ERROR << 16;
                        }
                        /* Place command on done queue. */
                        qla2100_done_q_put(ha, sp, done_q_first, done_q_last);
                    } else if( pkt->entry_type == COMMAND_A64_TYPE ||
                    pkt->entry_type == COMMAND_TYPE ) {
#ifdef QL_DEBUG_LEVEL_2
                        qla2100_print("qla2100_error_entry: ISP Invalid handle\n");
#endif
                        printk(KERN_WARNING "qla2x00: Error Entry invalid handle");
                    QLA2100_DPC_LOCK(ha);
	                ha->dpc_flags |= ISP_ABORT_NEEDED;
                    QLA2100_DPC_UNLOCK(ha);
                    }

#ifdef QL_DEBUG_LEVEL_3
                    LEAVE("qla2100_error_entry");
#endif
}

STATIC void
qla2x00_ms_entry(scsi_qla_host_t *ha, ms_iocb_entry_t *pkt,
    srb_t **done_q_first, srb_t **done_q_last)
{
	srb_t          *sp;

	DEBUG3(printk("qla2x00_ms_entry: entered. pkt=%p pkthandle=%d.\n",
	    pkt, pkt->handle);)

	/* Validate handle. */
	if (pkt->handle < MAX_OUTSTANDING_COMMANDS)
		sp = ha->outstanding_cmds[pkt->handle];
	else
		sp = NULL;

	if (sp == NULL) {
	        printk(KERN_WARNING "qla2x00: MS Entry invalid handle.\n");
	        ha->dpc_flags |= ISP_ABORT_NEEDED;
		return;
	}

	/* Free outstanding command slot. */
	ha->outstanding_cmds[pkt->handle] = 0;
	((scsi_qla_host_t *)(sp->cmd->host->hostdata))->actthreads--;
	sp->flags |= SRB_ISP_COMPLETED;

	DEBUG(sp->state = 0x8;)
	qla2100_callback(ha, sp->cmd);

	DEBUG3(printk("qla2x00_ms_entry: exiting.\n");)
}

/*
 *  qla2x00_restart_queues
 *	Restart device queues.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
void
qla2x00_restart_queues(scsi_qla_host_t *ha, uint8_t flush)
{
	os_tgt_t	*tq;
	os_lun_t	*lq;
	uint16_t	t, l;
	srb_t  *sp, *sp_next;
	srb_t  *done_q_first = (srb_t *) NULL;
	srb_t  *done_q_last = (srb_t *) NULL;
	int cnt = 0;
	unsigned long cpu_flags = 0;

	ENTER("qla2x00_restart_queues");

	ha->flags.restart_queues_needed = FALSE;
	/*
	 * start all queues working again.
	 */
	for (t = 0; t < ha->max_targets; t++) {
		/* Test for target. */
		if ((tq = TGT_Q(ha, t)) == NULL)
			continue;

		for (l = 0; l < ha->max_luns; l++) {
			lq = LUN_Q(ha, t, l);
			if (lq == NULL)
				continue;

			/* Acquire target queue lock. */
			TGT_LOCK(tq);

			lq->q_flag = lq->q_flag & ~LUN_QUEUE_SUSPENDED;

			if (lq->cmd.first != NULL)
				qla2x00_next(tq, lq);
			else
				/* Release target queue lock */
				TGT_UNLOCK(tq);
		}
	}

	/*
	 * Clear out our retry queue
	 */
	if( flush ) {
		QLA_RETRY_Q_LOCK(ha, cpu_flags);
		for (sp = ha->retry_q_first; (sp); sp = sp_next) {
			sp_next = sp->s_next;

			/* when time expire return request back to OS as BUSY */
			qla2100_timeout_remove(ha, sp);
			CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
			CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;

			qla2100_done_q_put(ha, sp, (srb_t **)&done_q_first,
			    (srb_t **)&done_q_last);
		}
		QLA_RETRY_Q_UNLOCK(ha, cpu_flags);

		cnt = qla2100_done(ha, (srb_t **)&done_q_first,
		    (srb_t **)&done_q_last);

		DEBUG2(printk("qla2100_restart_queues: callback %d "
		    "commands.\n", cnt);)
	}

#ifdef QL_DEBUG_LEVEL_3
	LEAVE("qla2x00_restart_queues");
#endif
}

/*
 *  qla2x00_abort_queues
 *	Abort all commands on queues on device
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Interrupt context.
 */
STATIC void
qla2x00_abort_queues(scsi_qla_host_t *ha, uint8_t doneqflg)
{

	os_tgt_t	*tq;
	os_lun_t	*lq;
    uint32_t    t, l;
    srb_t       *sp;
	ql_list_link_t	*link;
	ql_list_link_t	*next_link;

#ifdef QL_DEBUG_LEVEL_3
   ENTER("qla2100_abort_queues");
#endif
   ha->flags.abort_queue_needed = FALSE;
	/* Return all commands device queues. */
   for( t = 0; t < ha->max_targets; t++ ) {
		/* Test for target. */
		if ((tq = TGT_Q(ha, t)) == NULL)
			continue;

		for (l = 0; l < MAX_LUNS; l++) {
			lq = LUN_Q(ha, t, l);
			if (lq == NULL)
				continue;

			/* Acquire target queue lock. */
			TGT_LOCK(tq);

			/*
			 * Set BUS BUSY error status for
			 * all commands on LUN queue.
			 */
			for (link = lq->cmd.first; link != NULL;
			    link = next_link) {
				next_link = link->next;
				sp = link->base_address;

				if (sp->flags & SRB_ABORTED)
					continue;

				/* Remove srb from LUN queue. */
				qla2x00_remove_link(&lq->cmd, &sp->s_cmd);
            			lq->q_incnt--;

				/* Set ending status. */
                CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
                if ( doneqflg ) {
                   CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;
                   qla2100_done_q_put(ha, sp, (srb_t **)&ha->done_q_first, (srb_t **)&ha->done_q_last);
                } else {
                   qla2100_callback(ha,sp->cmd);
				}
			}
			/* Release target queue lock */
			TGT_UNLOCK(tq);
		}
	}
	/* Abort one command at a time for system. */

#ifdef QL_DEBUG_LEVEL_3
   LEAVE("qla2100_abort_queues");
#endif
}

#if MPIO_SUPPORT
/*
 * qla2x00_reset_lun_fo_counts
 *	Reset failover retry counts
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Interrupt context.
 */
void qla2x00_reset_lun_fo_counts(scsi_qla_host_t *ha, os_lun_t *lq)
{
   	srb_t       *tsp, *sp_next;
	ql_list_link_t	*link1;
	os_lun_t	*orig_lq;
        unsigned long cpu_flags = 0;

	/*
	* the varies input queues.
	*/
	for (link1 = lq->cmd.first;
  			link1 != NULL; link1 = link1->next) {
		tsp = link1->base_address;
		tsp->fo_retry_cnt = 0;
	}
	/*
 	* the retry queue.
 	*/
	QLA_RETRY_Q_LOCK(ha, cpu_flags);
	for( tsp = ha->retry_q_first; (tsp); tsp = sp_next ) {
		sp_next = tsp->s_next;
		orig_lq = tsp->lun_queue;
		if ( orig_lq == lq )
			tsp->fo_retry_cnt = 0;
	}
	QLA_RETRY_Q_UNLOCK(ha, cpu_flags);

	/*
 	* the done queue.
 	*/
        QLA_DONE_Q_LOCK(ha, cpu_flags);
        for( tsp = ha->done_q_first; (tsp); tsp = sp_next ) {
        	sp_next = tsp->s_next;
		orig_lq = tsp->lun_queue;
		if ( orig_lq == lq )
			tsp->fo_retry_cnt = 0;
	}
        QLA_DONE_Q_UNLOCK(ha, cpu_flags);
}

/*
 *  qla2x00_failover_cleanup
 *	Cleanup queues after a failover.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Interrupt context.
 */
STATIC void
qla2x00_failover_cleanup(scsi_qla_host_t *os_ha, os_tgt_t *tq, os_lun_t	*lq, srb_t *sp)
{

        CMD_RESULT(sp->cmd) = DID_BUS_BUSY << 16;
        CMD_HANDLE(sp->cmd) = (unsigned char *) NULL;
	/* turn-off all failover flags */
	sp->flags = sp->flags & ~(SRB_RETRY|SRB_FAILOVER|SRB_FO_CANCEL);
}


/*
 *  qla2x00_process_failover
 *	Process any command on the failover queue.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Interrupt context.
 */
STATIC void
qla2x00_process_failover(scsi_qla_host_t *ha)
{

	os_tgt_t	*tq;
	os_lun_t	*lq;
    	srb_t       *sp;
	ql_list_link_t	*link;
	ql_list_link_t	*next_link;
	fc_port_t *fcport;

   DEBUG(printk("qla2x00_process_failover - for hba %d\n",(int)ha->instance);)
	/*
	 * Process all the commands in the failover
	 * queue. Attempt to failover then either
	 * complete the command as is
	 * or requeue for retry.
	 */

	/* Prevent or allow acceptance of new I/O requests. */

	for (link = ha->failover_queue.first; link != NULL;
			    link = next_link) {
		next_link = link->next;
		sp = link->base_address;

		tq = sp->tgt_queue;
		lq = sp->lun_queue;
		fcport = lq->fclun->fcport; 

		/* Remove srb from failover queue. */
		qla2x00_remove_link(&ha->failover_queue, &sp->s_cmd);

			DEBUG2(printk("qla2x00_process_failover: pid %ld\n",sp->cmd->serial_number);)
			/* Select an alternate path */
		    	/* v5.21b16  - path has already been change by a previous request */
			/*
			 * if path changed after the bad path was detected
			 * then no need of changing paths again.
			 */
		    if ( sp->fclun != lq->fclun || 
			 ( (sp->err_id == 1 || sp->err_id == 3) && 
				fcport->loop_id != FC_NO_LOOP_ID) )  
			{
				qla2x00_failover_cleanup(ha,tq,lq,sp);
			} else if ( qla2x00_cfg_failover(ha, lq->fclun, 
					tq, sp) == NULL ) {
				/*
			 	* We ran out of paths, so just
			 	* post the status which is already
			 	* set in the cmd.
			 	*/
				printk(KERN_INFO "qla2x00_process_failover: Ran out of paths - pid %ld\n",sp->cmd->serial_number);
				sp->fstate = 4;
			}
			else {
				sp->fstate = 5;

				qla2x00_failover_cleanup(ha,tq,lq,sp);

        			lq->q_flag &= ~QLA2100_QSUSP;
			}
        	qla2100_callback(ha,sp->cmd);
	}



	qla2x00_restart_queues(ha,TRUE);
   DEBUG(printk("qla2x00_process_failover - done");)
}
#endif	/* end of MPIO_SUPPORT */

/*
 *  qla2100_loop_resync
 *      Resync with fibre channel devices.
 *
 * Input:
 *      ha = adapter block pointer.
 *
 * Returns:
 *      0 = success
 */
STATIC uint8_t
qla2100_loop_resync(scsi_qla_host_t *ha)
{
                    uint8_t   status;
#ifdef DPC_LOCK
                    unsigned long cpu_flags = 0;
#endif

#ifdef QL_DEBUG_LEVEL_3
                    ENTER("qla2100_loop_resync");
#endif

                    ha->loop_state = LOOP_UPDATE;
                    if( ha->flags.online ) {
                        if( !(status = qla2x00_fw_ready(ha)) ) {
                            do {
                                /* v2.19.05b6 */
                                ha->loop_state = LOOP_UPDATE;

                                /* Issue marker command. */
                                qla2x00_marker(ha, 0, 0, MK_SYNC_ALL);

                                /* Remap devices on Loop. */
                                QLA2100_DPC_LOCK(ha);
                                ha->dpc_flags &= ~LOOP_RESYNC_NEEDED;
                                QLA2100_DPC_UNLOCK(ha);

                                qla2100_configure_loop(ha);

                            }while( !ha->loop_down_timer &&
				!(ha->dpc_flags & ISP_ABORT_NEEDED) &&
                                (ha->dpc_flags & LOOP_RESYNC_NEEDED) ) ;
                        }
                        /* v2.19 - we don't want to call this if we are already
                         * in the loop resync code
                         */
                        qla2x00_restart_queues(ha,TRUE);
                    } else
                        status = 0;

		    if (ha->dpc_flags & ISP_ABORT_NEEDED) {
			return (1);
		    }

                    /* Restart queues that may have been stopped. */
                    /* 04/10 if( !ha->loop_down_timer ) {
                        qla2x00_restart_queues(ha,TRUE);
                    } */
                    /* v2.19 */

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
                    if( status )
                        qla2100_print("qla2100_loop_resync: **** FAILED ****\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
                    else
                        LEAVE("qla2100_loop_resync");
#endif
                    return(status);
}

/*
 * qla2100_debounce_register
 *      Debounce register.
 *
 * Input:
 *      port = register address.
 *
 * Returns:
 *      register value.
 */
STATIC uint16_t
qla2100_debounce_register(volatile uint16_t *addr)
{
                    volatile uint16_t ret;
                    volatile uint16_t ret2;

                    do {
                        ret = RD_REG_WORD(addr);
                        ret2 = RD_REG_WORD(addr);
                    }while( ret != ret2 );

                    return(ret);
}

/* qla2100_cmd_wait
 *	Stall driver until all outstanding commands are returned.
 *
 * Input:
 *	ha = adapter state pointer.
 *
 * Return;
 *  0 -- Done
 *  1 -- continue;
 *
 * Context:
 *	Kernel context.
 *
 * Note:
 *	This function must be called when io_request_lock and dpc lock
 *	are both held.
 */
STATIC void
qla2100_cmd_wait(scsi_qla_host_t *ha)
{

	uint32_t	cnt;
	uint16_t	index;
#ifdef DPC_LOCK
	unsigned long	cpu_flags = 0;
#endif
	srb_t		*psrb;
	fc_port_t	*fcport;
	unsigned long	time1;


	DEBUG3(printk("qla2100_cmd_wait(%ld): entered.\n", ha->host_no);)

	ha->dpc_flags &= ~COMMAND_WAIT_NEEDED;

	/* Wait for all outstanding commands to be returned. */
	cnt = 10;
	for (index = 0; index < MAX_OUTSTANDING_COMMANDS; index++) {
		if (ha->outstanding_cmds[index] == NULL) {
			continue;
		} else if (cnt != 0) {
			/*
			 * If the command is for a fabric device and
			 * device is NOT dead, continue to wait.
			 */
			psrb = ha->outstanding_cmds[index];
			fcport = psrb->lun_queue->fclun->fcport;
			if ((fcport->flags & FC_FABRIC_DEVICE) &&
			    (fcport->state != FC_DEVICE_DEAD)) {
				continue;
			}

			spin_unlock_irqrestore(&io_request_lock,ha->cpu_flags);

			/* Wait 1 second. */
			time1 = HZ;
			do {
				set_current_state(TASK_INTERRUPTIBLE);
				time1 = schedule_timeout(time1);
			} while (time1);

			spin_lock_irqsave(&io_request_lock,ha->cpu_flags);

			if ((ha->dpc_flags & COMMAND_WAIT_NEEDED) &&
			    (ha->dpc_flags & LOOP_RESYNC_NEEDED)) {
				DEBUG3(printk("qla2x00_cmd_wait(%ld): pending "
				    "loop resync\n", ha->instance);)
				/*
				 * No need to proceed. Let the next cmd_wait
				 * kick in and pick up the rest.
				 */
				break;
			}

			index = 0;
			cnt--;
		} else {
			DEBUG2(printk("qla2x00_cmd_wait(%ld): timeout\n",
			    ha->instance);)
			break;
		}
	}

	DEBUG3(printk("qla2100_cmd_wait(%ld): exiting.\n", ha->host_no);)
}


/*
 * qla2100_reset_chip
 *      Reset ISP chip.
 *
 * Input:
 *      ha = adapter block pointer.
 */
STATIC void
qla2100_reset_chip(scsi_qla_host_t *ha)
{
	device_reg_t	*reg = ha->iobase;
	uint32_t	cnt;
	unsigned long	mbx_flags = 0;

#ifdef QL_DEBUG_LEVEL_3
	ENTER("qla2100_reset_chip");
#endif

	/* Disable ISP interrupts. */
	qla2100_disable_intrs(ha);
	/* WRT_REG_WORD(&reg->ictrl, 0); */

#if 1
	/* Pause RISC. */
	WRT_REG_WORD(&reg->host_cmd, HC_PAUSE_RISC);
	if (ha->device_id == QLA2312_DEVICE_ID) {
		UDELAY(10);
	} else {
		for (cnt = 0; cnt < 30000; cnt++) {
			if ((RD_REG_WORD(&reg->host_cmd) & HC_RISC_PAUSE) != 0)
				break;
			else
				UDELAY(100);
		}
	}

	/* Select FPM registers. */
	WRT_REG_WORD(&reg->ctrl_status, 0x20);

	/* FPM Soft Reset. */
	WRT_REG_WORD(&reg->fpm_diag_config, 0x100);
#ifdef ISP2300
	WRT_REG_WORD(&reg->fpm_diag_config, 0x0); /* Toggle Fpm Reset */
#endif

	/* Select frame buffer registers. */
	WRT_REG_WORD(&reg->ctrl_status, 0x10);

	/* Reset frame buffer FIFOs. */
	WRT_REG_WORD(&reg->fb_cmd, 0xa000);

	/* Select RISC module registers. */
	WRT_REG_WORD(&reg->ctrl_status, 0);

	WRT_REG_WORD(&reg->semaphore, 0);

	WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
	WRT_REG_WORD(&reg->host_cmd, HC_CLR_HOST_INT);

	/* Reset ISP chip. */
	WRT_REG_WORD(&reg->ctrl_status, ISP_RESET);

	if (ha->device_id == QLA2312_DEVICE_ID) {
		UDELAY(10);
	} else {
		/* Wait for RISC to recover from reset. */
		for( cnt = 30000; cnt; cnt-- ) {
			if( !(RD_REG_WORD(&reg->ctrl_status) & ISP_RESET) )
				break;
			UDELAY(100);
		}
	}

	/* Reset RISC processor. */
	WRT_REG_WORD(&reg->host_cmd, HC_RESET_RISC);
	WRT_REG_WORD(&reg->host_cmd, HC_RELEASE_RISC);

	if (ha->device_id == QLA2312_DEVICE_ID) {
		UDELAY(10);
	} else {
		for (cnt = 0; cnt < 30000; cnt++) {
			/* ra 12/30/01 */
			if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
				QLA_MBX_REG_LOCK(ha);

			if (RD_REG_WORD(&reg->mailbox0) != MBS_BUSY) {
				if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
					QLA_MBX_REG_UNLOCK(ha);
				break;
			}

			if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
				QLA_MBX_REG_UNLOCK(ha);

			UDELAY(100);
		}
	}

#if defined(ISP2200) || defined(ISP2300)
	/* Disable RISC pause on FPM parity error. */
	WRT_REG_WORD(&reg->host_cmd, HC_DISABLE_PARITY_PAUSE);
#endif

#else
                    /* Insure mailbox registers are free. */
                    WRT_REG_WORD(&reg->semaphore, 0);
                    WRT_REG_WORD(&reg->host_cmd, HC_CLR_RISC_INT);
                    WRT_REG_WORD(&reg->host_cmd, HC_CLR_HOST_INT);

                    /* clear mailbox busy */
                    ha->flags.mbox_busy = FALSE;

                    /* Reset ISP chip. */
                    WRT_REG_WORD(&reg->ctrl_status, ISP_RESET);

                    /*
                    * Delay after reset, for chip to recover.
                    * Otherwise causes system PANIC
                    */
                    mdelay(2);

                    for( cnt = 30000; cnt; cnt-- ) {
                        if( !(RD_REG_WORD(&reg->ctrl_status) & ISP_RESET) )
                            break;
                        UDELAY(100);
                    }

                    /* Reset RISC processor. */
                    WRT_REG_WORD(&reg->host_cmd, HC_RESET_RISC);
                    WRT_REG_WORD(&reg->host_cmd, HC_RELEASE_RISC);
                    for( cnt = 30000; cnt; cnt-- ) {
			if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
				QLA_MBX_REG_LOCK(ha);
                        if( RD_REG_WORD(&reg->mailbox0) != MBS_BUSY ) {
				if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
				    QLA_MBX_REG_UNLOCK(ha);
                            break;
			}
			if (!(ha->dpc_flags & ABORT_ISP_ACTIVE))
				QLA_MBX_REG_UNLOCK(ha);
                        UDELAY(100);
                    }
#endif

#ifdef QL_DEBUG_LEVEL_3
                    LEAVE("qla2100_reset_chip");
#endif
}

/*
 * This routine will wait for fabric devices for
 * the reset delay.
 */
void qla2100_check_fabric_devices(scsi_qla_host_t *ha)
{
	uint16_t	fw_state;

	qla2x00_get_firmware_state(ha, &fw_state);
}

#if 0
/*
 * qla2100_extend_timeout
 *      This routine will extend the timeout to the specified value.
 *
 * Input:
 *      cmd = SCSI command structure
 *
 * Returns:
 *      None.
 */
static void qla2100_extend_timeout(Scsi_Cmnd *cmd, int timeout)
{
    if ( cmd->eh_timeout.function ) {
                del_timer(&cmd->eh_timeout);
                cmd->eh_timeout.expires = jiffies + timeout;
                add_timer(&cmd->eh_timeout);
    	cmd->done_late = 0;
    }
}
#endif

/*
* qla2100_display_fc_names
*      This routine will the node names of the different devices found
*      after port inquiry.
*
* Input:
*      cmd = SCSI command structure
*
* Returns:
*      None.
*/
STATIC void
qla2100_display_fc_names(scsi_qla_host_t *ha)
{
	uint16_t	tgt;
	os_tgt_t	*tq;

	/* Display the node name for adapter */
	printk(KERN_INFO
	    "scsi-qla%d-adapter-node=%02x%02x%02x%02x%02x%02x%02x%02x;\n",
	    (int)ha->instance,
	    ha->init_cb->node_name[0],
	    ha->init_cb->node_name[1],
	    ha->init_cb->node_name[2],
	    ha->init_cb->node_name[3],
	    ha->init_cb->node_name[4],
	    ha->init_cb->node_name[5],
	    ha->init_cb->node_name[6],
	    ha->init_cb->node_name[7]);

	/* display the port name for adapter */
	printk(KERN_INFO
	    "scsi-qla%d-adapter-port=%02x%02x%02x%02x%02x%02x%02x%02x;\n",
	    (int)ha->instance,
	    ha->init_cb->port_name[0],
	    ha->init_cb->port_name[1],
	    ha->init_cb->port_name[2],
	    ha->init_cb->port_name[3],
	    ha->init_cb->port_name[4],
	    ha->init_cb->port_name[5],
	    ha->init_cb->port_name[6],
	    ha->init_cb->port_name[7]);

	/* Print out device port names */
	for (tgt = 0; tgt < MAX_TARGETS; tgt++) {
		if ((tq = ha->otgt[tgt]) == NULL)
			continue;

		if (tq->vis_port == NULL)
			continue;

		printk(KERN_INFO
		    "scsi-qla%d-target-%d=%02x%02x%02x%02x%02x%02x%02x%02x;\n",
		    (int)ha->instance, tgt,
		    tq->fc_name[0], tq->fc_name[1],
		    tq->fc_name[2], tq->fc_name[3],
		    tq->fc_name[4], tq->fc_name[5],
		    tq->fc_name[6], tq->fc_name[7]);

#if VSA
		printk(KERN_INFO
		    "scsi-qla%d-target-%d-vsa=01;\n",
		    (int)ha->instance, tgt);
#endif
	}
}

/*
 * qla2100_find_propname
 *	Get property in database.
 *
 * Input:
 *	ha = adapter structure pointer.
 *      db = pointer to database
 *      propstr = pointer to dest array for string
 *	propname = name of property to search for.
 *	siz = size of property
 *
 * Returns:
 *	0 = no property
 *      size = index of property
 *
 * Context:
 *	Kernel context.
 */
STATIC uint8_t
qla2100_find_propname(scsi_qla_host_t *ha,
    char *propname, char *propstr, char *db, int siz)
{
	char	*cp;

	/* find the specified string */
	if (db) {
		/* find the property name */
		if( (cp = strstr(db,propname) ) != NULL ){
			while( (*cp)  && *cp != '=' ) 
				cp++;
			if ( *cp ) {
				strncpy(propstr, cp, siz+1);
				propstr[siz+1] = '\0';
				DEBUG(printk("qla2100_find_propname: found "
				    "property = {%s} \n",propstr);)
				return (siz);   /* match */
			}
		}
	}
	return (0);
}


/*
 * qla2100_get_prop_16chars
 *	Get an 8-byte property value for the specified property name by
 *      converting from the property string found in the configuration file.
 *      The resulting converted value is in big endian format (MSB at byte0).
 *
 * Input:
 *	ha = adapter state pointer.
 *	propname = property name pointer.
 *	propval  = pointer to location for the converted property val.
 *      db = pointer to database
 *
 * Returns:
 *	0 = value returned successfully.
 *
 * Context:
 *	Kernel context.
 */
static int
qla2100_get_prop_16chars(scsi_qla_host_t *ha,  char *propname,
    char *propval, char *db)
{
	char		*propstr;
	int		i, k;
	int		rval;
	uint8_t		nval;
	uint8_t		*pchar;
	uint8_t		*ret_byte;
	uint8_t		*tmp_byte;
	uint8_t		*retval = (uint8_t*)propval;
	uint8_t		tmpval[8] = {0, 0, 0, 0, 0, 0, 0, 0};
	uint16_t	max_byte_cnt = 8; /* 16 chars = 8 bytes */
	uint16_t	max_strlen = 16;
	char		buf[LINESIZE];

	rval = qla2100_find_propname(ha, propname, buf, db, max_strlen);

	propstr = &buf[0];
	if (*propstr == '=')
		propstr++;   /* ignore equal sign */

	if (rval == 0) {
		return (1);
	}


	/* Convert string to numbers. */

	pchar = (uint8_t *)propstr;
	tmp_byte = (uint8_t *)tmpval;

	rval = 0;
	for (i = 0; i < max_strlen; i++) {
		/*
		 * Check for invalid character, two at a time,
		 * then convert them starting with first byte.
		 */

		if ((pchar[i] >= '0') && (pchar[i] <= '9')) {
			nval = pchar[i] - '0';
		} else if ((pchar[i] >= 'A') && (pchar[i] <= 'F')) {
			nval = pchar[i] - 'A' + 10;
		} else if ((pchar[i] >= 'a') && (pchar[i] <= 'f')) {
			nval = pchar[i] - 'a' + 10;
		} else {
			/* invalid character */
			rval = 1;
			break;
		}

		if (i & BIT_0) {
			*tmp_byte = *tmp_byte | nval;
			tmp_byte++;
		} else {
			*tmp_byte = *tmp_byte | nval << 4;
		}
	}

	if (rval != 0) {
		/* Encountered invalid character. */
		return (rval);
	}

	/* Copy over the converted value. */

	ret_byte = retval;
	tmp_byte = tmpval;

	i = max_byte_cnt;
	k = 0;
	while (i--) {
		*ret_byte++ = *tmp_byte++;
	}

	/* big endian retval[0]; */
	return (0);
}

/*
* qla2100_get_properties
*	Find all properties for the specified adapeter in
*      command line.
*
* Input:
*	ha = adapter block pointer.
*	cmdline = pointer to command line string
*
* Context:
*	Kernel context.
*/
static void
qla2100_get_properties(scsi_qla_host_t *ha, char *cmdline)
{
	char		propbuf[LINESIZE];
	uint8_t		tmp_name[8];

	/* Adapter FC node names. */
	sprintf(propbuf, "scsi-qla%d-adapter-node", (int) ha->instance);
	qla2100_get_prop_16chars (ha, propbuf,
	    (caddr_t)(&ha->init_cb->node_name), cmdline);

	sprintf(propbuf, "scsi-qla%d-adapter-port", (int) ha->instance);

	/* DG 04/07 check portname of adapter */
	qla2100_get_prop_16chars (ha, propbuf,
	    (caddr_t)(tmp_name), cmdline);
	if (BCMP(&ha->init_cb->port_name[0], &tmp_name[0], 8) != 0) {
		/*
		 * Adapter port name is WWN, and cannot be changed.
		 * Inform users of the mismatch, then just continue driver
		 * loading using the original adapter port name in NVRAM.
		 */
		printk(KERN_WARNING
		    "qla2x00: qla%ld found mismatch in adapter port names.\n",
		    ha->instance);
		printk(KERN_INFO
		    "       qla%ld port name found in NVRAM -> %02x%02x%02x%02x%02x%02x%02x%02x\n",
		    ha->instance,
		    ha->init_cb->port_name[0],
		    ha->init_cb->port_name[1],
		    ha->init_cb->port_name[2],
		    ha->init_cb->port_name[3],
		    ha->init_cb->port_name[4],
		    ha->init_cb->port_name[5],
		    ha->init_cb->port_name[6],
		    ha->init_cb->port_name[7]);
		printk(KERN_INFO
		    "      qla%ld port name found on command line -> %02x%02x%02x%02x%02x%02x%02x%02x\n",
		    ha->instance,
		    tmp_name[0],
		    tmp_name[1],
		    tmp_name[2],
		    tmp_name[3],
		    tmp_name[4],
		    tmp_name[5],
		    tmp_name[6],
		    tmp_name[7]);
		printk(KERN_INFO
		    "      Using port name from NVRAM.\n");
	}

	qla2x00_cfg_persistent_binding(ha);
}

/*
 * qla2100_update_fc_database
 *      This routine updates the device data in the database.
 *
 * Input:
 *      ha = adapter block pointer.
 *      device = device data pointer.
 *
 * Returns:
 *      0 = success, if device found or added to database.
 *      BIT_0 = error
 *      BIT_1 = database was full and device was not configured.
 */
STATIC uint8_t
qla2100_update_fc_database(scsi_qla_host_t *ha, fcdev_t *device, uint8_t enable_slot_reuse)
{
    uint16_t    cnt, i;

    DEBUG(printk(
        "qla2x00: Found device - nodename=%02x%02x%02x%02x%02x%02x%02x%02x, "
        "portname=%02x%02x%02x%02x%02x%02x%02x%02x, port Id=%06x, "
        "loop id=%04x\n",
        device->name[0],device->name[1],
        device->name[2],device->name[3],
        device->name[4],device->name[5],
        device->name[6],device->name[7],
        device->wwn[0],device->wwn[1],
        device->wwn[2],device->wwn[3],
        device->wwn[4],device->wwn[5],
        device->wwn[6],device->wwn[7],
        device->d_id.b24, device->loop_id);)

    /* Look for device in database. */
    for (cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++) {
#if USE_PORTNAME
	if (BCMP(device->wwn, ha->fc_db[cnt].wwn, 8) == 0) {
#else
	if (BCMP(device->name, ha->fc_db[cnt].name, 8) == 0) {
#endif
    	    DEBUG(printk(
                "qla2x00: Reusing slot %d for device "
                "%02x%02x%02x%02x%02x%02x%02x%02x\n",
                cnt,
         	device->wwn[0],device->wwn[1],
         	device->wwn[2],device->wwn[3],
         	device->wwn[4],device->wwn[5],
         	device->wwn[6],device->wwn[7]);)
            if (device->flag == DEV_PUBLIC) {
                ha->fc_db[cnt].flag |= DEV_PUBLIC;
            } else {
                if (ha->fc_db[cnt].flag & DEV_PUBLIC) {
                    ha->fc_db[cnt].flag &= ~DEV_PUBLIC;
                    ha->fabricid[ha->fc_db[cnt].loop_id].in_use = FALSE;
                }
            }
            ha->fc_db[cnt].loop_id = device->loop_id;
            ha->fc_db[cnt].d_id.b24 = device->d_id.b24;
            return (0);
        }
    }

    /* Find a empty slot and add device into database. */
    for (i = 0; i < MAX_FIBRE_DEVICES; i++)
        if ((ha->fc_db[i].loop_id == PORT_UNUSED) ||
            (ha->fc_db[i].loop_id == PORT_NEED_MAP))
        {
    	    DEBUG(printk(
                "qla2x00: New slot %d for device "
                "%02x%02x%02x%02x%02x%02x%02x%02x\n",
                i,
         	device->wwn[0],device->wwn[1],
         	device->wwn[2],device->wwn[3],
         	device->wwn[4],device->wwn[5],
         	device->wwn[6],device->wwn[7]);)
            BCOPY(device->name, ha->fc_db[i].name, 8);
            BCOPY(device->wwn, ha->fc_db[i].wwn, 8);
            ha->fc_db[i].loop_id = device->loop_id;
            ha->fc_db[i].d_id.b24 = device->d_id.b24;
            if (device->flag == DEV_PUBLIC)
                ha->fc_db[i].flag |= DEV_PUBLIC;
            ha->flags.updated_fc_db = TRUE;
            return (0);
        }

    if (enable_slot_reuse)
    {
        for (i = 0; i < MAX_FIBRE_DEVICES; i++)
            if (ha->fc_db[i].loop_id == PORT_AVAILABLE)
            {
    	        DEBUG(printk(
                    "qla2x00: Assigned slot %d reuse for device "
                    "%02x%02x%02x%02x%02x%02x%02x%02x\n",
                    i, device->wwn[0],device->wwn[1],
                    device->wwn[2],device->wwn[3],
                    device->wwn[4],device->wwn[5],
                    device->wwn[6],device->wwn[7]);)
                BCOPY(device->name, ha->fc_db[i].name, 8);
                BCOPY(device->wwn, ha->fc_db[i].wwn, 8);
                ha->fc_db[i].loop_id = device->loop_id;
                ha->fc_db[i].d_id.b24 = device->d_id.b24;
                if (device->flag == DEV_PUBLIC)
                    ha->fc_db[i].flag |= DEV_PUBLIC;
                ha->flags.updated_fc_db = TRUE;
                return (0);
            }
    }
    return(BIT_1);
}


/*
 * qla2100_device_resync
 *	Marks devices in the database that needs resynchronization.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Kernel context.
 */
static void
qla2100_device_resync(scsi_qla_host_t *ha)
{
    uint16_t index;
    uint32_t mask;
    rscn_t dev;

#ifdef QL_DEBUG_LEVEL_3
    ENTER("qlc2200_device_resync: entered\n");
#endif

    while (ha->rscn_out_ptr != ha->rscn_in_ptr ||
           ha->flags.rscn_queue_overflow) {
        QLA2100_INTR_LOCK(ha);
	BCOPY((void *)&ha->rscn_queue[ha->rscn_out_ptr], (void *)&dev,
	    sizeof(rscn_t));
    	DEBUG(printk("qla%ld: device_resync: rscn_queue[%d], portID=%06x\n",
		ha->instance,
         	ha->rscn_out_ptr, 
		ha->rscn_queue[ha->rscn_out_ptr].d_id.b24);)
	ha->rscn_out_ptr++;
	if (ha->rscn_out_ptr == MAX_RSCN_COUNT)
	    ha->rscn_out_ptr = 0;
	/* Queue overflow, set switch default case. */
	if (ha->flags.rscn_queue_overflow) {
		DEBUG(printk("device_resync: rscn overflow\n");)
	    dev.format = 3;
	    ha->flags.rscn_queue_overflow = 0;
	}

	switch (dev.format) {
	case 0:
		mask = 0xffffff;
		break;
	case 1:
		mask = 0xffff00;
		break;
	case 2:
		mask = 0xff0000;
		break;
	default:
		mask = 0x0;
		dev.d_id.b24 = 0;
		ha->rscn_out_ptr = ha->rscn_in_ptr;
		break;
	}
        QLA2100_INTR_UNLOCK(ha);

	for (index = 0; index < MAX_FIBRE_DEVICES; index++) {
            if (ha->fc_db[index].flag & DEV_PUBLIC &&
	        (ha->fc_db[index].d_id.b24 & mask) == dev.d_id.b24) {
		if (ha->fc_db[index].loop_id <= LAST_SNS_LOOP_ID) {
	            ha->fc_db[index].loop_id |= PORT_LOST_ID;
    	    DEBUG(printk("qla%ld: RSCN: port lost @ slot %d %06x\n",
		ha->instance,
         	index, 
		ha->fc_db[index].d_id.b24);)
		}
	    }
	}
    }

#ifdef QL_DEBUG_LEVEL_3
    LEAVE("qlc2200_device_resync: exiting normally\n");
#endif
}

/*
 * qla2100_configure_fabric
 *      Setup SNS devices with loop ID's.
 *
 * Input:
 *      ha = adapter block pointer.
 *
 * Returns:
 *      0 = success.
 *      BIT_0 = error
 *      BIT_1 = database was full and device was not configured.
 */
#define MAX_PUBLIC_LOOP_IDS LAST_SNS_LOOP_ID + 1

STATIC uint8_t
qla2100_configure_fabric(scsi_qla_host_t *ha, uint8_t enable_slot_reuse)
{
    uint8_t     rval = 0;
    uint8_t     rval1;
    uint8_t     local_flags = 0;
    sns_data_t  *sns;
    uint8_t     tmp_name[8];
    fcdev_t     dev;
    uint16_t    i, index;
    dma_addr_t   phys_address = 0;
    uint16_t    new_dev_cnt;
    uint16_t    tmp_loop_id;
    uint16_t    tmp_topo;
    struct new_dev new_dev_list[MAX_FIBRE_DEVICES];

#ifdef QL_DEBUG_LEVEL_3
    ENTER("qla2x00_configure_fabric: entered\n");
#endif
    DEBUG(printk("scsi%ld: qla2x00_configure_fabric: hba=%p\n",ha->host_no, ha);)

    /* If FL port exists, then SNS is present */
    rval1 = qla2x00_get_port_name(ha, SNS_FL_PORT, tmp_name, 0);
    if ( rval1 || (tmp_name[0] == 0 && tmp_name[1] == 0 && tmp_name[2] == 0 &&
	tmp_name[3] == 0 && tmp_name[4] == 0 && tmp_name[5] == 0 &&
	tmp_name[6] == 0 && tmp_name[7] == 0) ) 
    {

#ifdef QL_DEBUG_LEVEL_2
        qla2100_print(
              "qla2100_configure_fabric: MBC_GET_PORT_NAME Failed, No FL Port\n");
#endif
	ha->device_flags &= ~SWITCH_FOUND;
        return (0);
    }

	ha->device_flags |= SWITCH_FOUND;

    /* Get adapter port ID. */
    rval = qla2x00_get_adapter_id(ha, &tmp_loop_id, &ha->d_id.b.al_pa,
        &ha->d_id.b.area, &ha->d_id.b.domain, &tmp_topo);


/* 4.10 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
   sns = KMALLOC_DMA(sizeof(sns_data_t));
   if ( sns != NULL ) {
       phys_address = VIRT_TO_BUS(sns);
   }
#else
   sns = pci_alloc_consistent(ha->pdev, sizeof(sns_data_t),
                                           &phys_address);
#endif
    if (sns == NULL)
    {

		printk(KERN_WARNING "qla(%d): Memory Allocation failed - sns.\n",(int)ha->host_no);
		ha->mem_err++;
        return (BIT_0);
    }

    BZERO((caddr_t)sns, sizeof(sns_data_t));

    /* Mark devices that need re-synchronization. */
    qla2100_device_resync(ha);


    do {
	if (qla2x00_register_fc4(ha, sns, phys_address)) {
		/* EMPTY */
		DEBUG2(printk("qla2x00_configure_fabric(%ld): register_fc4 "
		    "failed.\n", ha->host_no);)
	}

	rval = qla2100_find_all_fabric_devs(ha, sns, phys_address, new_dev_list,
	       &new_dev_cnt, &local_flags);

	if (rval != 0)
	    break;

	/*
	 * Logout all previous devices not currently in database
	 * and mark them available.
	 */
	for (index = 0; index < MAX_FIBRE_DEVICES && !ha->loop_down_timer
        && !(ha->dpc_flags & LOOP_RESYNC_NEEDED); index++) {
	    	if (ha->fc_db[index].loop_id & PORT_LOST_ID &&
	    	(ha->fc_db[index].flag & DEV_PUBLIC)) {
			qla2x00_fabric_logout(ha, ha->fc_db[index].loop_id &0xff);
			local_flags |= LOGOUT_PERFORMED;
	    	}
	}

    /*
     * Wait for all remaining IO's to finish if there was logout.
     */
     if (local_flags & LOGOUT_PERFORMED)
     {
	     local_flags &= ~LOGOUT_PERFORMED;

		if (ha->init_done) {
			QLA2100_DPC_LOCK(ha);
			if (!(ha->dpc_flags & COMMAND_WAIT_ACTIVE)) {
				ha->dpc_flags |= COMMAND_WAIT_ACTIVE;
				QLA2100_DPC_UNLOCK(ha);

				qla2100_cmd_wait(ha);

				QLA2100_DPC_LOCK(ha);
				ha->dpc_flags &= ~COMMAND_WAIT_ACTIVE;
			}
			QLA2100_DPC_UNLOCK(ha);
		}
	}

	/*
	 * Scan through our database and login entries already
	 * in our database.
	 */
	for (index = 0; index < MAX_FIBRE_DEVICES && !ha->loop_down_timer
        && !(ha->dpc_flags & LOOP_RESYNC_NEEDED); index++)
        {
	    if (!(ha->fc_db[index].loop_id & PORT_LOGIN_NEEDED))
	        continue;
	    ha->fc_db[index].loop_id &= ~PORT_LOGIN_NEEDED;
	    if (ha->fc_db[index].loop_id <= LAST_SNS_LOOP_ID) {
		/* loop_id reusable */
		dev.loop_id = ha->fc_db[index].loop_id & 0xff;
	    }
            else {
                for (i = ha->min_external_loopid; i < MAX_PUBLIC_LOOP_IDS; i++)
                {
                    if (!ha->fabricid[i].in_use)
                    {
                        ha->fabricid[i].in_use = TRUE;
                        dev.loop_id = i;
                        break;
                    }
                }
	        if (i == MAX_PUBLIC_LOOP_IDS)
		    break;
            }
            dev.d_id.b24 = ha->fc_db[index].d_id.b24;
	    /* login and update database */
	    if (qla2100_fabric_login(ha, &dev) == 0)
	        ha->fc_db[index].loop_id = dev.loop_id;
        }

	/*
	 * Scan through new device list and login and add to our
	 * database.
	*/
	for (index = 0; index < new_dev_cnt && !ha->loop_down_timer
        && !(ha->dpc_flags & LOOP_RESYNC_NEEDED); index++)
        {
	    BCOPY((void *)&new_dev_list[index], (void *)&dev,
	          sizeof(struct new_dev));
            dev.flag = DEV_PUBLIC;
            for (i = ha->min_external_loopid; i < MAX_PUBLIC_LOOP_IDS; i++)
            {
                if (!ha->fabricid[i].in_use)
                {
                    ha->fabricid[i].in_use = TRUE;
                    dev.loop_id = i;
                    break;
                }
            }
	    if (i == MAX_PUBLIC_LOOP_IDS)
		    break;
	    if (qla2100_fabric_login(ha, &dev) == 0) {
			if ( (rval = qla2100_update_fc_database(
				    ha, &dev, enable_slot_reuse)) ) {
		    	qla2x00_fabric_logout(ha, dev.loop_id);
                ha->fabricid[i].in_use = FALSE;
	            break;
            }
        }
	}

    } while(0);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
                KMFREE(sns, sizeof(sns_data_t));
#else
    pci_free_consistent(ha->pdev, sizeof(sns_data_t),
                                    sns, phys_address);
#endif


#ifdef QL_DEBUG_LEVEL_2
	if( rval != 0 ) 
	{
    	qla2100_print("qla2100_configure_fabric: exit, rval = ");
    	qla2100_output_number((uint32_t)rval, 10);
    	qla2100_print("\n");
	}
#endif
    return(rval);
}


/*
 * qla2100_find_all_fabric_devs
 *	Go through GAN list to find all fabric devices.  Will perform
 *	necessary logout of previously existed devices that have changed
 *	and save new devices in a new device list.
 *
 * Input:
 *	ha = adapter block pointer.
 *	dev = database device entry pointer.
 *
 * Returns:
 *	0 = success.
 *	BIT_0 = error.
 *
 * Context:
 *	Kernel context.
 */
static uint8_t
qla2100_find_all_fabric_devs(scsi_qla_host_t *ha, sns_data_t *sns, 
 DMA_ADDR phys_addr, struct new_dev *new_dev_list, uint16_t *new_dev_cnt,
 uint8_t *flags)
{
	fcdev_t		first_dev, dev;
	uint8_t		rval = 0;
	uint16_t	i;
	uint16_t	index;
	uint16_t	new_cnt;
	uint16_t	public_count;
	uint16_t	initiator;

	DEBUG(printk("scsi(%d): qla2100_find_all_fabric_devs entered.\n",
	    (int)ha->host_no);)
	ha->max_public_loop_ids = MAX_PUBLIC_LOOP_IDS;

	/*
	 * Loop getting devices from switch.
	 * Issue GAN to find all devices out there.
	 * Logout the devices that were in our database but changed
	 * port ID.
	 */
     /* Calculate the max number of public ports */
     public_count = ha->max_public_loop_ids - ha->min_external_loopid + 2;

	/* Set start port ID scan at adapter ID. */
	dev.d_id.b24 = 0;
	first_dev.d_id.b24 = 0;

	new_cnt = 0;	/* new device count */

        for (i = 0; i < public_count && !ha->loop_down_timer &&
		    !(ha->dpc_flags & LOOP_RESYNC_NEEDED); i++) {
	        /* Send GAN to the switch */
                rval = 0;
	        if (qla2100_gan(ha, sns, phys_addr, &dev)) {
	            rval = rval | BIT_0;
	            break;
	        }

	        /* If wrap on switch device list, exit. */
	        if (dev.d_id.b24 == first_dev.d_id.b24)
		     break;

		DEBUG(printk("qla%ld: found fabric(%d) - port Id=%06x\n", 
		    ha->instance, i, dev.d_id.b24);)

		if (first_dev.d_id.b24 == 0)
			first_dev.d_id.b24 = dev.d_id.b24;

		/* If port type not equal to N or NL port, skip it. */
		if (sns->p.rsp[16] != 1 && sns->p.rsp[16] != 2) {
			continue;		/* needed for McData switch */
		}

		/* Bypass if host adapter. */
		if (dev.d_id.b24 == ha->d_id.b24)
			continue;

		/* Bypass reserved domain fields. */
		if ((dev.d_id.b.domain & 0xf0) == 0xf0)
			continue;

		/* Bypass if same domain and area of adapter. */
		if ((dev.d_id.b24 & 0xffff00) == (ha->d_id.b24 & 0xffff00))
			continue;

		/* Bypass if initiator */
		initiator = FALSE;
		for (index = 0; index < ha->next_host_slot; index++) {
#if USE_PORTNAME
			if (BCMP(dev.wwn, ha->host_db[index].wwn, 8) == 0)
#else
			if (BCMP(dev.name, ha->host_db[index].name, 8) == 0)
#endif
			{
				initiator = TRUE;
				DEBUG(printk("qla%ld: host entry - nodename "
				    "%02x%02x%02x%02x%02x%02x%02x%02x "
				    "port Id=%06x\n", 
				    ha->instance,
				    dev.name[0], dev.name[1],
				    dev.name[2], dev.name[3],
				    dev.name[4], dev.name[5],
				    dev.name[6], dev.name[7], dev.d_id.b24);)
				break;
			}
		}

		if (initiator)
			continue;

	    /* Locate matching device in database. */
	    for (index = 0; index < MAX_FIBRE_DEVICES; index++) {
#if USE_PORTNAME
                if (BCMP(dev.wwn, ha->fc_db[index].wwn, 8) == 0)
#else
                if (BCMP(dev.name, ha->fc_db[index].name, 8) == 0)
#endif
		{
		    DEBUG(printk("qla%ld: found fabric dev %d in tgt %d db, "
		      "flags= 0x%x, loop_id=0x%04x, port=%06x, name=%02x"
		      "%02x%02x%02x%02x%02x%02x%02x\n", 
		      ha->instance,
		      i, index,
                      ha->fc_db[index].flag,
		      ha->fc_db[index].loop_id,
		      ha->fc_db[index].d_id.b24,
		      dev.wwn[0], dev.wwn[1],
		      dev.wwn[2], dev.wwn[3],
		      dev.wwn[4], dev.wwn[5],
		      dev.wwn[6], dev.wwn[7]);)

		    if (!(ha->fc_db[index].flag & DEV_PUBLIC)) {
		       /*
		        * This was in our database as a local
	            * device. Here we assume this device
		        * either has changed location so
		        * configure_local_loop has already
		        * done necessary clean up, or it's
		        * saved here due to persistent name
		        * binding. We'll just add it in
		        * as a fabric device.
		       */
		       /* Copy port id and name fields. */
                       ha->fc_db[index].flag |= DEV_PUBLIC;
                       ha->fc_db[index].d_id.b24 = dev.d_id.b24;
		       ha->fc_db[index].loop_id |= PORT_LOGIN_NEEDED;
		       break;
		   }
	   	   /* This was in our database as a fabric device. */
		   if ((ha->fc_db[index].d_id.b24 == dev.d_id.b24) &&
		   (ha->fc_db[index].loop_id <= LAST_SNS_LOOP_ID))
		   /* Device didn't change */
		       break;

		   if (ha->fc_db[index].loop_id == PORT_AVAILABLE) {
                        ha->fc_db[index].flag |= DEV_PUBLIC;
                        ha->fc_db[index].d_id.b24 = dev.d_id.b24;
			ha->fc_db[index].loop_id |= PORT_LOGIN_NEEDED;
			break;
		   }

		   /*
		    * Port ID changed or device was
		    * marked to be updated; logout and
		    * mark it for relogin later.
		    */
		   qla2x00_fabric_logout(ha,ha->fc_db[index].loop_id & 0xff);
                   ha->fc_db[index].flag |= DEV_PUBLIC;
                   ha->fc_db[index].d_id.b24 = dev.d_id.b24;
		   ha->fc_db[index].loop_id |= PORT_LOGIN_NEEDED;
		   ha->fc_db[index].loop_id &= ~PORT_LOST_ID;
		   *flags |= LOGOUT_PERFORMED;
		   break;
	        }
	    }

            if (index == MAX_FIBRE_DEVICES) {
		/*
		 * Did not find a match in our database.
		 * This is a new device.
		 */
		DEBUG3(printk("find_all: new device %02x%02x%02x%02x%02x%02x%02x%02x.\n",
		    dev.wwn[0], dev.wwn[1], dev.wwn[2], dev.wwn[3],
		    dev.wwn[4], dev.wwn[5], dev.wwn[6], dev.wwn[7]);)

		BCOPY((void *)&dev, (void *)&new_dev_list[new_cnt],
		    sizeof(struct new_dev));
		new_cnt++;
	    }
	}

	*new_dev_cnt = new_cnt;

    if( new_cnt >  0)
        ha->device_flags |= DFLG_FABRIC_DEVICES;

#ifdef QL_DEBUG_LEVEL_3
        qla2100_print("qla2100_find_all_fabric_devs: exit, rval = ");
        qla2100_output_number((uint32_t)rval, 10);
        qla2100_print("\n");
#endif
	return (rval);
}

/*
 * qla2x00_register_fc4
 *	Register adapter as FC4 device to the switch, so the switch won't
 *	need to login to us later which generates an RSCN event.
 *
 * Input:
 *	ha = adapter block pointer.
 *	sns = pointer to buffer for sns command.
 *	phys_addr = DMA buffer address.
 *
 * Context:
 *	Kernel context.
 */
static uint8_t
qla2x00_register_fc4(scsi_qla_host_t *ha, sns_data_t *sns, DMA_ADDR phys_addr)
{
	uint8_t rval;

	DEBUG3(printk("qla2x00_register_fc4: entered.\n");)

	/* Get port ID for device on SNS. */
	BZERO((caddr_t)sns, sizeof(sns_data_t));
	sns->p.req.hdr.buffer_length = 8;
	sns->p.req.hdr.buffer_address[0] = LS_64BITS(phys_addr);
	sns->p.req.hdr.buffer_address[1] = MS_64BITS(phys_addr);
	sns->p.req.hdr.subcommand_length = 22;
	sns->p.req.subcommand = 0x217;
	sns->p.req.param[6] = ha->d_id.b.al_pa;
	sns->p.req.param[7] = ha->d_id.b.area;
	sns->p.req.param[8] = ha->d_id.b.domain;
	sns->p.req.param[11] = 1;	/* SCSI - FCP */

	rval = BIT_0;
	if (!qla2x00_send_sns(ha, phys_addr, 14, sizeof(sns_data_t))) {
		if (sns->p.rsp[8] == 0x80 && sns->p.rsp[9] == 0x2) {

			DEBUG3(printk("qla2x00_register_fc4(%ld): exiting "
			    "normally.\n", ha->host_no);)
			rval = 0;

		}
	}

	if (rval != 0) {
		/* EMPTY */
		DEBUG2_3(printk("qla2x00_register_fc4(%ld): failed.\n",
		    ha->host_no);)
	}

	return (rval);
}

/*
 * qla2100_gan
 *	Issue Get All Next (GAN) Simple Name Server (SNS) command.
 *
 * Input:
 *	ha = adapter block pointer.
 *	sns = pointer to buffer for sns command.
 *	dev = FC device type pointer.
 *
 * Returns:
 *	qla2100 local function return status code.
 *
 * Context:
 *	Kernel context.
 */
static uint8_t
qla2100_gan(scsi_qla_host_t *ha, sns_data_t *sns, DMA_ADDR phys_addr, 
    fcdev_t *dev)
{
	uint8_t rval;

#ifdef QL_DEBUG_LEVEL_3
	ENTER("qla2100_gan: entered\n");
#endif
	/* Get port ID for device on SNS. */
	BZERO((caddr_t)sns, sizeof(sns_data_t));
	sns->p.req.hdr.buffer_length = 304;
	sns->p.req.hdr.buffer_address[0] = LS_64BITS(phys_addr);
	sns->p.req.hdr.buffer_address[1] = MS_64BITS(phys_addr);
	sns->p.req.hdr.subcommand_length = 6;
	sns->p.req.subcommand = 0x100;				/* GA_NXT */
	sns->p.req.param[6] = dev->d_id.b.al_pa;
	sns->p.req.param[7] = dev->d_id.b.area;
	sns->p.req.param[8] = dev->d_id.b.domain;

	rval = BIT_0;
	if (!qla2x00_send_sns(ha, phys_addr, 14, sizeof(sns_data_t))) {
		if (sns->p.rsp[8] == 0x80 && sns->p.rsp[9] == 0x2) {
			dev->d_id.b.al_pa = sns->p.rsp[19];
			dev->d_id.b.area = sns->p.rsp[18];
			dev->d_id.b.domain = sns->p.rsp[17];
			dev->flag = DEV_PUBLIC;

			/* Save FC name */
			BCOPY(&sns->p.rsp[284], dev->name, 8);

			/* Extract portname */
			BCOPY(&sns->p.rsp[20], dev->wwn, 8);

			DEBUG3(printk("qla2x00: gan entry - portname "
			    "%02x%02x%02x%02x%02x%02x%02x%02x port Id=%06x\n", 
			    sns->p.rsp[20], sns->p.rsp[21],
			    sns->p.rsp[22], sns->p.rsp[23],
			    sns->p.rsp[24], sns->p.rsp[25],
			    sns->p.rsp[26], sns->p.rsp[27], dev->d_id.b24);)
			rval = 0;
		}
	}

#ifdef QL_DEBUG_LEVEL_2
	if( rval != 0 )
		printk("qla2x00_gan: exit, rval = %d\n",rval);
#endif
	return (rval);
}


/*
 * qla2100_fabric_login
 *	Issue fabric login command.
 *
 * Input:
 *	ha = adapter block pointer.
 *	device = pointer to FC device type structure.
 *
 * Returns:
 *      0 - Login successfully
 *      1 - Login failed
 *      2 - Initiator device
 */
static uint8_t
qla2100_fabric_login(scsi_qla_host_t *ha, fcdev_t *device)
{
	uint16_t	status[3];

	for (;;) {
		DEBUG(printk( "Trying Fabric Login w/loop id= 0x%04x for "
		    " port %06x\n", device->loop_id, device->d_id.b24);)

		/* Login device on switch. */
		qla2x00_login_fabric(ha, device->loop_id, device->d_id.b.domain,
		    device->d_id.b.area, device->d_id.b.al_pa, &status[0],
		    BIT_0);

		if (status[0] == 0x4007) {
			ha->fabricid[device->loop_id].in_use = FALSE;
			device->loop_id = status[1];

			DEBUG(printk(
			    "Fabric Login: port in use - next loop id="
			    "0x%04x, port Id=%06x\n", 
			    device->loop_id, device->d_id.b24);)

			if (device->loop_id <= LAST_SNS_LOOP_ID)
				ha->fabricid[device->loop_id].in_use = TRUE;
			else
				return 1;

		} else if (status[0] == 0x4000) {
			if (status[1] & 0x0001) {
				/* It's initiator only. Save node name and
				 * ignore it.
				 */
				if (ha->next_host_slot < 8) {
					BCOPY(device->wwn, ha->host_db
					    [ha->next_host_slot].wwn, 8);
					BCOPY(device->name, ha->host_db
					    [ha->next_host_slot].name, 8);
					ha->host_db[ha->next_host_slot].
					    d_id.b24 = device->d_id.b24;
					ha->next_host_slot++;
				} else {
					printk(KERN_INFO
					    "qla%ld: Host table full.\n",
					    ha->instance);
				}
				return 2;
			}

			/* This is target capable device */
			qla2x00_get_port_database(ha, device, 0);
			DEBUG(printk("Fabric Login OK @ loop id=0x%04x, "
			    "port Id=%06x\n", device->loop_id,
			    device->d_id.b24);)
			return 0;

		} else if (status[0] == 0x4008) {
			if (device->loop_id++ <= LAST_SNS_LOOP_ID)
				ha->fabricid[device->loop_id].in_use = TRUE;
			else
				return 1;
		} else {
			return 1;
		}
	}
}


/*
 * qla2100_configure_loop
 *      Updates Fibre Channel Device Database with what is actually on loop.
 *
 * Input:
 *      ha                = adapter block pointer.
 *
 * Output:
 *      ha->fc_db = updated
 *
 * Returns:
 *      0 = success.
 *      1 = error.
 *      2 = database was full and device was not configured.
 */
STATIC uint8_t
qla2100_configure_loop(scsi_qla_host_t *ha)
{
	uint8_t  rval = 0;
	uint8_t  rval1 = 0;
	uint8_t  enable_slot_reuse = FALSE;
	uint16_t  cnt;
	uint32_t  flags;
#if MPIO_SUPPORT
	unsigned long cpu_flags = 0;
#endif

	DEBUG3(printk("qla2x00_configure_loop(%ld): entered\n", ha->host_no);)
	DEBUG(printk("scsi %ld: qla2100_configure_loop:\n",ha->host_no);)

	/* Get Initiator ID */
	if(  qla2x00_configure_hba(ha) ) {
		return(1);
	}

	QLA2100_DPC_LOCK(ha);
	flags = ha->dpc_flags;
	DEBUG(printk("qla2100_configure_loop: dpc flags =0x%x\n",flags);)
	ha->dpc_flags &= ~(LOCAL_LOOP_UPDATE | RSCN_UPDATE);
	QLA2100_DPC_UNLOCK(ha);

	ha->mem_err = 0 ; 

	/* Determine what we need to do */
	if (ha->current_topology == ISP_CFG_FL && 
	    (flags & LOCAL_LOOP_UPDATE)) {

		ha->flags.rscn_queue_overflow = TRUE; 
		flags = flags | RSCN_UPDATE;

	} else if (ha->current_topology == ISP_CFG_F && 
	    (flags & LOCAL_LOOP_UPDATE)) {

		ha->flags.rscn_queue_overflow = TRUE; 
		flags = flags | RSCN_UPDATE;
		flags = flags & ~LOCAL_LOOP_UPDATE;

	} else if (!ha->flags.online || 
	    (ha->dpc_flags & ABORT_ISP_ACTIVE)) {

		ha->flags.rscn_queue_overflow = TRUE;
		flags = flags | LOCAL_LOOP_UPDATE | RSCN_UPDATE;
	}

	do {
		if (flags & LOCAL_LOOP_UPDATE)
			rval = rval | qla2100_configure_local_loop(ha,
			    enable_slot_reuse);

		/* if (!(rval & BIT_0) && flags & RSCN_UPDATE) */
		if (flags & RSCN_UPDATE) {
			rval1 = qla2100_configure_fabric(ha, enable_slot_reuse);
		 	if ( (rval1 & BIT_0) && ha->sns_retry_cnt < 8 ) {
		 		ha->sns_retry_cnt++;
				ha->dpc_flags = ha->dpc_flags | flags;
			   	ha->device_flags |= LOGIN_RETRY_NEEDED;
		 	}
		}

		/* If devices not configured first time try reusing slots.*/
		if (enable_slot_reuse == FALSE && (rval & BIT_1) )
			enable_slot_reuse = TRUE;
		else
			enable_slot_reuse = FALSE;

		/* Isolate error status. */
		if (rval & BIT_0) {
			rval = 1;
		} else {
			rval = 0;
		}

	} while (enable_slot_reuse == TRUE && rval == 0);

	if (!ha->loop_down_timer && !(ha->dpc_flags & LOOP_RESYNC_NEEDED)) {

		/* Mark devices that are not present as DEV_ABSENCE */
		for (cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++) {
			if (ha->fc_db[cnt].loop_id & PORT_LOST_ID) {
				ha->fc_db[cnt].flag |= DEV_ABSENCE;
			} else {
				/* device returned */
				if (ha->fc_db[cnt].loop_id <= LAST_SNS_LOOP_ID
				    && ha->fc_db[cnt].flag & DEV_ABSENCE) {

					ha->fc_db[cnt].flag &= ~DEV_ABSENCE;
					ha->fc_db[cnt].flag |= DEV_RETURN;
					ha->fc_db[cnt].port_login_retry_count=8;
				}
			}

			if ( (flags & RSCN_UPDATE) && 0 && 	/* 5.21b7 */
			    (ha->fc_db[cnt].flag & DEV_ABSENCE) ) {
			/*
			 * This dev was not detected but its WWN
			 * is valid. To handle the case where
			 * the switch may not be giving us the
			 * device list correctly, schedule for
			 * a login retry later if not previously
			 * done so.
			 */
			if( ha->fc_db[cnt].port_login_retry_count ) {
				DEBUG(printk("qla2100_sns:Port login retry - "
				    "slot=%d, loop id 0x%x, count=%d\n",
				    cnt, ha->fc_db[cnt].loop_id,
				    ha->fc_db[cnt].port_login_retry_count); )
				ha->fc_db[cnt].port_login_retry_count--;
				ha->dpc_flags = ha->dpc_flags | flags;
				ha->device_flags |= LOGIN_RETRY_NEEDED;
			}

			}
		}

		rval1 = qla2x00_build_fcport_list(ha); 
		if (((rval1 & BIT_0) || ha->mem_err != 0)
		    && ha->sns_retry_cnt < 8 ) {
			ha->sns_retry_cnt++;
			ha->dpc_flags = ha->dpc_flags | flags;
			ha->device_flags |= LOGIN_RETRY_NEEDED;
		}

		if(!ha->flags.failover_enabled)
			qla2x00_config_os(ha);

		/* If we found all devices then go ready */
		if (!(ha->device_flags & LOGIN_RETRY_NEEDED)) {
			ha->loop_state = LOOP_READY;

#if	MPIO_SUPPORT
			if ( qla2100_failover ){
				DEBUG(printk("qla2100_configure_loop(%ld): "

				    "schedule FAILBACK EVENT\n", ha->host_no);)
                                QLA_FAILBACK_LOCK(ha, cpu_flags);
				ha->failback_flags |= FAILOVER_EVENT_NEEDED;
				ha->failback_delay = failbackTime;
                                QLA_FAILBACK_UNLOCK(ha, cpu_flags);
                                ha->dpc_flags |= COMMAND_WAIT_NEEDED;
				ha->failover_type = MP_NOTIFY_LOOP_UP;
			}
#endif

			for( cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++ )
				ha->fc_db[cnt].port_login_retry_count = 8;

			DEBUG(printk("qla2100_configure_loop(%ld): "
			    "LOOP READY\n", ha->host_no);)
		}
	} else {
		rval = 1;
	}

	if (rval) {
		DEBUG2_3(printk("qla2100_configure_loop(%ld): *** FAILED ***\n",
		    ha->host_no);)
	} else {
		DEBUG3(printk("qla2100_configure_loop: exiting normally\n");)
	}

	return(rval);
}


/*
 * qla2x00_config_os
 *	Setup OS target and LUN structures.
 *
 * Input:
 *	ha = adapter state pointer.
 *
 * Context:
 *	Kernel context.
 */
static void
qla2x00_config_os(scsi_qla_host_t *ha)
{
	fc_port_t	*fcport;
	fc_lun_t	*fclun;
	os_lun_t	*lq;
	uint16_t	t, l;


	DEBUG(printk("qla2x00_config_os(%ld): entered.\n",
	    ha->host_no);)

	for (fcport = ha->fcport; fcport != NULL; fcport = fcport->next) {
		/* Allocate target */
		if (fcport->loop_id == FC_NO_LOOP_ID)
			continue;

		/* Bind fcport to target number. */
		DEBUG(printk("qla2x00_config_os(%ld): fcport bind= %p\n",
		    ha->host_no,fcport);)
		if ((t = qla2x00_fcport_bind(ha, fcport)) == MAX_TARGETS)
			continue;

#if VSA
		if( (ha->fc_db[t].flag & DEV_FLAG_VSA) )
			fcport->flags |= FC_VSA;
#endif

		DEBUG5(printk("qla2x00_config_os(%ld): going to alloc lun for "
		    "tgt %d. mask=%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"
		    "%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x"
		    "%02x%02x%02x%02x%02x%02x%02x%02x%02x.\n",
		    ha->host_no, t, fcport->lun_mask.mask[0],
		    fcport->lun_mask.mask[1], fcport->lun_mask.mask[2],
		    fcport->lun_mask.mask[3], fcport->lun_mask.mask[4],
		    fcport->lun_mask.mask[5], fcport->lun_mask.mask[6],
		    fcport->lun_mask.mask[7], fcport->lun_mask.mask[8],
		    fcport->lun_mask.mask[9], fcport->lun_mask.mask[10],
		    fcport->lun_mask.mask[11], fcport->lun_mask.mask[12],
		    fcport->lun_mask.mask[13], fcport->lun_mask.mask[14],
		    fcport->lun_mask.mask[15], fcport->lun_mask.mask[16],
		    fcport->lun_mask.mask[17], fcport->lun_mask.mask[18],
		    fcport->lun_mask.mask[19], fcport->lun_mask.mask[20],
		    fcport->lun_mask.mask[21], fcport->lun_mask.mask[22],
		    fcport->lun_mask.mask[23], fcport->lun_mask.mask[24],
		    fcport->lun_mask.mask[25], fcport->lun_mask.mask[26],
		    fcport->lun_mask.mask[27], fcport->lun_mask.mask[28],
		    fcport->lun_mask.mask[29], fcport->lun_mask.mask[30],
		    fcport->lun_mask.mask[31]);)

		/* Allocate LUNs */
		for (fclun = fcport->fclun; fclun != NULL;
		    fclun = fclun->next) {

			l = fclun->lun;		/* Must not exceed MAX_LUN */

			/* Always alloc LUN 0 so kernel will scan
			 * past LUN 0.
			 */
			if (l != 0 &&
			    (EXT_IS_LUN_BIT_SET(&(fcport->lun_mask), l))) {
				/* mask this LUN */
				continue;
			}

			if ((lq = qla2x00_lun_alloc(ha, t, l)) == NULL)
				continue;
			lq->fclun = fclun;
		}
	}

	DEBUG(printk("qla2x00_config_os(%ld): exiting normally.\n",
	    ha->host_no);)
}

/*
 * qla2x00_fcport_bind
 *	Locates a target number for FC port.
 *
 * Input:
 *	ha = adapter state pointer.
 *	fcport = FC port structure pointer.
 *
 * Returns:
 *	target number
 *
 * Context:
 *	Kernel context.
 */
static uint16_t
qla2x00_fcport_bind(scsi_qla_host_t *ha, fc_port_t *fcport)
{
	uint8_t		*name;
	uint16_t	t;
	os_tgt_t	*tq;


	ENTER("qla2x00_fcport_bind");
	DEBUG(printk("qla2x00_fcport_bind\n");)
	name = &fcport->port_name[0];

	/* Check for tgt already allocated for persistent binding. */
	for (t = 0; t < MAX_TARGETS; t++) {
		if ((tq = TGT_Q(ha, t)) == NULL)
			continue;

		if (memcmp(name, tq->fc_name, WWN_SIZE) == 0) {
			tq->vis_port = fcport;
			break;
		}
	}

	if (t != MAX_TARGETS) {
		return (t);
	}

	/* Check for persistent binding not yet configured. */
	for (t = 0; t < MAX_TARGETS; t++) {
#if USE_PORTNAME
		if (memcmp(name, ha->fc_db[t].wwn, WWN_SIZE) == 0)
#else
		if (memcmp(name, ha->fc_db[t].name, WWN_SIZE) == 0)
#endif
		{
			/* found not-yet-allocated target at t */
			break;
		}
	}

	if (t == MAX_TARGETS) {
		/* Check if slot at loop ID is available. */
		t = fcport->loop_id;
		if (TGT_Q(ha, t) != NULL) {
			/* Locate first free target id in db for device. */
			for (t = 0; t < MAX_TARGETS; t++) {
				if (TGT_Q(ha, t) == NULL)
					break;
			}
		}
	}

	if (t != MAX_TARGETS && qla2x00_tgt_alloc(ha, t, name) == NULL)
		t = MAX_TARGETS;

	if (t == MAX_TARGETS) {
		/*EMPTY*/
		printk(KERN_WARNING "qla2x00_fcport_bind: **** FAILED ****");
	} else {
		if (!ha->flags.failover_enabled) {
			/* fcport IS the visible port in non-failover mode */
			tq = TGT_Q(ha, t);
			tq->vis_port = fcport;
		}

		DEBUG(tq = TGT_Q(ha, t);)
                DEBUG(printk("Assigning target ID %02x @ %p to loop id: %04x\n",
		    t,tq,fcport->loop_id);)

		LEAVE("qla2x00_fcport_bind: exiting normally");
	}

	return (t);
}

/*
 * qla2x00_build_fcport_list
 *	Updates device on list.
 *
 * Input:
 *	ha = adapter block pointer.
 *	fcport = port structure pointer.
 *
 * Return:
 *	0  - Success
 *  BIT_0 - error
 *
 * Context:
 *	Kernel context.
 */
static int qla2x00_build_fcport_list(scsi_qla_host_t *ha)
{
	fcdev_t		*dev;
	int	found = 0;
	int cnt, i;
	fc_port_t	*fcport;
	fc_port_t	*prev_fcport;

	ENTER("qla2x00_build_fcport_list");
	for (cnt = 0; cnt < MAX_FIBRE_DEVICES; cnt++) {
		dev = &ha->fc_db[cnt];

		/* Skip if zero port name */
		if ( qla2x00_is_wwn_zero( (uint8_t *) &dev->wwn[0]) ) {
			continue;
		}

		DEBUG2(printk("qla2x00_build_fcport_list(%ld): found tgt %d "
		    "in fc_db, wwn=%02x%02x%02x%02x%02x%02x%02x%02x.\n", ha->host_no, cnt,
                    dev->wwn[0],dev->wwn[1],dev->wwn[2],dev->wwn[3],
                    dev->wwn[4],dev->wwn[5],dev->wwn[6],dev->wwn[7]);)

		/* Check for matching device in port list. */
		found = 0;
		prev_fcport = NULL;
		for (i=0, fcport = ha->fcport; fcport != NULL;
		    fcport = fcport->next, i++) {

			if (BCMP(dev->wwn, fcport->port_name, WWN_SIZE) == 0 &&
			    BCMP(dev->name, fcport->node_name, WWN_SIZE) == 0) {

				DEBUG(printk("qla2x00_build_fcport_list: Found "
				    "matching port %06x, device flags= 0x%x\n",
				    dev->d_id.b24, dev->flag);)

				/* if device found is missing then mark it */
				if (dev->flag & DEV_ABSENCE) {
					DEBUG(printk(
					    "qla2100_buildfcportlist: Port "
					    "missing ---  (port_name) -> "
					    "%02x%02x%02x%02x%02x%02x%02x%02x, "
					    "loop id = 0x%04x\n",
					    fcport->port_name[0],
					    fcport->port_name[1],
					    fcport->port_name[2],
					    fcport->port_name[3],
					    fcport->port_name[4],
					    fcport->port_name[5],
					    fcport->port_name[6],
					    fcport->port_name[7],
					    fcport->loop_id);)
					fcport->loop_id = FC_NO_LOOP_ID;
					found++;
					break;
				}
			
				/* if device was missing but returned */
				if( fcport->loop_id == FC_NO_LOOP_ID ||
	        			!(dev->flag & DEV_PUBLIC) ||
				       fcport->state != FC_ONLINE  ) {
					DEBUG(printk(
					    "qla2100_buildfcportlist: Port "
					    "returned +++  (port_name) -> "
					    "%02x%02x%02x%02x%02x%02x%02x%02x, "
					    "loop id = 0x%04x\n",
					    fcport->port_name[0],
					    fcport->port_name[1],
					    fcport->port_name[2],
					    fcport->port_name[3],
					    fcport->port_name[4],
					    fcport->port_name[5],
					    fcport->port_name[6],
					    fcport->port_name[7],
					    fcport->loop_id);)
					fcport->loop_id = dev->loop_id;
					PORT_DOWN_RETRY(fcport) = 0;
					break;
				}

				DEBUG(printk(
				    "qla2100_buildfcportlist: Match -"
				    "fcport[%d] = fc_db[%d] (ignored) -> "
				    "%02x%02x%02x%02x%02x%02x%02x%02x, "
				    "loop id = 0x%04x\n",
				    i, cnt,
				    fcport->port_name[0],
				    fcport->port_name[1],
				    fcport->port_name[2],
				    fcport->port_name[3],
				    fcport->port_name[4],
				    fcport->port_name[5],
				    fcport->port_name[6],
				    fcport->port_name[7],
				    fcport->loop_id);)
				found++;
				break;
			}

			prev_fcport = fcport;
		}
		if ( found )
			continue;

		/* Add device to port list. */
		if (fcport == NULL) {
			DEBUG3(printk("qla2x00_build_fcport_list(%ld): adding "
			    "new device to list.\n", ha->host_no);)

			fcport = (fc_port_t *)KMALLOC(sizeof (fc_port_t));

			if (fcport == NULL)
				break;

			BZERO((caddr_t)fcport, sizeof(fc_port_t));

			/* copy fields into fcport */
			BCOPY(&dev->wwn[0], &fcport->port_name[0], WWN_SIZE);
			BCOPY(&dev->name[0], &fcport->node_name[0], WWN_SIZE);

			if (dev->flag & DEV_ABSENCE) {
				DEBUG(printk(
				    "qla2100_buildfcportlist: Port missing --- "
				    "(port_name) -> %02x%02x%02x%02x%02x%02x"
				    "%02x%02x, loop id = 0x%04x\n",
				    fcport->port_name[0],
				    fcport->port_name[1],
				    fcport->port_name[2],
				    fcport->port_name[3],
				    fcport->port_name[4],
				    fcport->port_name[5],
				    fcport->port_name[6],
				    fcport->port_name[7],
				    fcport->loop_id);)
				fcport->loop_id = FC_NO_LOOP_ID;
			} else {
				fcport->loop_id = dev->loop_id;
			}

			fcport->d_id = dev->d_id;

			DEBUG(printk(
			    "qla2100_buildfcportlist: New Device +++ "
			    "(port_name) -> %02x%02x%02x%02x%02x%02x%02x%02x, "
			    "loop id = 0x%04x\n",
			    fcport->port_name[0],
			    fcport->port_name[1],
			    fcport->port_name[2],
			    fcport->port_name[3],
			    fcport->port_name[4],
			    fcport->port_name[5],
			    fcport->port_name[6],
			    fcport->port_name[7],
			    fcport->loop_id);)

			/* flags */
			if ( dev->flag & DEV_PUBLIC )
				fcport->flags |= FC_FABRIC_DEVICE;

			if (dev->flag & DEV_INITIATOR)
				fcport->flags |= FC_INITIATOR_DEVICE;

			if (!qla2100_failover)
				qla2100_get_lun_mask_from_config(ha, fcport,
				    cnt, 0);

			if (prev_fcport == NULL) {
				/* nothing in fcport list yet */
				ha->fcport = fcport;
			} else {
				/*
				 * prev_fcport should be pointing to last
				 * port in list
				 */
				prev_fcport->next = fcport;
			}

		} else {
			DEBUG3(printk("qla2x00_build_fcport_list(%ld): "
			    "updating device to list.\n", ha->host_no);)
			fcport->loop_id = dev->loop_id;
		}

		if( qla2x00_update_fcport(ha, fcport, cnt) )
		{
			DEBUG2(printk("qla2x00_build_fcport_list(%ld): "
			    "update_fcport failed.\n", ha->host_no);)
			return BIT_0;
		}

		if (dev->flag & DEV_ABSENCE)
			fcport->state = FC_DEVICE_LOST;
	}

	LEAVE("qla2x00_build_fcport_list: exiting normally");
	return (0);
}


/*
 * qla2x00_update_fcport
 *	Updates device on list.
 *
 * Input:
 *	ha = adapter block pointer.
 *	fcport = port structure pointer.
 *
 * Return:
 *	0  - Success
 *  BIT_0 - error
 *
 * Context:
 *	Kernel context.
 */
static int 
qla2x00_update_fcport(scsi_qla_host_t *ha, fc_port_t *fcport, int index)
{
	ENTER("qla2x00_update_fcport");
	DEBUG4(printk("qla2x00_update_fcport: entered, loop_id = %d ",
	    fcport->loop_id);)

	fcport->port_login_retry_count = 8;
	fcport->state = FC_ONLINE;
	fcport->ha = ha;

	/* Do LUN discovery. */
	return ( qla2x00_lun_discovery(ha, fcport, index) );
}

/*
 * qla2x00_lun_discovery
 *	Issue SCSI inquiry command for LUN discovery.
 *
 * Input:
 *	ha = adapter block pointer.
 *	fcport = FC port structure pointer.
 *
 * Return:
 *	0  - Success
 *  BIT_0 - error
 *
 * Context:
 *	Kernel context.
 */
static int
qla2x00_lun_discovery(scsi_qla_host_t *ha, fc_port_t *fcport, int index)
{
	inq_cmd_rsp_t	*pkt;
	int		rval;
	uint16_t	lun;
	fc_lun_t	*fclun;
	dma_addr_t	phys_address = 0;
	int		disconnected;
	int		retry;
	fcdev_t		dev;

	ENTER("qla2x00_lun_discovery:");

/* 4.10 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	pkt = (inq_cmd_rsp_t *)KMALLOC_DMA(sizeof (inq_cmd_rsp_t));
	if (pkt != NULL) {
		phys_address = VIRT_TO_BUS(pkt);
	}
#else
	pkt = pci_alloc_consistent(ha->pdev, sizeof(inq_cmd_rsp_t),
	    &phys_address);
#endif

	if (pkt == NULL) {
		printk(KERN_WARNING "scsi(%d): Memory Allocation failed - INQ",(int)ha->host_no);
		ha->mem_err++;
		return BIT_0;
	}

	for (lun = 0; lun < MAX_FIBRE_LUNS; lun++) {
		retry = 2;
		do {
			BZERO((caddr_t)pkt, sizeof (inq_cmd_rsp_t));
			pkt->p.cmd.entry_type = COMMAND_TYPE;
			pkt->p.cmd.entry_count = 1;
			pkt->p.cmd.lun = lun;
			pkt->p.cmd.target = (uint8_t)fcport->loop_id;
			pkt->p.cmd.control_flags = CF_READ | CF_SIMPLE_TAG;
			pkt->p.cmd.scsi_cdb[0] = 0x12;
			pkt->p.cmd.scsi_cdb[4] = INQ_DATA_SIZE;
			pkt->p.cmd.dseg_count = 1;
			pkt->p.cmd.timeout = 10;

			pkt->p.cmd.byte_count = INQ_DATA_SIZE;
			LITTLE_ENDIAN_32(pkt->p.cmd.byte_count);

			pkt->p.cmd.dseg_0_address = (uint32_t)
			    (phys_address + sizeof (sts_entry_t));
			LITTLE_ENDIAN_32(pkt->p.cmd.dseg_0_address);

			pkt->p.cmd.dseg_0_length = INQ_DATA_SIZE;
			LITTLE_ENDIAN_32(pkt->p.cmd.dseg_0_length);

			DEBUG(printk("lun_discovery: Lun Inquiry - fcport= %p, "
			    "lun (%d)\n", fcport, lun);)

			rval = qla2x00_issue_iocb(ha, (caddr_t)pkt,
			    sizeof (inq_cmd_rsp_t));

			DEBUG4(printk("lun_discovery: lun (%d) "
			    "inquiry - inq[0]= 0x%x, comp status 0x%x, "
			    "scsi status 0x%x, rval=%d\n",
			    lun, pkt->inq[0], pkt->p.rsp.comp_status,
			    pkt->p.rsp.scsi_status, rval);)

			/* if port not logged in then try and login */
			if ( lun == 0 && 
			    pkt->p.rsp.comp_status == CS_PORT_LOGGED_OUT ) {
				BZERO((caddr_t)&dev, sizeof (dev));
				dev.d_id.b24 = ha->fc_db[index].d_id.b24;

				/* login and update database */
				if (qla2100_fabric_login(ha, &dev) == 0)
					ha->fc_db[index].loop_id = dev.loop_id;
			}
		} while ( (rval != QLA2X00_SUCCESS ||
		    pkt->p.rsp.comp_status != CS_COMPLETE) && retry--); 

		if (rval != QLA2X00_SUCCESS ||
		    pkt->p.rsp.comp_status != CS_COMPLETE ||
		    (pkt->p.rsp.scsi_status & SS_CHECK_CONDITION )) {
			DEBUG(printk("lun_discovery: Failed lun "
			    "inquiry - inq[0]= 0x%x, comp status 0x%x, "
			    "scsi status 0x%x. loop_id=%d\n",
			    pkt->inq[0], pkt->p.rsp.comp_status,
			    pkt->p.rsp.scsi_status, fcport->loop_id);)
			break;
		}

		disconnected = 0;
		if (pkt->inq[0] == 0 || pkt->inq[0] == 0xc)
			fcport->flags = fcport->flags & ~FC_TAPE_DEVICE;
		else if (pkt->inq[0] == 1)
			fcport->flags = fcport->flags | FC_TAPE_DEVICE;
		else if (pkt->inq[0] == 0x20 || pkt->inq[0] == 0x7f)
			disconnected++;
		else
			continue;

		/* Allocate LUN if not already allocated. */
		for (fclun = fcport->fclun; fclun != NULL;
		    fclun = fclun->next) {
			if (fclun->lun == lun)
				break;
		}

		if (fclun != NULL) {
			/* Found this lun already in our list */
			continue;
		}

		/* Add this lun to our list */
		fcport->lun_cnt++;

		fclun = (fc_lun_t *)KMALLOC(sizeof (fc_lun_t));
		if (fclun != NULL) {
			/* Setup LUN structure. */
			DEBUG4(printk("lun_discovery: Allocated fclun %p, "
			    "disconnected=%d\n", fclun,disconnected);)

			fclun->fcport = fcport;
			fclun->lun = lun;

			if( disconnected )
				fclun->flags |= FC_DISCON_LUN;

			fclun->next = fcport->fclun;
			fcport->fclun = fclun;
		} else {
			printk(KERN_WARNING
			    "scsi(%d): Memory Allocation failed - FCLUN\n",
			    (int)ha->host_no);
			ha->mem_err++;
			return BIT_0;
		}
	}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	KMFREE(pkt, sizeof (inq_cmd_rsp_t));
#else
        pci_free_consistent(ha->pdev, sizeof(inq_cmd_rsp_t),
	    pkt, phys_address);
#endif

	LEAVE("qla2x00_lun_discovery: exiting normally");
	return (0);
}

/*
 * qla2100_configure_local_loop
 *	Updates Fibre Channel Device Database with local loop devices.
 *
 * Input:
 *	ha = adapter block pointer.
 *	enable_slot_reuse = allows the use of PORT_AVAILABLE slots.
 *
 * Returns:
 *	0 = success.
 *	BIT_0 = error.
 *	BIT_1 = database was full and a device was not configured.
 */
static uint8_t
qla2100_configure_local_loop(scsi_qla_host_t *ha, uint8_t enable_slot_reuse)
{
	uint8_t  status = 0;
	uint8_t  rval;
	uint8_t  tmp_name[8];
	uint8_t  update_status = 0;
	uint16_t index, size;
	dma_addr_t  phys_address = 0;
	fcdev_t device;
	port_list_entry_t *gn_list, *port_entry;
	uint16_t localdevices = 0;
	uint8_t *cp = NULL;


	DEBUG3(printk("qla2100_configure_local_loop: entered\n");)

/* 4.10 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
	gn_list = KMALLOC_DMA(GN_LIST_LENGTH);
	if ( gn_list != NULL ) {
		phys_address = VIRT_TO_BUS(gn_list);
	}
#else
	gn_list = pci_alloc_consistent(ha->pdev, sizeof(GN_LIST_LENGTH),
                                           &phys_address);
#endif

	if (gn_list) 
	{
		BZERO((caddr_t)gn_list, sizeof(GN_LIST_LENGTH));
		if (!ha->loop_down_timer &&
		    !(ha->dpc_flags & LOOP_RESYNC_NEEDED)) {
			/* Mark all local devices PORT_LOST_ID first */
			for (index = 0; index < MAX_FIBRE_DEVICES; index++) {
				if (ha->fc_db[index].loop_id <=
				    LAST_SNS_LOOP_ID &&
				    !(ha->fc_db[index].flag & DEV_PUBLIC)) {

    	    DEBUG(printk("qla%ld: qla2x00_configure_local_loop: port lost @ slot %d %06x\n",
		ha->instance,
         	index, 
		ha->fc_db[index].d_id.b24);)
					ha->fc_db[index].loop_id |=
					    PORT_LOST_ID;
				}
			}

			/* Get port name list.*/
			if (!qla2x00_get_port_list(ha, gn_list, phys_address,
			    BIT_0, &size)) {
				port_entry = gn_list;

				/* dg: 10/29/99 for an empty list */
				if (size/sizeof( port_list_entry_t) != 0 ) { 
					for ( ;
					    size >= sizeof(port_list_entry_t);
					    size -= sizeof(port_list_entry_t),
					    port_entry++) {

						/* Skip the known ports. */
						if ((port_entry->loop_id ==
						    SNS_FL_PORT) ||
						    (port_entry->loop_id ==
						    FABRIC_CONTROLLER) ||
						    (port_entry->loop_id ==
						    SIMPLE_NAME_SERVER)) {
							continue;
						}

						device.loop_id =
						    port_entry->loop_id;

						/* Get port name */
						rval = qla2x00_get_port_name(ha,
						    device.loop_id, tmp_name,
						    0);
						if (!rval && (tmp_name[0] != 0 ||
						    tmp_name[1] != 0 ||
						    tmp_name[2] != 0 ||
						    tmp_name[3] != 0 ||
						    tmp_name[4] != 0 ||
						    tmp_name[5] != 0 ||
						    tmp_name[6] != 0 ||
						    tmp_name[7] != 0)) {

							BCOPY(tmp_name,
							    &device.wwn[0], 8);

#ifdef QL_DEBUG_LEVEL_3
							cp = (uint8_t *)
							    &device.wwn[0];

							printk("qla%ld configure_local: found portname -> %02x%02x%02x%02x%02x%02x%02x%02x\n",
							    ha->host_no, cp[0],
							    cp[1], cp[2], cp[3],
							    cp[4], cp[5], cp[6],							    cp[7]);
#endif

						} else {
							DEBUG2(printk("qla%ld "
							    "configure_local: "
							    "get_port_name "
							    "error.\n",
							    ha->host_no);)
							status = BIT_0;
							break;
						}

						/* Now get node name */
						BCOPY(&port_entry->name[0],
						    &device.name[0], WWN_SIZE);

						cp = (uint8_t *)&device.name[0];
						BIG_ENDIAN_64(*cp);

						DEBUG3(printk("qla%ld configure_local: found nodename -> %02x%02x%02x%02x%02x%02x%02x%02x\n",
						   ha->host_no, cp[0], cp[1],
						   cp[2], cp[3], cp[4], cp[5],
						   cp[6], cp[7]);)

						device.flag = 0;
						device.d_id.b24 = 0;
						update_status =
						    qla2100_update_fc_database(
						    ha, &device,
						    enable_slot_reuse);

						if( update_status ) {
							localdevices++;
							status |= update_status;
						}

					} /* for each port entry */
				}
			} else {
				status = BIT_0;
			}
		}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,18)
		KMFREE(gn_list, sizeof(GN_LIST_LENGTH));
#else
		pci_free_consistent(ha->pdev, sizeof(GN_LIST_LENGTH),
		    gn_list, phys_address);
#endif

	} else {
		printk(KERN_WARNING "scsi(%d): Memory Allocation failed - "
		    "port_list",(int)ha->host_no);
		ha->mem_err++;

#ifdef QL_DEBUG_LEVEL_2
		qla2100_print("qla2100_configure_local_loop: Failed to "
		    "allocate memory, No local loop\n");
#endif
		status = BIT_0;
	}

#if defined(QL_DEBUG_LEVEL_2) || defined(QL_DEBUG_LEVEL_3)
	if (status & BIT_0)
		qla2100_print("qla2100_configure_local_loop: *** FAILED ***\n");
#endif
#ifdef QL_DEBUG_LEVEL_3
	else
		LEAVE("qla2100_configure_local_loop: exiting normally\n");
#endif

	if (localdevices > 0 ) {
		ha->device_flags |= DFLG_LOCAL_DEVICES;
		ha->device_flags &= ~DFLG_RETRY_LOCAL_DEVICES;
	}

	return(status);
}

#ifdef NEW_SP
srb_t * qla2100_allocate_sp_from_ha( scsi_qla_host_t *ha, Scsi_Cmnd *cmd)
{
	int i, cnt;
	srb_t	*sp;

	ENTER("qla2100_allocate_sp_from_ha");
    i= ha->srb_cnt;
    cnt = 0;
    do {
	    /* handle warp-around */
	    i = (i == (MAX_CMDS_PER_LUN - 1) )? 0 : i;
	    sp = &ha->srb_pool[i];
	    if( sp->used == 0 ) {
	        sp->used = 0x81;
	        ha->srb_cnt = i;
	        DEBUG4(sprintf(debug_buff,"qla2100_allocate_sp: Used HA pool cmd=%p, sp=%p, index= %d\n",cmd,sp,i);)
		DEBUG4(qla2100_print(debug_buff);)
	        break;
	    }
	    cnt++;
	    i++;
    } while ( cnt < MAX_CMDS_PER_LUN );
    /* if no SP left in ha pool then give up */
    if ( cnt == MAX_CMDS_PER_LUN )
	sp = NULL;

   	CMD_SP(cmd) = (void *) sp;
	sp->s_cmd.base_address = sp;
	LEAVE("qla2100_allocate_sp_from_ha");
    return(sp);
}

/*
 * Get a sp for the specified lun.  The lun
 * structure has a list of possible sp pointers.
 */
srb_t * qla2100_allocate_sp( scsi_qla_host_t *ha, Scsi_Cmnd *cmd)
{
    uint32_t        t, l;
	srb_t	*sp;
    os_lun_t      *q;
	int i, cnt;

	ENTER("qla2100_allocate_sp");
	DEBUG4(sprintf(debug_buff,"qla2100_allocate_sp: Enter cmd=%p\n",cmd);)
	DEBUG4(qla2100_print(debug_buff);)
    t = SCSI_TCN_32(cmd);
    l = SCSI_LUN_32(cmd);

    if ( (q = GET_LU_Q(ha, t,l)) == NULL ) {
	    /* if we don't have a q pointer then use the ha pool */
	sp = qla2100_allocate_sp_from_ha(ha, cmd);
    }
    else {
	/* find a used srb ptr */
	i= q->srb_cnt;
	cnt = 0;
	do {
		/* handle warp-around */
		i = (i == (MAX_CMDS_PER_LUN - 1) )? 0 : i;
		/* sp = &q->srb_pool[i]; */
		sp = q->srb_pool[i];
		if( sp == NULL ) {
		  sp = (srb_t *)KMEM_ZALLOC(sizeof (srb_t), 30);
		  if (sp != NULL) {
        		DEBUG3(printk("Alloc Srb @ %p \n",sp);)
			q->srb_pool[i] = sp;
		   } else {
		   /*EMPTY*/
			printk(KERN_WARNING "scsi(%d): Memory Allocation failed - srb",(int)ha->host_no);
			ha->mem_err++;
			return( sp );
		   }
		}
		if( sp->used == 0 ) {
		   sp->used = 1;
		   q->srb_cnt = i;
		   DEBUG4(sprintf(debug_buff,"qla2100_allocate_sp: Used LUN pool cmd=%p, sp=%p, index= %d\n",cmd,sp,i);)
		   DEBUG4(qla2100_print(debug_buff);)
		   break;
		}
		cnt++;
		i++;
	} while ( cnt < MAX_CMDS_PER_LUN );

	if ( cnt == MAX_CMDS_PER_LUN )
		sp = NULL;
   	CMD_SP(cmd) = (void *) sp;
	sp->s_cmd.base_address = sp;
    }
	DEBUG4(sprintf(debug_buff,"qla2100_allocate_sp: Exit %p \n",sp);)
	DEBUG4(qla2100_print(debug_buff);)
	LEAVE("qla2100_allocate_sp");
	return( sp );
}


void qla2100_free_sp( Scsi_Cmnd *cmd, srb_t *sp)
{
	ENTER("qla2100_free_sp");
	if( ( sp->used ) ) {
		DEBUG4(sprintf(debug_buff,"qla2100: Freed sp=%p\n",sp);)
		DEBUG4(qla2100_print(debug_buff);)
    	CMD_SP(cmd) = NULL;
		sp->used = 0;
	}
	else
  	printk(KERN_INFO "qla2100_free_sp: Could not free sp %p it is already freed\n",sp);
	LEAVE("qla2100_free_sp");
}
#endif

/*
 * qla2x00_tgt_alloc
 *	Allocate and initialize target queue.
 *
 * Input:
 *	ha = adapter block pointer.
 *	t = SCSI target number.
 *	name = 8 byte node or port name.
 *
 * Returns:
 *	NULL = failure
 *
 * Context:
 *	Kernel context.
 */
os_tgt_t *
qla2x00_tgt_alloc(scsi_qla_host_t *ha, uint16_t t, uint8_t *name)
{
	os_tgt_t	*tq;

	DEBUG3(printk("qla2x00_tgt_alloc(%ld): entered for tgt %d.\n",
	    ha->host_no, t);)

	/*
	 * If SCSI addressing OK, allocate TGT queue and lock.
	 */
	if (t >= MAX_TARGETS) {
		DEBUG2(printk("qla2x00_tgt_alloc(%ld): *** FAILED exiting ***",
		    ha->host_no);)
		return (NULL);
	}

	tq = TGT_Q(ha, t);
	if (tq == NULL) {
		tq = (os_tgt_t *)KMALLOC(sizeof (os_tgt_t));
		if (tq != NULL) {
			BZERO((caddr_t)tq, sizeof(os_tgt_t));

			DEBUG(printk("Alloc Target %d @ %p \n", t, tq);)

			tq->flags = TGT_TAGGED_QUEUE;
			tq->ha = ha;

			TGT_Q(ha, t) = tq;
		} else {
			printk(KERN_WARNING
			    "scsi(%d): Failed to allocate target\n",
			    (int)ha->host_no);
			ha->mem_err++;
		}
	}

	if (tq != NULL) {
		/* Save fcport name for persistent binding. */
		BCOPY((void *)name, (void *)tq->fc_name, 8);

		tq->port_down_retry_count = ha->port_down_retry_count;

		LEAVE("qla2x00_tgt_alloc: exiting normally");
	} else {
		/*EMPTY*/
		DEBUG2(printk("qla2x00_tgt_alloc(%ld): *** FAILED exiting ***",
		    ha->host_no);)
	}

	return (tq);
}

/*
 * qla2x00_tgt_free
 *	Frees target and LUN queues.
 *
 * Input:
 *	ha = adapter block pointer.
 *	t = SCSI target number.
 *
 * Context:
 *	Kernel context.
 */
void
qla2x00_tgt_free(scsi_qla_host_t *ha, uint16_t t)
{
	os_tgt_t	*tq;
	uint16_t	l;

	/* ENTER("qla2x00_tgt_free: entered"); */

	/*
	 * If SCSI addressing OK, allocate TGT queue and lock.
	 */
	if (t >= MAX_TARGETS) {
		DEBUG2(printk("qla2x00_tgt_free: **** FAILED exiting ****");)
		return;
	}

	tq = TGT_Q(ha, t);
	if (tq != NULL) {
		TGT_Q(ha, t) = NULL;
        DEBUG(sprintf(debug_buff,"Dealloc target @ %p -- deleted\n",tq);)
        DEBUG(qla2100_print(debug_buff);)

		/* Free LUN structures. */
		for (l = 0; l < MAX_LUNS; l++)
			qla2x00_lun_free(ha, t, l);

		KMFREE(tq, sizeof (os_tgt_t));
	}

	/* LEAVE("qla2x00_tgt_free: exiting normally"); */

	return;
}

/*
 * qla2x00_lun_alloc
 *	Allocate and initialize LUN queue.
 *
 * Input:
 *	ha = adapter block pointer.
 *	t = SCSI target number.
 *	l = LUN number.
 *
 * Returns:
 *	NULL = failure
 *
 * Context:
 *	Kernel context.
 */
os_lun_t *
qla2x00_lun_alloc(scsi_qla_host_t *ha, uint16_t t, uint16_t l)
{
	os_lun_t	*lq;

	ENTER("qla2x00_lun_alloc: entered");

	/*
	 * If SCSI addressing OK, allocate LUN queue.
	 */
	if (t >= MAX_TARGETS || l >= MAX_LUNS || TGT_Q(ha, t) == NULL) {
		DEBUG2(sprintf(debug_buff,"qla2x00_lun_alloc: tgt=%d, tgt_q= %p,lun=%d,instance=%ld **** FAILED exiting ****\n",t,TGT_Q(ha,t),l,ha->instance);)
		DEBUG2(qla2100_print(debug_buff);)
		return (NULL);
	}

	lq = LUN_Q(ha, t, l);
	if (lq == NULL) {
		lq = (os_lun_t *)KMALLOC(sizeof (os_lun_t));
		if (lq != NULL) {
        		DEBUG2(sprintf(debug_buff,"Alloc Lun @ %p \n",lq);)
        		DEBUG2(qla2100_print(debug_buff);)
			BZERO((caddr_t)lq, sizeof (os_lun_t));
			LUN_Q(ha, t, l) = lq;
		} else {
		/*EMPTY*/
		DEBUG2(printk("qla2x00_lun_alloc: Failed to allocate lun %d ***\n",l);)
			printk(KERN_WARNING "scsi(%d): Memory Allocation failed - FCLUN",(int)ha->host_no);
			ha->mem_err++;
		}
	}

	if (lq == NULL) {
		/*EMPTY*/
		DEBUG2(printk("qla2x00_lun_alloc: **** FAILED exiting ****\n");)
	} else {
		/*EMPTY*/
		LEAVE("qla2x00_lun_alloc: exiting normally");
	}

	return (lq);
}

/*
 * qla2x00_lun_free
 *	Frees LUN queue.
 *
 * Input:
 *	ha = adapter block pointer.
 *	t = SCSI target number.
 *
 * Context:
 *	Kernel context.
 */
static void
qla2x00_lun_free(scsi_qla_host_t *ha, uint16_t t, uint16_t l)
{
	os_lun_t	*lq;
	srb_t 	*sp;
	int i;

	/* ENTER("qla2x00_lun_free: entered"); */

	/*
	 * If SCSI addressing OK, allocate TGT queue and lock.
	 */
	if (t >= MAX_TARGETS || l >= MAX_LUNS) {
		DEBUG2(printk("qla2x00_lun_free: **** FAILED exiting ****");)
		return;
	}

	if (TGT_Q(ha, t) != NULL && (lq = LUN_Q(ha, t, l)) != NULL) {
		LUN_Q(ha, t, l) = NULL;
		for( i = 0; i < MAX_CMDS_PER_LUN; i++ ){
			sp = lq->srb_pool[i];
			lq->srb_pool[i] = NULL;
			if( sp != NULL )
				KMEM_FREE(sp, sizeof (srb_t));
		}
		KMFREE(lq, sizeof (os_lun_t));
        DEBUG3(sprintf(debug_buff,"Dealloc lun @ %p -- deleted\n",lq);)
        DEBUG3(qla2100_print(debug_buff);)
	}

	/* LEAVE("qla2x00_lun_free: exiting normally"); */

	return;
}

/*
 * qla2x00_add_link_b
 *	Add link to the end of the chain.
 *
 * Input:
 *	head = Head of link list.
 *	link = link to be added.
 *	LOCK must be already obtained.
 *
 * Context:
 *	Kernel context.
 */
void
qla2x00_add_link_b(ql_list_head_t *head, ql_list_link_t *link)
{
	link->next = NULL;

	if ((link->prev = head->last) == NULL)
		head->first = link;
	else
		head->last->next = link;

	head->last = link;

}

/*
 * qla2x00_add_link_t
 *	Add link to the beginning of the chain.
 *
 * Input:
 *	head = Head of link list.
 *	link = link to be added.
 *	LOCK must be already obtained.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
void
qla2x00_add_link_t(ql_list_head_t *head, ql_list_link_t *link)
{
	link->prev = NULL;

	if ((link->next = head->first) == NULL)
		head->last = link;
	else
		head->first->prev = link;

	head->first = link;
}

/*
 * qla2x00_remove_link
 *	Remove a link from the chain.
 *
 * Input:
 *	head = Head of link list.
 *	link = link to be removed.
 *	LOCK must be already obtained.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
void
qla2x00_remove_link(ql_list_head_t *head, ql_list_link_t *link)
{
	if (link->prev != NULL) {
		if ((link->prev->next = link->next) == NULL)
			head->last = link->prev;
		else
			link->next->prev = link->prev;
	} else if ((head->first = link->next) == NULL) {
		head->last = NULL;
	} else {
		head->first->prev = NULL;
	}

}

/*
 * qla2x00_next
 *	Retrieve and process next job in the LUN queue.
 *
 * Input:
 *	tq = SCSI target queue pointer.
 *	lq = SCSI LUN queue pointer.
 *	TGT_LOCK must be already obtained.
 *
 * Output:
 *	Releases TGT_LOCK upon exit.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
void
qla2x00_next(os_tgt_t *tq, os_lun_t *lq)
{
	scsi_qla_host_t *ha;
	scsi_qla_host_t *vis_ha;
	fc_port_t	*fcport;
	srb_t		*sp;
	int		rval;
	ql_list_link_t	*link;
	ql_list_link_t	*next_link = NULL;

	/* ENTER("qla2x00_next"); */

	for (link = lq->cmd.first; link != NULL; link = next_link) {
		next_link = link->next;
		sp = link->base_address;
		fcport = lq->fclun->fcport;
		ha = fcport->ha;
		vis_ha = (scsi_qla_host_t *)sp->cmd->host->hostdata;
		/* Check if command can be started, exit if not. */
		if (
		    /* tq->flags & TGT_BUSY || */
		    (lq->q_flag & (LUN_BUSY | LUN_QUEUE_SUSPENDED)) ||
		    /* sp->flags & SRB_ABORTED || */
			LOOP_TRANSITION(ha)  ) {
			break;
		}


                qla2x00_remove_link(&lq->cmd, &sp->s_cmd);
            	lq->q_incnt--;

		lq->q_outcnt++;
		if (lq->q_outcnt >= ha->hiwat) {
			lq->q_flag = lq->q_flag | LUN_BUSY;
                	lq->q_bsycnt++;
		}

		/* Release target queue lock */
		TGT_UNLOCK(tq);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,3,18)
        if( ha->flags.enable_64bit_addressing )
			rval = qla2x00_64bit_start_scsi(sp);
		else
#endif
			rval = qla2x00_32bit_start_scsi(sp);

		/* Acquire target queue lock */
		TGT_LOCK(tq);

		if (rval != QLA2X00_SUCCESS) {
			/* Place request back on top of device queue */
			/* add to the top of queue */
			qla2100_putq_t(lq,sp);
			if (lq->q_outcnt != 0)
				lq->q_outcnt--;
			if (lq->q_outcnt < ha->hiwat) {
				lq->q_flag = lq->q_flag & ~LUN_BUSY;
                	        lq->q_bsycnt--;
                        }

			/* Restart queue later. */
			if (lq->q_outcnt == 0 && 
				lq->q_incnt > 0 &&
				vis_ha->queue_restart_timer == 0)
				vis_ha->queue_restart_timer = PORT_RETRY_TIME;

			break;
		}
	}
	/* Release target queue lock */
	TGT_UNLOCK(tq);

	/* LEAVE("qla2x00_next"); */
}
/*
 * qla2100_putq_t
 *      Add the standard SCB job to the top of standard SCB commands.
 *
 * Input:
 *      q  = SCSI LU pointer.
 *      sp = srb pointer.
 *      SCSI_LU_Q lock must be already obtained.
 */
static void qla2100_putq_t(os_lun_t *lq, srb_t *sp)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
            unsigned long cpu_flags = 0;
#endif

            ENTER("qla2100_putq_t");
            DRIVER_LOCK
            DEBUG5(sprintf(debug_buff,"Adding to queue 0x%p<-(0x%p)\n\r",lq,sp));
            DEBUG5(qla2100_print(debug_buff));

			qla2x00_add_link_t(&lq->cmd, &sp->s_cmd);
            lq->q_incnt++;
            DEBUG(sp->state = 2);

            DRIVER_UNLOCK
            LEAVE("qla2100_putq_t");
}

/*
 * qla2100_putq_b
 *      Add the standard SCB job to the bottom of standard SCB commands.
 *
 * Input:
 *      q  = SCSI LU pointer.
 *      sp = srb pointer.
 *      SCSI_LU_Q lock must be already obtained.
 */
static void qla2100_putq_b(os_lun_t *lq, srb_t *sp)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
            unsigned long cpu_flags = 0;
#endif

#ifdef QL_DEBUG_LEVEL_3
            /* ENTER("qla2100_putq_b"); */
#endif
            DRIVER_LOCK
            DEBUG5(sprintf(debug_buff,"Adding to queue 0x%p<-(0x%p)\n\r",lq,sp));
            DEBUG5(qla2100_print(debug_buff));

			qla2x00_add_link_b(&lq->cmd, &sp->s_cmd);
            lq->q_incnt++;
            DEBUG(sp->state = 2);

            DRIVER_UNLOCK
#ifdef QL_DEBUG_LEVEL_3
            /* LEAVE("qla2100_putq_b"); */
#endif
}

/*
 * qla2100_removeq
 *      Function used to remove a command block from the
 *      LU queue.
 *
 * Input:
 *      q  = SCSI LU pointer.
 *      sp = srb pointer.
 *      SCSI_LU_Q lock must be already obtained.
 */
static void qla2100_removeq(os_lun_t *lq, srb_t *sp)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,95)
            unsigned long cpu_flags = 0;
#endif
            DEBUG5(sprintf(debug_buff,"Removing from device_q (0x%p)->(0x%p)\n\r",lq,sp));
            DEBUG5(qla2100_print(debug_buff));
            DRIVER_LOCK

            lq->q_incnt--;
			qla2x00_remove_link(&lq->cmd, &sp->s_cmd);
            DRIVER_UNLOCK
}
/*
 * qla2x00_is_wwn_zero
 *
 * Input:
 *      ww_name = Pointer to WW name to check
 *
 * Returns:
 *      TRUE if name is 0 else FALSE
 *
 * Context:
 *      Kernel context.
 */
static int
qla2x00_is_wwn_zero(uint8_t *nn)
{
	int cnt;

	/* Check for zero node name */
	for (cnt = 0; cnt < WWN_SIZE ; cnt++, nn++) {
		if ( *nn != 0)
			break;
	}
	/* if zero return TRUE */
	if (cnt == WWN_SIZE)
		return(TRUE);
	else
		return(FALSE);
}

/*
 * qla2100_get_lun_mask_from_config
 *      Get lun mask from the configuration parameters.
 *      Bit order is little endian.
 *
 * Input:
 * ha  -- Host adapter
 * tgt  -- target/device number
 * port -- pointer to port
 */
void
qla2100_get_lun_mask_from_config(scsi_qla_host_t *ha, fc_port_t *port,
    uint16_t tgt, uint16_t dev_no)
{
	char		propbuf[60]; /* size of search string */
	int		rval, lun, l;
	lun_bit_mask_t	lun_mask, *mask_ptr = &lun_mask;

	/* Get "target-N-device-N-lun-mask" as a 256 bit lun_mask*/
	sprintf(propbuf, "scsi-qla%ld-tgt-%d-di-%d-lun-disabled",
	    ha->instance, tgt, dev_no);
	rval = qla2x00_get_prop_xstr(ha, propbuf, (uint8_t *)&lun_mask,
	    sizeof(lun_mask));

	if (rval != -1 && (rval == sizeof(lun_mask))) {

		DEBUG3(printk("qla2100_get_lun_mask_from_config(%ld): lun "
		    "mask for port %p from file:\n", ha->host_no, port);)
		DEBUG3(qla2100_dump_buffer((uint8_t *)&port->lun_mask,
		    sizeof(lun_bit_mask_t));)

		for (lun = 8*sizeof(lun_mask)-1, l = 0; lun >= 0; lun--, l++) {
			if (EXT_IS_LUN_BIT_SET(mask_ptr, lun))
				EXT_SET_LUN_BIT((&port->lun_mask),l);
			else
				EXT_CLR_LUN_BIT((&port->lun_mask),l);
		}

		DEBUG3(printk("qla2100_get_lun_mask_from_config(%ld): returning"
		    " lun mask for port %02x%02x%02x%02x%02x%02x%02x%02x:\n",
		    ha->host_no, port->port_name[0], port->port_name[1],
		    port->port_name[2], port->port_name[3],
		    port->port_name[4], port->port_name[5],
		    port->port_name[6], port->port_name[7]);)
		DEBUG3(qla2100_dump_buffer((uint8_t *)&port->lun_mask,
		    sizeof(lun_bit_mask_t));)
	}
}

/*
 * qla2x00_bstr_to_hex
 *	Convert hex byte string to number.
 *
 * Input:
 *	s = byte string pointer.
 *	bp = byte pointer for number.
 *	size = number of bytes.
 *
 * Context:
 *	Kernel/Interrupt context.
 */
static int
qla2x00_bstr_to_hex(char *s, uint8_t *bp, int size)
{
	int		cnt;
	uint8_t		n;

	ENTER("qla2x00_bstr_to_hex");
	for (cnt = 0; *s != '\0' && cnt / 2 < size; cnt++) {
		if (*s >= 'A' && *s <= 'F') {
			n = (*s++ - 'A') + 10;
		} else if (*s >= 'a' && *s <= 'f') {
			n = (*s++ - 'a') + 10;
		} else if (*s >= '0' && *s <= '9') {
			n = *s++ - '0';
		} else {
			cnt = 0;
			break;
		}

		if (cnt & BIT_0)
			*bp++ |= n;
		else
			*bp = n << 4;
	}
	/* fixme(dg) Need to swap data little endian */
	LEAVE("qla2x00_bstr_to_hex");

	return (cnt / 2);
}

/*
 * qla2100_get_prop_xstr
 *      Get a string property value for the specified property name and
 *      convert from the property string found in the configuration file,
 *      which are ASCII characters representing nibbles, 2 characters represent
 *      the hexdecimal value for a byte in the byte array.
 *      The byte array is initialized to zero.
 *      The resulting converted value is in big endian format (MSB at byte0).
 *
 * Input:
 *      ha = adapter state pointer.
 *      propname = property name pointer.
 *      propval  = pointer where to store converted property val.
 *      size = max or expected size of 'propval' array.
 *
 * Returns:
 *      0 = empty value string or invalid character in string
 *      >0 = count of characters converted
 *      -1 = property not found
 *
 * Context:
 *      Kernel context.
 */
int
qla2x00_get_prop_xstr(adapter_state_t *ha, char *propname, uint8_t *propval,
    int size)
{
	char	*propstr;
	int	rval = -1;
	char buf[LINESIZE];

	ENTER("qla2x00_get_prop_xstr");

	/* Get the requested property string */
	rval = qla2100_find_propname(ha, propname, buf, ha->cmdline, size*2);
	DEBUG3(printk("get_prop_xstr: Ret rval from find propname = %d\n", rval);)
	propstr = &buf[0];
	if ( *propstr == '=' )
 		propstr++;   /* ignore equal sign */

	if (rval == 0 ) {  /* not found */
	   LEAVE("qla2x00_get_prop_xstr");
		return (-1);
	}

	rval = qla2x00_bstr_to_hex(propstr, (uint8_t *)propval, size);
	if (rval == 0) {
		/* Invalid character in value string */
		printk(KERN_INFO "qla2x00_get_prop_xstr: %s Invalid hex string for property", propname);
		printk(KERN_INFO " Invalid string - %s ", propstr);
	}
	LEAVE("qla2x00_get_prop_xstr");
	return (rval);
}

/*
 * qla2x00_chg_endian
 *	Change endianess of byte array.
 *
 * Input:
 *	buf = array pointer.
 *	size = size of array in bytes.
 *
 * Context:
 *	Kernel context.
 */
void
qla2x00_chg_endian(uint8_t buf[], size_t size)
{
	uint8_t byte;
	size_t cnt1;
	size_t cnt;

	cnt1 = size - 1;
	for (cnt = 0; cnt < size / 2; cnt++) {
		byte = buf[cnt1];
		buf[cnt1] = buf[cnt];
		buf[cnt] = byte;
		cnt1--;
	}
}

/*
* Declarations for load module
*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#ifdef MODULE
            Scsi_Host_Template driver_template = QLA2100_LINUX_TEMPLATE;

#include "scsi_module.c"
#endif
#else   /* new kernel scsi initialization scheme */
static  Scsi_Host_Template driver_template = QLA2100_LINUX_TEMPLATE;
#include "scsi_module.c"
#endif

#ifdef QL_DEBUG_ROUTINES
/****************************************************************************/
/*                         Driver Debug Functions.                          */
/****************************************************************************/

/*
*  Get byte from I/O port
*/
STATIC uint8_t
qla2100_getbyte(uint8_t *port)
{
                uint8_t ret;

#ifdef MEMORY_MAPPED_IO
                ret = *port;
#else
                ret = inb((int)port);
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_getbyte: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)ret, 16);
                    qla2100_print("\n");
                }

                return(ret);
}

/*
*  Get word from I/O port
*/
STATIC uint16_t
qla2100_getword(uint16_t *port)
{
                uint16_t ret;

#ifdef MEMORY_MAPPED_IO
                ret = *port;
#else
                ret = inw((int)port);
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_getword: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)ret, 16);
                    qla2100_print("\n");
                }

                return(ret);
}

/*
*  Get double word from I/O port
*/
STATIC uint32_t
qla2100_getdword(uint32_t *port)
{
                uint32_t ret;

#ifdef MEMORY_MAPPED_IO
                ret = *port;
#else
                ret = inl((int)port);
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_getdword: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)ret, 16);
                    qla2100_print("\n");
                }

                return(ret);
}

/*
*  Send byte to I/O port
*/
STATIC void
qla2100_putbyte(uint8_t *port, uint8_t data)
{
#ifdef MEMORY_MAPPED_IO
                *port = data;
#else
                outb(data, (int)port);
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_putbyte: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)data, 16);
                    qla2100_print("\n");
                }
}

/*
*  Send word to I/O port
*/
STATIC void
qla2100_putword(uint16_t *port, uint16_t data)
{
#ifdef MEMORY_MAPPED_IO
                *port = data;
#else
#ifdef _LINUX_IOPORTS
                outw(data, (int)port);
#else
                outw((int)port, data);
#endif
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_putword: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)data, 16);
                    qla2100_print("\n");
                }
}

/*
*  Send double word to I/O port
*/
STATIC void
qla2100_putdword(uint32_t *port, uint32_t data)
{
#ifdef MEMORY_MAPPED_IO
                *port = data;
#else
#ifdef _LINUX_IOPORTS
                outl(data,(int)port);
#else
                outl((int)port, data);
#endif
#endif

                if( ql2x_debug_print ) {
                    qla2100_print("qla2100_putdword: address = ");
            qla2100_output_number((u_long)port, 16);
            qla2100_print(" data = 0x");
            qla2100_output_number((u_long)data, 16);
                    qla2100_print("\n");
                }
}

/*
* Dummy function to prevent warnings for
* declared and unused debug functions
*/
void
qla2100_debug(void)
{
                qla2100_getbyte(0);
                qla2100_getword(0);
                qla2100_getdword(0);
                qla2100_putbyte(0, 0);
                qla2100_putword(0, 0);
                qla2100_putdword(0, 0);
}

/*
*  Out character to COM1 port.
*      PORT must be at standard address for COM1 = 0x3F8
*      This port is available on both Lion and ia32 systems
*/
#define OUTB(addr,data)   outb((data),(addr))
STATIC void
qla2100_putc(int8_t c)
{
#ifdef QL_DEBUG_CONSOLE

                printk("%c",c);

#else /* QL_DEBUG_CONSOLE */

                uint8_t data;
                int     com_addr              = 0x3f8; /* COM1 */
                int     hardware_flow_control = 1;
                int     software_flow_control = 0;
                uint8_t loop                  = TRUE;

                do {

                    /* Wait for transmitter holding and shift registers for empty. */
                    do {
                        data = inb(com_addr+5);
                    }while( !(data & BIT_6) );

                    /*
                    * Set BAUD rate for COM1 to 9600
                    */

                    /* BAUD rate divisor LSB. */
                    OUTB(com_addr+3, 0x83);

                    /* BAUD rate divisor MSB. */
                    OUTB(com_addr, 0xc);           /* 0xC = 9600 baud */

                    /* Set No parity, 8 bits, 1 stop bit and
                    select interrupt enable register. */
                    OUTB(com_addr+3, 3);

                    /* Disable interrupts. */
                    OUTB(com_addr+1, 0);

                    /* Set data terminal ready and request to send */
                    OUTB(com_addr+1, 0);

                    if( hardware_flow_control ) {
                        /* Wait for clear-to-send and data-set-ready */
                        do {
                            data = inb(com_addr+6) & (BIT_5 + BIT_4);
                        }while( data != (BIT_5 + BIT_4) );
                    } else if( software_flow_control ) {
                        /* Test for data ready. */
                        data = inb(com_addr+5);
                        if( data & BIT_0 ) {
                            /* If XOFF */
                            data = inb(com_addr);
                            if( data == '\023' ) {
                                /* Wait for XON */
                                do {
                                    /* Wait for int8_t */
                                    do {
                                        data = inb(com_addr+5);
                                    }while( !(data & BIT_0) );
                                    data = inb(com_addr);
                                }while( data != '\021' );
                            }
                        }
                    }

                    /* Output character. */
                    OUTB(com_addr, c);

                    /* Add return. */
                    if( c == '\n' )
                        c = '\r';
                    else
                        loop = FALSE;

                }while( loop );
#endif  /* QL_DEBUG_CONSOLE */
}

/*
*  Out NULL terminated string to COM port.
*/
STATIC void
qla2100_print(int8_t *s)
{
#ifdef QL_DEBUG_CONSOLE
                    printk("%s",s);
#else
                    /* Output string. */
                    while( *s )
                        qla2100_putc(*s++);
#endif
                }

/*
*  Output long number to COM port.
*/
STATIC void
qla2100_output_number(u_long n, uint8_t base)
{
                int8_t str[12];
                int8_t *s      = &str[11];
                uint8_t output = 0;
                uint8_t hex    = FALSE;

                    if( base == 10 || base == 16 ) {
                        if( base == 16 && n > 9 )
                            hex = TRUE;

                        *s = 0;
                        do {
                            s--;
                            *s = n % base;
                            if( *s > 9 )
                                *s += 55;
                            else
                                *s += '0';
                            n /= base;
                        }while( n );

                        for( ; *s; s++ ) {
                            if( *s != '0' )
                                output = 1;
                            if( output )
                                qla2100_putc(*s);
                        }
                        if( !output )
                            qla2100_putc(*--s);

                        if( hex )
                            qla2100_putc('h');
                    }
                }

#ifdef QL_DEBUG_ROUTINES
STATIC void
qla2100_dump_buffer(uint8_t *b, uint32_t size)
{
                uint32_t cnt;
                uint8_t c;

                    qla2100_print(
                    " 0   1   2   3   4   5   6   7   8   9   Ah  Bh  Ch  Dh  Eh  Fh\n");
                    qla2100_print(
                    "---------------------------------------------------------------\n");

                    for( cnt = 0; cnt < size; ) {
                        c = *b++;
                        if( c < 16 )
                            qla2100_putc(' ');
                        qla2100_output_number((uint32_t)c, 16);
                        cnt++;
                        if( !(cnt % 16) )
                            qla2100_print("\n");
                        else if( c < 10 )
                            qla2100_print("  ");
                        else
                            qla2100_putc(' ');
                    }
                    if( cnt % 16 )
                        qla2100_print("\n");
                }
#endif

/**************************************************************************
*   ql2100_print_scsi_cmd
*
**************************************************************************/
void qla2100_print_scsi_cmd(Scsi_Cmnd *cmd)
{
                scsi_qla_host_t *ha;
                struct Scsi_Host  *host = cmd->host;
                srb_t           *sp;
        	os_lun_t	*lq;
		fc_port_t	*fcport;

                int i;
                ha = (scsi_qla_host_t *) host->hostdata;

                ql2x_debug_print = 1;
                sp = (srb_t *) CMD_SP(cmd);
        sprintf(debug_buff,"SCSI Command @= 0x%p, Handle=0x%08lx\n\r", cmd, (u_long)CMD_HANDLE(cmd));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"  chan=%d, target = 0x%02x, lun = 0x%02x, cmd_len = 0x%02x\n\r",
                cmd->channel, cmd->target, cmd->lun, cmd->cmd_len);
                qla2100_print(debug_buff);
                qla2100_print(" CDB = ");
                for( i = 0; i < cmd->cmd_len; i++ ) {
                    sprintf(debug_buff,"0x%02x ", cmd->cmnd[i]);
                    qla2100_print(debug_buff);
                }
                sprintf(debug_buff,"\n\r  seg_cnt =%d, retries=%d, serial_number_at_timeout=0x%lx\n\r",cmd->use_sg,cmd->retries,cmd->serial_number_at_timeout);
                qla2100_print(debug_buff);
        sprintf(debug_buff,"  request buffer=0x%p, request buffer len=0x%x\n\r",cmd->request_buffer,cmd->request_bufflen);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"  tag=%d, flags=0x%x, transfersize=0x%x \n\r",
                cmd->tag, cmd->flags,cmd->transfersize);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"  serial_number=%d, SP=0x%x\n\r", (int)cmd->serial_number, (int)CMD_SP(cmd));
                qla2100_print(debug_buff);
		if ( sp ) {
                sprintf(debug_buff,"  sp flags=0x%x, wdgtime=%d\n\r",sp->flags,sp->wdg_time);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"  r_start=0x%lx, u_start=0x%lx,state=%d\n\r",sp->r_start, sp->u_start, sp->state);
                qla2100_print(debug_buff);

		lq = sp->lun_queue;
		fcport = lq->fclun->fcport; 
                sprintf(debug_buff,"  fstate=%d, fo retry=%d, loopid =%d, port path=%d\n\r",sp->fstate,sp->fo_retry_cnt,fcport->loop_id, fcport->cur_path);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"f_start=0x%lx\n",sp->f_start);
                qla2100_print(debug_buff);
            }
}


void qla2100_print_q_info(os_lun_t *q)
{
	sprintf(debug_buff,
	    "Queue info: queue in =%d, queue out= %d, flags=0x%x\n\r",
	    q->q_incnt, q->q_outcnt, q->q_flag);
	qla2100_print(debug_buff);
}


/*
 *  Out NULL terminated string to COM port.
 */
static void
qla2x00_print_str(int8_t *s)
{
	/* Output string. */
	while (*s != 0)
		qla2100_putc(*s++);
}

/*
 *  Output 64 bit number to COM port.
 */
static void
qla2x00_output_number(uint64_t n, uint8_t base)
{
	switch(base) {
		case QDBG_DEC_NUM:
		sprintf(debug_buff,"%lld", n);
	qla2x00_print_str(debug_buff);
		break;
		case QDBG_HEX_NUM:
		sprintf(debug_buff,"0x%llx", n);
	qla2x00_print_str(debug_buff);
		break;
		default:
		sprintf(debug_buff,"%s", (char*)n);
	qla2x00_print_str(debug_buff);
	}
}

/*
 * qla2x00_print
 *	 Prints string plus number and newline.
 *
 * Input:
 *	 string  = Null terminated string (no newline at end).
 *	 number  = number.
 *	 radix   = QDBG_NO_NUM, QDBG_DEC_NUM or QDBG_HEX_NUM
 *	 newline = QDBG_NL or QDBG_NNL.
 */
void
qla2x00_print(char *string, uint64_t number, uint8_t radix, uint8_t newline)
{
	uint32_t  num = (uint32_t  )number;
	if (ql2x_debug_print != TRUE)
		return;

	qla2x00_print_str(string);

	if (radix != QDBG_NO_NUM) {
		if (radix == QDBG_PSTR)
			qla2x00_print_str((char *)&num);
		else
			qla2x00_output_number(number, radix);
	}
	if (newline == QDBG_NL)
		qla2100_putc('\n');
}

#ifdef QL_DEBUG_ROUTINES
/*
 * qla2x00_dump_buffer
 *	 Prints string plus buffer.
 *
 * Input:
 *	 string  = Null terminated string (no newline at end).
 *	 buffer  = buffer address.
 *	 wd_size = word size 8, 16, 32 or 64 bits
 *	 count   = number of words.
 */
void
qla2x00_dump_buffer(char *string, uint8_t *buffer, uint8_t wd_size,
    uint32_t count)
{
	uint32_t	cnt;
	uint16_t	*buf16;
	uint32_t	*buf32;
	uint64_t	*buf64;

	if (ql2x_debug_print != TRUE)
		return;

	if (strcmp(string, "") != 0) {
		qla2x00_print_str(string);
		qla2100_putc('\n');
	}

	switch (wd_size) {
	case 8:
		qla2x00_print_str(" 0    1    2    3    4    5    6    7");
		qla2x00_print_str(
		    "    8    9    Ah   Bh   Ch   Dh   Eh   Fh\n");
		qla2x00_print_str("---------------------------------------");
		qla2x00_print_str("---------------------------------------\n");

		for (cnt = 1; cnt <= count; cnt++, buffer++) {
			if (*buffer < 0x10)
				qla2100_putc(' ');

			qla2x00_output_number(*buffer, 16);
			if (cnt % 16 == 0)
				qla2100_putc('\n');
			else if (*buffer < 10)
				qla2x00_print_str("   ");
			else
				qla2x00_print_str("  ");
		}
		if (cnt % 16 != 0)
			qla2100_putc('\n');
		break;
	case 16:
		qla2x00_print_str("   0      2      4      6      8");
		qla2x00_print_str("      Ah     Ch     Eh\n");
		qla2x00_print_str("----------------------------");
		qla2x00_print_str("--------------------------\n");

		buf16 = (uint16_t *)buffer;
		for (cnt = 1; cnt <= count; cnt++, buf16++) {
			if (*buf16 < 0x10)
				qla2x00_print_str("   ");
			else if (*buf16 < 0x100)
				qla2x00_print_str("  ");
			else if (*buf16 < 0x1000)
				qla2100_putc(' ');

			qla2x00_output_number(*buf16, 16);

			if (cnt % 8 == 0)
				qla2100_putc('\n');
			else if (*buf16 < 10)
				qla2x00_print_str("   ");
			else
				qla2x00_print_str("  ");
		}
		if (cnt % 8 != 0)
			qla2100_putc('\n');
		break;
	case 32:
		qla2x00_print_str("       0          4");
		qla2x00_print_str("          8          Ch\n");
		qla2x00_print_str("----------------------");
		qla2x00_print_str("--------------------\n");

		buf32 = (uint32_t *)buffer;
		for (cnt = 1; cnt <= count; cnt++, buf32++) {
			if (*buf32 < 0x10)
				qla2x00_print_str("       ");
			else if (*buf32 < 0x100)
				qla2x00_print_str("      ");
			else if (*buf32 < 0x1000)
				qla2x00_print_str("     ");
			else if (*buf32 < 0x10000)
				qla2x00_print_str("    ");
			else if (*buf32 < 0x100000)
				qla2x00_print_str("   ");
			else if (*buf32 < 0x1000000)
				qla2x00_print_str("  ");
			else if (*buf32 < 0x10000000)
				qla2100_putc(' ');

			qla2x00_output_number(*buf32, 16);

			if (cnt % 4 == 0)
				qla2100_putc('\n');
			else if (*buf32 < 10)
				qla2x00_print_str("   ");
			else
				qla2x00_print_str("  ");
		}
		if (cnt % 4 != 0)
			qla2100_putc('\n');
		break;
	case 64:
		qla2x00_print_str("               0");
		qla2x00_print_str("                  4\n");
		qla2x00_print_str("-----------------------------------\n");

		buf64 = (uint64_t *)buffer;
		for (cnt = 1; cnt <= count; cnt++, buf64++) {
			if (*buf64 < 0x10)
				qla2x00_print_str("               ");
			else if (*buf64 < 0x100)
				qla2x00_print_str("              ");
			else if (*buf64 < 0x1000)
				qla2x00_print_str("             ");
			else if (*buf64 < 0x10000)
				qla2x00_print_str("            ");
			else if (*buf64 < 0x100000)
				qla2x00_print_str("           ");
			else if (*buf64 < 0x1000000)
				qla2x00_print_str("          ");
			else if (*buf64 < 0x10000000)
				qla2x00_print_str("         ");
			else if (*buf64 < 0x100000000)
				qla2x00_print_str("        ");
			else if (*buf64 < 0x1000000000)
				qla2x00_print_str("       ");
			else if (*buf64 < 0x10000000000)
				qla2x00_print_str("      ");
			else if (*buf64 < 0x100000000000)
				qla2x00_print_str("     ");
			else if (*buf64 < 0x1000000000000)
				qla2x00_print_str("    ");
			else if (*buf64 < 0x10000000000000)
				qla2x00_print_str("   ");
			else if (*buf64 < 0x100000000000000)
				qla2x00_print_str("  ");
			else if (*buf64 < 0x1000000000000000)
				qla2100_putc(' ');

			qla2x00_output_number(*buf64, 16);

			if (cnt % 2 == 0)
				qla2100_putc('\n');
			else if (*buf64 < 10)
				qla2x00_print_str("   ");
			else
				qla2x00_print_str("  ");
		}
		if (cnt % 2 != 0)
			qla2100_putc('\n');
		break;
	default:
		break;
	}
}
#endif


#endif




/**************************************************************************
*   ql2100_dump_regs
*
**************************************************************************/
static void qla2100_dump_regs(struct Scsi_Host *host)
{
                printk("Mailbox registers:\n");
                printk("qla2x00 : mbox 0 0x%04x \n", inw(host->io_port + 0x10));
                printk("qla2x00 : mbox 1 0x%04x \n", inw(host->io_port + 0x12));
                printk("qla2x00 : mbox 2 0x%04x \n", inw(host->io_port + 0x14));
                printk("qla2x00 : mbox 3 0x%04x \n", inw(host->io_port + 0x16));
                printk("qla2x00 : mbox 4 0x%04x \n", inw(host->io_port + 0x18));
                printk("qla2x00 : mbox 5 0x%04x \n", inw(host->io_port + 0x1a));
#ifdef TRACECODE
                sprintf(debug_buff,"qla2x00 : mbox 0 0x%04x \n\r", inw(host->io_port + 0x10));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"qla2x00 : mbox 1 0x%04x \n\r", inw(host->io_port + 0x12));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"qla2x00 : mbox 2 0x%04x \n\r", inw(host->io_port + 0x14));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"qla2x00 : mbox 3 0x%04x \n\r", inw(host->io_port + 0x16));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"qla2x00 : mbox 4 0x%04x \n\r", inw(host->io_port + 0x18));
                qla2100_print(debug_buff);
                sprintf(debug_buff,"qla2x00 : mbox 5 0x%04x \n\r", inw(host->io_port + 0x1a));
                qla2100_print(debug_buff);
#endif
}


#if  STOP_ON_ERROR
/**************************************************************************
*   ql2100_panic
*
**************************************************************************/
static void qla2100_panic(char *cp, struct Scsi_Host *host)
{
                scsi_qla_host_t *ha;
                long  *fp;

                ha = (scsi_qla_host_t *) host->hostdata;
                DEBUG2(ql2x_debug_print = 1;)
                sprintf(debug_buff,"qla2100 - PANIC:  %s\n",cp);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Current time=0x%lx\n", jiffies);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Number of pending commands =0x%lx\n", ha->actthreads);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Number of queued commands =0x%lx\n", ha->qthreads);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Number of free entries = (%d)\n",ha->req_q_cnt);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Request Queue @ 0x%lx, Response Queue @ 0x%lx\n",
                ha->request_dma,
                ha->response_dma);
                qla2100_print(debug_buff);
                sprintf(debug_buff,"Request In Ptr %d\n", ha->req_ring_index);
                qla2100_print(debug_buff);
                fp = (long *) &ha->flags;
                sprintf(debug_buff,"HA flags =0x%lx\n", *fp);
                qla2100_print(debug_buff);
                ql2100_dump_requests(ha);
                qla2100_dump_regs(host);
                cli();
                for( ;; ) {
                    QLA2100_DELAY(2); barrier();
                }
                sti();
}
#endif
/**************************************************************************
*   ql2100_dump_requests
*
**************************************************************************/
void
ql2100_dump_requests(scsi_qla_host_t *ha)
{

                Scsi_Cmnd       *cp;
                srb_t           *sp;
                int i;
#ifdef QL_DEBUG_ROUTINES
                qla2100_print("Outstanding Commands on controller:\n\r");
#else
                printk("Outstanding Commands on controller:\n");
#endif
                for( i=0; i < MAX_OUTSTANDING_COMMANDS; i++ ) {
                    if( (sp = ha->outstanding_cmds[i]) == NULL )
                        continue;
                    if( (cp = sp->cmd) == NULL )
                        continue;
                    sprintf(debug_buff,"(%d): Pid=%d, sp flags=0x%lx, cmd=0x%p\n\r", i, (int)sp->cmd->serial_number, (long)sp->flags,CMD_SP(sp->cmd));
#ifdef QL_DEBUG_ROUTINES
                    qla2100_print(debug_buff);
#else
                    printk("%s",debug_buff);
#endif
                }

}


/**************************************************************************
*   qla2100_setup
*
*   Handle Linux boot parameters. This routine allows for assigning a value
*   to a parameter with a ';' between the parameter and the value.
*   ie. qla2x00=arg0;arg1;...;argN;<properties .... properties>  OR
*   via the command line.
*   ie. qla2x00 ql2xopts=arg0;arg1;...;argN;<properties .... properties>
**************************************************************************/
void qla2100_setup(char *s, int *dummy)
{
                char *cp, *np;
                char	*slots[MAXARGS];
                char	**argv = &slots[0];
                char	buf[LINESIZE];
                int	argc, opts;

                /*
                * Determine if we have any properties.
                */
                cp = s;
                opts = 1;
                while( *cp && (np = qla2100_get_line(cp, buf)) != NULL ) {
                    if( strncmp("scsi-qla",buf,8) == 0 ) {
                        DEBUG(printk("qla2100: devconf=%s\n",cp);)
                        ql2xdevconf = cp;
                        (opts > 0)? opts-- : 0;
                        break;
                    }
                    opts++;
                    cp = np;
                }
                /*
                * Parse the args before the properties
                */
                if( opts ) {
                    opts = (opts > MAXARGS-1)? MAXARGS-1: opts;
                    argc = qla2100_get_tokens(s, argv, opts);
                    while (argc > 0 ) {
                        cp = *argv;
                        DEBUG(printk("scsi: found cmd arg =[%s]\n", cp);)
                        if( strcmp(cp, "verbose") == 0 ) {
                            DEBUG(printk("qla2100: verbose\n");)
                            qla2100_verbose++;
                        } else if (strcmp(cp, "quiet") == 0) {
                            qla2100_quiet = 1;
                        } else if( strcmp(cp, "reinit_on_loopdown") == 0 ) {
                            qla2100_reinit++;
                            DEBUG(printk("qla2100: reinit_on_loopdown\n");)
                        }
                        argc--, argv++;
                    }
                }

}

/********************** qla2100_get_line *********************
* qla2100_get_line
* Copy a substring from the specified string. The substring
* consists of any number of chars seperated by white spaces (i.e. spaces)
* and ending with a newline '\n' or a semicolon ';'.
*
* Enter:
* str - orig string
* line - substring
*
* Returns:
*   cp - pointer to next string
*     or
*   null - End of string
*************************************************************/
static char	*qla2100_get_line(char *str, char *line)
{
                register	char 	*cp = str;
                register	char 	*sp = line;

                /* skip preceeding spaces */
                while( *cp && *cp == ' ' )
                    ++cp;
                while ( (*cp) && *cp != '\n' && *cp != ';' )   /* end of line */
                    *sp++ = *cp++;

                *sp = '\0';
                DEBUG5(printk("qla2100_get_line: %s\n",line);)
                if( (*cp) ) {
                    cp++;
                    return( cp );
                }
                return( NULL );
}


/**************************** get_tokens *********************
* Parse command line into argv1, argv2, ... argvX
* Arguments are seperated by white spaces and colons and end
* with a NULL.
*************************************************************/
static int qla2100_get_tokens(char *line, char **argv, int maxargs )
{
                register	char 	*cp = line;
                int	count = 0;

                while( *cp && count < maxargs ) {
                    /* skip preceeding spaces */
                    while((*cp) && *cp == ' ')
                        ++cp;
                    /* symbol starts here */
                    argv[count++] = cp;
                    /* skip symbols */
                    while ( (*cp) && !( *cp == ' ' || *cp == ';' || *cp == ':' ) )
                        cp++;
                    /* replace comma or space with a null */
                    if( (*cp) && (*cp ==' ' ) && argv[count-1] != cp )
                        *cp++ = '\0';
                }
                return( count );
}

#if VSA
/*
 * qla2100_get_vsa_opt_from_config
 *      Get VSA option from the configuration parameters.
 *      Bit order is little endian.
 *
 * Input:
 * ha  -- Host adapter
 * tgt  -- target/device number
 */
void
qla2100_get_vsa_opt_from_config(scsi_qla_host_t *ha,
    uint16_t tgt, uint16_t dev_no)
{

	char		propbuf[60]; /* size of search string */
	int		rval;
	char		vsa; 

	/* Get "target-N-device-N-vsa" as a 1 bit value */
	sprintf(propbuf, "scsi-qla%ld-tgt-%d-di-%d-vsa",
	    ha->instance, tgt, dev_no);
	rval = qla2x00_get_prop_xstr(ha, propbuf, (uint8_t *)&vsa,1);

	if (rval != -1 && rval == 1) {
        ha->fc_db[tgt].flag |= DEV_FLAG_VSA;
        DEBUG(printk("cfg: scsi-qla%d-target-%d-vsa=1\n",
           		(int) ha->instance,  tgt);)
	}
}

#endif

/*
 * qla2x00_cfg_persistent_binding
 *	Get driver configuration file target persistent binding entries.
 *
 * Input:
 *	ha = adapter block pointer.
 *
 * Context:
 *	Kernel context.
 */
STATIC void
qla2x00_cfg_persistent_binding(adapter_state_t *ha)
{
	char		propbuf[LINESIZE];
	int		tmp_rval;
	uint8_t		name[8];
	uint16_t	tgt;
	uint16_t	dev_no = 0; /* not used */
	char		*cmdline = ha->cmdline;
	uint8_t		tmp_name[8];

	DEBUG3(printk("qla2x00_cfg_persistent_binding: entered.\n");)

	/* FC name for devices */
	for (tgt = 0; tgt < MAX_FIBRE_DEVICES; tgt++) { 
#if  USE_PORTNAME  /* updated for ioctl merge */
		sprintf(propbuf, "scsi-qla%d-tgt-%d-di-%d-port",
		    (int)ha->instance, tgt, dev_no);
#else
		sprintf(propbuf, "scsi-qla%d-tgt-%d-di-%d-node",
		    (int)ha->instance, tgt, dev_no);
#endif
		tmp_rval = qla2100_get_prop_16chars (ha, propbuf, name,
		    cmdline);

		if ( tmp_rval != 0 )
			continue;

		/* Got a valid name for this ID. */
#if  USE_PORTNAME  /* updated for ioctl merge */
		BCOPY(name, ha->fc_db[tgt].wwn, 8);

		/* copy node name if one is specified */
		sprintf(propbuf, "scsi-qla%d-tgt-%d-di-%d-node",
		    (int)ha->instance, tgt, dev_no);
		tmp_rval = qla2100_get_prop_16chars (ha, propbuf,
		    tmp_name, cmdline);
		if (tmp_rval == 0)
			BCOPY(tmp_name, ha->fc_db[tgt].name, 8);
#else
		BCOPY(name, ha->fc_db[tgt].name, 8);

		/* copy port name if one is specified */
		sprintf(propbuf, "scsi-qla%d-tgt-%d-di-%d-port",
		    (int)ha->instance, tgt, dev_no);
		tmp_rval = qla2100_get_prop_16chars (ha, propbuf,
		    tmp_name, cmdline);
		if (tmp_rval == 0)
			BCOPY(tmp_name, ha->fc_db[tgt].wwn, 8);
#endif
		ha->fc_db[tgt].loop_id = PORT_AVAILABLE;
		ha->fc_db[tgt].flag = 0;  /* v2.19.05b3 */
		ha->fc_db[tgt].flag |= DEV_CONFIGURED;

		qla2x00_tgt_alloc(ha, tgt, name);
		DEBUG(printk("Target %03d - configured by user: ",tgt);)
		DEBUG(printk("tgt-%03d=%02x%02x%02x%02x%02x%02x%02x%02x\n",
		    tgt,
		    name[0], name[1], name[2], name[3],
		    name[4], name[5], name[6], name[7]);)
			/* look for VSA */
#if VSA
		qla2100_get_vsa_opt_from_config(ha, tgt, dev_no);
#endif

	}

	DEBUG3(printk("qla2x00_cfg_persistent_binding: exiting normally.\n");)
}


/*
 * kmem_zalloc
 * Allocate and zero out the block of memory
 */
inline void *kmem_zalloc( int siz, int code, int id)
{
	uint8_t *bp; 
	
	if( (bp = kmalloc(siz, code)) != NULL ){
		memset(bp, 0, siz);
	}
#if QL_TRACE_MEMORY
	if ( mem_trace_ptr == 1000 )
		mem_trace_ptr = 0;
	mem_trace[mem_trace_ptr] = (u_long ) bp;
	mem_id[mem_trace_ptr++] = (u_long ) id;
#endif
	return( (void *) bp);
}

#if 0
/*
 * kmem_free
 * Deallocate the block of memory
 */
inline void kmem_free( void *ptr )
{
#if QL_TRACE_MEMORY
	int	i;

	for( i =0; i < 1000; i++ )
		if( mem_trace[i] == (unsigned long) ptr ) {
			mem_trace[i]  = (unsigned long) NULL;
			break;
		}
#endif
	kfree(ptr);
}
#endif

inline void qla_bcopy( unsigned char *dst, unsigned char *src, int siz)
{
    int     i;
	
    for ( i = 0 ; i < siz ; i++ ) {
	    *dst++ = *src++;
    }
}

#ifdef FC_IP_SUPPORT
            /* Include routines for supporting IP */
#include "qla2100ip.c"
#endif /* FC_IP_SUPPORT */

#if MPIO_SUPPORT
/*
 * Declarations for failover
 */
#include "qla_cfg.c"
#include "qla_fo.c"
#endif
#if  APIDEV
/****************************************************************************/
/* Create character driver "HbaApiDev" w dynamically allocated major number */
/* and create "/proc/scsi/qla2x00/HbaApiNode" as the device node associated */
/* with the major number.                                                   */
/****************************************************************************/

#define APIDEV_NODE  "HbaApiNode"
#define APIDEV_NAME  "HbaApiDev"

static int apidev_major = 0;
static struct Scsi_Host *apidev_host = 0;

static int apidev_open(struct inode *inode, struct file *file)
{
	printk (KERN_INFO "qla2100_apidev: open MAJOR number = %d, MINOR number = %d\n", MAJOR (inode->i_rdev), MINOR (inode->i_rdev));

    return 0;
}
static int apidev_close(struct inode *inode, struct file *file)
{
	printk (KERN_INFO "qla2100_apidev: closed\n");

    return 0;
}

static int apidev_ioctl(struct inode *inode, struct file *fp, unsigned int cmd, unsigned long arg)
{
    Scsi_Device fake_scsi_device;
    fake_scsi_device.host = apidev_host;
    return(qla2x00_ioctl(&fake_scsi_device, (int)cmd, (void*)arg));
}

static struct file_operations apidev_fops = {
    ioctl:    apidev_ioctl,
    open:     apidev_open,
    release:  apidev_close
};

static int apidev_init(struct Scsi_Host *host)
{

	if(apidev_host) {
		return 0;
	}

	if (0 > (apidev_major = register_chrdev(0, APIDEV_NAME, &apidev_fops))) {
		DEBUG(printk("qla2100_apidev: register_chrdev rc=%d\n",
		    apidev_major);)

		return apidev_major;
	}

	apidev_host = host;

	DEBUG(printk("qla2x00: Creating (%s) %s/%s major=%d\n",
	    host->hostt->proc_name,
	    host->hostt->proc_dir->name, APIDEV_NODE, apidev_major);)

	proc_mknod(APIDEV_NODE, 0777+S_IFCHR, host->hostt->proc_dir,
	    (kdev_t)MKDEV(apidev_major,0));

	return 0;
}

static int apidev_cleanup()
{
    if(!apidev_host) return 0;
    unregister_chrdev(apidev_major,APIDEV_NAME);
    remove_proc_entry(APIDEV_NODE,apidev_host->hostt->proc_dir);
    apidev_host = 0;
    return 0;
}
#endif /* APIDEV */

#ifdef QL_DEBUG_ROUTINES
#if  DEBUG_GET_FW_DUMP
#include  "x2300dbg.c"
#endif
#endif

/*
* Overrides for Emacs so that we almost follow Linus's tabbing style.
* Emacs will notice this stuff at the end of the file and automatically
* adjust the settings for this buffer only.  This must remain at the end
* of the file.
* ---------------------------------------------------------------------------
* Local variables:
* c-indent-level: 2
* c-brace-imaginary-offset: 0
* c-brace-offset: -2
* c-argdecl-indent: 2
* c-label-offset: -2
* c-continued-statement-offset: 2
* c-continued-brace-offset: 0
* indent-tabs-mode: nil
* tab-width: 8
* End:
*/

